//======================================================================================================
// GAME.H
// Copyright 2022 Everett Kaser
// All rights reserved
//======================================================================================================

#define	MSGHUU				5
#define	BARWUU				12
#define	MSGHPX				(5*UU)
#define	BARWPX				(BARWUU*UU)

#define	MAXBRDW	17
#define	MAXBRDH	12

#define	SPECIAL_TOTAL	1750

#define	MAXPUZZLENUM	65536

#define	MINSAVEGAMEREV		6

#define	HumanButtNum	(10*(CurPlayer.CurButtNum/6+2)+((CurPlayer.CurButtNum%6)+2))
#define	CalcPuzzW		(2+3*(2+CurPlayer.CurButtNum/6))
#define	CalcPuzzH		(6+(CurPlayer.CurButtNum%6)+(CurPlayer.CurButtNum==5?1:0))


#define	WIN_FIRE	0
#define	WIN_QUILT	1
#define	WIN_WEAVE	2
#define	WIN_BOUNCE	3
#define	WIN_RANDOM	4
#define	WIN_NONE	5

#define	MM_NOACTION		0
#define	MM_MOVETILE		1
#define	MM_FINDAVAIL	2
#define	MM_FINDTPBT		3
#define	MM_FINDALL		4
#define	MM_UNDO			5
#define	MM_REDO			6
#define	MM_NEXT_MOOD	7
#define	MM_PREV_MOOD	8

#define	MM_NUMACTIONS	9

typedef struct {
	DWORD	tim;
	WORD	swaps;
	WORD	puzznum;
	WORD	used;
	char	player[32];
} SCORE;

#define	NUMUNDOS	1024
typedef	struct {
	WORD	move;
} UNDO;

typedef struct {
// NOTE: The following 3 values MUST remain first,
//	so that LoadGame() can check them before loading
//	the saved game file.
	DWORD	SaveRev, SavePuzzType, SavePuzzNum;
	
	DWORD	StartTime;			// time when puzzle was started
	DWORD	LastTime;			// last time screen timer was updated (so only drawn once)
	DWORD	LastMoves;			// last MoveCnt written to screen
	DWORD	LastBest;			// last BestMoveCnt written to screen
	DWORD	PuzzW;
	DWORD	PuzzH;
	DWORD	TilesH;				// PuzzH-4 for all sizes except 27 which is PuzzH-5
	DWORD	Justify;			// 1 for top justify, 2 for bottom justify
	DWORD	MoveCnt;			// number of moves made solving puzzle
// NOTE: Put smallish non-array stuff above, as first 512 bytes
//	of SAVEGAME are written to save file UNcompressed. The remainder
//	
	DWORD	CurDo, DoCnt;
	UNDO	Undos[NUMUNDOS];
	WORD	board[MAXBRDH][MAXBRDW];
	BYTE	ColsDone[MAXBRDW];	// if OK should show on columns
} SAVEGAME;

// MouseMap activities:
#define	CLK_IDLE		0
#define	CLK_DOWN		1
#define	CLK_DRAG		2
#define	CLK_HOLD		3
#define	CLK_CLICK		4
#define	CLK_CLICKEND	5
#define	CLK_DCLICK		6

#define	CLK_START		1
#define	CLK_CHANGED		2
#define	CLK_END			3
#define	CLK_ABORT		4

void DrawBoard();
void InitGame();
BOOL IsPuzzleSolved();
void SetupDispVars();
void RandomizeTiles();
void PointerMove(long x, long y, DWORD flags);
BOOL IsPaused();
void SaveGame();
void SaveDo(long srcCol, long srcTopBot, long dstCol, long dstTopBot);
void UndoDo();
void RedoDo();
void RestartPuzzle();
void GameTimer();
void ReadBest();
void WriteBest();
void GetBest();
void SetBest(BOOL msgthem);
void NoBlink();
void BackSize();
DWORD RequestGG(char *imgname, long gus);
DWORD LoadSized(char *imgname, long gus);
BOOL AvailableGG(char *imgname, long gus);
void LoadGraphics(char *imgname);
void RunWinDisplay();
void TermWinDisplay();
void StartPuzzle(DWORD n);
void EndPuzzle();
void GameWheel(long x, long y, long dir);
void GameHWheel(long x, long y, long dir);
void NextMood();
void PrevMood();
void PlayNextWinMusic();
void PlayNextBackMusic();
void MovePauseAnimation();
void SetPause(DWORD on);
void NextUnsolvedPuzzle();
void FirstUnsolvedPuzzle();
void StartPuzzleType();
void StartMood(long m, long upwnd);
BYTE *GetBestPtr(long n);

extern KEYMAP	defkeymap[DEFKEYMAPCNT];
extern KEY		Keys[256];
extern long		KeysCnt;
extern SAVEGAME	Game;
extern DWORD	SaveDoCnt, SaveCurDo;
extern char		*SaveGameRename;
extern char		*iniSOLVEDTOTAL;
extern long		BlinkListCnt;
extern long		BlinkList[2*(MAXBRDW-2)];
extern long		PuzzSizes[16];
extern BOOL		IgnoreKeys;
extern long		GameWon;
extern DWORD	PauseTime;
extern BOOL		Paused;		// TRUE if game is paused
extern long		PauseX, PauseY, PauseW, PauseH, PauseI;
extern DWORD	hPauseBM;
extern BOOL		PuzzleActive;	// TRUE if in between StartPuzzle() and EndPuzzle()
extern long		BestMoveCnt;		// Game.MoveCnt for best solution for current puzzle
extern BOOL		SaveBest;			// TRUE if hBest list has been modified and needs storing to disk
extern BOOL		Retest;
extern DWORD	hBest;
extern long		BestEnd;			// len (end) of used bytes in hBest
extern BOOL		TestBest;
extern	BOOL	Retest;
extern	DWORD	hBest;
