//======================================================================================================
// MAIN.C
// Copyright 2022 Everett Kaser
// All rights reserved
//======================================================================================================
#include "lib.h"
#include "player.h"
#include "main.h"
#include "dlg.h"
#include "id.h"
#include "str.h"
#include "game.h"

//==============================================================
// NOTE: There can be NO actual data/code-generating declarations
// before the .KE HEADER. The .KE HEADER *MUST* *MUST* *MUST* be
// the FIRST thing compiled/assembled, as the HEADER *MUST* exist
// at address 0 in the .KE file!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//==============================================================

extern BYTE	EventBuf[1024];

//==============================================================
// .KE HEADER
//==============================================================
/*
	The following forms the 'header' at the start of a .KE file.
	The KINT interpreter depends upon these being there, in this
	particular order. They MUST all be INITIALIZED when they're
	declared here, so that they occur at the START of the .KE
	file. If not initialized, then they'd get relegated to the
	uninitialized data region (i.e., moved elsewhere behind the
	curtain. So, if you muck with these, be SURE to initialize
	any changes you make!
*/
QWORD	NULLPTR=0x0123456789ABCDEFULL;	// marker for checking for write to address 0
QWORD	CPU0=0;
QWORD	CPU1=0;
QWORD	CPU2=0;
QWORD	CPU3=0;
QWORD	CPU4=0;
QWORD	CPU5=0;
QWORD	CPU6=0;
QWORD	CPU7=0;
QWORD	CPU8=0;
QWORD	CPU9=0;
QWORD	CPU10=0;
QWORD	CPU11=0;
QWORD	CPU12=0;
QWORD	CPU13=0;
QWORD	CPU14=0;
QWORD	CPU15=0;
QWORD	CPUpc=0;
QWORD	CPUsb=0;
QWORD	CPUse=0;
QWORD	CPUsp=0;
QWORD	CPUfp=0;
QWORD	CPUmp=0;
QWORD	CPUscratch=0;
BYTE	CPUflags=0;
BYTE	EventVal=0;
BYTE	KINTresBYTE[6]={0,0,0,0,0,0};
QWORD	CPUreserved=0;
//					.KE MARKER "KINT6",NUL	Y Y	M D		REV		BUILD#
char	KINTrev[] = {'K','I','N','T','6',0,	0,0,0,0,	2,0,	0,0,0,0};
QWORD	KINTcodelen=0;
QWORD	KINTzerolen=0;
QWORD	KINTstacksize=64*1024;
QWORD	KINTheapsize=
/*
167772160
 30277632
  7340032
 53248000
  4194304
    11072
  8388698
  1048576
134349056
 16777216
---------
423406656
*/
				  4*8*1024*5*1024	// max hMain->hBM bitmap buffer (4 bytes/pixel, 8K x 5K pixels)
				+ 2*4*UIIMG_UW*32*UIIMG_UH*32	// two bitmaps (SOURCE and SIZED) of UI bitmap, 4 bytes/pixel, 42 units wide by max 32 pixels per unit, by 88 units tall by 32 pixels per unit
				+ 4*4*7*256*256		// game graphics SOURCE bitmap, 4 bytes/pixel, 4 rows, 7 columns, max tile width = 256 (overkill)
				+ 4*10*13*320*320	// game graphics SIZED bitmap, 4 bytes/pixel, 10 rows, 13 columns, max tile width = 320 (WAY overkill)
				+ 64*65536			// storage for a listbox with 64K items or two listboxes with 32K items with average string length of <44 bytes
				+ 4*(sizeof(DIALOG)+32*sizeof(BUTTONBM))	// up to 4 simultaneous dialogs with up to 32 controls in each dialog (overkill)
				+ 4*2*1024*1024		// max SysFontBM
				+ 4*512*512			// temp working bitmap for Help EXPORT
				+ 4*64+64*1024*(2+1024*2)	// Best Moves list (worst case)
				+ 16*1024*1024		// misc/slop
			;
QWORD	KINTconstn=0;	// initialized/overwritten by KINT
QWORD	KINTstartWndPos=0;
QWORD	KINTreserved[30]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

//=======================================================================
/*
	This is the function/point at which the interpreter always
	starts 'running' the .KE code. EVERYTHING gets started via
	a 'message' to this function, which then vectors control
	elsewhere according to the message passed. This includes
	startup/shutdown types of messages, as well as "real world"
	messages (such as keystrokes, mouse clicks, timers, etc.).
	The return value from EventProc() is message-dependent.
	Some messages will ignore it, some will use the returned
	value.
*/
DWORD EventProc(DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND	hH;
	long	x, y;
/*
long id;
if( msg!=KMSG_TIMER && msg!=KMSG_POINTER_MOVE ) {
//	SP()"EVENT IN(%d,%d,%d,%d) SP=0x%8.8X", msg,arg1,arg2,arg3,*(DWORD*)(13*8));
	SP()"EVENT IN(%d,%d,%d,%d\t%s)", msg,arg1,arg2,arg3,KMSGlist[msg]);
	WriteDebug(ScratBuf);
}
*/
	EventVal = 0;	// pre-set normal Event return value (only different for possibly KMSG_IDLE and KMSG_QUERY_QUIT)
	if( *(BYTE*)(0xC2) ) {
		if( msg>KMSG_END_STARTUP ) {
			if( *(BYTE*)(0xC2) & 1 ) {	// if running EXPORT HELP script
				switch( ScriptState ) {
				 case 0:	// unstarted
					if( !hModal ) {
						++ScriptState;
						MenuCurSelID = ID_MENU_HELP;
						SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
					}
					break;
				 case 1:
					++ScriptState;
					SendMsgID(ID_HELP_EXPORT, WM_ACTIVATE, 0, 0, 0);
					break;
				 case 2:
					++ScriptState;
					SendMsg(hModal, WM_DESTROY, ID_OK, 0, 0);
					break;
				 case 3:
					++ScriptState;
	//SP()"hModal=%d (%d) hFocus=%d (%d)",hModal,((WND*)HeapPtr(hModal))->id, hFocus, ((WND*)HeapPtr(hFocus))->id);SMS();
					SendMsg(hModal, WM_DESTROY, ID_OK, 0, 0);
					break;
				 case 4:
					++ScriptState;
					SaveMenuCtls();
					SendMsg(hMain, WM_QUERY_QUIT, 0, 0, 0);
					break;
				 case 5:
					*(BYTE*)(0xC2) &= ~1;
					ScriptState = 0;
					break;
				}
			}
		}
	}
	switch( msg ) {
	 case KMSG_BEGIN_STARTUP:
		OnBeginStartup();
		break;
	 case KMSG_END_STARTUP:
		OnEndStartup();
		break;
	 case KMSG_QUERY_QUIT:
		if( hMain ) EventVal = SendMsg(hMain, WM_QUERY_QUIT, 0, 0, 0);
		else EventVal = 1;
		if( EventVal ) _asm(term);	// quit
		break;
	 case KMSG_SHUTDOWN:
		break;
	 case KMSG_FOCUS:
		break;
	 case KMSG_SCREEN_CHANGE:
		SendMsg(hMain, WM_SCREEN_CHANGE, 0, 0, 0);
		break;
	 case KMSG_WINDOW_HIDDEN:
		break;
	 case KMSG_WINDOW_MOVED:
		SendMsg(hMain, WM_NOTIFY, WN_MODIFIED, arg1, arg2);
		break;
	 case KMSG_WINDOW_SIZED:
		ViewW = arg1;
		ViewH = arg2;
		ViewMR= arg3;
		SendMsg(hMain, WM_PARENTSIZED, arg1, arg2, arg3);
		break;
	 case KMSG_MUSIC_DONE:
		SendMsg(hMain, WM_MUSIC_DONE, 0, 0, 0);
		break;
	 case KMSG_TIMER:
		OnTimer();
		break;
	 case KMSG_KEY_DOWN:
		// 0xE0 and above are OEM keys that can cause us grief
		if( arg1 < 0x0E0 || arg1 > 0x0FF ) OnKeyDown(arg1, arg2, arg3);
		break;
	 case KMSG_KEY_UP:
		if( arg1 < 0xE0 || arg1 > 0x0FF ) {
			if( !hMenuDlg && Paused && arg1!=K_SHIFT && arg1!=K_CTL ) SetPause(0);
			else OnKeyUp(arg1, arg2, arg3);
		}
		break;
	 case KMSG_CHAR:
		*(DWORD*)EventBuf = arg1;
		*(DWORD*)(EventBuf+4) = arg2;
		EventBuf[8] = 0;	// backup NUL, in case of unplanned excessosity
		OnChar(EventBuf, arg3);
		break;
	 case KMSG_POINTER_MOVE:
		x = arg1;
		y = arg2;
		hH = FindPtWnd(x, y);
	
		if( hH!=hHover ) {
			SendMsg(hHover, WM_HOVER, 0, 0, 0);
			hHover = hH;
			SendMsg(hHover, WM_HOVER, 1, 0, 0);
		}
		MouseMsg(WM_PTRMOVE, x, y, arg3);
		break;
	 default:
		sprintf((char*)EventBuf, "Unknown KMSG: %d. Terminating. So sorry!", msg);
		SystemMessage((char*)EventBuf);
		EventVal = 1;
		_asm(term);	// quit
	}
	if( NeedDrawAll ) mainDrawAll();
	UpdateInvalid();
/*
if( msg!=KMSG_TIMER && msg!=KMSG_POINTER_MOVE ) {
//	SP()"EVENT OUT(%d,%d,%d,%d) SP=0x%8.8X hMainBM=%d *=0x%8.8X", msg,arg1,arg2,arg3,*(DWORD*)(13*8), hMainBM, *(((DWORD*)HeapPtr(hMainBM))+7));
	SP()"EVENT OUT(%d,%d,%d,%d\t%s)", msg,arg1,arg2,arg3,KMSGlist[msg]);
	WriteDebug(ScratBuf);
}
*/
	return 0;
}

//==============================================
// EventBuf[] is a general resource for EventProc() and NO ONE ELSE!
BYTE	EventBuf[1024];

//=======================================================================
// arg1 = keycode
// arg2 = repeat cnt & flags (OS dependent)
// arg3 = x/y for mouse buttons
void OnKeyDown(DWORD arg1, DWORD arg2, DWORD arg3) {
	
	long	x, y, w;
	short	k;

	k = GetKeyStatus(K_ALT);
	if( k<0 ) KeyFlags |= MK_ALT;
	else KeyFlags &= ~MK_ALT;
	
	w = (long)arg2;			// save whole thing for WHEEL
	arg2 &= 0x0FFFF;	// keep only repeat cnt for everyone else
	x = arg3 & 0x0FFFF;
	y = arg3 >> 16;	// for mouse events

	if( !hMenuDlg && Paused ) return;
	switch( arg1 ) {
	 case K_LBUTT:
		KeyFlags |= MK_LBUTT;
		MouseMsg(WM_LBUTTDN, x, y, arg2);
		return;
	 case K_MBUTT:
		KeyFlags |= MK_MBUTT;
		MouseMsg(WM_MBUTTDN, x, y, arg2);
		return;
	 case K_RBUTT:
		KeyFlags |= MK_RBUTT;
		MouseMsg(WM_RBUTTDN, x, y, arg2);
		return;
// wParam is 0x00010020 when BACK is pressed, 0x00010000 when BACK is released.
// wParam is 0x00020040 when FWD is pressed,  0x00020000 when FWD is released.
	 case K_BACKBUTT:
		KeyFlags |= MK_BBUTT;
		MouseMsg(WM_BBUTTDN, x, y, arg2);
		return;
	 case K_FWDBUTT:
		KeyFlags |= MK_FBUTT;
		MouseMsg(WM_FBUTTDN, x, y, arg2);
		return;
	 case K_WHEEL:
		if( (hMenuDlg?tmpMood.flags:CurMood.flags) & PF_REV_WHEEL ) w = -w;
		MouseMsg(WM_WHEEL, x, y, w);
		return;
	 case K_HWHEEL:
		if( (hMenuDlg?tmpMood.flags:CurMood.flags) & PF_REV_TILT ) w = -w;
		MouseMsg(WM_HWHEEL, x, y, w);
		return;
	 case K_F4:
		if( !(KeyFlags & MK_ALT) ) break;
		if( SendMsg(hMain, WM_QUERY_QUIT, 0, 0, 0) ) _asm(TERM);
		return;
	 case K_LFCURS:
		if( KeyFlags & MK_ALT ) {
			MouseMsg(WM_BBUTTDN, MMlastX, MMlastY, MMlastFlags | MK_BBUTT);	// emulate BACK button
			return;
		}
		break;
	 case K_RTCURS:
		if( KeyFlags & MK_ALT ) {
			MouseMsg(WM_FBUTTDN, MMlastX, MMlastY, MMlastFlags | MK_FBUTT);	// emulate FWD button
			return;
		}
		break;
	 case K_ALT:
		if( KeyFlags & MK_ALT ) return;
		KeyFlags |= MK_ALT;
		break;
	 case K_SHIFT:
		KeyFlags |= MK_SHIFT;
		break;
	 case K_CTL:
		KeyFlags |= MK_CTL;
		break;
	}
	SendMsg(hCapture?hCapture:(hFocus?hFocus:hMain), WM_KEYDN, arg1, arg2, arg3);
}

//=======================================================================
void OnKeyUp(DWORD arg1, DWORD arg2, DWORD arg3) {
	
	long	x, y, w;

	w = arg2;			// save whole thing for WHEEL
	arg2 &= 0x0FFFF;
	x = arg3 & 0x0FFFF;
	y = arg3 >> 16;	// for mouse events
	if( x & 0x8000 ) x -= 65536;
	if( y & 0x8000 ) y -= 65536;
	
	switch( arg1 ) {
	 case K_LBUTT:
		KeyFlags &= ~MK_LBUTT;
		MouseMsg(WM_LBUTTUP, x, y, arg2);
		return;
	 case K_MBUTT:
		KeyFlags &= ~MK_MBUTT;
		MouseMsg(WM_MBUTTUP, x, y, arg2);
		return;
	 case K_RBUTT:
		KeyFlags &= ~MK_RBUTT;
		MouseMsg(WM_RBUTTUP, x, y, arg2);
		return;
	 case K_BACKBUTT:
		KeyFlags &= ~MK_BBUTT;
		MouseMsg(WM_BBUTTUP, x, y, arg2);
		return;
	 case K_FWDBUTT:
		KeyFlags &= ~MK_FBUTT;
		MouseMsg(WM_FBUTTUP, x, y, arg2);
		return;
	 case K_WHEEL:
		MouseMsg(WM_WHEEL, x, y, w);
		return;
	 case K_HWHEEL:
		MouseMsg(WM_HWHEEL, x, y, w);
		return;
	 case K_LFCURS:
		if( KeyFlags & MK_ALT ) {
			MouseMsg(WM_BBUTTUP, MMlastX, MMlastY, MMlastFlags & ~MK_BBUTT);	// emulate BACK button
			return;
		}
		break;
	 case K_RTCURS:
		if( KeyFlags & MK_ALT ) {
			MouseMsg(WM_FBUTTUP, MMlastX, MMlastY, MMlastFlags & ~MK_FBUTT);	// emulate FWD button
			return;
		}
		break;
	 case K_ALT:
		KeyFlags &= ~MK_ALT;
		break;
	 case K_SHIFT:
		KeyFlags &= ~MK_SHIFT;
		break;
	 case K_CTL:
		KeyFlags &= ~MK_CTL;
		break;
	}
	SendMsg(hCapture?hCapture:(hFocus?hFocus:hMain), WM_KEYUP, arg1, arg2, arg3);
}

//=======================================================================
void OnChar(BYTE *pCharBuf, DWORD repflags) {

	SendMsg(hCapture?hCapture:(hFocus?hFocus:hMain), WM_CHAR, (DWORD)pCharBuf, repflags, 0);
}

//================================================
void MouseMsg(DWORD msg, long x, long y, DWORD flags) {

	DWORD	hW;
	
	MMlastFlags = flags;
	MMlastX = x;
	MMlastY = y;	// save these so ALT-LFcurs and ALT-RTcurs (BACK/FWD button emulation) works

	hW = FindPtWnd(x, y);
	SendMsg(hW, msg, x, y, flags);
}

//====================================================================================
//====================================================================================
//====================================================================================
//====================================================================================
DWORD mainProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD FrodoProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD MenuButtProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);

char 	PgmTitle[]		= "Crocotile";
char	Copyright[]		= "Copyright 2022 Everett Kaser";
char	*PgmFolder		= "CT";
#define	GAMEVER			"200"	// for iniSaveHeader()
char	*logname		= "CTLOG.CSV";
char	*iniFILENAME	= "ct.ini";

char	*FMTd			= "%d";
char	*FMTdSP			= "%d ";
char	*FMTd5			= "%d %d %d %d %d";
char	*FMT8x			= "0x%8.8X";
char	*FMT8x8x		= "0x%8.8X 0x%8.8X";
char	*FMT8xSP		= "0x%8.8X ";

char	*iniREV			= "irev";
char	*pgmREV			= "prev";
char	*iniCURPLAYER	= "curplayer";
char	*iniNUMPLAYERS	= "numplayers";
char	*iniMAXID		= "maxid";

char	*iniPLAYER		= "PLAYER%d";
char	*iniID			= "id";
char	*iniNAME		= "name";
char	*iniMOODCNT		= "moodcnt";
char	*iniCURMOOD		= "curmood";

char	*iniHELPFLAGS	= "htmlflags";
char	*iniHELPCOLORS	= "htmlcolors";
char	*iniHELPFONT	= "htmlfont";
char	*iniHELPWIDTH	= "htmlwidth";
char	*iniHELPLMARGIN = "htmlleftmargin";
char	*iniHELPCONTENT = "htmlcontentbar";
char	*iniRULESPACE	= "htmlrulespace";

char	*iniMOOD		= "PLAYER%dMOOD%d";
char	*iniUIIMAGE		= "ui";
char	*iniIMAGE		= "image";
char	*iniHELP		= "help";
char	*iniSTRINGS		= "strings";
char	*iniKEYMAPS		= "keymaps";
char	*iniCURBUTT		= "curbutt";
char	*iniCURPUZZNUM	= "curpuzznum";
char	*iniWNDPOS		= "window";
char	*iniACTIONS		= "actions";
char	*iniFLAGS		= "flags";
char	*iniUU			= "UIsize";
char	*iniMUSICBACK	= "musicback";
char	*iniMUSICWIN	= "musicwin";
char	*iniWINDISP		= "windisp";
char	*iniPUZZSIZE	= "puzzpercent";
char	*iniCSVSEP		= "csvsep";

BYTE	defactions[20]={
	MM_MOVETILE, MM_FINDTPBT, MM_FINDAVAIL, MM_UNDO, MM_REDO,
	MM_NOACTION, MM_FINDALL,  MM_NOACTION, MM_NOACTION, MM_NOACTION,
	MM_FINDTPBT, MM_NOACTION, MM_NOACTION, MM_NOACTION, MM_NOACTION,
	MM_NOACTION, MM_NOACTION, MM_NOACTION, MM_NOACTION, MM_NOACTION,
};
DWORD	defflags		= 
	  PF_SHOW_TOOLTIPS	| PF_SHOWTIMER		| PF_SHOWTILESMOVED
	| PF_PROMPTNEXT		| PF_PROMPTRESTART	| PF_GAMESOUNDS
	| PF_ERRORSOUNDS	| PF_SHOWBESTCOUNT	| PF_ENABLECLICK;
DWORD	defsndflags		= 0;

KEYMAP	defkeymap[DEFKEYMAPCNT]={
	{K_K, 0, KF_KEYMAP_DLG},
	{K_O, 0, KF_OPTIONS_DLG},
	{K_M, 0, KF_MOODS_DLG},
	{K_M, MK_CTL, KF_NEXT_MOOD},
	{K_M, MK_SHIFT, KF_PREV_MOOD},
	{K_F, MK_CTL, KF_NEXT_MOOD},
	{K_F, MK_SHIFT, KF_PREV_MOOD},
	{K_P, 0, KF_PLAYER_DLG},
	{K_P, MK_CTL, KF_NEXT_PLAYER},
	{K_P, MK_SHIFT, KF_PREV_PLAYER},
	{K_UNDERSUB, 0, KF_PREV_PUZZ},
	{K_NUMPADSUB, 0, KF_PREV_PUZZ},
	{K_NUMPADADD, 0, KF_NEXT_PUZZ},
	{K_EQPLUS, 0, KF_NEXT_PUZZ},
	{K_EQPLUS, MK_SHIFT, KF_NEXT_PUZZ},
	{K_SPACE, 0, KF_PAUSE},
	{K_F1, 0, KF_MENU_DLG},
	{K_N, 0, KF_NEXT_UNSOLVED},
	{K_N, MK_SHIFT, KF_FIRST_UNSOLVED},
	{K_Z, MK_CTL, KF_UNDO},
	{K_Y, MK_CTL, KF_REDO},
	{K_R, MK_CTL, KF_RESTART},
	{K_S, MK_CTL, KF_SAVEGAME},
// when you add more here, don't forget to bump DEFKEYMAPCNT in PLAYER.H
};

KEY		Keys[256];
long	KeysCnt;
DWORD	LastDownKey=0xFFFFFFFF;	// arg1 keycode of last key pressed (except for SHIFT/CTL)
DWORD	LastDownKeyFlags;		// KeyFlags at time of LastDownKey

char	*defbackmusic	= "BACHMIN1.MID";
char	*defwinmusic	= "HALLELUJ.MID";
char	*defuiname		= "EKS_Tan3D.png";
char	*defimgname		= "Crocotile.png";
char	*defHelp		= "HLP_English.txt";
char	*defStrings		= "STR_English.txt";

//#define	SIZEDTYPE		0	// 0 for BMP (fastest, most responsive), 1 for PNG (smallest file sizes)
char	*sizedUIname	= "SIZED_U%d_%s.bmp";

HWND	hMain=0;	// handle to MAIN window
DWORD	hMainBM=0;	// handle of MAIN window's BM (created when hMain created, destroyed when hMain destroyed, otherwise always there and globally accessible)
DWORD	hSplashBM=0;
HWND	hMsgBar=0;	// handle of message bar dialog
HWND	hButtBar=0;	// handle of button bar dialog

HWND	hTitle=0;	// title/copyright dialog

char	MsgText[256];	// text buffer for MsgBar message
DWORD	MsgBackClr, MsgTextClr;

char	UIBMOname[128];	// name of ORIGINAL UI graphics file
DWORD	hUIBMO;		// handle of ORIGINAL UI graphics
long	UIBMUO;		// unit size of hUIBMO UI elements
QWORD	UIBMtime;	// time of last modification of UIBMO image set

DWORD	hUIBM;		// handle of UI graphics SIZED
long	UIBMU;		// unit size of hUIBM UI elements, 1/4 of standard button w/h
long	UIBMU2;
long	UIBMU3;
long	UIBMU4;		// 4*UIBMU, size of standard button in hUIBM
long	UIBMU5;
long	UIBMU8;		// 8*UIBMU, width of CF_DBLWIDE button in hUIBM
long	UIBMU10;

char	GBMOname[128];	// name of ORIGINAL game graphics file
DWORD	hGBMO;		// handle of ORIGINAL game graphics
long	GBMOwh;		// width/height of one tile in ORIGINAL game graphics
QWORD	GBMtime;	// time of last modification of GBMO image set
DWORD	ClrHeadDark, ClrHeadLight;	// colors of top/bottom column header 'frame'
BOOL	DrawNumFirst;	// if image set wants NUMBERS drawn first on tile before CROC image

DWORD	hGBM;		// handle of GAME graphics SIZED
long	GU, GU15, GU2, GU25, GU6, GU8, GU9, GU10;	// various partial tile-widths (total tile is GU10xGU10, edge is GU, face is GU9xGU9)
DWORD	hPuppyBM;	// handle of special win display graphics

long	ScreenW, ScreenH;			// pixel w/h of current display
long	MaxScreenW, MaxScreenH;		// pixel w/h of virtual display (all monitors together)
long	ViewW, ViewH, ViewMR;		// pixel w/h of current window's client space

char	ScratBuf[8192];		// for general consumption

WNDPOS	wndPosNormal={0,0,0,0,2};	// a predefined WNDPOS for restoring the window from minimized
BOOL	MinimizePause;				// if game is PAUSED because of window minimize

BOOL	NeedDrawAll;

DWORD	NumPlayers, maxID;
PLAYER	CurPlayer;
MOOD	CurMood;

DWORD	LastSoundTime;	// time last .WAV file was played, to keep them from walking all over each other

DWORD	timeExpireNotify, timeNotify, NotifyBackClr, NotifyTextClr;
DWORD	hNotify;

BOOL	ShuttingDown=FALSE;	// TRUE when hMain is being destroyed
BOOL	NeedNewGame;		// if "go to next puzzle" is pending because of PromptNewGame

long	ScriptState;		// 0 = nothing to do (see GameTimer() in game.c if any scripts are supported)

BOOL	IsMainUp=FALSE;		// set at end of OnEndStartup()

//=======================================================================
void OnBeginStartup() {
	
	QWORD	ss;
	WORD	*pWP;
	long	i;
	
	CPUreserved = (QWORD)0;
	
	Decode(str_mod_help0);
	Decode(str_mod_help1);
	Decode(str_mod_help2);
	Decode(str_mod_help3);
	Decode(str_mod_help4);
	Decode(str_mod_help5);
	Decode(LicStr);
	Decode(DemoStr);
	Decode(AllRights);
	
	ss = GetScreenSize();
	ScreenW = (long)(ss & 0x0FFFFFFFF);
	ScreenH = (long)((ss>>32) & 0x0FFFFFFFF);
	
	// return desired X/Y/W/H of starting window in KINTstartWndPos
	pWP = (WORD*)&KINTstartWndPos;

	*pWP++ = ScreenW/20;
	*pWP++ = ScreenH/20;
	*pWP++ = 9*ScreenW/10;
	*pWP++ = 9*ScreenH/10;
}

//=======================================================================
void OnEndStartup() {

	long	i, j;
	DWORD	f, w, h;
	INFOBM	ibm;
	QWORD	qw;
	WND		*pW;

	ScriptState = 0;
	
	qw = GetViewSize();
	ViewW = (long)(qw & 0x0FFFFFFFF);
	ViewH = (long)((qw>>32) & 0x0FFFFFFFF);
	
	LastSoundTime = TimeMilli()-1000;	// make sure first .WAV file will play okay
	SetPgmTitle(PgmTitle);
	
	hSplashBM = LoadBM("about.png");

readoldcfg:
	if( !iniRead(iniFILENAME) ) {// if error or non-existent
startover:
		iniDelAll();			// make sure it's all emptied (probably not needed, but safer than sorrier)
		iniSaveHeader();		// create game ini header
		LoadStrings(defStrings);// DefaultCurPlayer() needs default player name string

		NumPlayers = 1;			// the number of PLAYERs created
		maxID = 0;				// the highest valid PLAYER number
		if( ReadCT1CFG() ) goto readoldcfg;

		DefaultCurPlayer();		// fill in default values (this sets us up with a valid UI file)
		CurPlayer.id = 0;		// set the Player's number
		iniSavePlayer();

		StartPlayer(FALSE);
		
		hMain = CreateWnd(ID_MAIN, WT_WND, BARWPX, MSGHPX, ViewW-BARWPX, ViewH-MSGHPX, CF_TABSTOP, 0, mainProc);

		SetupDispVars();
		LoadGraphics(CurMood.images);

		ShowWnd(hMain, SW_SHOW);

		hMsgBar = CreateDialog(&MsgBarDlg[0], hMain);
		hButtBar = CreateDialog(&ButtBarDlg[0], hMain);
		
		hPuppyBM = 	LoadBM("ekssie");
		f = GetPixelBM(hPuppyBM, 0, 0);
		SetAlphaBM(hPuppyBM, 0, 0, 100, 100, f, 0);

		SetTimer(50);
		AddTimerWnd(hMain);

		mainDrawAll();

		SetFocus(hMain);
		
		DoModal(TitleDlg, hMain);
		// Bring up the menu to the RENAME PLAYER menu in 'special' mode that REQUIRES a name acceptance

			// disable ALL of the menu items EXCEPT for PLAYER RENAME, HELP, and EXIT
			for(i=1; MenuDlg[i].id; i++) if( MenuDlg[i].id!=ID_MENU_PLYR_REN && MenuDlg[i].id!=ID_MENU_HELP && MenuDlg[i].id!=ID_MENU_EXIT ) MenuDlg[i].flags |= CF_DISABLED;
		
			// set current menu to RENAME PLAYER and activate the menu
			MenuCurSelID = ID_MENU_PLYR_REN;
			// NOTE: execution will not return from the following call until the MENU dialog closes!
			SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
			// re-enable the menu items
			for(i=1; MenuDlg[i].id; i++) MenuDlg[i].flags &= ~CF_DISABLED;
		
		// The menu dialog has closed, so set OPTIONS as the default menu selection for the next entry
		MenuCurSelID = ID_MENU_OPTIONS;
	} else {
		if( !iniFindSec(PgmFolder) ) SystemMessage("OES:iFS");
		if( !iniFindVar(iniREV) ) goto startover;
		iniGetNum(&f);
		if( f<INIREV ) goto startover;
		if( iniFindVar(iniCURPLAYER) && iniGetNum(&f) ) CurPlayer.id = f;
		if( iniFindVar(iniNUMPLAYERS) && iniGetNum(&f) ) NumPlayers = f;
		if( iniFindVar(iniMAXID) && iniGetNum(&f) ) maxID = f;

		j = -1;
		sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
		if( !iniFindSec(ScratBuf) ) {	// didn't find that player, try to find another
			for(i=0; i<=maxID; i++) {
				sprintf(ScratBuf, iniPLAYER, i);
				if( iniFindSec(ScratBuf) ) {	// found one
					CurPlayer.id = i;
					j = i;
					break;
				}
			}
		} else j = CurPlayer.id;

		// Either have desired player, first found player, or no player (j==-1)
		if( j<0 ) goto startover;	// didn't find a previous player cfg, something happened to the .ini

		StartPlayer(FALSE);

		hMain = CreateWnd(ID_MAIN, WT_WND, BARWPX, MSGHPX, ViewW-BARWPX, ViewH-MSGHPX, 0, 0, mainProc);

		SetupDispVars();
		LoadGraphics(CurMood.images);

		ShowWnd(hMain, SW_SHOW);

		hMsgBar = CreateDialog(&MsgBarDlg[0], hMain);
		hButtBar = CreateDialog(&ButtBarDlg[0], hMain);

// Load special win tile graphics
		hPuppyBM = 	LoadBM("ekssie");
		f = GetPixelBM(hPuppyBM, 0, 0);
		SetAlphaBM(hPuppyBM, 0, 0, 100, 100, f, 0);
		
		SetTimer(50);
		AddTimerWnd(hMain);
		
		// if not running a script AND (not bypassing the title screen OR strings are bad), show TITLE screen
		if( *((BYTE*)0xC2)==0 && (!(CurMood.flags & PF_BYPASS_TITLE) || (TitleDlg[1].flags & WF_SHOW)) ) {
			SendMsg(hMain, WM_DRAWBACKGROUND, 0, 0, 0);
			DoModal(TitleDlg, hMain);
		}
	}
	if( hGBM==0 ) {
		SetupDispVars();
		LoadGraphics(CurMood.images);
	}
	hPauseBM = LoadBM("pause.png");
	GetInfoBM(hPauseBM, &ibm);
	PauseW = ibm.w;
	PauseH = ibm.h/8;
	PauseI = 0;
	SetAlphaBM(hPauseBM, 0, 0, ibm.w, ibm.h, GetPixelBM(hPauseBM, 0, 0), 0);
	
	StartPuzzleType();	// setup for CurPlayer.CurButtNum and StartPuzzle(CurPlayer.CurPuzzNum)

	SetFocus(hMain);
	
	IsMainUp = TRUE;
}

//=======================================================================
// Initialize CurMood to default conditions
void DefaultCurMood() {
	
	FILELIST	*pM;
	DWORD		hM, hMnext;

	// free up any previous MUSIC blocks
	for(hM=CurMood.hBackMusic; hM; hM=hMnext) {
		pM = (FILELIST*)HeapPtr(hM);
		hMnext = pM->hNext;
		HeapFree(hM);
	}
	for(hM=CurMood.hWinMusic; hM; hM=hMnext) {
		pM = (FILELIST*)HeapPtr(hM);
		hMnext = pM->hNext;
		HeapFree(hM);
	}

	memset((BYTE*)&CurMood, 0, sizeof(MOOD));
	strcpy(CurMood.name, KString(BLK_DLG, STR_DEFMOODNAME));
	strcpy(CurMood.UIimage, defuiname);
	strcpy(CurMood.images, defimgname);
	strcpy(CurMood.HelpName, defHelp);
	strcpy(CurMood.StringsName, defStrings);
	memcpy(CurMood.Actions, defactions, sizeof(CurMood.Actions));

	GetWndPos(&CurMood.wndPos);
	CurMood.wndPos.x = ScreenW/20;
	CurMood.wndPos.y = ScreenH/20;
	CurMood.wndPos.w = 9*ScreenW/10;
	CurMood.wndPos.h = 9*ScreenH/10;

	CurMood.flags = defflags | defsndflags;
	CurMood.flags2 = 0;

	CurMood.WinDisp = WIN_RANDOM;
	CurMood.WinDelay = 30;
	CurMood.WinSpeed = 1;
	
	CurMood.LimitPuzzSize = 90;
	CurMood.CSVsep = (DWORD)',';
	
	// allocate default MUSIC blocks
	CurMood.hBackMusic = HeapAlloc(sizeof(FILELIST));
	pM = (FILELIST*)HeapPtr(CurMood.hBackMusic);
	strcpy(pM->name, defbackmusic);
	pM->hNext = pM->flags = 0;
	CurMood.BackMusicCnt = 1;

	CurMood.hWinMusic = HeapAlloc(sizeof(FILELIST));
	pM = (FILELIST*)HeapPtr(CurMood.hWinMusic);
	strcpy(pM->name, defwinmusic);
	pM->hNext = pM->flags = 0;
	CurMood.WinMusicCnt = 1;
}

//=======================================================================
// Initialize CurPlayer to default conditions
void DefaultCurPlayer() {
	
	MOOD	*pM;
	long	i;

	// initialize everything to defaults
	memset((BYTE*)&CurPlayer, 0, sizeof(CurPlayer));	// zero the current PLAYER struct
	
	DefaultCurMood();	// init the current MOOD struct
	
	CurPlayer.hFirstMood = HeapAlloc(sizeof(MOOD));	// allocate a MOOD struct for the PLAYER
	memcpy((BYTE*)HeapPtr(CurPlayer.hFirstMood), (BYTE*)&CurMood, sizeof(MOOD));	// copy the default current MOOD into the PLAYER's first MOOD
	CurPlayer.MoodCnt = 1;
	CurPlayer.CurMoodNum = 0;
	strcpy(CurPlayer.name, KString(BLK_DLG, STR_SECRET_ID));
	
	CurPlayer.HelpExportFlags = 0;
	CurPlayer.HelpExportColors[HEC_BACK]	 = 0xFFED2E1E;
	CurPlayer.HelpExportColors[HEC_TEXT]	 = 0xFF000000;
	CurPlayer.HelpExportColors[HEC_H1]		 = 0xFFFF0000;
	CurPlayer.HelpExportColors[HEC_H2]		 = 0xFF000000;
	CurPlayer.HelpExportColors[HEC_H3]		 = 0xFF000000;
	CurPlayer.HelpExportColors[HEC_CONTBACK] = 0xFFE31192;
	CurPlayer.HelpExportColors[HEC_CONTTEXT] = 0xFF000000;
	CurPlayer.HelpExportColors[HEC_CONTLINK] = 0xFF97FF54;
	CurPlayer.HelpExportColors[HEC_HEADBACK] = 0xFF47A6A8;
	CurPlayer.HelpExportColors[HEC_HEADTEXT] = 0xFFCA0000;
	CurPlayer.HelpExportColors[HEC_HEADFRAME]= 0xFF000000;

	CurPlayer.HelpExportFonts[HEFONT_TEXT] = 1;
	CurPlayer.HelpExportFonts[HEFONT_H1] = 5;
	CurPlayer.HelpExportFonts[HEFONT_H2] = 6;
	CurPlayer.HelpExportFonts[HEFONT_H3] = 7;
	
	CurPlayer.HelpExportWidth = 40;
	CurPlayer.HelpExportLmargin = 2;
	CurPlayer.HelpExportContentWidth = 12;
	CurPlayer.HelpExportContentSpacing = 5;
	
	CurPlayer.HelpExportRule1Top = 10;
	CurPlayer.HelpExportRule1Bot = 5;
	CurPlayer.HelpExportRule2Top = 6;
	CurPlayer.HelpExportRule2Bot = 10;
	
	// it's up to the programmer to set the id and name of the player upon return
}

//=======================================================================
void UpdateTitleBar() {

	sprintf(ScratBuf, "%s - %s [%s]", PgmTitle, CurPlayer.name, hMenuDlg?tmpMood.name:CurMood.name);
	SetPgmTitle(ScratBuf);
}

//=======================================================================
void BeepCool() {

	if( TimeMilli() - LastSoundTime < 200 ) return;
	if( CurMood.flags & PF_GAMESOUNDS ) 
		PlaySound("COOL.WAV");
	LastSoundTime = TimeMilli();
}

//=======================================================================
void BeepIll() {

	if( TimeMilli() - LastSoundTime < 200 ) return;
	if( CurMood.flags & PF_ERRORSOUNDS ) 
		PlaySound("ILLCLICK.WAV");
	LastSoundTime = TimeMilli();
}

//=======================================================================
void BeepMistake() {

	if( CurMood.flags & PF_ERRORSOUNDS ) 
		PlaySound("MISTAKE.WAV");
	LastSoundTime = TimeMilli();
}

//=======================================================================
DWORD mainProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	WND		*pW, *pX;
	RECT	invrect, wndrect, rc;
	long	i, x, y;
	DWORD	dw, f;
	QWORD	qw;

	pW = (WND*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
	// NOTE!!! hW IS 'hMain' at this point, BUT!!! 'hMain' has not yet been set!
	// It doesn't get set until CreateWnd() returns to OnEndStartup()!

		// Do any default window creation stuff
		DefWndProc(hW, msg, arg1, arg2, arg3);

		// === Now do our specific setup stuff ===

		// Create bitmap for main
		qw = GetMaxScreen();
		MaxScreenW = (long)(qw & 0x0FFFFFFFF);
		MaxScreenH = (long)((qw>>32) & 0x0FFFFFFFF);

		hMainBM = CreateBM(MaxScreenW, MaxScreenH);

		// save the BM handle in our WND struct
		pW = (WND*)HeapPtr(hW);
		pW->hBM = hMainBM;

		// default set the focus to us
		SetFocus(hW);

		return 0;
	//**********************
	// if ARG1 != 0 then the machine is shutting down and we need to SAVE and save NOW!
	 case WM_QUERY_QUIT:
		if( MinimizePause ) {
			SetWndPos(&wndPosNormal);	// restore
		}

		if( NeedNewGame ) {
			EndPuzzle();
			if( (CurPlayer.CurPuzzNum+1) < MAXPUZZLENUM ) {
				StartPuzzle(CurPlayer.CurPuzzNum+1);
				NeedNewGame = FALSE;
			}
		}

		if( hMenuDlg ) {
			SaveMoodChanges();
		}
		SendMsg(hW, WM_DESTROY, 0, 0, 0);
		_asm(term);	// quit

		return 0;
	 case WM_DESTROY:
		iniSaveHeader();
		EndPlayer();
		iniWrite(iniFILENAME);
		iniClose();
		// free whatever resources you need to
		hModal = 0;		// we're dying, prevent DefDlgProc from seeing any existing modal dialog (like MENU) and returning THERE instead of dying
		
		SetTimer(0);	// stop timer

		ShuttingDown = TRUE;
		
		DefWndProc(hW, msg, arg1, arg2, arg3);

// or	ScrollWndProc(pW, msg, arg1, arg2, arg3);

// NOTE: the Window structure is GONE at this point!!!
// anything that it's needed for needs to be done
// before calling DefWndProc() !!!!

		hMain = 0;
		hMainBM = 0;
		hMsgBar = 0;

		if( hUIBM ) DeleteBM(hUIBM);
		hUIBM = 0;
		
		if( hSplashBM ) DeleteBM(hSplashBM);
		hSplashBM = 0;
		
		if( hPauseBM ) DeleteBM(hPauseBM);
		hPauseBM = 0;

		// delete any "old" sized game graphics files
		{
			DWORD	dietime, filetime;
			long	isdir, isdata;
			
			// remove any that are more than X days older than the CFG file (that was just written)
			dietime = ftime(iniFILENAME) - 14*86400;
			if( -1!=FindFirst("SIZED*", ScratBuf, &isdir, &isdata) ) {
				do {
					filetime = ftime(ScratBuf);
					if( !(isdir & A_DIR) & isdata && filetime < dietime ) fremove(ScratBuf);
				} while( -1 != FindNext(ScratBuf, &isdir, &isdata) );
				FindClose();
			}
		}
		return 0;

	 case WM_SETFOCUS:
	 case WM_KILLFOCUS:
		return 0;		// don't call DefWndProc() so it doesn't force an Invalidate/redraw on us
		
	 case WM_PAINT:		// more complex than usually necessary in order to support panning/scrolling
		if( GetRectInvalid(&invrect) ) {
			GetRectWnd(hW, &wndrect);
			if( IntersectRect(&invrect, &wndrect) ) {
				ClipRectBM(0, invrect, &rc);

				DrawBM(0, invrect.x, invrect.y, pW->hBM, invrect.x-pW->xoff, invrect.y-pW->yoff, invrect.w, invrect.h, 0);

				ClipRectBM(0, rc, NULL);
			}
		}
		NotifyChildren(hW, WM_PAINT, 0, 0, 0);
		return 0;
	 case WM_DRAWBACKGROUND:
		RepDrawBM(hMainBM, 0, 0, hUIBMO, X_WNDBACK*UIBMUO, Y_WNDBACK*UIBMUO, W_WNDBACK*UIBMUO, H_WNDBACK*UIBMUO, 0, ViewW-BARWPX, ViewH-MSGHPX);
		return 0;
	 case WM_HOVER:
		return 0;
	 case WM_CHAR:
		return 0;
	 case WM_KEYUP:
		if( GameWon ) {
			TermWinDisplay();
			return 0;
		}
		if( IgnoreKeys ) return 0;
		if( arg1==K_SHIFT || arg1==K_CTL ) {

			break;
		}
		// in case focus left window between down and up strokes, and we've now returned to the window
		if( LastDownKey!=0xFFFFFFFF && LastDownKey!=arg1 ) {
			LastDownKey = 0xFFFFFFFF;
			break;
		}
		dw = LastDownKey==0xFFFFFFFF ? KeyFlags : LastDownKeyFlags;

		for(i=0; i<CurMood.KeyMapsCnt; i++) {
			if( LastDownKey==(DWORD)CurMood.KeyMaps[i].key && (dw & (MK_SHIFT | MK_CTL))==(DWORD)CurMood.KeyMaps[i].flags ) {
				f = CurMood.KeyMaps[i].func;
				switch( f ) {
				 case KF_KEYMAP_DLG:
					MenuCurSelID = ID_MENU_KEYS;
					SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_PREV_PUZZ:	// did on the DOWN (for repeat)
					break;
				 case KF_NEXT_PUZZ:	// did on the DOWN (for repeat)
					break;
				 case KF_OPTIONS_DLG:
					MenuCurSelID = ID_MENU_OPTIONS;
					SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_MOODS_DLG:
					if( CurPlayer.MoodCnt>1 ) {
						MenuCurSelID = ID_MENU_MOOD_SEL;
						SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
					}
					break;
				 case KF_NEXT_MOOD:
					NextMood();
					break;
				 case KF_PREV_MOOD:
					PrevMood();
					break;
				 case KF_PLAYER_DLG:
					MenuCurSelID = ID_MENU_PLYR_SEL;
					SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_NEXT_PLAYER:
					NextPlayer();
					break;
				 case KF_PREV_PLAYER:
					PrevPlayer();
					break;
				 case KF_PAUSE:
					if( !Paused ) {
						if( !(CurMood.flags & PF_REMOVE_PAUSE) ) {
							if( TestBest ) CheckBests();
							else if( !hMenuDlg ) SetPause(TRUE);	// gets turned off by ANY key in KMSG_KEY_DOWN
						}
					}
					break;
				 case KF_MENU_DLG:
					SendMsgID(ID_MENU_BUTT, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_NEXT_UNSOLVED:
					NextUnsolvedPuzzle();
					break;
				 case KF_FIRST_UNSOLVED:
					FirstUnsolvedPuzzle();
					break;
				 case KF_UNDO:
					break;
				 case KF_REDO:
					break;
				 case KF_RESTART:
					SendMsgID(ID_RESTART, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_SAVEGAME:
					SendMsgID(ID_SAVEGAME, WM_ACTIVATE, 0, 0, 0);
					break;
				 default:
					SystemMessage("Unknown key mapping");
					break;
				}
			}
		}
		LastDownKey = 0xFFFFFFFF;
		return 0;
	 case WM_KEYDN:
		if( GameWon ) {
			TermWinDisplay();
			return 0;
		}
		if( IgnoreKeys ) return 0;
		if( arg1==K_SHIFT || arg1==K_CTL ) {

			break;
		}

		// if FIRST downkey of a sequence, save KeyFlags
		if( LastDownKey==0xFFFFFFFF ) LastDownKeyFlags = KeyFlags;
		// save most recent downkey
		LastDownKey = arg1;
		for(i=0; i<CurMood.KeyMapsCnt; i++) {
			if( arg1==(DWORD)CurMood.KeyMaps[i].key && (KeyFlags & (MK_SHIFT | MK_CTL))==(DWORD)CurMood.KeyMaps[i].flags ) {
				f = CurMood.KeyMaps[i].func;
				switch( f ) {
				 case KF_KEYMAP_DLG:
					break;
				 case KF_PREV_PUZZ:
					if( Paused ) {
						if( !hMenuDlg ) SetPause(0);
					} else SendMsgID(ID_PREVPUZZ, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_NEXT_PUZZ:
					if( Paused ) {
						if( !hMenuDlg ) SetPause(0);
					} else SendMsgID(ID_NEXTPUZZ, WM_ACTIVATE, 0, 0, 0);
					break;
				 case KF_OPTIONS_DLG:
					return 0;
				 case KF_MOODS_DLG:
					return 0;
				 case KF_NEXT_MOOD:
					return 0;
				 case KF_PREV_MOOD:
					return 0;
				 case KF_PLAYER_DLG:
					return 0;
				 case KF_NEXT_PLAYER:
					return 0;
				 case KF_PREV_PLAYER:
					return 0;
				 case KF_PAUSE:
					return 0;
				 case KF_MENU_DLG:
					return 0;
				 case KF_NEXT_UNSOLVED:
					return 0;
				 case KF_FIRST_UNSOLVED:
					return 0;
				 case KF_UNDO:
					if( Paused ) {
						if( !hMenuDlg ) SetPause(0);
					} else SendMsgID(ID_UNDO, WM_ACTIVATE, 0, 0, 0);
					return 0;
				 case KF_REDO:
					if( Paused ) {
						if( !hMenuDlg ) SetPause(0);
					} else SendMsgID(ID_REDO, WM_ACTIVATE, 0, 0, 0);
					return 0;
				 case KF_RESTART:
					return 0;
				 case KF_SAVEGAME:
					return 0;
				 default:
					SystemMessage("Unknown key mapping");
					break;
				}
			}	
		}
		return 0;
	 case WM_SCREEN_CHANGE:
	 case WM_PARENTSIZED:
		if( !ShuttingDown && (pW->w != ViewW || pW->h != ViewH) ) {
			// MainWndProc must NOT pass WM_PARENTSIZED on to DefWndProc, as handling
			// is different for MainWnd than for all other windows.

			qw = GetScreenSize();
			ScreenW = (long)(qw & 0x0FFFFFFFF);
			ScreenH = (long)((qw>>32) & 0x0FFFFFFFF);

			SetupUUvars();
			
			pW->xoff = BARWPX;
			pW->yoff = MSGHPX;
			pW->w = ViewW - BARWPX;
			pW->h = ViewH - MSGHPX;


			if( pW->uu != UU ) {
				pW->uu = UU;

				SizeUI();
			}
			// if the MENU is open, then the important settings are in tmpMood, not CurMood.
			if( hMenuDlg ) {
				GU = UU;
				InitGUvars();
				GetWndPos(&tmpMood.wndPos);
				LoadGraphics(tmpMood.images);
			} else {
				SetupDispVars();
				GetWndPos(&CurMood.wndPos);
				UpdateMood();
				LoadGraphics(CurMood.images);
			}

			mainDrawAll();
			NotifyChildren(hW, msg, arg1, arg2, arg3);
			if( hMenuDlg && MenuCurSelID==ID_MENU_ABOUT ) {
				sprintf(AboutWindow, KString(BLK_DLG, STR_WINDOW), ViewW,ViewH);
				InvalidateAllWnd(FindWndID(ID_ABOUT_WINDOW));
			}
		}
		return 0;
	 case WM_NOTIFY:
		switch( arg1 ) {
		 case WN_MODIFIED:
			GetWndPos(hMenuDlg ? &tmpMood.wndPos : &CurMood.wndPos);
			UpdateMood();
			break;
		}
		break;
	 case WM_TIMER:
		GameTimer();
		return 0;
	 case WM_WHEEL:
		GameWheel(arg1,arg2,arg3);
		return 0;
	 case WM_HWHEEL:
		GameHWheel(arg1,arg2,arg3);
		return 0;
	 case WM_LBUTTDN:
		if( hFocus != hW ) SetFocus(hW);
		return MouseMap(hW, arg1, arg2, arg3, 0, CLK_START);
	 case WM_LBUTTUP:
		if( IgnoreMouseUp ) {
			IgnoreMouseUp = FALSE;
			return 0;
		}
		return MouseMap(hW, arg1, arg2, arg3, 0, CLK_END);
	 case WM_MBUTTDN:
		if( hFocus!=hW ) SetFocus(hW);
		return MouseMap(hW, arg1, arg2, arg3, 1, CLK_START);
	 case WM_MBUTTUP:
		if( IgnoreMouseUp ) {
			IgnoreMouseUp = FALSE;
			return 0;
		}
		return MouseMap(hW, arg1, arg2, arg3, 1, CLK_END);
	 case WM_RBUTTDN:
		if( hFocus!=hW ) SetFocus(hW);
		return MouseMap(hW, arg1, arg2, arg3, 2, CLK_START);
	 case WM_RBUTTUP:
		if( IgnoreMouseUp ) {
			IgnoreMouseUp = FALSE;
			return 0;
		}
		return MouseMap(hW, arg1, arg2, arg3, 2, CLK_END);
	 case WM_BBUTTDN:
		if( hFocus!=hW ) SetFocus(hW);
		return MouseMap(hW, arg1, arg2, arg3, 3, CLK_START);
	 case WM_BBUTTUP:
		if( IgnoreMouseUp ) {
			IgnoreMouseUp = FALSE;
			return 0;
		}
		return MouseMap(hW, arg1, arg2, arg3, 3, CLK_END);
	 case WM_FBUTTDN:
		if( hFocus!=hW ) SetFocus(hW);
		return MouseMap(hW, arg1, arg2, arg3, 4, CLK_START);
	 case WM_FBUTTUP:
		if( IgnoreMouseUp ) {
			IgnoreMouseUp = FALSE;
			return 0;
		}
		return MouseMap(hW, arg1, arg2, arg3, 4, CLK_END);
	 case WM_PTRMOVE:
		SetCursor(CURSOR_NORMAL);
		PointerMove(arg1, arg2, arg3);
		return 0;
	 case WM_MUSIC_DONE:
		if( GameWon ) TermWinDisplay();
		else StopMusic();
		return 0;
	 default:
		break;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
// Triggered via PAUSE key after HelpTitleWindow 6 RIGHT-clicks, 1 LEFT-click.
void CheckBests() {
	
	BYTE	*pBST = ((BYTE*)HeapPtr(hBest))+0x0100;
	long	pn=0, i;
	long	m, dr, dc, dtb, sr, sc, stb, dcnt;
	BYTE	*pD, *pS, *pH;
	long	oldlen, delta, tb, ib;

	Retest = TRUE;
	EndPuzzle();
	while( pn<MAXPUZZLENUM-1 ) {
		BestMoveCnt = *((WORD*)pBST);
		while( !BestMoveCnt && pn+1<MAXPUZZLENUM ) {
			++pn;
			pBST += 2+2*(*(WORD*)pBST);
			BestMoveCnt = *((WORD*)pBST);
		}
		StartPuzzle(pn);

		if( BestMoveCnt ) {
			while( Game.CurDo < Game.DoCnt ) {
				m = Game.Undos[Game.CurDo].move;

				stb = (m&0x0080)?2:1;
				sc = m&0x0F;
				if( stb==1 ) for(sr=1; sr<Game.PuzzH-1 && Game.board[sr][sc]==255; ++sr) ;
				else for(sr=Game.PuzzH-2; sr && Game.board[sr][sc]==255; --sr) ;

				dtb = (m&0x8000)?2:1;
				dc = (m>>8)&0x0F;
				if( dtb==1 ) {
					cdy = BrdY+GU9;
					for(dr=1; dr<Game.PuzzH-1 && Game.board[dr][dc]==255; ++dr) ;
				} else {
					cdy = BrdY + (Game.PuzzH-2)*GU9;
					for(dr=Game.PuzzH-2; dr && Game.board[dr][dc]==255; --dr) ;
				}
				dcnt = Game.DoCnt;
				if( !GetTargetDrop(sr, sc, stb, &dr, &dc) ) {
					sprintf(ScratBuf, "Failed on puzzle# %d", pn+1);
					SystemMessage(ScratBuf);
					
					
					oldlen = (long)(DWORD)*(WORD*)pBST;
					delta = 2*oldlen;

					// close up the entry, making it shorter
					pH = (BYTE*)HeapPtr(hBest);
					pD = pBST;
					pS = pD + delta;

					memmove(pD, pS, BestEnd-(pS-pH));
					BestEnd -= delta;
					*(WORD*)pD = 0;
					// adjust all block ptrs above this one
					tb = pn/1024;
					for(ib=tb+1; ib<64; ib++) *(DWORD*)(pH+4*ib) -= delta;

					SaveBest = TRUE;
					BestMoveCnt = dcnt = Game.CurDo = 0;
				}
				Game.DoCnt = dcnt;
				CheckOKs();
			}
		}
		EndPuzzle();
		if( pn+1>=MAXPUZZLENUM ) break;
		++pn;
		pBST += 2+2*(*(WORD*)pBST);
		BestMoveCnt = *((WORD*)pBST);
		UpdateInvalid();
	}
	Retest = FALSE;
	StartPuzzle(0);
	PlaySound("EXP1.WAV");
	SystemMessage("TestBest done!");
}

//===========================================================================
// NOTE: This must NOT depend upon the hMain existing!
// However, it CAN depend upon the system window existing
// and ViewW/ViewH being set.
void SetupUUvars() {
	
	MOOD	*pM;

	pM = (hMenuDlg ? &tmpMood : &CurMood);
	// 200 = max UU width of window
	// 109 = max UU height of window
	if( pM->UU ) UU = pM->UU+3;
	else UU = min(ViewW/200, ViewH/109);

	if( UU<4 ) UU = 4;
	
	// All bitmapped buttons in hUIBM are GBMU4 in the narrowest width
	UU2 = 2*UU;
	UU3 = 3*UU;
	UU4 = 4*UU;	// set primary UI control 'size' (height and/or width)
	UU5 = 5*UU;
	UU10=10*UU;

	TEXTYSPACE = UU4/15;
	if( TEXTYSPACE<2 ) TEXTYSPACE = 2;
	TEXTH = (UU4-2*TEXTYSPACE);

	HTSPACE = UU4/15;
	HTH = (UU4-2*TEXTYSPACE);
}

//===========================================================================
void mainDrawAll() {

	SendMsg(hMain, WM_DRAWBACKGROUND, 0, 0, 0);
	DrawBoard();
	InvalidateAllWnd(hMain);
	NeedDrawAll = FALSE;
}

DWORD	notifyBackClr, notifyTextClr;

//===========================================================================
DWORD NotifyProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND		*pW;
	RECT	r;
	DWORD	*pClrs;

	switch( msg ) {
	 case WM_CREATE:
		hNotify = hW;
		AddTimerWnd(hW);
		CaptureMouse(hW);
		break;
	 case WM_DESTROY:
		hNotify = 0;
		RemoveTimerWnd(hW);
		ReleaseMouse();
		break;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		pClrs[2] = notifyBackClr;
		return 0;
	 case WM_TIMER:
		if( timeExpireNotify ) {
			if( timeExpireNotify+timeNotify <= TimeMilli() ) {
				SendMsg(hW, WM_DESTROY, ID_CANCEL, 0, 0);
			}
		}
		return 0;
	 case WM_CHAR:
		if( arg1==27 ) return SendMsg(hW, WM_DESTROY, ID_CANCEL, 0, 0);
	 case WM_RBUTTDN:
	 case WM_MBUTTDN:
	 case WM_BBUTTDN:
	 case WM_FBUTTDN:
	 case WM_KEYDN:
	 case WM_LBUTTDN:
		return 0;
	 case WM_LBUTTUP:
	 case WM_RBUTTUP:
	 case WM_MBUTTUP:
	 case WM_BBUTTUP:
	 case WM_FBUTTUP:
		if( IgnoreMouseUp ) IgnoreMouseUp = FALSE;
		else return SendMsg(hW, WM_DESTROY, ID_CANCEL, 0, 0);
		return 0;
	}
	return DefDlgProc(hW, msg, arg1, arg2, arg3);
}

char	NotifyBuf[256];

//=======================================================================
DLGRES NotifyDlg[] = {
	{ID_NOTIFY_DLG,	WT_WND,		 -1,-1,100,10, WF_SHOW, NotifyProc, 0, 0, 0},
	{ID_NOTIFY_MSG,	WT_TEXT,	  1, 3, 98, 4, WF_SHOW+CF_CENTER+CF_STR, TextProc, NotifyBuf, 0, 0},
	
	{0}
};

//===========================================================================
void Notify(char *str, DWORD ExpireTime, DWORD bclr, DWORD tclr) {

	if( hNotify ) SendMsg(hNotify, WM_DESTROY, 0, 0, 0);
	
	notifyBackClr = bclr;
	notifyTextClr = tclr;
	
	timeExpireNotify = ExpireTime;
	timeNotify = TimeMilli();

	if( strlen(str)<sizeof(NotifyBuf)-1 ) strcpy(NotifyBuf, str);
	else SystemMessage("PO:notify");

	CreateDialog(NotifyDlg, hMain);
}

//===========================================================================
void Size6x6(long ux, long uy) {
	
	// top of 6x6 UI object
	StretchBM(hUIBM, ux*UU, uy*UU, 6*UU, UU, hUIBMO, ux*UIBMUO, uy*UIBMUO, 6*UIBMUO, UIBMUO);
	// left side
	StretchBM(hUIBM, ux*UU, (uy+1)*UU, UU, 4*UU, hUIBMO, ux*UIBMUO, (uy+1)*UIBMUO, UIBMUO, 4*UIBMUO);
	// right side
	StretchBM(hUIBM, (ux+5)*UU, (uy+1)*UU, UU, 4*UU, hUIBMO, (ux+5)*UIBMUO, (uy+1)*UIBMUO, UIBMUO, 4*UIBMUO);
	// bottom
	StretchBM(hUIBM, ux*UU, (uy+5)*UU, 6*UU, UU, hUIBMO, ux*UIBMUO, (uy+5)*UIBMUO, 6*UIBMUO, UIBMUO);
	// middle
	StretchBM(hUIBM, (ux+1)*UU, (uy+1)*UU, 4*UU, 4*UU, hUIBMO, (ux+1)*UIBMUO, (uy+1)*UIBMUO, 4*UIBMUO, 4*UIBMUO);
}

//===========================================================================
void SizeHscroll(long ux, long uy) {
	
	// left
	StretchBM(hUIBM, ux*UU, uy*UU, 3*UU, 3*UU, hUIBMO, ux*UIBMUO, uy*UIBMUO, 3*UIBMUO, 3*UIBMUO);
	// middle
	StretchBM(hUIBM, (ux+3)*UU, uy*UU, 3*UU, 3*UU, hUIBMO, (ux+3)*UIBMUO, uy*UIBMUO, 3*UIBMUO, 3*UIBMUO);
	// right
	StretchBM(hUIBM, (ux+6)*UU, uy*UU, 3*UU, 3*UU, hUIBMO, (ux+6)*UIBMUO, uy*UIBMUO, 3*UIBMUO, 3*UIBMUO);
}

//===========================================================================
void SizeVscroll(long ux, long uy) {
	
	// top
	StretchBM(hUIBM, ux*UU, uy*UU, 3*UU, 3*UU, hUIBMO, ux*UIBMUO, uy*UIBMUO, 3*UIBMUO, 3*UIBMUO);
	// middle
	StretchBM(hUIBM, ux*UU, (uy+3)*UU, 3*UU, 3*UU, hUIBMO, ux*UIBMUO, (uy+3)*UIBMUO, 3*UIBMUO, 3*UIBMUO);
	// bottom
	StretchBM(hUIBM, ux*UU, (uy+6)*UU, 3*UU, 3*UU, hUIBMO, ux*UIBMUO, (uy+6)*UIBMUO, 3*UIBMUO, 3*UIBMUO);
}

//===========================================================================
// Create new hUIBM bitmap with UI 'unit' size of UU from hUIBMO with 'unit' size of UIBMUS
void SizeUI() {
	
	long	i, j, W, H, x, y, w, h;
	DWORD	clrrev, hFile;
	QWORD	t;
	
	// delete the old SIZED bitmap
	if( hUIBM ) DeleteBM(hUIBM);
	hUIBM = 0;

	// set SIZED var values
	UIBMU = UU;
	UIBMU2 = 2*UIBMU;
	UIBMU3 = 3*UIBMU;
	UIBMU4 = 4*UIBMU;
	UIBMU5 = 5*UIBMU;
	UIBMU8 = 8*UIBMU;
	UIBMU10=10*UIBMU;
	
	sprintf(ScratBuf+1024, sizedUIname, UU, UIBMOname);
	if( fsize(ScratBuf+1024) != -1 ) {
		t = ftime(ScratBuf+1024);
		if( t > UIBMtime ) {
			hUIBM = LoadBM(ScratBuf+1024);
			clrrev = GetPixelBM(hUIBM, 0, 1);
			if( clrrev==UISIZED_REV ) {
				// Update time/date stamp of 'sized' graphics file, so we don't delete it
				if( -1 != (hFile=fopen(ScratBuf+1024, O_WRBIN)) ) {
					if( 4==fread(hFile, &clrrev, 4) ) {
						fseek(hFile, SEEK_SET, 0);
						fwrite(hFile, &clrrev, 4);
					}
					fclose(hFile);
				}
				return;
			}
			DeleteBM(hUIBM);
			hUIBM = 0;
		}
	}

	// create the new SIZED bitmap
	W = UIIMG_UW*UU;
	H = UIIMG_UH*UU;
	hUIBM = CreateBM(W, H);
	
	// set entire SIZED bitmap to TRANSPARENT
	DrawRectBM(hUIBM, 0, 0, W, H, CLR_BLACK, CLR_NONE);
	SetAlphaBM(hUIBM, 0, 0, W, H, CLR_BLACK, 0);
	
	// record the revision of this sized graphics image
	SetPixelBM(hUIBM, 0, 1, UISIZED_REV);
	
	// write the COLOR pixels
	for(i=0; i<64; i++) SetPixelBM(hUIBM, i, 0, ((i<2)?CLR_BLACK:Colors[i]));
	
	// resize the MENU LINE graphics
	StretchBM(hUIBM, X_SEPLINE*UU, Y_SEPLINE*UU, 8*UU, 3*UU, hUIBMO, X_SEPLINE*UIBMUO, Y_SEPLINE*UIBMUO, 8*UIBMUO, 3*UIBMUO);
	
	// resize CHECKBOXES and RADIOBUTTONS
	for(y=0; y<2; y++) {
		for(x=0; x<6; x++) {
			StretchBM(hUIBM, (X_CHECKBOX+4*x)*UU, (Y_CHECKBOX+4*y)*UU, UIBMU4, UIBMU4, hUIBMO, (X_CHECKBOX+4*x)*UIBMUO, (Y_CHECKBOX+4*y)*UIBMUO, UIBMUO*4, UIBMUO*4);
		}
	}
	
	// resize 6x6 elements
	Size6x6(X_GROUPBOX, Y_GROUPBOX);
	Size6x6(X_DIALOG, Y_DIALOG);
	for(y=0; y<2; y++) {
		for(x=0; x<3; x++) {
			Size6x6(X_BUTTON+6*x, Y_BUTTON+6*y);
		}
	}
	Size6x6(X_INBOX, Y_INBOX);
	Size6x6(X_LISTBOX, Y_LISTBOX);
	
	// resize SCROLLBARS
	for(i=0; i<3; i++) {
		SizeVscroll(X_VSCROLLBAR+3*i, Y_VSCROLLBAR);
		SizeHscroll(X_HSCROLLBAR, Y_VSCROLLBAR+3*i);
		StretchBM(hUIBM, X_HSCROLLBOX*UU, (Y_HSCROLLBOX+3*i)*UU, 6*UU, 3*UU, hUIBMO, X_HSCROLLBOX*UIBMUO, (Y_HSCROLLBOX+3*i)*UIBMUO, 6*UIBMUO, 3*UIBMUO);
		StretchBM(hUIBM, (X_VSCROLLBOX+3*i)*UU, Y_VSCROLLBOX*UU, 3*UU, 6*UU, hUIBMO, (X_VSCROLLBOX+3*i)*UIBMUO, Y_VSCROLLBOX*UIBMUO, 3*UIBMUO, 6*UIBMUO);
	}
	
	// resize MENU buttons
	for(x=0; x<2; x++) {
		StretchBM(hUIBM, (X_MENUBUTT+12*x)*UU, Y_MENUBUTT*UU, UIBMU*12, UIBMU*5, hUIBMO, (X_MENUBUTT+12*x)*UIBMUO, Y_MENUBUTT*UIBMUO, UIBMUO*12, UIBMUO*5);
	}

	// resize COLOR PICKER buttons
	for(i=0; i<2; i++) {
		StretchBM(hUIBM, X_COPY*UU, (4*i+Y_COPY)*UU, 8*UU, 4*UU, hUIBMO, X_COPY*UIBMUO, (4*i+Y_COPY)*UIBMUO, 8*UIBMUO, 4*UIBMUO);
		StretchBM(hUIBM, X_PASTE*UU, (4*i+Y_PASTE)*UU, 8*UU, 4*UU, hUIBMO, X_PASTE*UIBMUO, (4*i+Y_PASTE)*UIBMUO, 8*UIBMUO, 4*UIBMUO);
	}
	
	//============= Game specific buttons
	
	// resize GAME BUTTONS
	for(j=0; j<12; j++) {
		for(i=0; i<4; i++) {
			StretchBM(hUIBM, (24+i*5)*UU, 5*j*UU, UIBMU5, UIBMU5, hUIBMO, (24+i*5)*UIBMUO, 5*j*UIBMUO, UIBMUO*5, UIBMUO*5);
		}
	}
	StretchBM(hUIBM, 34*UU, 30*UU, UIBMU10, UIBMU5, hUIBMO, 34*UIBMUO, 30*UIBMUO, UIBMUO*10, UIBMUO*5);
	StretchBM(hUIBM, 34*UU, 35*UU, UIBMU10, UIBMU5, hUIBMO, 34*UIBMUO, 35*UIBMUO, UIBMUO*10, UIBMUO*5);

	//=============== Toolbar background
	StretchBM(hUIBM, X_TOOLBACK*UU, Y_TOOLBACK*UU, W_TOOLBACK*UU, H_TOOLBACK*UU, hUIBMO, X_TOOLBACK*UIBMUO, Y_TOOLBACK*UIBMUO, W_TOOLBACK*UIBMUO, H_TOOLBACK*UIBMUO);
	
	//=============== Game background
	StretchBM(hUIBM, X_WNDBACK*UU, Y_WNDBACK*UU, W_WNDBACK*UU, H_WNDBACK*UU, hUIBMO, X_WNDBACK*UU, Y_WNDBACK*UIBMUO, W_WNDBACK*UIBMUO, H_WNDBACK*UIBMUO);
	
	sprintf(ScratBuf, sizedUIname, UU, UIBMOname);
	SaveBM(hUIBM, ScratBuf, SIZEDTYPE);
}

//===========================================================================
void SizeSB() {
	
	long	i, j, tu, tu4, tu6, xu, yu;
	QWORD	qw;
	
	qw = GetMaxScreen();
	MaxScreenW = (long)(qw & 0x0FFFFFFFF);
	MaxScreenH = (long)((qw>>32) & 0x0FFFFFFFF);
	numButtSize = min(MaxScreenW/200, MaxScreenH/106)-2;
	wButtSize = ((200-5)*4 - 31*(numButtSize+2))/(6*numButtSize);
	wButtSize = min(wButtSize, numButtSize+2);
	if( wButtSize==0 ) wButtSize = 1;
	wButtSize *= 6;

	tu = wButtSize/6;
	tu4=4*tu;
	tu6 = wButtSize;
	
	// if max-sized size buttons bitmap was created before, delete it
	if( hSBBM ) DeleteBM(hSBBM);
	// create the proper bitmap for the max-sized buttons
	hSBBM = CreateBM(3*wButtSize, 2*wButtSize);
	// now stretch the original text button graphics to the max-sized bitmap
	for(j=0; j<2; j++) {
		for(i=0; i<3; i++) {
			xu = X_BUTTON + 6*i;
			yu = Y_BUTTON + 6*j;
			// top of 6x6 UI object
			StretchBM(hSBBM, (6*i)*tu, (6*j)*tu, tu6, tu, hUIBMO, xu*UIBMUO, yu*UIBMUO, 6*UIBMUO, UIBMUO);
			// left side
			StretchBM(hSBBM, (6*i)*tu, (6*j+1)*tu, tu, tu4, hUIBMO, xu*UIBMUO, (yu+1)*UIBMUO, UIBMUO, 4*UIBMUO);
			// right side
			StretchBM(hSBBM, (6*i+5)*tu, (6*j+1)*tu, tu, tu4, hUIBMO, (xu+5)*UIBMUO, (yu+1)*UIBMUO, UIBMUO, 4*UIBMUO);
			// bottom
			StretchBM(hSBBM, (6*i)*tu, (6*j+5)*tu, tu6, tu, hUIBMO, xu*UIBMUO, (yu+5)*UIBMUO, 6*UIBMUO, UIBMUO);
			// middle
			StretchBM(hSBBM, (6*i+1)*tu, (6*j+1)*tu, tu4, tu4, hUIBMO, (xu+1)*UIBMUO, (yu+1)*UIBMUO, 4*UIBMUO, 4*UIBMUO);
		}
	}
}

//===========================================================================
void LoadUI(char *UIname) {
	
	INFOBM	ibm;
	long	i;
	DWORD	eClr[22];

	// if there's already one loaded, get rid of it
	if( hUIBMO ) DeleteBM(hUIBMO);
	hUIBMO = 0;

	strcpy(UIBMOname, UIname);
	sprintf(ScratBuf, "UIIMAGES/%s", UIname);
	if( UIBMOname[strlen(UIBMOname)-4]=='.' ) UIBMOname[strlen(UIBMOname)-4] = 0;

	UIBMtime = ftime(ScratBuf);
	hUIBMO = LoadBM(ScratBuf);
	GetInfoBM(hUIBMO, &ibm);

	// Figure out the UIBM 'unit' size of the UI graphics for location/sizing info

	// NOTE: UIimages *MUST* be a multiple of 56 units wide, with a MINIMUM unit width of 12,
	// giving a MINIMUM bitmap width of 12*56=672. The game's default UIimage set uses a
	// unit width of 16, resulting in a bitmap width of 16*56=896. It's 84 units high,
	// giving a bitmap height of 16*84=1344.
	//
	// The unit width CAN be larger, in multiples of 56 pixels, with everything scaled up appropriately,
	// except for the color pixels in the upper-left sector, which must remain where they are,
	// pixel-by-pixel. So, UIimages image widths could be 672, 728, 784, 840, 896, 952, etc., but
	// I'd recommend scaling up in nice binary steps, with 'unit' size of 16, 20, 24, 28, 32, 36, etc.,
	// giving bitmap widths of 896, 1120, 1344, 1568, 1792, 2016, etc. I'll leave it to you to
	// multiply out 84 times your chosen unit size, should you choose to buck the system...
	
	UIBMUO = ibm.w/UIIMG_UW;	// size of one UI 'unit' in the SOURCE UI bitmap

	// get the colors of things
	for(i=0; i<sizeof(Colors)/sizeof(Colors[0]); i++) Colors[i] = GetPixelBM(hUIBMO, i, 0);
	
	// make transparency colors transparent UNLESS transparency color is BLACK (no transparency besides what might be inherent in the PNG)
	if( Colors[ClrTrans1]!=CLR_BLACK ) SetAlphaBM(hUIBMO, 0, 0, ibm.w, ibm.h, Colors[ClrTrans1], 0);
	if( Colors[ClrTrans2]!=CLR_BLACK ) SetAlphaBM(hUIBMO, 0, 0, ibm.w, ibm.h, Colors[ClrTrans2], 0);

	MsgBackClr = Colors[ClrMSGfillNORM];
	
	SizeUI();
	SizeSB();
}

//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
// INI FILE STUFF
//===========================================================================
// NOTE: 'str' MUST NOT BE LONGER THAN 256 bytes!!!
void iniSet1Str(char *nam, char *str) {
	
	char	setbuf[260];
	long	d=0;
	
	setbuf[d++] = '"';
	while( d<256 && *str ) {
		if( *str=='"' ) setbuf[d++] = '\\';
		setbuf[d++] = *str++;
	}
	setbuf[d++] = '"';
	setbuf[d] = 0;
	iniSetVar(nam, setbuf);
}

//===========================================================================
// Make sure GAME (ini file 'header') information is up to date
void iniSaveHeader() {
	
	iniSetSec(PgmFolder);
	sprintf(ScratBuf, FMTd, INIREV);
	iniSetVar(iniREV, ScratBuf);
	iniSetVar(pgmREV, (char*)GAMEVER);
	sprintf(ScratBuf, FMTd, CurPlayer.id);
	iniSetVar(iniCURPLAYER, ScratBuf);
	sprintf(ScratBuf, FMTd, NumPlayers);
	iniSetVar(iniNUMPLAYERS, ScratBuf);
	sprintf(ScratBuf, FMTd, maxID);
	iniSetVar(iniMAXID, ScratBuf);
}

//===========================================================================
void iniSaveMood(long m) {
	
	long	i;
	DWORD	hM, hMUS;
	MOOD	*pM;
	FILELIST	*pMUS;
	
	// find ptr to mood 'm'
	for(i=0, hM=CurPlayer.hFirstMood; i<m && hM; i++, hM=((MOOD*)HeapPtr(hM))->hNext) ;
	if( !hM ) return;
	pM = (MOOD*)HeapPtr(hM);
	
	sprintf(ScratBuf, iniMOOD, CurPlayer.id, i);
	iniSetSec(ScratBuf);
	iniSet1Str(iniNAME, pM->name);
	iniSet1Str(iniUIIMAGE, pM->UIimage);
	iniSet1Str(iniIMAGE, pM->images);
	iniSet1Str(iniHELP, pM->HelpName);
	iniSet1Str(iniSTRINGS, pM->StringsName);
	
	iniDelVar(iniKEYMAPS);
	for(i=0; i<pM->KeyMapsCnt; i++) {
		iniAddVarNum(iniKEYMAPS, pM->KeyMaps[i].key, 0);
		iniAddVarNum(iniKEYMAPS, pM->KeyMaps[i].flags, 0);
		iniAddVarNum(iniKEYMAPS, pM->KeyMaps[i].func, 0);
	}
	sprintf(ScratBuf, "%d %d %d", pM->WinDisp, pM->WinDelay, pM->WinSpeed);
	iniSetVar(iniWINDISP, ScratBuf);
	
	sprintf(ScratBuf, FMTd5, pM->wndPos.x, pM->wndPos.y, pM->wndPos.w, pM->wndPos.h, pM->wndPos.showcmd);
	iniSetVar(iniWNDPOS, ScratBuf);

	ScratBuf[0] = 0;
	for(i=0; i<sizeof(pM->Actions); i++) {
		sprintf(ScratBuf+strlen(ScratBuf), FMTdSP, pM->Actions[i]);
	}
	iniSetVar(iniACTIONS, ScratBuf);
	
	sprintf(ScratBuf, FMT8x8x, pM->flags, pM->flags2);
	iniSetVar(iniFLAGS, ScratBuf);
	
	sprintf(ScratBuf, FMTd, pM->UU);
	iniSetVar(iniUU, ScratBuf);

	sprintf(ScratBuf, FMTd, pM->LimitPuzzSize);
	iniSetVar(iniPUZZSIZE, ScratBuf);

	sprintf(ScratBuf, FMTd, pM->CSVsep);
	iniSetVar(iniCSVSEP, ScratBuf);

	iniSetVar(iniMUSICBACK, "");
	for(hMUS=pM->hBackMusic; hMUS!=0 && (pMUS=(FILELIST*)HeapPtr(hMUS)); hMUS=pMUS->hNext) {
		iniAddVarNum(iniMUSICBACK, pMUS->flags, 0);
		iniAddVarStr(iniMUSICBACK, pMUS->name);
	}
	
	iniSetVar(iniMUSICWIN, "");
	for(hMUS=pM->hWinMusic; hMUS!=0 && (pMUS=(FILELIST*)HeapPtr(hMUS)); hMUS=pMUS->hNext) {
		iniAddVarNum(iniMUSICWIN, pMUS->flags, 0);
		iniAddVarStr(iniMUSICWIN, pMUS->name);
	}
}

//===========================================================================
// Make sure PLAYER information is up to date
void iniSavePlayer() {
	
	long	i;
	DWORD	hM;

	sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
	iniSetSec(ScratBuf);

	sprintf(ScratBuf, FMTd, CurPlayer.id);
	iniSetVar(iniID, ScratBuf);

	iniSet1Str(iniNAME, CurPlayer.name);
	
	iniDelVar(iniCURPUZZNUM);
	for(i=0; i<24; i++) iniAddVarNum(iniCURPUZZNUM, CurPlayer.CurButtPuzzNum[i], 0);
	
	sprintf(ScratBuf, FMTd, CurPlayer.SolvedTotal);
	iniSetVar(iniSOLVEDTOTAL, ScratBuf);

	sprintf(ScratBuf, FMT8x, CurPlayer.HelpExportFlags);
	iniSetVar(iniHELPFLAGS, ScratBuf);

	ScratBuf[0] = 0;
	for(i=0; i<HEC_NUMCOLORS; i++) sprintf(ScratBuf+strlen(ScratBuf), FMT8xSP, CurPlayer.HelpExportColors[i]);
	iniSetVar(iniHELPCOLORS, ScratBuf);

	ScratBuf[0] = 0;
	for(i=0; i<HEFONT_NUM; i++) sprintf(ScratBuf+strlen(ScratBuf), FMTdSP, CurPlayer.HelpExportFonts[i]);
	iniSetVar(iniHELPFONT, ScratBuf);

	sprintf(ScratBuf, FMTd, CurPlayer.HelpExportWidth);
	iniSetVar(iniHELPWIDTH, ScratBuf);
	
	sprintf(ScratBuf, FMTd, CurPlayer.HelpExportLmargin);
	iniSetVar(iniHELPLMARGIN, ScratBuf);
	
	sprintf(ScratBuf, "%d %d", CurPlayer.HelpExportContentWidth, CurPlayer.HelpExportContentSpacing);
	iniSetVar(iniHELPCONTENT, ScratBuf);
	
	sprintf(ScratBuf, "%d %d %d %d",
		CurPlayer.HelpExportRule1Top, CurPlayer.HelpExportRule1Bot,
		CurPlayer.HelpExportRule2Top, CurPlayer.HelpExportRule2Bot
		);
	iniSetVar(iniRULESPACE, ScratBuf);
	
	sprintf(ScratBuf, FMTd, CurPlayer.MoodCnt);
	iniSetVar(iniMOODCNT, ScratBuf);

	sprintf(ScratBuf, FMTd, CurPlayer.CurMoodNum);
	iniSetVar(iniCURMOOD, ScratBuf);

	for(i=0, hM=CurPlayer.hFirstMood; hM && i<CurPlayer.MoodCnt; i++, hM=((MOOD*)HeapPtr(hM))->hNext) ;
	
	for(i=0; i<CurPlayer.MoodCnt; i++) iniSaveMood(i);
}

//=======================================================================
// Load the CurPlayer.id settings and everything associated with them.
// Load a saved game, if appropriate, else init game.
// Whatever.
// 'fullstart' is FALSE during OnEndStartup(), TRUE at all other times.
void StartPlayer(BOOL fullstart) {
	
	long		i, m;
	DWORD		hMOOD, hMOODnext, hMOODlast, hM, hMnext, hMlast;
	FILELIST	*pM;
	MOOD		*pMOOD;
	DWORD		key, flag, f, mcnt;

	// find CurPlayer.id's section in .ini file
	sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
	if( !iniFindSec(ScratBuf) ) {
		SystemMessage("Oops! Didn't find PLAYER in .ini, terminating. So sorry!");
		_asm(term);
	}
	// read in the players info
	iniFindVar(iniNAME);
	iniGetStr(CurPlayer.name, sizeof(CurPlayer.name));
	
	iniFindVar(iniCURBUTT);
	iniGetNum(&CurPlayer.CurButtNum);

	iniFindVar(iniCURPUZZNUM);
	for(i=0; i<24; i++) iniGetNum(&CurPlayer.CurButtPuzzNum[i]);
	CurPlayer.CurPuzzNum = CurPlayer.CurButtPuzzNum[CurPlayer.CurButtNum];

	iniFindVar(iniSOLVEDTOTAL);
	iniGetNum(&CurPlayer.SolvedTotal);	// get count of total tiles solved (for special win display)

	iniFindVar(iniHELPFLAGS);
	iniGetNum(&CurPlayer.HelpExportFlags);

	iniFindVar(iniHELPCOLORS);
	for(i=0; i<HEC_NUMCOLORS; i++) iniGetNum(&CurPlayer.HelpExportColors[i]);

	iniFindVar(iniHELPFONT);
	for(i=0; i<HEFONT_NUM; i++) iniGetNum(&CurPlayer.HelpExportFonts[i]);

	iniFindVar(iniHELPWIDTH);
	iniGetNum(&CurPlayer.HelpExportWidth);

	iniFindVar(iniHELPLMARGIN);
	iniGetNum(&CurPlayer.HelpExportLmargin);

	iniFindVar(iniHELPCONTENT);
	iniGetNum(&CurPlayer.HelpExportContentWidth);
	iniGetNum(&CurPlayer.HelpExportContentSpacing);

	iniFindVar(iniRULESPACE);
	iniGetNum(&CurPlayer.HelpExportRule1Top);
	iniGetNum(&CurPlayer.HelpExportRule1Bot);
	iniGetNum(&CurPlayer.HelpExportRule2Top);
	iniGetNum(&CurPlayer.HelpExportRule2Bot);

	iniFindVar(iniMOODCNT);
	iniGetNum(&mcnt);	// get CurPlayer.MoodCnt, but don't change it until we delete any old ones

	iniFindVar(iniCURMOOD);
	iniGetNum(&CurPlayer.CurMoodNum);	// get CurPlayer's CurMood
	
	CurPlayer.hFirstMood = 0;
	CurPlayer.MoodCnt = 0;
	
	// NOW, read in the player's MOODs
	for(m=0; m<mcnt; m++) {
		sprintf(ScratBuf, iniMOOD, CurPlayer.id, m);	// format name of mood section
		if( iniFindSec(ScratBuf) ) {
			++CurPlayer.MoodCnt;
			// create a MOOD structure to hold this mood, and link it into the CurPlayer's mood-chain
			hMOOD = HeapAlloc(sizeof(MOOD));	// allocate storage for it
			if( CurPlayer.hFirstMood==0 ) CurPlayer.hFirstMood = hMOOD;
			else ((MOOD*)HeapPtr(hMOODlast))->hNext = hMOOD;
			hMOODlast = hMOOD;
			pMOOD = (MOOD*)HeapPtr(hMOOD);
			pMOOD->hNext = 0;

			// now read the MOOD values from the .ini file into the structure
			iniFindVar(iniNAME);	// get the MOOD name
			iniGetStr(pMOOD->name, sizeof(pMOOD->name));

			i = iniFindVar(iniUIIMAGE);	// get the UI image file name
			iniGetStr(pMOOD->UIimage, sizeof(pMOOD->UIimage));
			
			iniFindVar(iniIMAGE);	// get the game image set name
			iniGetStr(pMOOD->images, sizeof(pMOOD->images));
			
			iniFindVar(iniHELP);	// get the Help file name
			iniGetStr(pMOOD->HelpName, sizeof(pMOOD->HelpName));
			
			iniFindVar(iniSTRINGS);	// get the Strings file name
			iniGetStr(pMOOD->StringsName, sizeof(pMOOD->StringsName));

			if( iniFindVar(iniKEYMAPS) ) {
				for(pMOOD->KeyMapsCnt=0; pMOOD->KeyMapsCnt<sizeof(pMOOD->KeyMaps)/sizeof(pMOOD->KeyMaps[0]); pMOOD->KeyMapsCnt++) {
					if( iniGetNum(&key) ) {
						iniGetNum(&flag);
						iniGetNum(&f);
						pMOOD->KeyMaps[pMOOD->KeyMapsCnt].key   = (WORD)key;
						pMOOD->KeyMaps[pMOOD->KeyMapsCnt].flags = (BYTE)flag;
						pMOOD->KeyMaps[pMOOD->KeyMapsCnt].func  = (BYTE)f;
					} else break;
				}
			} else {
				memcpy((BYTE*)&pMOOD->KeyMaps[0], (BYTE*)&defkeymap, sizeof(defkeymap));
				pMOOD->KeyMapsCnt = sizeof(defkeymap)/sizeof(defkeymap[0]);
			}

			if( iniFindVar(iniWINDISP) ) {
				iniGetNum((DWORD*)&pMOOD->WinDisp);
				iniGetNum(&pMOOD->WinDelay);
				iniGetNum((DWORD*)&pMOOD->WinSpeed);
			} else {
				pMOOD->WinDisp = WIN_RANDOM;
				pMOOD->WinDelay = 30;
				pMOOD->WinSpeed = 1;
			}
			
			iniFindVar(iniWNDPOS);
			iniGetNum((DWORD*)&pMOOD->wndPos.x);
			iniGetNum((DWORD*)&pMOOD->wndPos.y);
			iniGetNum((DWORD*)&pMOOD->wndPos.w);
			iniGetNum((DWORD*)&pMOOD->wndPos.h);
			iniGetNum((DWORD*)&pMOOD->wndPos.showcmd);

			iniFindVar(iniPUZZSIZE);
			iniGetNum((DWORD*)&pMOOD->LimitPuzzSize);
			if( pMOOD->LimitPuzzSize<50 ) pMOOD->LimitPuzzSize = 50;
			if( pMOOD->LimitPuzzSize>100 ) pMOOD->LimitPuzzSize = 100;

			if( iniFindVar(iniCSVSEP) ) iniGetNum((DWORD*)&pMOOD->CSVsep);
			else pMOOD->CSVsep = (DWORD)',';

			iniFindVar(iniACTIONS);
			for(i=0; i<sizeof(pMOOD->Actions); i++) iniGetNum((DWORD*)&pMOOD->Actions[i]);
			
			iniFindVar(iniFLAGS);
			iniGetNum(&pMOOD->flags);
			iniGetNum(&pMOOD->flags2);
			
			iniFindVar(iniUU);
			iniGetNum((DWORD*)&pMOOD->UU);

			// now read in this player's MUSIC settings
			pMOOD->hBackMusic = pMOOD->hWinMusic = pMOOD->BackMusicCnt = pMOOD->WinMusicCnt = 0;
			if( iniFindVar(iniMUSICBACK) ) {
				while( !iniIsEndVar() ) {
					iniGetNum((DWORD*)&i);
					iniGetStr(ScratBuf, sizeof(pM->name));
					hM = HeapAlloc(sizeof(FILELIST));
					pM = (FILELIST*)HeapPtr(hM);
					pM->hNext = 0;
					pM->flags = i;
					strcpy(pM->name, ScratBuf);
					if( pMOOD->hBackMusic==0 ) pMOOD->hBackMusic = hM;
					else ((FILELIST*)HeapPtr(hMlast))->hNext = hM;
					hMlast = hM;
					++pMOOD->BackMusicCnt;
				}
			}
			if( iniFindVar(iniMUSICWIN) ) {
				while( !iniIsEndVar() ) {
					iniGetNum((DWORD*)&i);
					iniGetStr(ScratBuf, sizeof(pM->name));
					hM = HeapAlloc(sizeof(FILELIST));
					pM = (FILELIST*)HeapPtr(hM);
					pM->hNext = 0;
					pM->flags = i;
					strcpy(pM->name, ScratBuf);
					if( pMOOD->hWinMusic==0 ) pMOOD->hWinMusic = hM;
					else ((FILELIST*)HeapPtr(hMlast))->hNext = hM;
					hMlast = hM;
					++pMOOD->WinMusicCnt;
				}
			}
		}
	}
	
	// make sure CurPlayer.CurMoodNum exists in the list of moods
	for(i=0, hMOOD=CurPlayer.hFirstMood; i<CurPlayer.CurMoodNum && hMOOD; ++i, hMOOD=((MOOD*)HeapPtr(hMOOD))->hNext) ;
	// if not, default back to the first mood
	if( !hMOOD ) CurPlayer.CurMoodNum = 0;
	
	// start/load the mood
	StartMood(CurPlayer.CurMoodNum, fullstart);

	if( fullstart ) {
		StartPuzzleType();	// setup for CurPlayer.CurButtNum and StartPuzzle(CurPlayer.CurPuzzNum)
		if( hMenuDlg ) {
			PauseTime = TimeMilli();
		}
		mainDrawAll();
		if( hModal ) InvalidateAllWnd(hModal);
	}
}

//=======================================================================
// Save the current player settings to .ini
// Save current puzzle, if appropriate
// Do whatever's needed in prep for either shutting down or switching players.
void EndPlayer() {
	
	HWND		hMOOD, hMOODnext;
	MOOD		*pMOOD;
	DWORD		hM, hMnext;
	FILELIST	*pM;

	if( PuzzleActive ) EndPuzzleType();
	
	iniSavePlayer();	// save current player
	
	EndMood();	// delete/free-up any mood stuff
	
	// Delete all of player's moods and music blocks
	for(hMOOD=CurPlayer.hFirstMood; hMOOD; hMOOD=hMOODnext) {
		pMOOD = (MOOD*)HeapPtr(hMOOD);
		hMOODnext = pMOOD->hNext;
		for(hM=pMOOD->hBackMusic; hM; hM=hMnext) {
			pM = (FILELIST*)HeapPtr(hM);
			hMnext = pM->hNext;
			HeapFree(hM);
		}
		for(hM=pMOOD->hWinMusic; hM; hM=hMnext) {
			pM = (FILELIST*)HeapPtr(hM);
			hMnext = pM->hNext;
			HeapFree(hM);
		}
		HeapFree(hMOOD);
	}
	CurPlayer.hFirstMood = CurPlayer.MoodCnt = 0;
}

//==========================================================================
void UpdateMood() {
	
	DWORD	i, hMOOD;
	
	// find the CurPlayer.mood that is the matching mood for CurMood
	for(i=0, hMOOD=CurPlayer.hFirstMood; i<CurPlayer.CurMoodNum && hMOOD; ++i, hMOOD=((MOOD*)HeapPtr(hMOOD))->hNext) ;
	if( !hMOOD ) {	// shouldn't ever happen, but just in case...
		hMOOD = CurPlayer.hFirstMood;
		CurPlayer.CurMoodNum = 0;
	}
	// and copy CurMood to it for eventual storage in .ini
	if( hMOOD ) memcpy((BYTE*)HeapPtr(hMOOD), (BYTE*)&CurMood, sizeof(MOOD));
}

//==========================================================================
DWORD MsgBarDlgProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	DIALOG	*pW;
	RECT	r, rc, ri;
	long	i, xs, ys, ws, hs, dx, dy, dw, dh;
	DWORD		msgbuf[4], *pClrs, clrs[4];
	
	pW = (DIALOG*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		DefDlgProc(hW, msg, arg1, arg2, arg3);
		pW->uu = UU+1;	// force resize
		SendMsg(hW, WM_PARENTSIZED, 0, 0, 0);
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			dx = r.x;
			dy = r.y;
			dw = r.w;
			dh = r.h;

			DefDlgProc(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);
			DrawRectBM(0, dx, dy, dw, dh, MsgBackClr, Colors[ClrMSGedgeNORM]);
			
			DrawTextBM(0, MsgText, dx+UIBMU, dy+UIBMU/2, dw-UIBMU2, dh-UIBMU, CF_LEFT, MsgBackClr, MsgTextClr);

			ClipRectBM(0, rc, NULL);
		}
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		return 0;
	 case WM_PARENTSIZED:
		if( pW->uu && pW->uu!=UU ) {
			InvalidateAllWnd(hW);
		}
		dw = pW->w;
		pW->xoff = 0;
		pW->yoff = -MSGHPX;
		pW->w = ViewW-BARWPX;
		pW->h = pW->pRES->uh*UU;
		if( dw != pW->w ) {
			pW->uu = UU;
			InvalidateAllWnd(hW);
		}
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		return 0;
	}
	return DefDlgProc(hW, msg, arg1, arg2, arg3);
}

//==========================================================================
DWORD ButtBarDlgProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	DIALOG	*pW;
	RECT	r, rc, ri;
	long	i, xs, ys, ws, hs, dx, dy, dw, dh;
	DWORD		msgbuf[4], *pClrs, clrs[4];
	
	pW = (DIALOG*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		DefDlgProc(hW, msg, arg1, arg2, arg3);
		pW->uu = UU+1;	// force resize
		SendMsg(hW, WM_PARENTSIZED, 0, 0, 0);
		return 0;
	 case WM_PAINT:
		pW->h = ViewH;
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			dx = r.x;
			dw = r.w;
			dh = r.h;

			dy = r.y+r.h-H_TOOLBACK*UU;
			DrawBM(0, dx, dy, hUIBM, X_TOOLBACK*UIBMU, Y_TOOLBACK*UIBMU, BARWUU*UIBMU, H_TOOLBACK*UIBMU, 0);
			while( dy > r.y ) {
				dy -= (H_TOOLBACK/2)*UU;
				DrawBM(0, dx, dy, hUIBM, X_TOOLBACK*UIBMU, Y_TOOLBACK*UIBMU, BARWUU*UIBMU, (H_TOOLBACK/2)*UIBMU, 0);
			}
			DrawRectBM(0, dx, r.y, dw, dh, CLR_NONE, Colors[ClrButtBarFrame]);

			ClipRectBM(0, rc, NULL);
		}
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		return 0;
	 case WM_PTRMOVE:
		SetCursor(CURSOR_NORMAL);
		break;
	 case WM_PARENTSIZED:
		if( pW->uu && pW->uu!=UU ) {
			InvalidateAllWnd(hW);
		}
		dh = pW->h;
		pW->xoff = pW->pRES->ux*UU;
		pW->yoff = pW->pRES->uy*UU;
		pW->w = pW->pRES->uw*UU;
		pW->h = ViewH;
		
		if( dh != pW->h ) {
			pW->uu = UU;
			
			InvalidateAllWnd(hW);
		}
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		return 0;
	}
	return DefDlgProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
void SetMsgText(char *str, DWORD backclr, DWORD textclr) {
	
	strcpy(MsgText, str);
	MsgBackClr = backclr;
	MsgTextClr = textclr;
	InvalidateAllWnd(hMsgBar);
}

//===========================================================================
DWORD BarButtBMProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	BUTTONBM	*pW;
	RECT		r, rc, ri;
	long		i, xs, ys, ws, hs;
	DWORD		*pClrs, clrs[2];
	QWORD		qw;

	pW = (BUTTONBM*)HeapPtr(hW);
	switch( msg ) {
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) && (pW->id!=ID_PAUSE || !(CurMood.flags & PF_REMOVE_PAUSE)) ) {
			ClipRectBM(0, ri, &rc);

			xs = pW->UIBMUx*UU;
			ys = pW->UIBMUy*UU;
			ws = pW->UIBMUw*UU;
			hs = pW->UIBMUh*UU;

			// All ButtonBar buttons must follow these rules in the UIBM:
			// The 'normal' button must be at (pW->UIBMUx, pW->UIBMUy) in the UIBM.
			// The 'down' button must be at pW->UIBMUw to the right of the 'normal' button.
			// If the button can be CF_DISABLED, then the disabled 'normal' and 'down'
			//		buttons must be pW->UIBMUh below the non-disabled buttons.
			
			if( pW->pRES->uw > pW->pRES->uh ) {	// if double-wide button
				ws = 2*ws;
				if( pW->flags & CF_DOWN ) ys += hs;	// double-wide buttons are vertically stacked rather than horizontally
			} else {
				if( pW->flags & CF_DOWN ) xs += ws;
				if( pW->flags & CF_DISABLED ) ys += hs;
			}
			
			// draw the button
			DrawBM(0, r.x, r.y, hUIBM, xs, ys, ws, hs, 0);

			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_HOVER:
		SetMsgText( (arg1 && (CurMood.flags & PF_SHOW_TOOLTIPS)) ? (char*)pW->text : "", Colors[ClrMSGfillNORM], Colors[ClrMSGtextNORM]);
		break;
	 case WM_SETFOCUS:
		hFocus = hMain;	// don't allow focus to stay on toolbar button
		break;
	 case WM_NOTIFY:
		switch( arg1 ) {
		 case WN_STRINGS_RELOADED:
			if( pW && pW->pRES ) pW->text = (DWORD)GetCtlString(pW->pRES);
			break;
		}
		break;
	}
	return ButtonBMProc(hW, msg, arg1, arg2, arg3);
}

DWORD	TitleKeyCnt;
char str_mod_warn1[]	={0x2B,0x2E,0x40,0x2A,0x3B,0x2E,0x23,0x09,0x3D,0x51,0x4E,0x09,0x5D,0x4E,0x61,0x5D,0x09,0x5C,0x5D,0x5B,0x52,0x57,0x50,0x5C,0x09,0x52,0x57,0x09,0x5D,0x51,0x52,0x5C,0};
char str_mod_warn2[]	={0x59,0x5B,0x58,0x50,0x5B,0x4A,0x56,0x09,0x51,0x4A,0x5F,0x4E,0x09,0x4B,0x4E,0x4E,0x57,0x09,0x56,0x58,0x4D,0x52,0x4F,0x52,0x4E,0x4D,0x0A,0};
char str_mod_warn3[]	={0x2C,0x58,0x57,0x5D,0x4A,0x4C,0x5D,0x09,0x4E,0x5F,0x4E,0x5B,0x4E,0x5D,0x5D,0x29,0x54,0x4A,0x5C,0x4E,0x5B,0x17,0x4C,0x58,0x56,0};

//=========================================================================
void Decode(char *s) {
	
	while( *s ) {
		*s += 23;
		++s;
	}
}

//=========================================================================
DWORD TitleDlgProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	DIALOG		*pW;
	RECT		r, mr, ri;
	long		i, x, y, w, h;
	INFOBM		bmi;

	pW = (DIALOG*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		hTitle = hW;
		TitleKeyCnt = 0;
		if( str_mod_warn1[0]!='B' ) {
			Decode(str_mod_warn1);
			Decode(str_mod_warn2);
			Decode(str_mod_warn3);
		}
		return 0;
	 case WM_DESTROY:
		hTitle = 0;
		break;
	 case WM_PARENTSIZED:
		++BlockPaint;	// for speed and clean redrawing
		InvalidateAllWnd(hW);
		DefDlgProc(hW, msg, arg1, arg2, arg3);
		--BlockPaint;
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &mr);

			GetInfoBM(hSplashBM, &bmi);
			StretchBM(0, r.x, r.y, r.w, r.h, hSplashBM, 0, 0, bmi.w, 3*bmi.w/5);
			DrawRectBM(0, r.x, r.y, r.w, r.h, CLR_NONE, CLR_BLACK);

			ClipRectBM(0, mr, NULL);
		}
		NotifyChildren(hW, WM_PAINT, arg1, arg2, arg3);
		return 0;
	}	
	return DefDlgProc(hW, msg, arg1, arg2, arg3);
}

//==========================================================================
DWORD TitleGroupboxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	RECT	r;
	
	switch( msg ) {
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		DrawRectBM(0, r.x, r.y, r.w, r.h, CLR_RED, CLR_NONE);
		break;
	}
	return GroupboxProc(hW, msg, arg1, arg2, arg3);
}

//==========================================================================
DWORD TitleDemoProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	RECT	r;
	DWORD	dw;
	
	switch( msg ) {
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		dw = GetPixelBM(hUIBMO, X_GROUPBOX*UIBMUO+UIBMUO-1, Y_GROUPBOX*UIBMUO+UIBMUO-1);
		DrawRectBM(0, r.x, r.y, r.w, r.h, dw, CLR_NONE);
		break;
	}
	return GroupboxProc(hW, msg, arg1, arg2, arg3);
}

//==========================================================================
DWORD TitleContinueProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	switch( msg ) {
	 case WM_KEYDN:
		++TitleKeyCnt;
		break;
	 case WM_KEYUP:
		if( ++TitleKeyCnt < 2 ) break;
	 case WM_ACTIVATE:
		SendMsg(hTitle, WM_DESTROY, 0, 0, 0);
		return 0;
	}
	return ButtonProc(hW, msg, arg1, arg2, arg3);
}

//=======================================================================
DLGRES TitleDlg[] = {
	{ID_TITLE_DLG,		WT_WND,			 -1, -1,160, 96, WF_SHOW, TitleDlgProc, 0, 0, 0},

	{ID_MODWARN1,		WT_GROUPBOX,	108, 72, 48, 14, 0, TitleGroupboxProc, 0, 0, 0},
	{ID_MODWARN2,		WT_TEXT,		110, 73, 45,  4, 0+CF_LEFT+CF_STR, 0, str_mod_warn1, 0, 0},
	{ID_MODWARN3,		WT_TEXT,		110, 77, 45,  4, 0+CF_LEFT+CF_STR, 0, str_mod_warn2, 0, 0},
	{ID_MODWARN4,		WT_TEXT,		110, 81, 45,  4, 0+CF_LEFT+CF_STR, 0, str_mod_warn3, 0, 0},
	{ID_CONTINUE,		WT_BUTTONTEXT,	108, 86, 48,  6, WF_SHOW+WF_NSC+CF_TABSTOP+CF_CENTER+CF_GROUP+CF_FOCUS, TitleContinueProc, MAKESTR(BLK_MISC, STR_CONTINUE), 0, 0},

	{0}
};

