//======================================================================================================
// PLAYER.H
// Copyright 2022 Everett Kaser
// All rights reserved
//======================================================================================================
#ifndef PLAYERINC
#define PLAYERINC

//=======================================================================
// PLAYER flags definitions

#define	PF_GAMESOUNDS				0x00000001	// if normal sound effects are enabled
#define	PF_ERRORSOUNDS				0x00000002	// if error sound effects are enabled
//3
#define	PF_SHOW_TOOLTIPS			0x00000020	// if show tooltips in message window
#define	PF_PROMPTRESTART			0x00000040	// if 
#define	PF_SHOWBESTCOUNT			0x00000080	// if 
#define	PF_ENABLECLICK				0x00000100	// if 
#define	PF_PROMPTNEXT				0x00000200	// if 
#define	PF_STARTBOTTOM				0x00000400	// if 
#define	PF_SCRAMBLEORDER			0x00000800	// if randomize tile colors used
#define	PF_SHOWTILESMOVED			0x00001000	// if 
#define	PF_REMOVE_PAUSE				0x00002000	// if remove PAUSE feature
#define	PF_BACK_MUSIC				0x00004000	// if BACKGROUND music is enabled
#define	PF_WIN_MUSIC				0x00008000	// if WIN music is enabled
#define	PF_RAND_MUSIC				0x00010000	// if play music list in RANDOM sequence
#define	PF_SHOWTIMER				0x00020000	// if 
#define PF_REV_WHEEL				0x00040000	// if reverse action of mouse wheel roll
#define	PF_REV_TILT					0x00080000	// if reverse action of mouse wheel tilt
//3
#define	PF_BYPASS_TITLE				0x00800000	// if bypass title screen
//5
#define	PF_ENABLE_WHEEL				0x20000000	// if enable mouse wheel to move clue selection
#define	PF_SHIFT_WHEEL				0x40000000	// if SHIFT needed to activate mouse wheel
#define	PF_REVERSE_WHEEL			0x80000000	// if reverse direction of wheel effects

#define	PF2_SEEN_QUICKSTART	0x00000001	// if quick-start Help has been displayed once
#define	PF2_MIRROR_MOUSE	0x00000002	// if mirror mouse changes
#define	PF2_MIRROR_KEYS		0x00000004	// if mirror key changes
#define	PF2_MIRROR_OPTIONS	0x00000008	// if mirror options changes
#define	PF2_MIRROR_GRAPHICS	0x00000010	// if mirror graphics changes
#define	PF2_MIRROR_LANGUAGE	0x00000020	// if mirror language changes
#define	PF2_MIRROR_INTERFACE 0x00000040	// if mirror interface changes
#define	PF2_MIRROR_MUSIC	0x00000080	// if mirror music changes

//=======================================================================
// associated KEY NAME and KEY CODE
typedef struct {
	char	name[32];
	DWORD	code;
} KEY;

// mapping of a KEY and shift/ctl flags to a FUNCTION
typedef struct {
	WORD	key;
	BYTE	flags;
	BYTE	func;
} KEYMAP;

#define	DEFKEYMAPCNT	23

#define	KF_FUNC			0	// always the first KF_function

#define KF_KEYMAP_DLG		0
#define KF_OPTIONS_DLG		1
#define KF_MOODS_DLG		2
#define KF_NEXT_PUZZ		3
#define KF_PREV_PUZZ		4
#define	KF_NEXT_MOOD		5
#define	KF_PREV_MOOD		6
#define	KF_PAUSE			7
#define	KF_MENU_DLG			8
#define	KF_NEXT_UNSOLVED	9
#define	KF_FIRST_UNSOLVED	10
#define	KF_UNDO				11
#define	KF_REDO				12
#define	KF_RESTART			13
#define	KF_SAVEGAME			14
#define	KF_PLAYER_DLG		15
#define	KF_NEXT_PLAYER		16
#define	KF_PREV_PLAYER		17

#define	KF_NUMFUNCS		15	// always 1 more than the last KF_ #define above

#define	MAXKEYMAP		256

//=======================================================================
typedef struct {
	DWORD	hNext;			// ptr to next file memblk (or 0)
	DWORD	flags;			// file flags (music played/playing/etc flags)
	char	name[128];		// file name
} FILELIST;

#define	MUF_PLAYED		1		// set if this music file has already been played

//========================================================================
typedef struct {
	DWORD	hNext;			// handle of next MOOD (or 0)
	
	char	name[64];		// name of mood - max length = 31 bytes plus nul
	char	UIimage[128];	// name of UI graphics file
	char	images[128];	// name of images file
	char	HelpName[128];	// name of help file
	char	StringsName[128];//name of strings file

	KEYMAP	KeyMaps[MAXKEYMAP];	// keys->functions mappings
	DWORD	KeyMapsCnt;
	
	WNDPOS	wndPos;			// for saving OS-dependant window position info

	BYTE	Actions[20];	// action to take on mouse clicks [5_buttons * normal/shift/ctl/shift-ctl]
	
	DWORD	flags;			// miscellaneous configuration flags (see PF_ above)
	DWORD	flags2;			// see PF2_ defines above
	
	DWORD	LimitPuzzSize;	// force drawn puzzle to be smaller (or not)
	DWORD	CSVsep;			// separator character for .CSV
	
	DWORD	WinDelay;		// how long to flash toolbar before starting win display
	long	WinDisp;		// which win display to run
	long	WinSpeed;		// how fast to run the win display
	
	long	UU;				// 0 = auto, else override "pixels per UI unit"
	
	DWORD	BackMusicCnt; 	// number of background music files
	DWORD	hBackMusic;		// handle of first background music file
	DWORD	WinMusicCnt; 	// number of win music files
	DWORD	hWinMusic;		// handle of first win music file
} MOOD;

//========================================================================
typedef struct {
	DWORD	id;				// player #
	char	name[64];		// name of player - max length = 31 bytes plus nul
	DWORD	MoodCnt;		// how many MOODs this player has (list is NULL-terminated, but this may be handy)
	DWORD	CurMoodNum;		// # of current mood for player
	DWORD	hFirstMood;		// handle of first MOOD
	DWORD	HelpExportFlags;// flags for Help Export dialog
#define	HEF_BACK			0x00000001	// include BACK color in Help EXPORT
#define	HEF_TEXT			0x00000002	// include TEXT color in Help EXPORT
#define	HEF_H1				0x00000004	// include TEXT color in Help EXPORT
#define	HEF_H2				0x00000008	// include TEXT color in Help EXPORT
#define	HEF_H3				0x00000010	// include TEXT color in Help EXPORT
#define	HEF_CONTBACK		0x00000020	// include CONTBACK color in Help EXPORT
#define	HEF_CONTTEXT		0x00000040	// include CONTTEXT color in Help EXPORT
#define	HEF_CONTLINK		0x00000080	// include CONTLINK color in Help EXPORT
#define	HEF_HEADBACK		0x00000100	// include HEADBACK color in Help EXPORT
#define	HEF_HEADTEXT		0x00000200	// include HEADTEXT color in Help EXPORT
#define	HEF_HEADFRAME		0x00000400	// include HEADFRAME color in Help EXPORT

#define	HEF_HRULE			0x00001000	// if insert HORIZONTAL RULE line before each section head
#define	HEF_LINKCONTENTS	0x00002000	// if insert link to table of contents at ends of each section
#define	HEF_HRULE2			0x00004000	// if insert HORIZONTAL RULE line after each section head
#define	HEF_CP_HEX			0x00008000	// if COLOR PICKER should display HEX values

#define	HEC_NUMCOLORS	11
	DWORD	HelpExportColors[HEC_NUMCOLORS];	// HSL color for HTML background, normal text, h1, h2, h3
#define	HEC_BACK		0
#define	HEC_TEXT		1
#define	HEC_H1			2
#define	HEC_H2			3
#define	HEC_H3			4
#define	HEC_CONTBACK	5
#define	HEC_CONTTEXT	6
#define	HEC_CONTLINK	7
#define	HEC_HEADBACK	8
#define	HEC_HEADTEXT	9
#define	HEC_HEADFRAME	10


#define	HEFONT_NUM	4
	DWORD	HelpExportFonts[HEFONT_NUM];	// font for HTML export
#define	HEFONT_TEXT	0
#define	HEFONT_H1	1
#define	HEFONT_H2	2
#define	HEFONT_H3	3

	DWORD	HelpExportWidth;// width of help text in 'em's (0 means default full-width)
	DWORD	HelpExportLmargin;// width of help left margin of help text in 'em's (0 means default, no extra margin)
	DWORD	HelpExportContentWidth;		// width of CONTENT BAR in ems.
	DWORD	HelpExportContentSpacing;	// extra vertical spacing between CONTENT BAR entries in 1/10th ems.
	DWORD	HelpExportRule1Top, HelpExportRule1Bot;	// spacing above/below <hr> BEFORE section heading
	DWORD	HelpExportRule2Top, HelpExportRule2Bot;	// spacing above/below <hr> AFTER  section heading
	
	
//------------- GAME SPECIFIC ----------------
	DWORD	CurButtNum;		// PLAYERnTYPEn number of current puzzle type
	DWORD	CurButtPuzzNum[24];	// CurPuzzNum for each puzzle type
	DWORD	CurPuzzNum;		// dynamic, not stored in .ini current puzzle number for current map
	
	DWORD	SolvedTotal;	// for special win display
	
} PLAYER;

#endif
