#import "kintMAIN.h"
#import "kintPMACH.h"
#include "spng.h"
#include <math.h>

extern DWORD	PCpreJSR;	// in case of crashes, gets us back one JSR from where the crash occurred. Might help.

FILE *KMZ_open(const char* fname, char* flags);
int KMZ_close(void *file);

QWORD	LineStyle=0xFFFFFFFFFFFFFFFF;

KBM		*pKBMD;
KLONG	dWidth, dHeight, dStepX, dStepY, dClipX, dClipY, dClipW, dClipH;
KLONG	hKBMD=-1;
BYTE	*pARGBD;

KBM		*pKBMS;
KLONG	sWidth, sHeight, sStepX, sStepY, sClipX, sClipY, sClipW, sClipH;
KLONG	hKBMS=-1;
BYTE	*pARGBS;

SKBM	SysFontBM;

#define	SCREEN_ARGB	1		// set to 1 for 3-byte (no alpha channel on SCREEN (0) and FONT (-2) bitmaps

///=================================
BOOL CreateSBM(SKBM *pX, WORD w, WORD h) {

#if SCREEN_ARGB
	pX->linelen = 4*w;
#else
	pX->linelen = 4*((w*3+3)/4);
#endif
	pX->w = w;
	pX->h = h;
	pX->clipX = pX->clipY = 0;
	pX->clipW = w;
	pX->clipH = h;

    pX->linelen = 4*w;    // force line of pixels to take even number of DWORDS
    pX->pBits = (BYTE *)malloc(pX->linelen*h);
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    pX->cgImage = NULL;
    pX->bitmapContext = CGBitmapContextCreate(pX->pBits, w, h, 8, pX->linelen, colorSpace, (CGBitmapInfo)kCGImageAlphaPremultipliedLast);
    pX->w = w;
    pX->h = h;
    CGColorSpaceRelease(colorSpace);
    pX->clipX = 0;
    pX->clipY = 0;
    pX->clipW = w;
    pX->clipH = h;

	return TRUE;;
}

///=================================
/// Returns the AllocHeap handle to the KBM structure
DWORD CreateBM(KLONG w, KLONG h) {

	KBM			*pX;
	DWORD		hHP, hBM;

#if KDEBUG
{
DWORD    hFile;
sprintf(ScratBuf,"CreateBM ctl(%d,%d)\n", w, h);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	hHP = AllocHeap(sizeof(KBM));
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"CreateBM data\n");
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	hBM = AllocHeap(4*w*h);
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"CreateBM ctl=%d data=%d\n",hHP,hBM);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	pX = (KBM*)GetHeapPtr(hHP);
		pX->linelen = (KLONG)4*w;
		pX->w = w;
		pX->h = h;
		pX->clipX = pX->clipY = 0;
		pX->clipW = w;
		pX->clipH = h;
		pX->hBM = hBM;
	RelHeapPtr(hHP);

#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"CreateBM return %d\n",hHP);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	return hHP;
}

///=================================
void DeleteSBM(SKBM *pX) {

//	if( pX->cgImage != NULL ) DeleteObject(pX->hBitmap);
}

///=================================
void DeleteBM(DWORD hHP) {

	KBM			*pX;

	if( hHP ) {
		pX = (KBM*)GetHeapPtr(hHP);
		RelHeapPtr(hHP);
		if( pX->hBM ) FreeHeap(pX->hBM);
		FreeHeap(hHP);
	}
}
KLONG    dirtyX1=30000, dirtyY1, dirtyX2=-1, dirtyY2;

//**********************************************************************
void InvalidateScreen(KLONG x, KLONG y, KLONG w, KLONG h) {
    
    if (windowReady ) {
        if( w<0 ) {x += w; w = -w;}
        if( h<0 ) {y += h; h = -h;}
        if( x<dirtyX1 ) dirtyX1 = x;
        if( y<dirtyY1 ) dirtyY1 = y;
        if( x+w>dirtyX2 ) dirtyX2 = x+w;
        if( y+h>dirtyY2 ) dirtyY2 = y+h;
    }
}

///=================================================================
void StartDST(DWORD hBM) {

	if( hKBMD!=0xFFFFFFFF ) {
		SystemMessage("StartDST() already started");
		exit(1);
	}
	hKBMD = hBM;
	if( hBM==0 ) {	/// drawing to SCREEN bitmap
		pARGBD = scrBM.pBits;
		dWidth = scrBM.w;
		dHeight= scrBM.h;
#if SCREEN_ARGB
		dStepX = 4;
#else
		dStepX = 3;
#endif
		dStepY = scrBM.linelen;
		dClipX = scrBM.clipX;
		dClipY = scrBM.clipY;
		dClipW = scrBM.clipW;
		dClipH = scrBM.clipH;
	} else if( hBM==-2 ) {	/// drawing to SysFontBM bitmap
		pARGBD = SysFontBM.pBits;
		dWidth = SysFontBM.w;
		dHeight= SysFontBM.h;
#if SCREEN_ARGB
		dStepX = 4;
#else
		dStepX = 3;
#endif
		dStepY = SysFontBM.linelen;
		dClipX = SysFontBM.clipX;
		dClipY = SysFontBM.clipY;
		dClipW = SysFontBM.clipW;
		dClipH = SysFontBM.clipH;
	} else {		/// drawing to memory bitmap
		pKBMD  = (KBM*)GetHeapPtr(hBM);
		if( pKBMD->hBM==0 ) SystemMessage("StartDST: hBM==0");
		pARGBD = (BYTE*)GetHeapPtr(pKBMD->hBM);
			dWidth = pKBMD->w;
			dHeight= pKBMD->h;
			dStepX = 4;
			dStepY = pKBMD->linelen;
			dClipX = pKBMD->clipX;
			dClipY = pKBMD->clipY;
			dClipW = pKBMD->clipW;
			dClipH = pKBMD->clipH;
	}
}

///=================================================================
void EndDST(void) {

	if( hKBMD>0 ) {	/// drawing to memory bitmap
		RelHeapPtr(pKBMD->hBM);
		RelHeapPtr(hKBMD);
	}
	hKBMD = 0xFFFFFFFF;
}

///=================================================================
void StartSRC(DWORD hBM) {

	if( hKBMS!=0xFFFFFFFF ) {
		SystemMessage("StartSRC() already started");
		exit(1);
	}
	hKBMS = hBM;
	if( hBM==0 ) {
		pARGBS = scrBM.pBits;
		sWidth = scrBM.w;
		sHeight= scrBM.h;
#if SCREEN_ARGB
		sStepX = 4;
#else
		sStepX = 3;
#endif
		sStepY = scrBM.linelen;
		sClipX = scrBM.clipX;
		sClipY = scrBM.clipY;
		sClipW = scrBM.clipW;
		sClipH = scrBM.clipH;
	} else if( hBM==-2 ) {
		pARGBS = SysFontBM.pBits;
		sWidth = SysFontBM.w;
		sHeight= SysFontBM.h;
#if SCREEN_ARGB
		sStepX = 4;
#else
		sStepX = 3;
#endif
		sStepY = SysFontBM.linelen;
		sClipX = SysFontBM.clipX;
		sClipY = SysFontBM.clipY;
		sClipW = SysFontBM.clipW;
		sClipH = SysFontBM.clipH;
	} else {
		pKBMS  = (KBM*)GetHeapPtr(hBM);
		if( pKBMS->hBM==0 ) SystemMessage("StartSRC: hBM==0");
		pARGBS = (BYTE*)GetHeapPtr(pKBMS->hBM);
			sWidth = pKBMS->w;
			sHeight= pKBMS->h;
			sStepX = 4;
			sStepY = pKBMS->linelen;
			sClipX = pKBMS->clipX;
			sClipY = pKBMS->clipY;
			sClipW = pKBMS->clipW;
			sClipH = pKBMS->clipH;
	}
}

///=================================================================
void EndSRC(void) {

	if( hKBMS>0 ) {	/// drawing to memory bitmap
		RelHeapPtr(pKBMS->hBM);
		RelHeapPtr(hKBMS);
	}
	hKBMS = 0xFFFFFFFF;
}

void StartSRCDST(DWORD hBMS, DWORD hBMD) {

	StartDST(hBMD);
	if( hBMS==hBMD ) {
		hKBMS = hBMS;
		pKBMS = pKBMD;
		sStepX = dStepX;
		sStepY = dStepY;
		sClipX = dClipX;
		sClipY = dClipY;
		sClipW = dClipW;
		sClipH = dClipH;
		pARGBS = pARGBD;
	} else StartSRC(hBMS);
}

void EndSRCDST(void) {

	if( hKBMS==hKBMD ) {
		EndDST();
		hKBMS = 0xFFFFFFFF;
	} else {
		EndSRC();
		EndDST();
	}
}

///=================================================================
/// Goes through bitmap 'hBM' rectangle (x,y) to (x+w,y+h) and sets the ALPHA-channel to 'alpha'
/// for all pixels who's RGB components match 'clr'.
/// If qlr is 0xFFFFFFFFFFFFFFFF, then ALL pixels are modified, and 'clr' is ignored
QWORD SetAlpha(DWORD pArgs) {

	BYTE	*pB;
	KLONG	x, y, w, h;
	DWORD	rgb, crgb;
	DWORD	hBM, clr, alpha;
	KLONG	tx, ty;
	QWORD	qlr, cnt;

	hBM  = GETMEMD(pArgs); pArgs += 4;
	x    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	y    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	w    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	h    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	qlr  = GETMEMQ(pArgs); pArgs += 8;		/// pixel RGB value 0x00000000-0x00FFFFFF
	alpha= (GETMEMD(pArgs)<<24);			/// alpha level 0-255
#if SCREEN_ARGB
#else
		if( (KLONG)hBM<=0 ) return 0;
#endif
	cnt = 0;
	StartSRC(hBM);

		if( x<sClipX ) x = sClipX;
		if( x+w>sClipX+sClipW ) w = sClipX+sClipW-x;
		if( y<sClipY ) y = sClipY;
		if( y+h>sClipY+sClipH ) h = sClipY+sClipH-y;

		if( qlr==0xFFFFFFFFFFFFFFFF ) {		/// set ALL pixels to this ALPHA
			for(ty=y; ty<y+h; ty++) {
				for(tx=x, pB = pARGBS + sStepY*ty + sStepX*tx; tx<x+w; tx++, pB+=sStepX) {
					if( (*(DWORD*)pB & 0xFF000000) != alpha ) {
						*(DWORD*)pB = (*(DWORD*)pB & 0x00FFFFFF) | alpha;
						++cnt;
					}
				}
			}
		} else {							/// set only MATCHING pixels to this ALPHA
			crgb = (DWORD)(qlr & 0x00FFFFFF);
			crgb = (crgb & 0xFF00FF00) | ((crgb<<16)&0x00FF0000) | ((crgb>>16)&0x000000FF);
			clr  = crgb | alpha;
			for(ty=y; ty<y+h; ty++) {
				for(tx=x, pB = pARGBS + sStepY*ty + sStepX*tx; tx<x+w; tx++, pB+=sStepX) {
					rgb = *(DWORD*)pB & 0x00FFFFFF;
					if( rgb==crgb ) {
						*(DWORD*)pB = clr;
						++cnt;
					}
				}
			}
		}
		if( hBM==0 ) InvalidateScreen(x, y, w, h);

	EndSRC();
	return cnt;
}

///==================================================================
DWORD GetPixelBM(DWORD hBM, KLONG x, KLONG y) {

	DWORD	clr;

	StartSRC(hBM);

#if SCREEN_ARGB
#else
		if( (KLONG)hBM<=0 ) {
			if( x>=0 && x<sWidth && y>=0 && y<sHeight ) clr = (*(DWORD*)(pARGBS + y*sStepY + x*sStepX)) & 0x00FFFFFF;
			else clr = 0;
		} else
#endif
		if( x>=0 && x<sWidth && y>=0 && y<sHeight ) {
			clr = *(DWORD*)(pARGBS + y*sStepY + x*sStepX);
			clr = (clr & 0xFF00FF00) | ((clr>>16)&0x000000FF) | ((clr<<16)&0x00FF0000);
		} else clr = 0;

	EndSRC();
	return clr;
}

///==================================================================
void SetPixelBM(DWORD hBM, KLONG x, KLONG y, DWORD clr) {

	BYTE	*pDST, r, g, b, a, a1;

	StartSRC(hBM);

		if( x>=0 && x<sWidth && y>=0 && y<sHeight ) {
			a = (BYTE)(clr >> 24);
			a1= 255-a;
			r = (BYTE)(clr >> 16);
			g = (BYTE)(clr >> 8);
			b = (BYTE)(clr);
			pDST = pARGBS + y*sStepY + x*sStepX;
				*pDST = (BYTE)((((WORD)a1*(WORD)(* pDST   ))/255) + ((WORD)r*(WORD)a)/255);
			*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)g*(WORD)a)/255);
			*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)b*(WORD)a)/255);
#if SCREEN_ARGB
#else
			if( (KLONG)hBM>0 )
#endif
			*(pDST+3) = (BYTE)((((WORD)a1*(WORD)(*(pDST+3)))/255) + ((WORD)a*(WORD)a)/255);
		}
		if( hBM==0 ) InvalidateScreen(x, y, 1, 1);

	EndSRC();
}

typedef struct tagBITMAPFILEHEADER { // bmfh
    WORD    bfType;
    DWORD   bfSize;
    WORD    bfReserved1;
    WORD    bfReserved2;
    DWORD   bfOffBits;
} __attribute__((packed)) BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{ // bmih
    DWORD  biSize;
    int32_t   biWidth;
    int32_t   biHeight;
    WORD   biPlanes;
    WORD   biBitCount;
    DWORD  biCompression;
    DWORD  biSizeImage;
    int32_t   biXPelsPerMeter;
    int32_t   biYPelsPerMeter;
    DWORD  biClrUsed;
    DWORD  biClrImportant;
} __attribute__((packed)) BITMAPINFOHEADER;

typedef struct tagRGBQUAD { // rgbq
    BYTE    rgbBlue;
    BYTE    rgbGreen;
    BYTE    rgbRed;
    BYTE    rgbReserved;
} __attribute__((packed)) RGBQUAD;

typedef struct tagBITMAPINFO { // bmi
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD          bmiColors[1];
} __attribute__((packed)) BITMAPINFO;

///==================================================================
/// Load a MS Windows/OS2 .BMP format file into a device-independent
/// memory array.  BMP file must be a 24-bit (RGB pixel values), and can
/// NOT be compressed, and is assumed to be in the "bottom-up" format
/// (bmiHeader.biHeight > 0).
DWORD LoadBMP(char *fname) {

/// NOTE: If porting to non-MS Windows platform, duplicate the Windows
///	typedefs for BITMAPFILEHEADER and BITMAPINFO at top of this file

	BITMAPFILEHEADER	bmfh;
	BITMAPINFO			*pbmi;
	DWORD				dwInfoSize, dwBitsSize, BytesRead;
	DWORD				hHP;
//	KBM					*pHP;
    BYTE				*pB, *pL, *pO;//, *pBITS;
	DWORD				hFile;
	KLONG				x, y, w, h, r, g, b;
	DWORD				linlen;

	/// Open the file: read access, prohibit write access
	if( -1==(hFile = Kopen(fname, O_RDBIN)) ) return 0;

	/// Read in the BITMAPFILEHEADER
	BytesRead = Kread(hFile, (BYTE *)&bmfh, sizeof(BITMAPFILEHEADER));
	if( BytesRead!=sizeof(BITMAPFILEHEADER) || (bmfh.bfType != *(WORD *)"BM") ) {
		Kclose(hFile);
		return 0;
	}
	/// Allocate memory for the information structure & read it in
	dwInfoSize = bmfh.bfOffBits - sizeof (BITMAPFILEHEADER);

	if( NULL==(pbmi=(struct tagBITMAPINFO *)malloc(dwInfoSize)) ) {
		Kclose(hFile);
		return 0;
	}
	BytesRead = Kread(hFile, (BYTE *)pbmi, dwInfoSize);
	if( BytesRead != dwInfoSize )	{
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	if( pbmi->bmiHeader.biBitCount<24 || pbmi->bmiHeader.biCompression!=0 ) {	// don't handle < 24-bit color bitmaps OR compressed bitmaps
		Kclose(hFile);
		free(pbmi);
		return 0;
	}

	/// Create the memory bitmap
	w = pbmi->bmiHeader.biWidth;
	h = pbmi->bmiHeader.biHeight;
	if( 0==(hHP=(KLONG)CreateBM((KLONG)w, (KLONG)h)) ) {
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	/// allocate temp storage for .BMP file RGB/palette-index bytes
	dwBitsSize = bmfh.bfSize - bmfh.bfOffBits;
	if( NULL==(pB=(BYTE *)malloc(dwBitsSize)) ) {
		DeleteBM(hHP);
		Kclose(hFile);
		free(pbmi);
		return 0;
	}
	BytesRead = Kread(hFile, pB, dwBitsSize);
	Kclose(hFile);
	if( BytesRead != dwBitsSize ) {
		free(pB);
		DeleteBM(hHP);
		free(pbmi);
		return 0;
	}

	if( pbmi->bmiHeader.biSizeImage==0 ) pbmi->bmiHeader.biSizeImage = 4*((w*3+3)/4)*h;
	linlen = pbmi->bmiHeader.biSizeImage / h;

	StartSRC(hHP);

if( pbmi->bmiHeader.biBitCount==24 ) {
		for(y=0; y<h; y++) {
			pL = pB + linlen*(h-y-1);	// common .BMPs are inverted vertically
			pO = pARGBS + sStepY*y;
			for(x=0; x<w; x++) {
				r = *pL++;
				g = *pL++;
				b = *pL++;

				*pO++ = r;
				*pO++ = g;
				*pO++ = b;
				*pO++ = 0xFF;	/// alpha
			}
		}
} else {
		for(y=0; y<h; y++) {
			pL = pB + linlen*(h-y-1);	// common .BMPs are inverted vertically
			pO = pARGBS + sStepY*y;
			memcpy(pO, pL, 4*w);
		}
}
	EndSRC();
	free(pB);
	free(pbmi);

	return hHP;
}

///=================================================================
DWORD SaveBMP(DWORD hBM, char *fname) {

	BITMAPFILEHEADER	bmfh;
	BITMAPINFOHEADER	bmih;
	BYTE				*pB;
	DWORD				hFile, retval=FALSE, isize;
	KLONG				h;

	if( -1!=(hFile=Kopen(fname, O_WRBINNEW)) ) {
		StartSRC(hBM);

			isize = sStepY*sHeight;
			bmfh.bfType = 0x4D42;	// 'BM'
			bmfh.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFO)+isize-4;
			bmfh.bfReserved1 = bmfh.bfReserved2 = 0;
			bmfh.bfOffBits = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);

			bmih.biSize = sizeof(BITMAPINFOHEADER);
			bmih.biWidth = sWidth;
			bmih.biHeight = sHeight;
			bmih.biPlanes = 1;
#if SCREEN_ARGB
			bmih.biBitCount = 32;
#else
			bmih.biBitCount = ((KLONG)hBM<=0) ? 24 : 32;
#endif
			bmih.biCompression = 0;
			bmih.biSizeImage = isize;
			bmih.biXPelsPerMeter = 2947;
			bmih.biYPelsPerMeter = 2947;
			bmih.biClrUsed = 0;
			bmih.biClrImportant = 0;

			if( sizeof(BITMAPFILEHEADER)==Kwrite(hFile, (BYTE *)&bmfh, sizeof(BITMAPFILEHEADER)) ) {
				if( sizeof(BITMAPINFOHEADER)==Kwrite(hFile, (BYTE *)&bmih, sizeof(BITMAPINFOHEADER)) ) {

					pB = pARGBS+(sHeight-1)*sStepY;
					for(h=0; h<sHeight; h++) {
						if( sStepY!=Kwrite(hFile, pB, sStepY) ) break;
						pB -= sStepY;
					}
					if( h>=sHeight ) retval = TRUE;
				}
			}

		EndSRC();
		Kclose(hFile);
	}
	return retval;
}

///=================================================================
/// Block-transfer a region from one bitmap to another.
/// If either bitmap handle is 0, redirect it to the screen bitmap
/// In the source (masked) bitmap, the pBits memory block has 4 bytes
/// per pixel containing the ARGB information to be forced into the
/// destination. The SRC is mixed into the DST according to the SRC
/// ALPHA-channel value for each pixel.
///=================================================================
void DrawBM(DWORD hDST, KLONG xd, KLONG yd, DWORD hSRC, KLONG xs, KLONG ys, KLONG ws, KLONG hs, KLONG invert, KLONG rep2x, KLONG rep2y) {

	BYTE	*pDST, *pSRC, a, a1;
    KLONG	x, y, limx, limy, resetx, resety;

	StartSRCDST(hSRC, hDST);
		resetx = xd;
		resety = yd;
		if( invert ) {
			do {
				limy = yd+hs;
				if( rep2y ) limy = min(limy, rep2y);
				for(y=yd; y<limy; y++) {
					if( y>=dClipY && y<dClipY+dClipH && (rep2y==0 || y<rep2y) ) {
						xd = resetx;
						do {
							pSRC = pARGBS + sStepY*(ys+y-yd) + sStepX*xs;
							pDST = pARGBD + dStepY*y + dStepX*xd;
							limx = xd+ws;
							if( rep2x ) limx = min(limx, rep2x);
							for(x=xd; x<limx; x++) {
								if( x>=dClipX && x<dClipX+dClipW && (rep2x==0 || x<rep2x) ) {
									a = *(pSRC+3);
									if( a ) {
										a1= 255-a;
											*pDST = (BYTE)((((WORD)a1*(WORD)(*(pDST+0)))/255) + ((WORD)(0xFF ^ (*(pSRC+0)))*(WORD)a)/255);
										*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)(0xFF ^ (*(pSRC+1)))*(WORD)a)/255);
										*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)(0xFF ^ (*(pSRC+2)))*(WORD)a)/255);
					#if SCREEN_ARGB
										*(pDST+3) = min(255, (WORD)*(pSRC+3)+(WORD)*(pDST+3));
					#else
										if( (KLONG)hDST>0 ) {
											if( (KLONG)hSRC>0 )
															*(pDST+3) = min(255, (WORD)*(pSRC+3)+(WORD)*(pDST+3));
											else *(pDST+3) = 0xFF;
										}
					#endif
									}
								}
								pSRC += sStepX;
								pDST += dStepX;		// move dest ptr to next pixel
							}
							xd += ws;
						} while( rep2x && xs<rep2x );
					}
				}
				yd += hs;
			} while( rep2y && yd<rep2y );
		} else {
			do {
				limy = yd+hs;
				if( rep2y ) limy = min(limy, rep2y);
				for(y=yd; y<limy; y++) {
					if( (y>=dClipY && y<dClipY+dClipH && (rep2y==0 || y<rep2y)) ) {
						xd = resetx;
						do {
							pSRC = pARGBS + sStepY*(ys+y-yd) + sStepX*xs;
							pDST = pARGBD + dStepY*y + dStepX*xd;
							limx = xd+ws;
							if( rep2x ) limx = min(limx, rep2x);
							for(x=xd; x<limx; x++) {
								if( (x>=dClipX && x<dClipX+dClipW && (rep2x==0 || x<rep2x)) ) {
									a = *(pSRC+3);
									if( a ) {
										a1= 255-a;
											*pDST = (BYTE)((((WORD)a1*(WORD)(*(pDST+0)))/255) + ((WORD)(*(pSRC+0))*(WORD)a)/255);
										*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)(*(pSRC+1))*(WORD)a)/255);
										*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)(*(pSRC+2))*(WORD)a)/255);
					#if SCREEN_ARGB
										*(pDST+3) = (BYTE)min((WORD)255, (WORD)*(pDST+3)+(WORD)*(pSRC+3));
					#else
										if( (KLONG)hDST>0 ) {
											if( (KLONG)hSRC>0 )
															*(pDST+3) = min(255, (WORD)*(pSRC+3)+(WORD)*(pDST+3));
											else *(pDST+3) = 0xFF;
										}
					#endif
									}
								}
								pSRC += sStepX;
								pDST += dStepX;		// move dest ptr to next pixel
							}
							xd = limx;
						} while( rep2x && xd<rep2x );
					}
				}
				yd = limy;
			} while( rep2y && yd<rep2y );
		}
		if( hDST==0 ) InvalidateScreen(resetx, resety, rep2x?(rep2x-resetx):ws, rep2y?(rep2y-resety):hs);
bail:
	EndSRCDST();
}

#if 0
This is the original floating-point version of StretchBM, now replaced by the integral version below
///=====================================================================================================
/// Stretch or compress the source bitmap into the destination bitmap
/// averaging source pixels to generate destination pixels (smooth resize)
void StretchBM(DWORD pArgs) {

	DWORD hBMD, hBMS;
    KLONG xd, yd, wd, hd, xs, ys, ws, hs, rep2x, rep2y;

	BYTE	*pD, *pS;
	WORD	a, a1, r, g, b;
    KLONG	x, y, resetx, resety;
	double	xsdoub=(double)xs, ysdoub=(double)ys, sxl, sxr, syt, syb;
	double	tmpa, xrat=(double)ws/(double)wd, yrat=(double)hs/(double)hd, psy, psx, pa, pr, pg, pb, tsa, psa;

	hBMD = GETMEMD(pArgs); pArgs += 4;
	xd   = GETMEMD(pArgs); pArgs += 4;
	yd   = GETMEMD(pArgs); pArgs += 4;
	wd   = GETMEMD(pArgs); pArgs += 4;
	hd   = GETMEMD(pArgs); pArgs += 4;
	hBMS = GETMEMD(pArgs); pArgs += 4;
	xs   = GETMEMD(pArgs); pArgs += 4;
	ys   = GETMEMD(pArgs); pArgs += 4;
	ws   = GETMEMD(pArgs); pArgs += 4;
	hs   = GETMEMD(pArgs); pArgs += 4;
	rep2x= GETMEMD(pArgs); pArgs += 4;
	rep2y= GETMEMD(pArgs); pArgs += 4;

	xsdoub = (double)xs;
	ysdoub = (double)ys;
	xrat   = (double)ws/(double)wd;
	yrat   = (double)hs/(double)hd;

	if( wd<=0 || hd<=0 || ws<=0 || hs<=0 ) return;
	if( wd==ws && hd==hs ) {
		DrawBM(hBMD, xd, yd, hBMS, xs, ys, ws, hs, FALSE, rep2x, rep2y);
		return;
	}

	StartSRCDST(hBMS, hBMD);

		resetx = xd;
		resety = yd;
		do {
			/// for each DST pixel
			for(y=yd; y<yd+hd; y++) {
				if( y>=dClipY && y<dClipY+dClipH && (rep2y==0 || y<rep2y) ) {
					syt = ysdoub + (double)(y-yd)*yrat;	/// Src Y Top
					syb = syt + yrat;					/// Src Y Bottom
					xd = resetx;
					do {
						for(x=xd; x<xd+wd; x++) {
							if( x>=dClipX && x<dClipX+dClipW && (rep2x==0 || x<rep2x) ) {
								/// calculate scaled top-left and bottom-right coordinates in SRC bitmap of DST 'pixel'
								sxl = xsdoub + (double)(x-xd)*xrat;	/// Src X Left
								sxr = sxl + xrat;					/// Src X Right
								/// for each partial SRC pixel included in the DST pixel, calculate its proportional area (psa) in the DST pixel
								/// and add the (psa/tsa)*SRCpixelCLR to the DSTpixelCLR (pa,pr,pg,pb).
								pa = pr = pg = pb = 0.0;			/// Proportional ARGB values, running totals for this DST pixel
								tsa = (sxr-sxl)*(syb-syt);	/// Total Source Area
								for(psy=syt; psy<syb; psy=floor(psy+1.0)) {		/// for partial Y sub-pixels
									for(psx=sxl; psx<sxr; psx=floor(psx+1.0)) {		/// for partial X sub-pixels
										pS = pARGBS + sStepY*(KLONG)psy + sStepX*(KLONG)psx;
										/// calculate the portion of the DST pixel that this partial SRC pixel covers
										psa = (min(floor(psx+1.0), sxr)-psx)*(min(floor(psy+1.0), syb)-psy)/tsa;
										/// now add that fraction of the total DST pixel's "SRC pixel size"
										tmpa = psa*(double)(*(pS+3));
										if( (KLONG)tmpa ) {
				#if SCREEN_ARGB
											pa += psa*(double)(*(pS+3));
				#else
											if( (KLONG)hBMS>0 ) pa += psa*(double)(*(pS+3));
											else pa = 255;
				#endif
											pb += psa*(double)(*(pS+2));
											pg += psa*(double)(*(pS+1));
											pr += psa*(double)(*(pS+0));
										}
									}
								}
								/// we now have the "color of the SRC pixel(s)" to blend into the DST pixel
								pD = pARGBD + dStepY*y + dStepX*x;

								a = min(255, (KLONG)pa);
								r = min(255, (KLONG)pr);
								g = min(255, (KLONG)pg);
								b = min(255, (KLONG)pb);
								if( a ) {
/*									if( a>200 ) {
										*pD = (BYTE)r;
										*(pD+1) = (BYTE)g;
										*(pD+2) = (BYTE)b;
										*(pD+3) = 255;
									} else {
*/
									a1= 255-a;
										*pD = (BYTE)(((a1*(* pD   ))/255) + ((r)*a)/255);
									*(pD+1) = (BYTE)(((a1*(*(pD+1)))/255) + ((g)*a)/255);
									*(pD+2) = (BYTE)(((a1*(*(pD+2)))/255) + ((b)*a)/255);
				#if SCREEN_ARGB
				#else
									if( (KLONG)hBMD>0 )
				#endif
									*(pD+3) = (BYTE)min(255,(((a1*(*(pD+3)))/255) + ((a)*a)/255));
//									}

								}
							}
						}
						xd += wd;
					} while( rep2x && xd<rep2x );
				}
			}
			yd += hd;
		} while( rep2y && yd<rep2y );
	EndSRCDST();
	if( hBMD==0 ) InvalidateScreen(resetx, resety, rep2x?(rep2x-resetx):wd, rep2y?(rep2y-resety):hd);
}
#endif

/// FSH is the "Fractional SHift", the distance the integral coordinates are shifted left to allow bits for holding "partial pixel coordinates"
/// defining all the 'long' (32-bit) variables as "long long" (64-bit) and increasing FSH to 16 has no apparent affect on the 'artifacts' that
/// we have to account for, caused by rounding errors, when all sub-pixels have the same a,r,g,b component values.
#define	FSH	5

///=====================================================================================================
/// Compress the source bitmap into the destination bitmap
/// averaging source pixels to generate destination pixels (smooth resize)
void StretchBM(DWORD pArgs) {

	BYTE	*pD, *pS;
	WORD	a, a1, r, g, b;
    DWORD	hBMD, hBMS;//, clr, clra;
    KLONG	x, y, resetx, resety, xd, yd, wd, hd, xs, ys, ws, hs, rep2x, rep2y;
    KLONG	t16curxs, t16curys, t16endxs, t16endys, suma, sumr, sumg, sumb;
    KLONG	t16SRCarea, p16SRCarea, t16DSTarea, alla, allr, allg, allb;

	hBMD = GETMEMD(pArgs); pArgs += 4;
	xd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	yd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	wd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hBMS = GETMEMD(pArgs); pArgs += 4;
	xs   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	ys   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	ws   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hs   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	rep2x= (KLONG)GETMEMD(pArgs); pArgs += 4;
//	*(DWORD*)(mainKE+0x1E0) =
	rep2y= (KLONG)GETMEMD(pArgs); pArgs += 4;

/*
{
DWORD	hFile;
char	bbuf[128];
sprintf(bbuf,"====== StretchBM(%d, %d,%d,%d,%d, %d, %d,%d,%d,%d, %d,%d) PC=0x%8.8lX PCpreJSR=0x%8.8lX\n", hBMD, xd, yd, wd, hd, hBMS, xs, ys, ws, hs, rep2x, rep2y, *(DWORD*)(mainKE+0x088), PCpreJSR);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, bbuf, strlen(bbuf));
Kclose(hFile);
}
*/

	if( wd<=0 || hd<=0 || ws<=0 || hs<=0 ) return;
	if( wd==ws && hd==hs ) {
		DrawBM(hBMD, xd, yd, hBMS, xs, ys, ws, hs, FALSE, rep2x, rep2y);
		return;
	}

	StartSRCDST(hBMS, hBMD);

		resetx = xd;
		resety = yd;
		t16DSTarea = hd*wd;
		do {
			t16endys = ys<<FSH;
			/// for each DST pixel
			for(y=yd; y<yd+hd; y++) {
				if( y>=dClipY && y<dClipY+dClipH && (rep2y==0 || y<rep2y) ) {
					xd = resetx;
					do {
						t16endxs = xs<<FSH;
						for(x=xd; x<xd+wd; x++) {
							if( x>=dClipX && x<dClipX+dClipW && (rep2x==0 || x<rep2x) ) {
								suma = sumr = sumg = sumb = 0;
								pD = pARGBD + dStepY*y + dStepX*x;
								t16curys = (ys<<FSH) + (y-yd)*(hs<<FSH)/hd;
								t16endys = (ys<<FSH) + (y-yd+1)*(hs<<FSH)/hd;
								t16SRCarea = (t16endys-t16curys)*(	(x-xd+1)*(ws<<FSH)/wd	- (x-xd)*(ws<<FSH)/wd	);

								t16curxs = (xs<<FSH) + (x-xd)*(ws<<FSH)/wd;			// t16endxs;
								pS = pARGBS + sStepY*(t16curys>>FSH) + sStepX*(t16curxs>>FSH);
								allr = *(pS+0);
								allg = *(pS+1);
								allb = *(pS+2);
								alla = *(pS+3);

								for(; t16curys<t16endys; t16curys = (t16curys+(1<<FSH)) & ~((1<<FSH)-1)) {
									t16curxs = (xs<<FSH) + (x-xd)*(ws<<FSH)/wd;			// t16endxs;
									t16endxs = (xs<<FSH) + (x-xd+1)*(ws<<FSH)/wd;		//(t16curxs + (1<<FSH)) & ~((1<<FSH)-1);
									for(; t16curxs<t16endxs; t16curxs = (t16curxs+(1<<FSH)) & ~((1<<FSH)-1)) {
										pS = pARGBS + sStepY*(t16curys>>FSH) + sStepX*(t16curxs>>FSH);
										p16SRCarea = (min((t16curxs+(1<<FSH)) & (~((1<<FSH)-1)), t16endxs)-t16curxs) * (min((t16curys+(1<<FSH)) & (~((1<<FSH)-1)), t16endys)-t16curys);
										if( p16SRCarea > t16SRCarea ) p16SRCarea = t16SRCarea;
										if( p16SRCarea < 0 ) p16SRCarea = 0;
/// this line is here because, without it, something is causing a write to bad memory. Compiler optimization problem? Don't know, will look later.
///++*(DWORD*)(mainKE+0x1E0);
										/// if not all src pixels have the same (a r g b), we have to calculate the 'average' of the ones that are different
										/// so when a difference is detected, we set that one to 256 to indicate to use the average
										r = (WORD)(*(pS+0));
										g = (WORD)(*(pS+1));
										b = (WORD)(*(pS+2));
										a = (WORD)(*(pS+3));

										if( allr!=r ) allr = 256;
										if( allg!=g ) allg = 256;
										if( allb!=b ) allb = 256;
										if( alla!=a ) alla = 256;

										sumr += ((p16SRCarea*(KLONG)r/t16SRCarea));
										sumg += ((p16SRCarea*(KLONG)g/t16SRCarea));
										sumb += ((p16SRCarea*(KLONG)b/t16SRCarea));
				#if SCREEN_ARGB
										suma += ((p16SRCarea*(KLONG)a/t16SRCarea));
				#else
										if( (KLONG)hBMS>0 ) pa += psa*(double)(*(pS+3));
										else pa = 255;
				#endif
									}
								}
								/// we now have the "color of the SRC pixel(s)" to blend into the DST pixel
								r = (allr>255) ? min(255, sumr) : (BYTE)allr;	/// either the 'average' red or the original red
								g = (allg>255) ? min(255, sumg) : (BYTE)allg;	/// either the 'average' green or the original green
								b = (allb>255) ? min(255, sumb) : (BYTE)allb;	/// either the 'average' blue or the original blue
								a = (alla>255) ? min(255, suma) : (BYTE)alla;	/// either the 'average' alpha or the original alpha
								/// NOTE: the 'averages', since we're not using floating point math, can vary just enough to cause "shadow stripes", hence this kludge to use the original if all sub-pixels agree
								if( a ) {
									a1= 255-a;
										*pD = (BYTE)(((a1*(* pD   ))/255) + ((r)*a)/255);
									*(pD+1) = (BYTE)(((a1*(*(pD+1)))/255) + ((g)*a)/255);
									*(pD+2) = (BYTE)(((a1*(*(pD+2)))/255) + ((b)*a)/255);
				#if SCREEN_ARGB
				#else
									if( (KLONG)hBMD>0 )
				#endif
									*(pD+3) = (BYTE)min(255,(((a1*(*(pD+3)))/255) + ((a)*a)/255));
								}
							}
						}
						xd += wd;
					} while( rep2x && xd<rep2x );
				}
			}
			yd += hd;
		} while( rep2y && yd<rep2y );
	EndSRCDST();
	if( hBMD==0 ) InvalidateScreen(resetx, resety, rep2x?(rep2x-resetx):wd, rep2y?(rep2y-resety):hd);
}

///=====================================================================================================
/// Enlarge the source bitmap into the destination bitmap
/// stretching and averaging (sometimes) source pixels to generate destination pixels (smooth resize)

void EnlargeBM(DWORD pArgs) {

	DWORD hBMD, hBMS, clr;
    KLONG xd, yd, wd, hd, xs, ys, ws, hs, rep2x, rep2y;

	BYTE	*pD, *pS;
	WORD	a, a1, r, g, b;
    KLONG	x, y, resetx, resety;
    KLONG	t16curxs, t16curys, t16endxs, t16endys, suma, sumr, sumg, sumb;
    KLONG	t16SRCarea, p16SRCarea, t16DSTarea, tmp, alla=0;

	hBMD = GETMEMD(pArgs); pArgs += 4;
	xd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	yd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	wd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hd   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hBMS = GETMEMD(pArgs); pArgs += 4;
	xs   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	ys   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	ws   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hs   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	rep2x= (KLONG)GETMEMD(pArgs); pArgs += 4;
	rep2y= (KLONG)GETMEMD(pArgs); pArgs += 4;

	if( wd<=0 || hd<=0 || ws<=0 || hs<=0 ) return;
	if( wd==ws && hd==hs ) {
		DrawBM(hBMD, xd, yd, hBMS, xs, ys, ws, hs, FALSE, rep2x, rep2y);
		return;
	}

	StartSRCDST(hBMS, hBMD);

		resetx = xd;
		resety = yd;
		t16DSTarea = hd*wd;
		t16SRCarea = ((hs<<FSH)*(ws<<FSH))/t16DSTarea;
		do {
			t16endys = ys<<FSH;
			/// for each DST pixel
			for(y=yd; y<yd+hd; y++) {
				if( y>=dClipY && y<dClipY+dClipH && (rep2y==0 || y<rep2y) ) {
					xd = resetx;
					do {
						t16endxs = xs<<FSH;
						for(x=xd; x<xd+wd; x++) {
							if( x>=dClipX && x<dClipX+dClipW && (rep2x==0 || x<rep2x) ) {
								suma = sumr = sumg = sumb = 0;
								pD = pARGBD + dStepY*y + dStepX*x;

								t16curys = (ys<<FSH) + (y-yd)*(hs<<FSH)/hd;
								t16endys = (ys<<FSH) + (y-yd+1)*(hs<<FSH)/hd;
								t16curxs = (xs<<FSH) + (x-xd)*(ws<<FSH)/wd;			// t16endxs;
								clr = *(DWORD*)(pARGBS + sStepY*(t16curys>>FSH) + sStepX*(t16curxs>>FSH));
								tmp = 1;
								for(; tmp && t16curys<t16endys; t16curys = (t16curys+(1<<FSH)) & ~((1<<FSH)-1)) {
									t16curxs = (xs<<FSH) + (x-xd)*(ws<<FSH)/wd;			// t16endxs;
									t16endxs = (xs<<FSH) + (x-xd+1)*(ws<<FSH)/wd;		//(t16curxs + (1<<FSH)) & ~((1<<FSH)-1);
									for(; tmp && t16curxs<t16endxs; t16curxs = (t16curxs+(1<<FSH)) & ~((1<<FSH)-1)) {
										if( clr != *(DWORD*)(pARGBS + sStepY*(t16curys>>FSH) + sStepX*(t16curxs>>FSH)) ) tmp = 0;
									}
								}
								/// if all partial src pixels are the same color, then the target IS that color
								if( tmp ) {
									a = (clr>>24);
									r = clr & 0x0FF;
									g = (clr>>8)&0x0FF;
									b = (clr>>16)&0x0FF;
									if( a ) {
										a1= 255-a;
											*pD = (BYTE)(((a1*(* pD   ))/255) + ((r)*a)/255);
										*(pD+1) = (BYTE)(((a1*(*(pD+1)))/255) + ((g)*a)/255);
										*(pD+2) = (BYTE)(((a1*(*(pD+2)))/255) + ((b)*a)/255);
					#if SCREEN_ARGB
					#else
										if( (KLONG)hBMD>0 )
					#endif
										*(pD+3) = (BYTE)min(255,(((a1*(*(pD+3)))/255) + ((a)*a)/255));
									}
								/// else, we need to come up with an 'average' color for the pixel
								} else {
									t16curys = (ys<<FSH) + (y-yd)*(hs<<FSH)/hd;
									t16endys = (ys<<FSH) + (y-yd+1)*(hs<<FSH)/hd;
			t16SRCarea = (t16endys-t16curys)*(	(x-xd+1)*(ws<<FSH)/wd	- (x-xd)*(ws<<FSH)/wd	);
									for(; t16curys<t16endys; t16curys = (t16curys+(1<<FSH)) & ~((1<<FSH)-1)) {
										t16curxs = (xs<<FSH) + (x-xd)*(ws<<FSH)/wd;			// t16endxs;
										t16endxs = (xs<<FSH) + (x-xd+1)*(ws<<FSH)/wd;		//(t16curxs + (1<<FSH)) & ~((1<<FSH)-1);
										alla = *(pARGBS + sStepY*(t16curys >> FSH) + sStepX*(t16curxs>>FSH)+3);
										for(; t16curxs<t16endxs; t16curxs = (t16curxs+(1<<FSH)) & ~((1<<FSH)-1)) {
											pS = pARGBS + sStepY*(t16curys>>FSH) + sStepX*(t16curxs>>FSH);
											p16SRCarea = (min((t16curxs+(1<<FSH)) & (~((1<<FSH)-1)), t16endxs)-t16curxs) * (min((t16curys+(1<<FSH)) & (~((1<<FSH)-1)), t16endys)-t16curys);
											if( p16SRCarea > t16SRCarea ) p16SRCarea = t16SRCarea;
											if( p16SRCarea < 0 ) p16SRCarea = 0;
											if( alla!=(KLONG)(*(pS+3)) ) alla = -1;	/// if not all src pixels have the same alpha, we have to calculate the 'average' alpha
				#if SCREEN_ARGB
											suma += ((p16SRCarea*(KLONG)(*(pS+3))/t16SRCarea));
				#else
											if( (KLONG)hBMS>0 ) pa += psa*(double)(*(pS+3));
											else pa = 255;
				#endif
											sumb += ((p16SRCarea*(KLONG)(*(pS+2))/t16SRCarea));
											sumg += ((p16SRCarea*(KLONG)(*(pS+1))/t16SRCarea));
											sumr += ((p16SRCarea*(KLONG)(*(pS+0))/t16SRCarea));
										}
									}
									/// we now have the "color of the SRC pixel(s)" to blend into the DST pixel
									a = (alla==-1) ? min(255, suma) : (BYTE)alla;	/// either the 'average' alpha or the original alpha
									/// NOTE: the 'average' alpha, since we're not using floating point math, can vary just enough to cause "shadow stripes", hence this kludge to use the original alpha if possible
									r = min(255, sumr);
									g = min(255, sumg);
									b = min(255, sumb);
									if( a ) {
										a1= 255-a;
											*pD = (BYTE)(((a1*(* pD   ))/255) + ((r)*a)/255);
										*(pD+1) = (BYTE)(((a1*(*(pD+1)))/255) + ((g)*a)/255);
										*(pD+2) = (BYTE)(((a1*(*(pD+2)))/255) + ((b)*a)/255);
					#if SCREEN_ARGB
					#else
										if( (KLONG)hBMD>0 )
					#endif
										*(pD+3) = (BYTE)min(255,(((a1*(*(pD+3)))/255) + ((a)*a)/255));
									}
								}
							}
						}
						xd += wd;
					} while( rep2x && xd<rep2x );
				}
			}
			yd += hd;
		} while( rep2y && yd<rep2y );
	EndSRCDST();
	if( hBMD==0 ) InvalidateScreen(resetx, resety, rep2x?(rep2x-resetx):wd, rep2y?(rep2y-resety):hd);
}

///=================================================================
void DrawLineBM(DWORD hBM, KLONG x1, KLONG y1, KLONG x2, KLONG y2, DWORD clr) {

	BYTE	*pDST, r, g, b, a, a1;
    KLONG	x, y, d=0, w, s=0;

	StartDST(hBM);
		a = (BYTE)(clr >> 24);
		a1= 255-a;
		r = (BYTE)(clr >> 16);
		g = (BYTE)(clr >> 8);
		b = (BYTE)(clr);
		if( x1==x2 ) {	/// VERTICAL line
			if( x1<dClipX || x1>=dClipX+dClipW ) goto bail;
			if( y1<y2 ) {
				d = 1;
				if( y1>=dClipY+dClipH || y2<dClipY ) goto bail;
				if( y1<dClipY ) y1 = dClipY;
				if( y2>=dClipY+dClipH ) y2 = dClipY+dClipH-1;
				w = y2-y1+1;
				s = dStepY;
			} else if( y1>y2 ) {
				d = -1;
				if( y2>=dClipY+dClipH || y1<dClipY ) goto bail;
				if( y2<dClipY ) y2 = dClipY;
				if( y1>=dClipY+dClipH ) y1 = dClipY+dClipH-1;
				w = y1-y2+1;
				s = -dStepY;
			} else w = 1;
			pDST = pARGBD + dStepY*y1 + dStepX*x1;
			for(y=y1; w--; y += d) {
				if( LineStyle & 0x8000000000000000 ) {
						*pDST = (BYTE)((((WORD)a1*(WORD)(* pDST   ))/255) + ((WORD)r*(WORD)a)/255);
					*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)g*(WORD)a)/255);
					*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)b*(WORD)a)/255);
#if SCREEN_ARGB
#else
					if( (KLONG)hBM>0 )
#endif
					*(pDST+3) = (BYTE)((((WORD)a1*(WORD)(*(pDST+3)))/255) + ((WORD)a*(WORD)a)/255);
					LineStyle = (LineStyle<<1) | 1;
				} else LineStyle <<= 1;
				pDST += s;
			}
			++x2;	/// for invalidate
		} else if( y1==y2 ) {	/// HORIZONTAL line
			if( y1<dClipY || y1>=dClipY+dClipH ) goto bail;
		// if end-points are same, just plot single point
			if( x1<x2 ) {
				d = 1;
				if( x1>=dClipX+dClipW || x2<dClipX ) goto bail;
				if( x1<dClipX ) x1 = dClipX;
				if( x2>=dClipX+dClipW ) x2 = dClipX+dClipW-1;
				w = x2-x1+1;
				s = dStepX;
			} else if( x1>x2 ) {
				d = -1;
				if( x2>=dClipX+dClipW || x1<dClipX ) goto bail;
				if( x2<dClipX ) x2 = dClipX;
				if( x1>=dClipX+dClipW ) x1 = dClipX+dClipW-1;
				w = x1-x2+1;
				s = -dStepX;
			} else w = 1;
			pDST = pARGBD + dStepY*y1 + dStepX*x1;
			for(x=x1; w--; x += d) {
				if( LineStyle & 0x8000000000000000 ) {
						*pDST = (BYTE)((((WORD)a1*(WORD)(* pDST   ))/255) + ((WORD)r*(WORD)a)/255);
					*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)g*(WORD)a)/255);
					*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)b*(WORD)a)/255);
#if SCREEN_ARGB
#else
					if( (KLONG)hBM>0 )
#endif
					*(pDST+3) = (BYTE)((((WORD)a1*(WORD)(*(pDST+3)))/255) + ((WORD)a*(WORD)a)/255);
					LineStyle = (LineStyle<<1) | 1;
				} else LineStyle <<= 1;
				pDST += s;
			}
			++y2;	// for invalidate
		} else {	/// DIAGONAL line
			int dX, dY, Xincr, Yincr, dPr, dPru, P, Ax, Ay;

			pDST = pARGBD + dStepY*y1 + dStepX*x1;
			if( x1 > x2 ) {
				Xincr = -1;
				dStepX = -dStepX;
			} else Xincr = 1;

			if( y1 > y2 ) {
				Yincr = -1;
				dStepY = -dStepY;
			} else Yincr = 1;
			Ax = x1;
			Ay = y1;
			dX = abs(x2-x1);
			dY = abs(y2-y1);
			if( dX >= dY ) {
				dPr = dY<<1;
				dPru = dPr - (dX<<1);
				P = dPr - dX;

				for(; dX >= 0; dX--) {
					if( Ax>=dClipX && Ax<dClipX+dClipW && Ay>=dClipY && Ay<dClipY+dClipH ) {
						if( LineStyle & 0x8000000000000000 ) {
								*pDST = (BYTE)((((WORD)a1*(WORD)(* pDST   ))/255) + ((WORD)r*(WORD)a)/255);
							*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)g*(WORD)a)/255);
							*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)b*(WORD)a)/255);
#if SCREEN_ARGB
#else
							if( (KLONG)hBM>0 )
#endif
							*(pDST+3) = (BYTE)((((WORD)a1*(WORD)(*(pDST+3)))/255) + ((WORD)a*(WORD)a)/255);
							LineStyle = (LineStyle<<1) | 1;
						} else LineStyle <<= 1;
					}
					if( P > 0 ) {
						Ay += Yincr;
						pDST += dStepY;
						P += dPru;
					} else P += dPr;
					Ax += Xincr;
					pDST += dStepX;
				}
			} else {
				dPr = dX<<1;
				dPru = dPr - (dY<<1);
				P = dPr - dY;
				for(; dY >= 0; dY--) {
					if( Ax>=dClipX && Ax<dClipX+dClipW && Ay>=dClipY && Ay<dClipY+dClipH ) {
						if( LineStyle & 0x8000000000000000 ) {
								*pDST = (BYTE)((((WORD)a1*(WORD)(* pDST   ))/255) + ((WORD)r*(WORD)a)/255);
							*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)g*(WORD)a)/255);
							*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)b*(WORD)a)/255);
#if SCREEN_ARGB
#else
							if( (KLONG)hBM>0 )
#endif
							*(pDST+3) = (BYTE)((((WORD)a1*(WORD)(*(pDST+3)))/255) + ((WORD)a*(WORD)a)/255);
							LineStyle = (LineStyle<<1) | 1;
						} else LineStyle <<= 1;
					}
					if( P > 0 ) {
						Ax += Xincr;
						pDST += dStepX;
						P += dPru;
					} else P += dPr;
					Ay += Yincr;
					pDST += dStepY;
				}
			}
		}
		if( hBM==0 ) InvalidateScreen(x1, y1, x2-x1, y2-y1);
bail:
	EndDST();
}

///=================================================================
void DrawLines(DWORD hBM, KLONG *pPoints, KLONG num, DWORD clr) {

    KLONG	x, y;

	while( --num ) {
		x = *pPoints++;
		y = *pPoints++;
		DrawLineBM(hBM, x, y, *pPoints, *(pPoints+1), clr);
	}
}

///=================================================================
/// Draw a rectangle in the destination bitmap.  The rectangle may be
/// SOLID, OUTLINED, or BOTH, depending upon the values of 'fill' and 'edge'.
/// if 'fill' is not transparent (ALPHA==0), fill the rectangle with that color
/// if 'edge' is not transparent (ALPHA==0), outline the rectangle with that color
void DrawRect(DWORD hBM, KLONG x1, KLONG y1, KLONG wi, KLONG hi, DWORD fill, DWORD edge) {

	BYTE	*pDST, *pD, r, g, b, a, a1;
    KLONG	x, y, w, h, ww, edged;

if( x1==0 && y1==50 && wi==2000 && hi==1400 ) {
	a = 0;
}
	StartDST(hBM);

	a = (BYTE)(fill >> 24);
	a1= 255-a;
	r = (BYTE)(fill >> 16);
	g = (BYTE)(fill >> 8);
	b = (BYTE)(fill);

	edged = (edge & 0xFF000000)?1:0;
	if( fill & 0xFF000000 ) {
		x = Kmax(dClipX, x1+edged);
		y = Kmax(dClipY, y1+edged);
		w = Kmin(dClipX+dClipW, x1+wi-edged)-x;
		h = Kmin(dClipY+dClipH, y1+hi-edged)-y;
		if( w>0 && h>0 ) {
			pD = pARGBD + dStepY*y + dStepX*x;
			while( h-- ) {
				pDST = pD;
				for(ww=w; ww--; ) {
						*pDST = (BYTE)((((WORD)a1*(WORD)(* pDST   ))/255) + ((WORD)r*(WORD)a)/255);
					*(pDST+1) = (BYTE)((((WORD)a1*(WORD)(*(pDST+1)))/255) + ((WORD)g*(WORD)a)/255);
					*(pDST+2) = (BYTE)((((WORD)a1*(WORD)(*(pDST+2)))/255) + ((WORD)b*(WORD)a)/255);
#if SCREEN_ARGB
#else
					if( (KLONG)hBM>0 )
#endif
					*(pDST+3) = (BYTE)((((WORD)a1*(WORD)(*(pDST+3)))/255) + ((WORD)a*(WORD)a)/255);
					pDST += dStepX;
				}
				pD += dStepY;
			}
		}
	}
	EndDST();
	if( edged ) {
		--wi;
		--hi;
		DrawLineBM(hBM, x1, y1, x1+wi, y1, edge);
		DrawLineBM(hBM, x1+wi, y1, x1+wi, y1+hi, edge);
		DrawLineBM(hBM, x1+wi, y1+hi, x1, y1+hi, edge);
		DrawLineBM(hBM, x1, y1+hi, x1, y1, edge);
		++wi;
		++hi;
	}
	if( hBM==0 ) InvalidateScreen(x1, y1, wi, hi);
}

///===============================================================================
/// Fills a rectangle with a gradient hue, over a vertical range of lightness
/// and a horizontal range of saturation.
void RectGradientBM(DWORD pArgs) {
//DWORD hBM, long xo, long yo, long wo, long ho, WORD hue, WORD satmin, WORD satmax, WORD lightmin, WORD lightmax

	DWORD	hBM;
    KLONG	xo, yo, wo, ho;
	BYTE	hue, satmin, satmax, lightmin, lightmax;
	BYTE	*pDST, *pD, r, g, b;
	KLONG		i, x, y, w, h, cy, cx, hs, ws;
	double	li, sa, hu;
	double	f, p, q, t;

	hBM = GETMEMD(pArgs); pArgs += 4;
	xo  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	yo  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	wo  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	ho  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hue = (BYTE)GETMEMD(pArgs); pArgs += 4;
	satmin = (BYTE)GETMEMD(pArgs); pArgs += 4;
	satmax = (BYTE)GETMEMD(pArgs); pArgs += 4;
	lightmin = (BYTE)GETMEMD(pArgs); pArgs += 4;
	lightmax = (BYTE)GETMEMD(pArgs);

	if( ho < 0 ) {ho = -ho; hs = 1;}
	else hs = 0;
	if( wo < 0 ) {wo = -wo; ws = 1;}
	else ws = 0;
	x = xo;
	y = yo;
	w = wo;
	h = ho;
	if( w>0 && h>0 ) {
		StartDST(hBM);
				hu = (251.0-84.0-(double)hue);
				if( hu<0 ) hu += 252;
				hu /= 42.0;
				for(cy=0, pD=pARGBD+dStepY*y+dStepX*x;  cy<h;  cy++, pD+=dStepY) {
					li = (double)(hs?(h-cy-1):cy)/(double)(h-1);
					for(cx=0, pDST=pD; cx<w; cx++, pDST+= dStepX) {
						if( xo+cx>=dClipX && xo+cx<dClipX+dClipW && yo+cy>=dClipY && yo+cy<dClipY+dClipH ) {
							sa = (double)(ws?(w-cx-1):cx)/(double)(w-1);
							if( sa==0 ) r = g = b = (BYTE)(255.0*li);
							else {
								i = (int)floor(hu);
								f = hu-(double)i;
								p = li*(1.0-sa);
								q = li*(1.0-sa*f);
								t = li*(1.0-sa*(1.0-f));
								switch( (int)i ) {
								 case 0:
									r = (BYTE)(255.0*li);
									g = (BYTE)(255.0*t);
									b = (BYTE)(255.0*p);
									break;
								 case 1:
									r = (BYTE)(255.0*q);
									g = (BYTE)(255.0*li);
									b = (BYTE)(255.0*p);
									break;
								 case 2:
									r = (BYTE)(255.0*p);
									g = (BYTE)(255.0*li);
									b = (BYTE)(255.0*t);
									break;
								 case 3:
									r = (BYTE)(255.0*p);
									g = (BYTE)(255.0*q);
									b = (BYTE)(255.0*li);
									break;
								 case 4:
									r = (BYTE)(255.0*t);
									g = (BYTE)(255.0*p);
									b = (BYTE)(255.0*li);
									break;
								 default:
									r = (BYTE)(255.0*li);
									g = (BYTE)(255.0*p);
									b = (BYTE)(255.0*q);
									break;
								}
							}
							*pDST = (BYTE)r;
							*(pDST+1) = (BYTE)g;
							*(pDST+2) = (BYTE)b;
#if SCREEN_ARGB
#else
							if( (KLONG)hBM>0 )
#endif
							*(pDST+3) = 0xFF;
						}
					}
				}

		EndDST();
	}
	if( hBM==0 ) InvalidateScreen(xo, yo, wo, ho);
}

///=============================================================================
/// Draw an oval in the destination bitmap.  The oval may be
/// SOLID, OUTLINED, or BOTH, depending upon the values of 'fill' and 'edge'.
/// if 'fill' is not alpha==0, fill the oval with that color
/// if 'edge' is not alpha==0, outline the rectangle with that color

#if SCREEN_ARGB
	#define OvalPoint(pD, x, y, a, a1, r, g, b) {\
				if( x>=dClipX && x<dClipX+dClipW && y>=dClipY && y<dClipY+dClipH ) {\
						*pD = (BYTE)((((WORD)a1*(WORD)(* pD   ))/255) + ((WORD)r*(WORD)a)/255);\
					*(pD+1) = (BYTE)((((WORD)a1*(WORD)(*(pD+1)))/255) + ((WORD)g*(WORD)a)/255);\
					*(pD+2) = (BYTE)((((WORD)a1*(WORD)(*(pD+2)))/255) + ((WORD)b*(WORD)a)/255);\
					*(pD+3) = (BYTE)((((WORD)a1*(WORD)(*(pD+3)))/255) + ((WORD)a*(WORD)a)/255);\
				}\
			}
#else
	#define OvalPoint(pD, x, y, a, a1, r, g, b) {\
				if( x>=dClipX && x<dClipX+dClipW && y>=dClipY && y<dClipY+dClipH ) {\
						*pD = (BYTE)((((WORD)a1*(WORD)(* pD   ))/255) + ((WORD)r*(WORD)a)/255);\
					*(pD+1) = (BYTE)((((WORD)a1*(WORD)(*(pD+1)))/255) + ((WORD)g*(WORD)a)/255);\
					*(pD+2) = (BYTE)((((WORD)a1*(WORD)(*(pD+2)))/255) + ((WORD)b*(WORD)a)/255);\
					if( (KLONG)hBM>0 )\
						*(pD+3) = (BYTE)((((WORD)a1*(WORD)(*(pD+3)))/255) + ((WORD)a*(WORD)a)/255);\
				}\
			}
#endif

///=========================================================================
///DWORD hBM, long x1, long y1, long w, long h, DWORD fill, DWORD edge)
void DrawOvalBM(DWORD pArgs) {

	DWORD	hBM, fill, edge;
	KLONG	x1, y1, w, h;

	BYTE	*pDSTnw, *pDSTne, *pDSTsw, *pDSTse, *pDfn, *pDfs, re, ge, be, rf, gf, bf, ae, ae1, af, af1;
	KLONG	x, y, dx, dy, a_mul, b_mul, tt, xw, xe, yn, ys, xf;
	KLONG	row, as, bs, as2, bs2, as4, bs4, dd, xo, yo, bs6, as6;
	DWORD	edge_color;

	hBM = GETMEMD(pArgs); pArgs += 4;
	x1  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	y1  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	w   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	h   = (KLONG)GETMEMD(pArgs); pArgs += 4;
	fill= GETMEMD(pArgs); pArgs += 4;
	edge= GETMEMD(pArgs);

	ae  = (BYTE)(edge >> 24);
	ae1 = 255-ae;
	af  = (BYTE)(fill >> 24);
	af1 = 255-af;

	edge_color = ((edge&0xFF000000)==0)?fill:edge;
	if( w==0 || h==0 ) {
		DrawLineBM(hBM, x1, y1, x1+w, y1+h, edge_color);
		return;
	}
	StartDST(hBM);

		dx = w;
		dy = h;
		xo = (dx+1) & 1;		// whether to add an extra line down center of ellipse
		yo = (dy+1) & 1;		// whether to add an extra line across center of ellipse
		x  = x1+(dx-xo)/2;				// center of ellipse
		y  = y1+(dy-yo)/2;
		as4 = (dx-1-xo)*(dx-1-xo);	// 4*A^2
		as2 = as4/2;				// 2*A^2
		as  = as2/2;				// A^2
		bs4 = (dy-1-yo)*(dy-1-yo);	// 4*B^2
		bs2 = bs4/2;				// 2*B^2
		bs  = bs2/2;				// B^2
		bs6 = bs4+bs2;
		as6 = as4+as2;

		row = (dy-yo)/2;		// starting y-offset from center
		a_mul= as4*row;			// amount to adjust decision variable when step in Y
		b_mul= 0;				// amount to adjust decision variable when step in X
		dd  = as - as2*row + bs2;	// initial value of decision variable
		tt  = bs - as4*row + as2;

		rf = (BYTE)(fill & 0x000000FF);
		gf = (BYTE)((fill & 0x0000FF00) >> 8);
		bf = (BYTE)((fill & 0x00FF0000) >> 16);
		re = (BYTE)(edge_color & 0x000000FF);
		ge = (BYTE)((edge_color & 0x0000FF00) >> 8);
		be = (BYTE)((edge_color & 0x00FF0000) >> 16);

		xw = x;
		xe = x+xo;
		yn = y-row;
		ys = y+row+yo;
		pDSTnw = pARGBD + dStepY*yn + dStepX*xw;
		pDSTne = pARGBD + dStepY*yn + dStepX*xe;
		pDSTsw = pARGBD + dStepY*ys + dStepX*xw;
		pDSTse = pARGBD + dStepY*ys + dStepX*xe;

		OvalPoint(pDSTnw, xw, yn, ae, ae1, re, ge, be);
		if( xw!=xe ) OvalPoint(pDSTne, xe, yn, ae, ae1, re, ge, be);
		if( yn!=ys ) {
			OvalPoint(pDSTsw, xw, ys, ae, ae1, re, ge, be);
			if( xw!=xe ) OvalPoint(pDSTse, xe, ys, ae, ae1, re, ge, be);
		}
		do {
			if( dd<0 ) {
				dd += b_mul + bs6;
				tt += b_mul + bs4;
				b_mul += bs4;
				pDSTnw -= dStepX;
				pDSTne += dStepX;
				pDSTsw -= dStepX;
				pDSTse += dStepX;
				--xw;
				++xe;
			} else if( tt<0 ) {
				dd += b_mul + bs6 - a_mul + as4;
				tt += b_mul + bs4 - a_mul + as6;
				a_mul -= as4;
				b_mul += bs4;
				--row;
				++yn;
				--ys;
				xf = xw;
				--xw;
				++xe;
				pDSTnw += dStepY-dStepX;
				pDSTne += dStepY+dStepX;
				pDSTsw -= dStepY+dStepX;
				pDSTse -= dStepY-dStepX;

				pDfn = pDSTnw+dStepX;
				pDfs = pDSTsw+dStepX;
				if( fill & 0xFF000000 ) {
					 while( pDfn!=pDSTne ) {
						OvalPoint(pDfn, xf, yn, af, af1, rf, gf, bf);
						if( yn!=ys ) OvalPoint(pDfs, xf, ys, af, af1, rf, gf, bf);
						pDfn += dStepX;
						pDfs += dStepX;
						++xf;
					}
				}
			} else {
				dd -= a_mul-as4;
				tt -= a_mul-as6;
				a_mul -= as4;
				--row;
				++yn;
				--ys;
				xf = xw+1;
				pDSTnw += dStepY;
				pDSTne += dStepY;
				pDSTsw -= dStepY;
				pDSTse -= dStepY;

				pDfn = pDSTnw+dStepX;
				pDfs = pDSTsw+dStepX;
				if( (fill & 0xFF000000) && ys>=yn ) {
					while( xf<xe ) {
						OvalPoint(pDfn, xf, yn, af, af1, rf, gf, bf);
						if( yn!=ys ) OvalPoint(pDfs, xf, ys, af, af1, rf, gf, bf);
						pDfn += dStepX;
						pDfs += dStepX;
						++xf;
					}
				}
			}
			if( ys>=yn ) {	// never write the same dot twice in case AlphaBlend is active
				OvalPoint(pDSTnw, xw, yn, ae, ae1, re, ge, be);
				if( xw!=xe) OvalPoint(pDSTne, xe, yn, ae, ae1, re, ge, be);
				if( ys>yn ) {
					OvalPoint(pDSTsw, xw, ys, ae, ae1, re, ge, be);
					if( xw!=xe ) OvalPoint(pDSTse, xe, ys, ae, ae1, re, ge, be);
				}
			}
		} while( ys>yn );

	EndDST();
	if( hBM==0 ) InvalidateScreen(x1, y1, w+1, w+1);
}

///========================================================
DWORD LoadPNG(char *fname) {

    FILE					*png;
    spng_ctx				*ctx = NULL;
    BYTE					*out = NULL;//, *pout;
    struct spng_ihdr		ihdr;
    struct spng_plte		plte = {0};
    uint32_t				out_size, out_width;
/**    Output format, does not depend on source PNG format except for
       SPNG_FMT_PNG, which is the PNG's format in host-endian or
       big-endian for SPNG_FMT_RAW.
       Note that for these two formats <8-bit images are left byte-packed.
*/
    int						fmt = SPNG_FMT_PNG;
//    struct spng_row_info	row_info = {0};
    int						ret;
	KBM						*pHP;
	DWORD					hHP;
//	KLONG					x, y, t;
	BYTE					*pB;


    if( NULL==(png = KMZ_open(fname, "rb")) ) return 0;
	if( NULL==(ctx = spng_ctx_new(0)) ) {
error1:
		KMZ_close(png);
		return 0;
	}

    spng_set_crc_action(ctx, SPNG_CRC_USE, SPNG_CRC_USE);
    uint32_t limit = 1024 * 1024 * 64;	// set memory limits
    spng_set_chunk_limits(ctx, limit, (uint32_t)limit);
    spng_set_png_file(ctx, png);	// set PNG source

    if( (ret = spng_get_ihdr(ctx, &ihdr)) ) {
error2:
		spng_ctx_free(ctx);
		goto error1;
    }

/*
    if(ihdr.color_type == SPNG_COLOR_TYPE_GRAYSCALE) clr_type_str = "grayscale";
    else if(ihdr.color_type == SPNG_COLOR_TYPE_TRUECOLOR) clr_type_str = "truecolor";
    else if(ihdr.color_type == SPNG_COLOR_TYPE_INDEXED) clr_type_str = "indexed color";
    else if(ihdr.color_type == SPNG_COLOR_TYPE_GRAYSCALE_ALPHA) clr_type_str = "grayscale with alpha";
    else clr_type_str = "truecolor with alpha";

    printf("width: %" PRIu32 "\nheight: %" PRIu32 "\n"
           "bit depth: %" PRIu8 "\ncolor type: %" PRIu8 " - %s\n",
           ihdr.width, ihdr.height,
           ihdr.bit_depth, ihdr.color_type, clr_type_str);
    printf("compression method: %" PRIu8 "\nfilter method: %" PRIu8 "\n"
           "interlace method: %" PRIu8 "\n",
           ihdr.compression_method, ihdr.filter_method,
           ihdr.interlace_method);
*/

    if( SPNG_ECHUNKAVAIL!=(ret=spng_get_plte(ctx, &plte)) && ret ) goto error2;


 //   if( ihdr.color_type == SPNG_COLOR_TYPE_INDEXED )
	fmt = SPNG_FMT_RGBA8;
	if( (ret=spng_decoded_image_size(ctx, fmt, (uint32_t*)&out_size)) ) goto error2;

    out_width = out_size / ihdr.height;
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"LoadPNG(%d)\n", out_size);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	out = (BYTE*)malloc(out_size);
	if( (ret = spng_decode_image(ctx, out, out_size, SPNG_FMT_RGBA8, 0)) ) {
error3:
		goto error2;
	}

    spng_ctx_free(ctx);
	KMZ_close(png);

	/// we had to do this bogus malloc()/free() (via spng_ctx_new/free) because
	/// the system heap was getting confused if our CreateBM() caused us to
	/// realloc() the HeapPtrs[] array. Sheesh. Who doesn't hate memory management???

	if( 0==(hHP=CreateBM(ihdr.width, ihdr.height)) ) return 0;

	pHP = (KBM*)GetHeapPtr(hHP);			/// get a ptr to the KBM struct
	pB = (BYTE*)GetHeapPtr(pHP->hBM);		/// get a ptr to the bitmap bits
	RelHeapPtr(pHP->hBM);
	RelHeapPtr(hHP);

	memcpy(pB, out, out_size);				/// copy the bitmap bits

	/// PNG order seems to be ARGB, whereas we're using ABGR, so we reverse them to get the right colors.
    /// That was Windows. On the MAC, things are okay, so  no reversal.
/*	for(y=0; y<ihdr.height; y++) {
		for(x=0; x<ihdr.width; x++) {
			pout = pB + y*pHP->linelen + x*4;
			t = *pout;
			*pout = *(pout+2);
			*(pout+2) = t;
		}
	}
*/
	free(out);

    return hHP;
}

///======================================================
DWORD SavePNG(DWORD hBM, char *fname) {

    int						fmt = SPNG_FMT_PNG;
//    struct spng_row_info	row_info = {0};
    struct spng_ihdr		ihdr;
    int						ret;
//	KBM						*pHP;
//	DWORD					hHP;
	KLONG					x, y, t;
	BYTE					*pB;
    DWORD					hFile, retval=FALSE;//, isize;
//    BYTE					*out = NULL;
    struct spng_plte		plte = {0};
//    size_t					out_size, out_width;

#if SCREEN_ARGB
#else
	if( hBM<=0 ) SystemMessage("Can't save to PNG from screen bitmap");
#endif
//	if( -1!=(hFile=Kopen(fname, O_WRBINNEW)) ) {
		StartSRC(hBM);

			/// PNG order seems to be ARGB, whereas we're using ABGR, so we reverse them to get the right colors.
			for(y=0; y<sHeight; y++) {
				for(x=0; x<sWidth; x++) {
					pB = pARGBS + y*sStepY + x*4;
					t = *pB;
					*pB = *(pB+2);
					*(pB+2) = t;
				}
			}

			/// Creating an encoder context requires a flag
			spng_ctx *enc = spng_ctx_new(SPNG_CTX_ENCODER);

			/// Encode to internal buffer managed by the library
			spng_set_option(enc, SPNG_ENCODE_TO_BUFFER, 1);

			/// Alternatively you can set an output FILE* or stream with spng_set_png_file() or spng_set_png_stream()

			/// In this case we're reencoding to the same PNG format
			ihdr.width = sWidth;
			ihdr.height = sHeight;
			ihdr.bit_depth = 8;
			ihdr.color_type = SPNG_COLOR_TYPE_TRUECOLOR_ALPHA;
			ihdr.compression_method = 0;
			ihdr.filter_method = 0;
			ihdr.interlace_method = 0;
/**
struct spng_ihdr
{
    uint32_t width;
    uint32_t height;
    uint8_t bit_depth;
    uint8_t color_type;
    uint8_t compression_method;
    uint8_t filter_method;
    uint8_t interlace_method;
};
*/

			spng_set_ihdr(enc, &ihdr);

			/// Copy the palette from the source file
			if(plte.n_entries > 0) spng_set_plte(enc, &plte);

			/// SPNG_FMT_PNG is a special value that matches the format in ihdr
			fmt = SPNG_FMT_PNG;

			/// SPNG_ENCODE_FINALIZE will finalize the PNG with the end-of-file marker
			if( spng_encode_image(enc, pARGBS, sStepY*sHeight, fmt, SPNG_ENCODE_FINALIZE) ) goto encode_error;

            uint32_t png_size;
			void *png_buf = NULL;

			/// Get the internal buffer of the finished PNG
			png_buf = spng_get_png_buffer(enc, (uint32_t *)&png_size, &ret);

			if( png_buf!=NULL )	{
				if( -1!=(hFile=Kopen(fname, O_WRBINNEW)) ) {
					if( png_size==Kwrite(hFile, png_buf, (DWORD)(uint32_t)png_size) ) {
						retval = TRUE;
					}
					Kclose(hFile);
				}
				/// User owns the buffer after a successful call
				free(png_buf);
			}
encode_error:
			spng_ctx_free(enc);

			/// PNG order seems to be ARGB, whereas we're using ABGR, so we reverse them to get the right colors.
			for(y=0; y<sHeight; y++) {
				for(x=0; x<sWidth; x++) {
					pB = pARGBS + y*sStepY + x*4;
					t = *pB;
					*pB = *(pB+2);
					*(pB+2) = t;
				}
			}

		EndSRC();
//	}
	return retval;
}

///================================================================
/// Currently, supports loading a BMP or PNG file.
DWORD LoadImg(char *fname) {

	DWORD		hFile;
	BYTE		head[20];

	/// read start of file to determine its type
	if( -1==(hFile = Kopen(fname, O_RDBIN)) ) return 0;
	if( sizeof(head)!=Kread(hFile, head, sizeof(head)) ) {
		Kclose(hFile);
		return 0;
	}
	Kclose(hFile);

	if( *(WORD*)head == *(WORD *)"BM" ) {
		hFile = LoadBMP(fname);			/// if .BMP header
		return hFile;
	}
	if( *(QWORD*)head == 0x0A1A0A0D474E5089 ) {
		return LoadPNG(fname);	/// if .PNG header
	}

	return 0;
}

///================================================================
/// Currently, supports saving a BMP or PNG file.
DWORD SaveImg(DWORD hBM, char *fname, KLONG format) {

	if( format==0 ) return SaveBMP(hBM, fname);	/// save as .BMP file
	if( format==1 ) return SavePNG(hBM, fname);	/// save as .PNG file
	return 0;
}

///================================================================
/// convert rgb values from 0-255 to 0.0-1.0, then convert to HSL, then convert back to 0-251 for H, 0-255 for S & L.
DWORD RGB2HSL(DWORD argb) {

	double	r, g, b, h, s, l, cmin, cmax, delta;

	b = (double)((argb>>16)&0x0FF)/255.0;
	g = (double)((argb>> 8)&0x0FF)/255.0;
	r = (double)((argb    )&0x0FF)/255.0;

	cmin = min(r, min(g, b));
	cmax = max(r, max(g, b));
	l = cmax;
	delta = cmax-cmin;
	if( delta==0 ) {	/// r=g=b=0
		s = 0;
		h = 0;	/// doesn't matter, SAT=0 (black)
	} else {
		s = delta/cmax;
		if( r==cmax ) h = (g-b)/delta;			/// between yellow & magenta
		else if( g==cmax ) h = 2.0 + (b-r)/delta;	/// between cyan & yellow
		else h = 4.0 + (r-g)/delta;				/// between magenta & cyan

		while( h<0 ) h += 6.0;
		while( h>=6.0 ) h -= 6.0;
		h *= 252.0/6.0;	/// 6*42=252, for 0-252
	}
	return (argb&0xFF000000) + (((DWORD)(255.0*l))<<16) + (((DWORD)(255.0*s))<<8) + (DWORD)h;
}

///================================================================
/// convert hsl values from 0-252 for H and 0-255 for S & L to 0.0-1.0, then convert to RGB, then convert back to 0-255.
DWORD HSL2RGB(DWORD ahsl) {

	double	r, g, b, h, s, l, p, q, t, f;
	KLONG	i;

	l = (double)((ahsl>>16)&0x0FF)/255.0;
	s = (double)((ahsl>> 8)&0x0FF)/255.0;
	h = (double)((ahsl    )&0x0FF)/(252.0/6.0);

	if( s==0 ) r = g = b = l;	/// greyscale
	else {
		i = (KLONG)floor(h);		/// integer part of h (0-5)
		f = h - (double)i;		/// fractional part of h
		p = l*(1.0-s);
		q = l*(1-s*f);
		t = l*(1-s*(1-f));
		switch( i ) {
		 case 0:
		 	r = l;
		 	g = t;
		 	b = p;
			break;
		 case 1:
		 	r = q;
		 	g = l;
		 	b = p;
			break;
		 case 2:
		 	r = p;
		 	g = l;
		 	b = t;
			break;
		 case 3:
		 	r = p;
		 	g = q;
		 	b = l;
			break;
		 case 4:
		 	r = t;
		 	g = p;
		 	b = l;
			break;
		 case 5:
		 	r = l;
		 	g = p;
		 	b = q;
			break;
         default:
            r = 0;
            g = 0;
            b = 0;
            break;
		}
	}
	return (ahsl&0xFF000000) + (((DWORD)(255.0*b))<<16) + (((DWORD)(255.0*g))<<8) + (DWORD)(255.0*r);
}

///========================================================
void GetInfoBM(DWORD hBM, INFOBM *pIBM) {

	StartSRC(hBM);
        pIBM->w = sWidth;
		pIBM->h = sHeight;
		pIBM->clipX = sClipX;
		pIBM->clipY = sClipY;
		pIBM->clipW = sClipW;
		pIBM->clipH = sClipH;
	EndSRC();
}

///========================================================
void ClipBM(DWORD hBM, KLONG x, KLONG y, KLONG w, KLONG h, KRECT *pRect) {

	StartSRC(hBM);
		if( hBM==0 ) {
			if( pRect!=NULL ) {
				pRect->x = scrBM.clipX;
				pRect->y = scrBM.clipY;
				pRect->w = scrBM.clipW;
				pRect->h = scrBM.clipH;
			}
			scrBM.clipX = min(max(0, x), scrBM.w);
			scrBM.clipY = min(max(0, y), scrBM.h);
			scrBM.clipW = max(min(scrBM.w-scrBM.clipX, w), 0);
			scrBM.clipH = max(min(scrBM.h-scrBM.clipY, h), 0);
		} else {
			if( pRect!=NULL ) {
				pRect->x = pKBMS->clipX;
				pRect->y = pKBMS->clipY;
				pRect->w = pKBMS->clipW;
				pRect->h = pKBMS->clipH;
			}
			pKBMS->clipX = min(max(0, x), pKBMS->w);
			pKBMS->clipY = min(max(0, y), pKBMS->h);
			pKBMS->clipW = max(min(pKBMS->w-pKBMS->clipX, w), 0);
			pKBMS->clipH = max(min(pKBMS->h-pKBMS->clipY, h), 0);
		}
	EndSRC();
}

///========================================================
QWORD SetLineStyle(QWORD ls){

	QWORD	oldLS=LineStyle;

	LineStyle = ls;
	return oldLS;
}

///====================================================================================================
/// hue and sat are values from 0-255, and if >= 0 simply replace the previous values for each pixel.
/// light is from -255 to +255, and is a "percentage" of previous value added to the previous value for each pixel.
void ColorizeBM(DWORD pArgs) {

	DWORD hBM;
    KLONG x, y, w, h, hue, sat, light, contrast;

	BYTE	*pD, *pDST;
    KLONG	tw, th, r, g, b;
    KLONG	mx, mn, v, s, hh, rc, gc, bc, f, p, q, t;

	hBM  = GETMEMD(pArgs); pArgs += 4;
	x    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	y    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	w    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	h    = (KLONG)GETMEMD(pArgs); pArgs += 4;
	hue  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	sat  = (KLONG)GETMEMD(pArgs); pArgs += 4;
	light= (KLONG)GETMEMD(pArgs); pArgs += 4;
	contrast = (KLONG)GETMEMD(pArgs);

	StartSRC(hBM);

		if( x<sClipX ) {
			w -= (sClipX-x);
			x = sClipX;
		}
		if( y<sClipY ) {
			h -= (sClipY-y);
			y = sClipY;
		}
		if( (x+w)>=sClipX+sClipW ) w = sClipX+sClipW-x;
		if( (y+h)>=sClipY+sClipH ) h = sClipY+sClipH-y;
		if( w<=0 || h<=0 ) goto bail;
		pD = pARGBS + sStepY*y + sStepX*x;
		for(th=h; th; pD += sStepY, --th) {
			pDST = pD;
			for(tw=w; tw; pDST += sStepX, --tw) {
				b = (KLONG)*(pDST+2);
				g = (KLONG)*(pDST+1);
				r = (KLONG)*pDST;
				mx= (r>=g && r>=b)?r:((g>=r && g>=b)?g:b);
				mn= (r<=g && r<=b)?r:((g<=r && g<=b)?g:b);
				v = mx;	/// value
				if( mx ) s = 255*(mx-mn)/mx;	/// saturation
				else s = 0;
				if( s==0 ) hh = 0;
				else {
					rc = 255*(mx-r)/(mx-mn);
					gc = 255*(mx-g)/(mx-mn);
					bc = 255*(mx-b)/(mx-mn);

					if( r==mx ) hh = (bc-gc)/6;
					else if( g==mx ) hh = 85+(rc-bc)/6;
					else hh = 171+(gc-rc)/6;
				}
				if( hue>=0 ) hh = hue;
				hh &= 255;

				if( sat>=0 ) s = sat;
				if( s<0 ) s = 0;
				s &= 255;

				if( contrast ) {
					v += ((v-128)*contrast)/255;
				}
				if( light ) {
					v += (light*v)/255;
				}
				if( v>255 ) v = 255;
				else if( v<0 ) v = 0;

				if( s==0 ) {
					*(pDST+2) = *(pDST+1) = *pDST = (BYTE)v;
				} else {
					if( hh>255 ) hh = 0;
					f = 6*(hh % 42);
					hh= hh / 42;
					if( hh>5 || hh<0 ) hh = 0;
					p = (v*(255-s))/255;
					q = (f==0)?v:((v*(255*246-s*f))/(255*246));
					t = (v*(255*246-(s*(246-f))))/(255*266);
					switch( hh ) {
					 case 0:
						*(pDST+0) = (BYTE)v;
						*(pDST+1) = (BYTE)t;
						*(pDST+2) = (BYTE)p;
						break;
					 case 1:
						*(pDST+0) = (BYTE)q;
						*(pDST+1) = (BYTE)v;
						*(pDST+2) = (BYTE)p;
						break;
					 case 2:
						*(pDST+0) = (BYTE)p;
						*(pDST+1) = (BYTE)v;
						*(pDST+2) = (BYTE)t;
						break;
					 case 3:
						*(pDST+0) = (BYTE)p;
						*(pDST+1) = (BYTE)q;
						*(pDST+2) = (BYTE)v;
						break;
					 case 4:
						*(pDST+0) = (BYTE)t;
						*(pDST+1) = (BYTE)p;
						*(pDST+2) = (BYTE)v;
						break;
					 case 5:
						*(pDST+0) = (BYTE)v;
						*(pDST+1) = (BYTE)p;
						*(pDST+2) = (BYTE)q;
						break;
					}
				}
			}
		}
	if( hBM==0 ) InvalidateScreen(x, y, w, h);
bail:
	EndSRC();
}

///***************************************************************
/// SysFont FLAGS (NOTE: These MUST be identical to the CF_/TF_ defines used in the .KE C code
#define	CF_LEFT			0x00000000	// set nothing for left-justified TEXT control
#define	CF_CENTER		0x00020000	// center TEXT control (this bit is shared with CF_VERSCROLL)
#define	CF_RIGHT		0x00040000	// right-justify TEXT control (this bit is shared with CF_HORSCROLL)
#define	CF_FIXED		0x00200000	// for controls with text, use FIXED-point font
#define	CF_BOLD			0x00400000	// for controls with text, use BOLD font
#define	CF_SERIF		0x01000000	// for controls with text, use SERIF font
#define	CF_ITALIC		0x08000000	// for controls with text, use ITALIC font
#define	CF_UNDERLINE	0x10000000	// for controls with text, UNDERLINE text
#define	TF_BREAK_CHAR	0x20000000	// only set during call to DrawTextBM(), breaks text on a character (at right edge)
#define	TF_BREAK_SPACE	0x40000000	// only set during call to DrawTextBM(), breaks text on a SPACE character (at right edge)
#define	TF_GET_WIDTH	0x80000000	// only set during call to DrawTextBM(), doesn't draw, just returns width of what WOULD be drawn

#define	CF_FONT_CHARS	(CF_FIXED | CF_BOLD | CF_ITALIC | CF_UNDERLINE | CF_SERIF)

#define    SYSFONTNUM    18

#define    NUMSYSFONT    100
typedef struct {
    int        h, flags;
    int        prevFont, nextFont;
} SYSFONT;
//CTFontRef    hFont[NUMSYSFONT];
NSFont    *hFont[NUMSYSFONT];

SYSFONT    SysFonts[NUMSYSFONT];
KLONG    firstSysFont;

int        SysFontW;    // width of SysFontBuf
int        SysFontH;    // height of SysFontBuf
//KLONG    SysFontBMB=-1;
/*
#define	SYSFONTNUM	18

#define	NUMSYSFONT	100
typedef struct {
	HFONT	hFont;
	int		h;
	DWORD	flags;
	int		prevFont, nextFont;
} SYSFONT;

SYSFONT	SysFonts[NUMSYSFONT];
long	firstSysFont;

int		SysFontW;	// width of SysFontBuf
int		SysFontH;	// height of SysFontBuf
WCHAR	*SysFontWbuf=NULL;
int		SysFontWbufLen;
HDC		SFdc=NULL;
HBITMAP	oldSFbm;
*/
///***************************************************************
/// Creates a font such that "text" will be as large as possible
/// while still fitting within a rectangle of width "w" and height "h".
//HFONT FitSysFont(WCHAR *text, int w, int h, DWORD flags, SIZE *fontsize) {
NSFont *FitSysFont(NSString *text, int w, int h, DWORD flags, CGSize *fontsize) {
        
    int        i, len;
    NSFont    *fnt;
    CGSize    size;
    NSString *fntName;//, *fntAttribs;

    i = h;
    len = (int)[text length];
    if( flags & CF_FIXED ) {
        if( flags & CF_BOLD ) {
            if( flags & CF_ITALIC ) fntName = @"Courier-BoldOblique";
            else fntName = @"Courier-Bold";
        } else if( flags & CF_ITALIC ) {
            fntName = @"Courier-Oblique";
        } else fntName = @"Courier";
    } else {
        if( flags & CF_SERIF ) {
            if( flags & CF_BOLD ) {
                if( flags & CF_ITALIC ) fntName = @"Times-BoldItalic";
                else fntName = @"Times-Bold";
            } else if( flags & CF_ITALIC ) {
                fntName = @"Times-Italic";
            } else fntName = @"Times";
        } else {
            if( flags & CF_BOLD ) {
                if( flags & CF_ITALIC ) fntName = @"Helvetica-BoldOblique";
                else fntName = @"Helvetica-Bold";
            } else if( flags & CF_ITALIC ) {
                fntName = @"Helvetica-Oblique";
            } else fntName = @"Helvetica";
        }
    }
    NSMutableDictionary *attribs = [[NSMutableDictionary alloc] init];

    do {
        ++i;
        fnt = [NSFont fontWithName:fntName size:i];
        [attribs setObject:fnt forKey:NSFontAttributeName];
        size = [text sizeWithAttributes:attribs];
    } while( len && size.width<w && size.height<h );
    do {
        --i;
        fnt = [NSFont fontWithName:fntName size:i];
        [attribs setObject:fnt forKey:NSFontAttributeName];
        size = [text sizeWithAttributes:attribs];

        if( i<=4 || (size.width<w && size.height<h) ) break;
    } while( i>4 && (size.width>=w || size.height>=h) );

    if( len==0 || size.width>=w || size.height>=h ) {
        fnt = [NSFont fontWithName:fntName size:i];
        [attribs setObject:fnt forKey:NSFontAttributeName];
        size = [text sizeWithAttributes:attribs];
    }

    if( fontsize!=NULL ) {
        fontsize->width = size.width;
        fontsize->height = size.height;
    }

	return fnt;
}

//**************************************
void InitSysFonts(void) {
#if 0
	HDC		hDC;

	firstSysFont = -1;
	memset(&SysFonts[0], 0xFF, sizeof(SysFonts));
	SysFontW = scrBM.w;
	SysFontH = scrBM.h/15;
	CreateSBM(&SysFontBM, SysFontW, SysFontH);
	SysFontWbufLen = 2048;
	SysFontWbuf = (WCHAR*)malloc(SysFontWbufLen * sizeof(WCHAR));

	hDC = GetDC(NULL);
	if( NULL!=(SFdc=CreateCompatibleDC(hDC)) ) {
		SetBkMode(SFdc, TRANSPARENT);
		SetTextAlign(SFdc, TA_LEFT | TA_TOP);
		oldSFbm = (HBITMAP)SelectObject(SFdc, SysFontBM.hBitmap);
	}
	ReleaseDC(NULL, hDC);
#endif
    
    firstSysFont = -1;
    memset(&SysFonts[0], 0xFF, sizeof(SysFonts));
    SysFontW = scrBM.w;
    SysFontH = scrBM.h/15;
    CreateSBM(&SysFontBM, SysFontW, SysFontH);
//    SysFontWbufLen = 2048;
//    SysFontWbuf = (WCHAR*)malloc(SysFontWbufLen * sizeof(WCHAR));


}

///=============================================================
void FreeSysFonts(void) {

	int		i;
#if 0
	SelectObject(SFdc, oldSFbm);
	DeleteDC(SFdc);
	SFdc = NULL;

	while( firstSysFont >= 0 ) {
		i = SysFonts[firstSysFont].nextFont;
		DeleteObject(SysFonts[firstSysFont].hFont);
		SysFonts[firstSysFont].prevFont = SysFonts[firstSysFont].nextFont = -1;
		firstSysFont = i;
	}

	DeleteSBM(&SysFontBM);

	if( SysFontWbuf!=NULL ) {
		free(SysFontWbuf);
		SysFontWbuf = NULL;
	}
#endif
    while( firstSysFont >= 0 ) {
        i = SysFonts[firstSysFont].nextFont;
        hFont[firstSysFont] = nil;//DeleteObject(SysFonts[firstSysFont].hFont);
        SysFonts[firstSysFont].prevFont = SysFonts[firstSysFont].nextFont = -1;
        firstSysFont = i;
    }
    
//    if( SysFontBMB>=0 ) DeleteBMB(SysFontBMB);
//    SysFontBMB = -1;

}

///=============================================
#if SCREEN_ARGB
void Untrans(KLONG x, KLONG y, KLONG w, KLONG h) {

	KLONG	tx, ty, limx, limy;
	BYTE	*pB;

	limx = min(x+w, SysFontW);
	limy = min(y+h, SysFontH);
	for(ty=y; ty<limy; ty++) {
		for(tx=x, pB = SysFontBM.pBits + SysFontBM.linelen*ty + 4*tx; tx<limx; tx++, pB+=4) {
			*(DWORD*)pB |= 0xFF000000;
		}
	}
}
#else
	#define	Untrans(x,y,w,h)
#endif // SCREEN_RGB

///=============================================
/// RETURN # of CHARs consumed from pText in lower WORD, pixel-width of characters drawn in upper WORD
DWORD DrawSysFontText(DWORD pArgs) {

	DWORD hBM;
	char *pText;
	KLONG x, y, w, h;
	DWORD bclr, fclr, flags;
	char		c;
	int			wcnt=0, n, p, fontix, dx, dy;
	int			i, j, ww, savend, curend;
	SYSFONT		*pSF;
    NSString    *tmpWbuf;

	hBM = GETMEMD(pArgs); pArgs += 4;
	pText = (char*)(mainKE + GETMEMD(pArgs)); pArgs += 4;
/*
if( !strcmp(pText, "YourMoodName") ) {
	x = 0;
}
*/
	x = (int)GETMEMD(pArgs); pArgs += 4;
	y = (int)GETMEMD(pArgs); pArgs += 4;
	w = (int)GETMEMD(pArgs); pArgs += 4;
	h = (int)GETMEMD(pArgs); pArgs += 4;
	flags = GETMEMD(pArgs); pArgs += 4;
	bclr = GETMEMD(pArgs); pArgs += 4;
	fclr = GETMEMD(pArgs); //pArgs += 4;
//	pW = (long*)((GETMEMD(pArgs)==0) ? NULL : (mainKE + GETMEMD(pArgs)));

	/// bail if nothing to do
	if( w==0 || h==0 || *pText==0 ) return 0;
    if( SysFontBM.bitmapContext==NULL || w==0 || h==0 || *pText==0 ) return 0;
    if( bclr==0xFFFFFFFF ) flags &= ~128;    // don't allow rotate if we're not filling the background
    
    /// find or create the desired font
    for(i=fontix=(int)firstSysFont, j=0; i>=0 && (SysFonts[i].h!=h || SysFonts[i].flags!=(flags & CF_FONT_CHARS)); j++) {
        fontix = i;    // save last 'good' SysFont found (the least-used one, in case we're out of slots)
        i = SysFonts[i].nextFont;
    }
    /// if the desired font hasn't been created yet, create it
    if( i<0 ) {    /// font doesn't already exist
        if( j >= NUMSYSFONT ) {    /// there's NO room for another
            pSF = &SysFonts[fontix];    /// point to the least-used font
            if( pSF->nextFont >= 0 ) SysFonts[pSF->nextFont].prevFont = pSF->prevFont;    /// unlink it from the list
            if( pSF->prevFont >= 0 ) SysFonts[pSF->prevFont].nextFont = pSF->nextFont;
            hFont[fontix] = nil;// DeleteObject(SysFonts[fontix].hFont);    /// delete it
        } else {    /// there's room for another font, somewhere
            /// find a free slot
            for(fontix=0; fontix<NUMSYSFONT; fontix++) if( SysFonts[fontix].h<0 ) break;
        }
        /// fontix now indexes to an empty SysFonts[] slot, create the font there
        
        pSF = &SysFonts[fontix];
        
        tmpWbuf = [[NSString alloc] initWithUTF8String:"\320\215y"];
        hFont[fontix] = FitSysFont(tmpWbuf, 2048, h, flags, NULL);
        
        /// link the new font into the front of the list, as it is now most-recent
        pSF->h = h;
        pSF->flags = flags & CF_FONT_CHARS;
        pSF->prevFont = -1;
        pSF->nextFont = (int)firstSysFont;
        if( pSF->nextFont >= 0 ) SysFonts[pSF->nextFont].prevFont = fontix;    /// point previous 'first' font at this new 'first' font
        firstSysFont = fontix;
    } else {    /// font exists, move it to the front of the list (most-recently used) if it's not already there
        fontix = i;
        if( fontix != firstSysFont ) {    /// it's not first
            pSF = &SysFonts[fontix];
            n = pSF->nextFont;
            p = pSF->prevFont;
            if( n >= 0 ) SysFonts[n].prevFont = p;    /// unlink it from the list
            if( p >= 0 ) SysFonts[p].nextFont = n;
            /// we know firstSysFont != -1, because THIS font already exists and was found
            /// so just link this font into the front of the list (now most-recently used)
            pSF->prevFont = -1;
            if( firstSysFont >= 0 ) SysFonts[firstSysFont].prevFont = fontix;
            pSF->nextFont = (int)firstSysFont;
            firstSysFont = fontix;
        }
    }
    tmpWbuf = [[NSString alloc] initWithUTF8String:pText];
    CGSize    size;// = [tmpWbuf sizeWithAttributes:@{NSFontAttributeName: hFont[fontix]}];
    NSPoint    point;
     
    point = NSMakePoint(0,0);
    NSMutableDictionary *attribs = [[NSMutableDictionary alloc] init];
    [attribs setObject:hFont[fontix] forKey:NSFontAttributeName];

    [attribs setObject:[NSColor colorWithRed:(CGFloat)((fclr>>16)&0xFF)/255.0 green:(CGFloat)((fclr>>8)&0xFF)/255.0 blue:(CGFloat)(fclr&0xFF)/255.0 alpha:1.0] forKey:NSForegroundColorAttributeName];
    if( flags & CF_UNDERLINE ) [attribs setObject:[NSNumber numberWithInt:NSUnderlineStyleSingle] forKey:NSUnderlineStyleAttributeName];

    size = [tmpWbuf sizeWithAttributes:attribs];

    wcnt = (int)[tmpWbuf length];
    
    curend = wcnt;    /// because wcnt includes the NUL char, apparently
    /// now draw the string (possibly multiple times if it's too long for the space)
    ww = -1;    /// flag this is the first time through
    
    dy = SysFontH-size.height-[hFont[fontix] descender];
    [NSGraphicsContext saveGraphicsState];
    [NSGraphicsContext setCurrentContext:[NSGraphicsContext graphicsContextWithCGContext:SysFontBM.bitmapContext flipped:FALSE]];

    if( bclr==0xFFFFFFFF ) {
        DrawBM(-2, 0, 0, hBM, x, y, w, h, 0, 0, 0);
    } else {
        [[NSColor colorWithCalibratedRed:(CGFloat)(bclr & 0xFF)/255.0 green:(CGFloat)((bclr>>8)&0xFF)/255.0 blue:(CGFloat)((bclr>>16)&0xFF)/255.0 alpha:1.0] set];
        NSRectFill(NSMakeRect(0, SysFontBM.h-h, (flags & CF_ITALIC) ? min(w+h, 2048) : w, h));
    }
    NSMutableString *frag = [@"" mutableCopy];
    
    if( size.width>w ) {
        if( (flags & TF_BREAK_SPACE) ) {    // if breaking on SPACE
            curend = 0;
            int lastSpace=0;
            for(curend=0; curend<wcnt; curend++) {
                [frag appendString:[tmpWbuf substringWithRange:NSMakeRange(curend, 1)]];
                if( [frag characterAtIndex:curend]==' ' ) lastSpace = curend;
                if( [frag sizeWithAttributes:attribs].width > w ) {
                    if( curend==0 ) return 0;    // if not room to output even a single char
                    if( lastSpace==0 && (flags & TF_BREAK_CHAR) ) {
                        frag = [@"" mutableCopy];
                        goto breakany;
                    }
    // the following line should be commented out!  It causes breaks in the middle of words
    //                if( lastSpace==0 ) lastSpace = curend-1;
                    while( curend && lastSpace>=0 && curend > lastSpace ) {
                        [frag deleteCharactersInRange:NSMakeRange(curend--, 1)];
                    }
                    if( [frag length]==1 && [frag characterAtIndex:0]==' ' ) return 0;
                    break;
                }
            }
        } else if( (flags & TF_BREAK_CHAR) ) {    // if breaking on ANY CHAR
breakany:
            for(curend=0; curend<wcnt; curend++) {
                [frag appendString:[tmpWbuf substringWithRange:NSMakeRange(curend, 1)]];
                if( [frag sizeWithAttributes:attribs].width > w ) {
                    if( curend==0 ) return 0;    // if not room to output even a single char
                    [frag deleteCharactersInRange:NSMakeRange(curend, 1)];
                    break;
                }
            }
        } else {    // no break
            frag = [tmpWbuf mutableCopy];
        }
    } else frag = [tmpWbuf mutableCopy];
    dx = 0;
//    [frag characterAtIndex:0]==' ' ? h/6 : 0;    // if starts with space, give previous char a little more room so we don't clip it
    ww = [frag sizeWithAttributes:attribs].width;
    [frag drawWithRect:NSMakeRect(dx, dy, SysFontBM.w, size.height) options:0 attributes:attribs];
    

    ww = [frag sizeWithAttributes:attribs].width;
    if( curend && (flags & CF_ITALIC) ) {    // italics
        ww += h/6;    // hack! a little slop for last char of italics
    }
    if( !(flags & TF_GET_WIDTH) ) {    // if not just doing a "get text width," copy the drawn text
        if( flags & CF_CENTER ) {            /// CENTERED
            if( (bclr & 0xFF000000)!=0xFF000000 ) {    /// if transparent
                DrawBM(-2, 0, 0, hBM, x+(w-ww)/2, y, ww, h, 0, 0, 0);
            } else {    /// else opaque
                DrawRect(-2, 0, 0, w, h, bclr, 0);
            }
            [frag drawWithRect:NSMakeRect(dx, dy, SysFontBM.w, size.height) options:0 attributes:attribs];
//            TextOutW(SFdc, 0, 0, SysFontWbuf, curend);
            Untrans(0, 0, ww, h);
            DrawBM(hBM, x+(w-ww)/2, y, -2, 0, 0, ww, h, 0, 0, 0);
        } else if( flags & CF_RIGHT ) {    /// RIGHT JUSTIFIED
            if( (bclr&0xFF000000)!=0xFF000000 ) {    /// if transparent
                DrawBM(-2, 0, 0, hBM, x+w-ww, y, ww, h, 0, 0, 0);
            } else {    /// else opaque
                DrawRect(-2, 0, 0, w, h, bclr, 0);
            }
            [frag drawWithRect:NSMakeRect(dx, dy, SysFontBM.w, size.height) options:0 attributes:attribs];
//            TextOutW(SFdc, 0, 0, SysFontWbuf, curend);
            Untrans(0, 0, ww, h);
            DrawBM(hBM, x+w-ww, y, -2, 0, 0, ww, h, 0, 0, 0);
        } else {                    /// LEFT JUSTIFIED
            if( (bclr&0xFF000000)!=0xFF000000 ) {    /// if transparent
                DrawBM(-2, 0, 0, hBM, x, y, ww, h, 0, 0, 0);
            } else {    /// else opaque
                DrawRect(-2, 0, 0, w, h, bclr, 0);
            }
            [frag drawWithRect:NSMakeRect(dx, dy, SysFontBM.w, size.height) options:0 attributes:attribs];
//            TextOutW(SFdc, 0, 0, SysFontWbuf, curend);
            Untrans(0, 0, min(w,ww), h);
            DrawBM(hBM, x, y, -2, 0, 0, min(w,ww), h, 0, 0, 0);
        }
    }
    [NSGraphicsContext restoreGraphicsState];

    for(i=0, savend=curend; savend && pText[i]; --savend) {
        c = pText[i];
        if( c & 0x80 ) {
            if( (c & 0xE0)==0xC0 ) {        /// 2-byte code
                c = pText[++i];
                if( (c & 0xC0)==0x80 ) ++i;
            } else if( (c & 0xF0)==0xE0 ) {    /// 3-byte code
                c = pText[++i];
                if( (c & 0xC0)==0x80 ) {
                    c = pText[++i];
                    if( (c & 0xC0)==0x80 ) ++i;
                }
            } else if( (c & 0xF8)==0xF0 ) {    /// 4-byte code
                c = pText[++i];
                if( (c & 0xC0)==0x80 ) {
                    c = pText[++i];
                    if( (c & 0xC0)==0x80 ) {
                        c = pText[++i];
                        if( (c & 0xC0)==0x80 ) ++i;
                    }
                }
            }
        } else ++i;                            /// 1-byte code
    }
    /// return length drawn (in lower word) & pixel-width of drawn text (in upper word)
//    Dwrite("DrawSysFontText EXIT");
    return (ww<<16)+i;
}

