%{

/*
41   ARGB
73   INDIRB
584  JUMPV
711  VREGP

4117 CNSTI4
4118 CNSTU4
4133 ARGI4
4134 ARGU4
4149 ASGNI4
4150 ASGNU4
4161 INDIRF4
4165 INDIRI4
4166 INDIRU4
4167 INDIRP4

4213 CVFI4

4310 CALLU4
4326 LOADU4
4359 ADDRGP4
4375 ADDRFP4
4391 ADDRLP4

4405 ADDI4
4407 ADDP4
4421 SUBI4
4437 LSHI4
4470 RSHU4
4486 BANDU4

4501 BCOMI4
4582 EQU4

8257 INDIRF8

8305 CVFF8

8657 MULF8
*/

/*
NOTE: Within the various blocks of an .MD file, you can
use "%{\n" to start a literal block (just as THIS block
is started in that way above), and terminate the literal
block with "%}\n" (without the quotes and where '\n'
represents the EOL newline char. These blocks are not
interpreted by LBURG, but rather are copied as-is. So,
if you wish to COMMENT out something that will appear
as a COMMENTED BLOCK in the .c output code, then you'll
also need to nest it between SLASH-STAR and STAR-SLASH
within the "%{\n" "%}\n" block (as is this comment).

So, there's nothing special about THIS block being first
in the file, it's just a block of code/comments that are
included in the output file "as-is". You can use this
method to include comments within the 'rules' section
as well, although those comments will appear "up-front"
in the .c output file, since the rules themselves get
turned into several arrays and switch statements, which
makes it pretty much impossible to include the comment
"in place" where it appears in the .md file.

With the modified KLBURG.EXE, you can also comment out
single lines within the 'rules' section using the
standard '//' C "single-line comment". A line thus
commented-out will NOT appear, even as a comment, in
the .c output file, whereas the above-mentioned comment
BLOCK method WILL appear in the output file, albeit
"out of place."
*/

/*
NOTE: Book pgs 418-419:
ralloc frees the input registers before allocating the output register,
which allows it to reuse an input register as the output register. This
econonmy is always safe when the node is implemented by a single instruction,
but it can be unsafe if a node is implemented by a sequence of instructions:
If the output register is also one of the input registers, and if the
sequence changes the output register before reading the corresponding input
register, then the read fetches a corrupt value. We take care that all rules
that emit instruction sequences set their output register ONLY after they
finish reading all input registers. Most templates emit just one instruction,
so this assumption is a good default, but it does require considerable care
with multi-instruction sequences.
*/
/*
	R7 is used as a scratch register
	R6 is used to hold return values from funcions
	R5-R0 are general purpose registers
*/


enum {	R6=0, R0=1, R1=2, R2=3, R3=4, R4=5, R5=6 };

#include <io.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/stat.h>

#include "c.h"
#define NODEPTR_TYPE Node
#define OP_LABEL(p) ((p)->op)
#define LEFT_CHILD(p) ((p)->kids[0])
#define RIGHT_CHILD(p) ((p)->kids[1])
#define STATE_LABEL(p) ((p)->x.state)
static void address(Symbol, Symbol, long);
static void blkfetch(long, long, long, long);
static void blkloop(long, long, long, long, long, long[]);
static void blkstore(long, long, long, long);
static void defaddress(Symbol);
static void defconst(long, long, Value);
static void defstring(long, char *);
static void defsymbol(Symbol);
static void doarg(Node);
static int emit2(Node);
static void export(Symbol);
static void clobber(Node);
static void function(Symbol, Symbol [], Symbol [], long);
static void global(Symbol);
static void import(Symbol);
static void local(Symbol);
static void progbeg(long, char **);
static void progend(void);
static void segment(long);
static void space(long);
static void target(Node);
static Symbol charreg[32], shortreg[32], longreg[32], longlongreg[32];
static Symbol fltreg[32];
static int memop(Node p);
static int sametree(Node p, Node q);

static Symbol charregw, shortregw, longregw, longlongregw, fltregw, asmresult;

static long cseg;

//static Symbol quo, rem;

%}
%start stmt

%term CNSTF4=4113
%term CNSTF8=8209
%term CNSTI1=1045
%term CNSTI2=2069
%term CNSTI4=4117
%term CNSTI8=8213

%term CNSTP4=4119

%term CNSTU1=1046
%term CNSTU2=2070
%term CNSTU4=4118
%term CNSTU8=8214

%term ARGB=41
%term ARGI2=2085
%term ARGU2=2086
%term ARGF4=4129
%term ARGF8=8225
%term ARGI4=4133
%term ARGI8=8229
%term ARGP4=4135
%term ARGU4=4134
%term ARGU8=8230

%term ASGNB=57
%term ASGNF4=4145
%term ASGNF8=8241
%term ASGNI1=1077
%term ASGNI2=2101
%term ASGNI4=4149
%term ASGNI8=8245
%term ASGNP4=4151
%term ASGNU1=1078
%term ASGNU2=2102
%term ASGNU4=4150
%term ASGNU8=8246

%term INDIRB=73
%term INDIRF4=4161
%term INDIRF8=8257
%term INDIRI1=1093
%term INDIRI2=2117
%term INDIRI4=4165
%term INDIRI8=8261
%term INDIRP4=4167
%term INDIRU1=1094
%term INDIRU2=2118
%term INDIRU4=4166
%term INDIRU8=8262

%term CVFF4=4209
%term CVFF8=8305
%term CVFI2=2165
%term CVFI4=4213
%term CVFI8=8309

%term CVII1=1157
%term CVII2=2181
%term CVII4=4229
%term CVII8=8325
%term CVIU1=1158
%term CVIU2=2182
%term CVIU4=4230
%term CVIU8=8326
%term CVIF4=4225
%term CVIF8=8321

%term CVUI1=1205
%term CVUI2=2229
%term CVUI4=4277
%term CVUI8=8373
%term CVUU1=1206
%term CVUU2=2230
%term CVUU4=4278
%term CVUU8=8374
%term CVUP4=4279
%term CVUF4=4273
%term CVUF8=8369

%term CVPP4=4247
%term CVPU4=4246
%term CVPU8=8342

%term NEGI1=1221
%term NEGI2=2245
%term NEGI4=4293
%term NEGI8=8389
%term NEGF4=4289
%term NEGF8=8385

%term CALLB=217
%term CALLI2=2261
%term CALLU2=2262
%term CALLF4=4305
%term CALLF8=8401
%term CALLI4=4309
%term CALLI8=8405
%term CALLP4=4311
%term CALLU4=4310
%term CALLU8=8406
%term CALLV=216

%term RETI1=1269
%term RETU1=1270
%term RETI2=2293
%term RETU2=2294
%term RETI4=4341
%term RETU4=4342
%term RETP4=4343
%term RETI8=8437
%term RETU8=8438
%term RETF4=4337
%term RETF8=8433
%term RETV=248

%term ADDRFP4=4375
%term ADDRGP4=4359
%term ADDRLP4=4391

%term ADDI1=1333
%term ADDU1=1334
%term ADDI2=2357
%term ADDU2=2358
%term ADDI4=4405
%term ADDU4=4406
%term ADDI8=8501
%term ADDU8=8502
%term ADDP4=4407
%term ADDF4=4401
%term ADDF8=8497

%term SUBI1=1349
%term SUBU1=1350
%term SUBI2=2373
%term SUBU2=2374
%term SUBI4=4421
%term SUBU4=4422
%term SUBI8=8517
%term SUBU8=8518
%term SUBP4=4423
%term SUBF4=4417
%term SUBF8=8513

%term DIVI1=1477
%term DIVU1=1478
%term DIVI2=2501
%term DIVU2=2502
%term DIVI4=4549
%term DIVU4=4550
%term DIVI8=8645
%term DIVU8=8646
%term DIVF4=4545
%term DIVF8=8641

%term MULI1=1493
%term MULU1=1494
%term MULI2=2517
%term MULU2=2518
%term MULI4=4565
%term MULU4=4566
%term MULI8=8661
%term MULU8=8662
%term MULF4=4561
%term MULF8=8657

%term LSHI1=1365
%term LSHU1=1366
%term LSHI2=2389
%term LSHU2=2390
%term LSHI4=4437
%term LSHU4=4438
%term LSHI8=8533
%term LSHU8=8534

%term MODI1=1381
%term MODU1=1382
%term MODI2=2405
%term MODU2=2406
%term MODI4=4453
%term MODU4=4454
%term MODI8=8549
%term MODU8=8550

%term RSHI1=1397
%term RSHU1=1398
%term RSHI2=2421
%term RSHU2=2422
%term RSHI4=4469
%term RSHU4=4470
%term RSHI8=8565
%term RSHU8=8566

%term BANDI1=1413
%term BANDU1=1414
%term BANDI2=2437
%term BANDU2=2438
%term BANDI4=4485
%term BANDU4=4486
%term BANDI8=8581
%term BANDU8=8582

%term BCOMI1=1429
%term BCOMU1=1430
%term BCOMI2=2453
%term BCOMU2=2454
%term BCOMI4=4501
%term BCOMU4=4502
%term BCOMI8=8597
%term BCOMU8=8598

%term BORI1=1445
%term BORU1=1446
%term BORI2=2469
%term BORU2=2470
%term BORI4=4517
%term BORU4=4518
%term BORI8=8613
%term BORU8=8614

%term BXORI1=1461
%term BXORU1=1462
%term BXORI2=2485
%term BXORU2=2486
%term BXORI4=4533
%term BXORU4=4534
%term BXORI8=8629
%term BXORU8=8630

%term EQI1=1509
%term EQU1=1510
%term EQI2=2533
%term EQU2=2534
%term EQI4=4581
%term EQU4=4582
%term EQI8=8677
%term EQU8=8678
%term EQF4=4577
%term EQF8=8673

%term GEI1=1525
%term GEU1=1526
%term GEI2=2549
%term GEU2=2550
%term GEI4=4597
%term GEU4=4598
%term GEI8=8693
%term GEU8=8694
%term GEF4=4593
%term GEF8=8689

%term GTI1=1541
%term GTU1=1542
%term GTI2=2565
%term GTU2=2566
%term GTI4=4613
%term GTU4=4614
%term GTI8=8709
%term GTU8=8710
%term GTF4=4609
%term GTF8=8705

%term LEI1=1557
%term LEU1=1558
%term LEI2=2581
%term LEU2=2582
%term LEI4=4629
%term LEU4=4630
%term LEI8=8725
%term LEU8=8726
%term LEF4=4625
%term LEF8=8721

%term LTI1=1573
%term LTU1=1574
%term LTI2=2597
%term LTU2=2598
%term LTI4=4645
%term LTU4=4646
%term LTI8=8741
%term LTU8=8742
%term LTF4=4641
%term LTF8=8737

%term NEI1=1589
%term NEU1=1590
%term NEI2=2613
%term NEU2=2614
%term NEI4=4661
%term NEU4=4662
%term NEI8=8757
%term NEU8=8758
%term NEF4=4657
%term NEF8=8753

%term JUMPV=584

%term LABELV=600

%term ASM0=720
%term ASM1=736
%term ASM2=752
%term ASMRETU4=4870
%term ASMRETU8=8966

%term LOADB=233
%term LOADI1=1253
%term LOADU1=1254
%term LOADI2=2277
%term LOADU2=2278
%term LOADI4=4325
%term LOADU4=4326
%term LOADP4=4327
%term LOADI8=8421
%term LOADU8=8422
%term LOADF4=4321
%term LOADF8=8417


%term VREGP=711
%%

con1:	CNSTI1	"%a"
con1:	CNSTU1	"%a"
con2:	CNSTI2	"%a"
con2:	CNSTU2	"%a"
con4:	CNSTI4	"%a"
con4:	CNSTU4	"%a"
con4:	CNSTP4	"%a"
con8:	CNSTI8	"%a"
con8:	CNSTU8	"%a"
con4:	CNSTF4	"%a"
con8:	CNSTF8	"%a"

con:	con1	"%a"
con:	con2	"%a"
con:	con4	"%a"
con:	con8	"%a"

conv0:	CNSTI1	"0"	range(a,0,0)	
conv0:	CNSTI2	"0"	range(a,0,0)
conv0:	CNSTI4	"0"	range(a,0,0)
conv0:	CNSTI8	"0"	range(a,0,0)
conv0:	CNSTU1	"0"	range(a,0,0)
conv0:	CNSTU2	"0"	range(a,0,0)
conv0:	CNSTU4	"0"	range(a,0,0)
conv0:	CNSTU8	"0"	range(a,0,0)

conv1:	CNSTI1	"1"	range(a,1,1)	
conv1:	CNSTI2	"1"	range(a,1,1)
conv1:	CNSTI4	"1"	range(a,1,1)
conv1:	CNSTI8	"1"	range(a,1,1)
conv1:	CNSTU1	"1"	range(a,1,1)
conv1:	CNSTU2	"1"	range(a,1,1)
conv1:	CNSTU4	"1"	range(a,1,1)
conv1:	CNSTU8	"1"	range(a,1,1)

conv2:		CNSTI4 "2"		range(a, 2, 2)
conv2:		CNSTU4 "2"		range(a, 2, 2)
conv4:		CNSTI4 "4"		range(a, 4, 4)
conv4:		CNSTU4 "4"		range(a, 4, 4)
conv8:		CNSTI4 "8"		range(a, 8, 8)
conv8:		CNSTU4 "8"		range(a, 8, 8)
conv16:		CNSTI4 "16"		range(a, 16, 16)
conv16:		CNSTU4 "16"		range(a, 16, 16)
conv32:		CNSTI4 "32"		range(a, 32, 32)
conv32:		CNSTU4 "32"		range(a, 32, 32)
conv64:		CNSTI4 "64"		range(a, 64, 64)
conv64:		CNSTU4 "64"		range(a, 64, 64)
conv128:	CNSTI4 "128"	range(a, 128, 128)
conv128:	CNSTU4 "128"	range(a, 128, 128)
conv256:	CNSTI4 "256"	range(a, 256, 256)
conv256:	CNSTU4 "256"	range(a, 256, 256)

reg:	INDIRI1(VREGP)		"#\n"
reg:	INDIRU1(VREGP)		"#\n"
reg:	INDIRI2(VREGP)		"#\n"
reg:	INDIRU2(VREGP)		"#\n"
reg:	INDIRI4(VREGP)		"#\n"
reg:	INDIRU4(VREGP)		"#\n"
reg:	INDIRP4(VREGP)		"#\n"
reg:	INDIRI8(VREGP)		"#\n"
reg:	INDIRU8(VREGP)		"#\n"
reg:	INDIRF4(VREGP)		"#\n"
reg:	INDIRF8(VREGP)		"#\n"

// changed all of these from "x ? mov %c,%0\n" to "#\n"
reg:	LOADI1(reg)			"#\n"  1
reg:	LOADU1(reg)			"#\n"  1
reg:	LOADI2(reg)			"#\n"  1
reg:	LOADU2(reg)			"#\n"  1
reg:	LOADI4(reg)			"#\n"  1
reg:	LOADU4(reg)			"#\n"  1
reg:	LOADP4(reg)			"#\n"  1
reg:	LOADI8(reg)			"#\n"  1
reg:	LOADU8(reg)			"#\n"  1

reg:	LOADI4(adr)			"\td mov %c,%0\n"	1
reg:	LOADU4(ad)			"\td lea %c,%0\n"	1
reg:	LOADU4(INDIRU8(ad))	"\td lea %c,%0\n"	0

reg:	LOADU2(LOADU4(reg))	"#\n" 0

reg:	con1				"\tb mov %c,=%0\n"	1
reg:	con2				"\tw mov %c,=%0\n"	1
reg:	con4				"\td mov %c,=%0\n"	1
reg:	con8				"\tq mov %c,=%0\n"	1

stmt:	reg	""

ca:		ADDRGP4				"=%a"
ca:		reg					"%0"

ad:		ADDRGP4				"[%a]"
ad:		ADDRFP4				"[FP][%a]"
ad:		ADDRLP4				"[FP][%a]"

adr:	reg					"[%0]"
adr:	ad					"%0"

iadr1:	INDIRI1(adr)		"%0"
iadr1:	INDIRU1(adr)		"%0"
iadr2:	INDIRI2(adr)		"%0"
iadr2:	INDIRU2(adr)		"%0"
iadr4:	INDIRI4(adr)		"%0"
iadr4:	INDIRU4(adr)		"%0"
iadr4:	INDIRP4(adr)		"%0"
iadr8:	INDIRI8(adr)		"%0"
iadr8:	INDIRU8(adr)		"%0"
iadr4:	INDIRF4(adr)		"%0"
iadr8:	INDIRF8(adr)		"%0"

iadr:	iadr1				"%0"
iadr:	iadr2				"%0"
iadr:	iadr4				"%0"
iadr:	iadr8				"%0"
iadr:	INDIRB( adr)		"%0"

eadr1:	INDIRI1(ad)			"%0"
eadr1:	INDIRU1(ad)			"%0"
eadr2:	INDIRI2(ad)			"%0"
eadr2:	INDIRU2(ad)			"%0"
eadr4:	INDIRI4(ad)			"%0"
eadr4:	INDIRU4(ad)			"%0"
eadr4:	INDIRP4(ad)			"%0"
eadr8:	INDIRI8(ad)			"%0"
eadr8:	INDIRU8(ad)			"%0"
eadr4:	INDIRF4(ad)			"%0"
eadr8:	INDIRF8(ad)			"%0"

eadr:	eadr1				"%0"
eadr:	eadr2				"%0"
eadr:	eadr4				"%0"
eadr:	eadr8				"%0"
eadr:	INDIRB(ad)			"%0"

reg:	iadr1				"\tb mov %c,%0\n"	1
reg:	iadr2				"\tw mov %c,%0\n"	1
reg:	iadr4				"\td mov %c,%0\n"	1
reg:	iadr8				"\tq mov %c,%0\n"	1

reg:	ad					"\td lea %c,%0\n"	1
//rc:		reg					"%0"
//rc:		con					"=%a"

dm:	reg						"%0"
dm:	iadr					"%0"
dm:	ADDRGP4					"%a"
dm:	ADDRFP4					"[FP][%a]"
dm:	ADDRLP4					"[FP][%a]"

am:	reg						"%0"
am:	con						"=%a"
am:	iadr					"%0"
am:	ADDRGP4					"=%a"
am:	ADDRFP4					"[FP][%a]"
am:	ADDRLP4					"[FP][%a]"

stmt:	ASGNI1(VREGP,reg)		"# write register\n"
stmt:	ASGNU1(VREGP,reg)		"# write register\n"
stmt:	ASGNI2(VREGP,reg)		"# write register\n"
stmt:	ASGNU2(VREGP,reg)		"# write register\n"
stmt:	ASGNI4(VREGP,reg)		"# write register\n"
stmt:	ASGNU4(VREGP,reg)		"# write register\n"
stmt:	ASGNP4(VREGP,reg)		"# write register\n"
stmt:	ASGNF4(VREGP,reg)		"# write register\n"
stmt:	ASGNI8(VREGP,reg)		"# write register\n"
stmt:	ASGNU8(VREGP,reg)		"# write register\n"
stmt:	ASGNF8(VREGP,reg)		"# write register\n"

stmt:	ASGNP4(VREGP, ADDRGP4)	"\td mov %c,%1\n" 20

stmt:	ASGNI1(adr, am)			"\tb mov %0,%1\n"	1
stmt:	ASGNU1(adr, am)			"\tb mov %0,%1\n"	1
stmt:	ASGNI2(adr, am)			"\tw mov %0,%1\n"	1
stmt:	ASGNU2(adr, am)			"\tw mov %0,%1\n"	1
stmt:	ASGNI4(adr, am)			"\td mov %0,%1\n"	1
stmt:	ASGNU4(adr, am)			"\td mov %0,%1\n"	1

stmt:	ASGNU2(INDIRP4(VREGP), LOADU2(reg))	"\tw mov [%a],%0\n"	1
reg:	ASGNU2(INDIRP4(VREGP), LOADU2(reg))	"\tw mov %a,%0\n"	0
stmt:	ASGNU1(adr, LOADU1(INDIRU4(ad)))	"\tw mov %0,%1\n"	1
stmt:	ASGNU2(adr, LOADU2(INDIRU4(ad)))	"\tw mov %0,%1\n"	1
stmt:	ASGNI1(adr, LOADI1(INDIRI4(ad)))	"\tw mov %0,%1\n"	1
stmt:	ASGNI2(adr, LOADI2(INDIRI4(ad)))	"\tw mov %0,%1\n"	1

stmt:	ASGNP4(adr, reg)		"\td mov %0,%1\n" 	1
stmt:	ASGNP4(adr, con)		"\td mov %0,=%1\n" 	1
stmt:	ASGNP4(adr, iadr)		"\td mov %0,%1\n" 	1
stmt:	ASGNP4(adr, ad)			"\td lea %0,%1\n" 	1

stmt:	ASGNI8(adr, am)			"\tq mov %0,%1\n"	1
stmt:	ASGNU8(adr, am)			"\tq mov %0,%1\n"	1
stmt:	ASGNF4(adr, am)			"\td mov %0,%1\n"	1
stmt:	ASGNF8(adr, am)			"\tq mov %0,%1\n"	1

stmt:	ASGNI1(adr, ADDI1(INDIRI1(adr), am))	"\tb add %0,%2\n"		memop(a)
stmt:	ASGNU1(adr, ADDU1(INDIRU1(adr), am))	"\tb add %0,%2\n"		memop(a)
stmt:	ASGNI2(adr, ADDI2(INDIRI2(adr), am))	"\tw add %0,%2\n"		memop(a)
stmt:	ASGNU2(adr, ADDU2(INDIRU2(adr), am))	"\tw add %0,%2\n"		memop(a)
stmt:	ASGNI4(adr, ADDI4(INDIRI4(adr), am))	"\td add %0,%2\n"		memop(a)
stmt:	ASGNU4(adr, ADDU4(INDIRU4(adr), am))	"\td add %0,%2\n"		memop(a)
stmt:	ASGNI8(adr, ADDI8(INDIRI8(adr), am))	"\tq add %0,%2\n"		memop(a)
stmt:	ASGNU8(adr, ADDU8(INDIRU8(adr), am))	"\tq add %0,%2\n"		memop(a)

stmt:	ASGNI4(adr, SUBI4(INDIRI4(adr), am))	"\td sub %0,%2\n"		memop(a)
stmt:	ASGNU4(adr, SUBU4(INDIRU4(adr), am))	"\td sub %0,%2\n"		memop(a)

stmt:	ASGNI4(adr, RSHI4(INDIRI4(adr), am))	"\td rshi %0,%2\n"	memop(a)
stmt:	ASGNU4(adr, RSHU4(INDIRU4(adr), am))	"\td rshu %0,%2\n"	memop(a)
stmt:	ASGNI4(adr, LSHI4(INDIRI4(adr), am))	"\td lsh %0,%2\n"		memop(a)
stmt:	ASGNU4(adr, LSHU4(INDIRU4(adr), am))	"\td lsh %0,%2\n"		memop(a)
stmt:	ASGNI4(adr, BANDI4(INDIRI4(adr), am))	"\td and %0,%2\n"		memop(a)
stmt:	ASGNU4(adr, BANDU4(INDIRU4(adr), am))	"\td and %0,%2\n"		memop(a)
stmt:	ASGNI4(adr, BORI4(INDIRI4(adr), am))	"\td or %0,%2\n"		memop(a)
stmt:	ASGNU4(adr, BORU4(INDIRU4(adr), am))	"\td or %0,%2\n"		memop(a)
stmt:	ASGNI4(adr, BXORI4(INDIRI4(adr), am))	"\td xor %0,%2\n"		memop(a)
stmt:	ASGNU4(adr, BXORU4(INDIRU4(adr), am))	"\td xor %0,%2\n"		memop(a)

stmt:	ASGNB(adr, iadr)	"\td mov [0x80],=%a\n\tb memmov %0,%1\n"	2
stmt:	ARGB(iadr)			"\tpumem =%a,%0\n"	1

reg:	ADDI1(reg, am)		"?\tb mov %c,%0\n\tb add  %c,%1\n"	2
reg:	ADDU1(reg, am)		"?\tb mov %c,%0\n\tb add  %c,%1\n"	2
reg:	ADDI2(reg, am)		"?\tw mov %c,%0\n\tw add  %c,%1\n"	2
reg:	ADDU2(reg, am)		"?\tw mov %c,%0\n\tw add  %c,%1\n"	2
reg:	ADDI4(reg, am)		"?\td mov %c,%0\n\td add  %c,%1\n"	2
reg:	ADDU4(reg, am)		"?\td mov %c,%0\n\td add  %c,%1\n"	2
reg:	ADDI8(reg, am)		"?\tq mov %c,%0\n\tq add  %c,%1\n"	2
reg:	ADDU8(reg, am)		"?\tq mov %c,%0\n\tq add  %c,%1\n"	2
reg:	ADDF4(reg, am)		"?\td mov %c,%0\n\td addf %c,%1\n"	2
reg:	ADDF8(reg, am)		"?\tq mov %c,%0\n\tq addf %c,%1\n"	2

reg:	ADDP4(reg, ca)		"?\td mov %c,%0\n\td add  %c,%1\n"	2

reg:	SUBI1(reg, am)		"?\tb mov %c,%0\n\tb sub  %c,%1\n"	2
reg:	SUBU1(reg, am)		"?\tb mov %c,%0\n\tb sub  %c,%1\n"	2
reg:	SUBI2(reg, am)		"?\tw mov %c,%0\n\tw sub  %c,%1\n"	2
reg:	SUBU2(reg, am)		"?\tw mov %c,%0\n\tw sub  %c,%1\n"	2
reg:	SUBI4(reg, am)		"?\td mov %c,%0\n\td sub  %c,%1\n"	2
reg:	SUBU4(reg, am)		"?\td mov %c,%0\n\td sub  %c,%1\n"	2
reg:	SUBP4(reg, am)		"?\td mov %c,%0\n\td sub  %c,%1\n"	2
reg:	SUBI8(reg, am)		"?\tq mov %c,%0\n\tq sub  %c,%1\n"	2
reg:	SUBU8(reg, am)		"?\tq mov %c,%0\n\tq sub  %c,%1\n"	2
reg:	SUBF4(reg, am)		"?\td mov %c,%0\n\td subf %c,%1\n"	2
reg:	SUBF8(reg, am)		"?\tq mov %c,%0\n\tq subf %c,%1\n"	2

reg:	MULU4(con, LOADU4(eadr))	"\td mov %c,=%0\n\td mul %c,%1\n"	2

reg:	MULI1(reg, am)		"?\tb mov %c,%0\n\tb mul  %c,%1\n"	3
reg:	MULU1(reg, am)		"?\tb mov %c,%0\n\tb mul  %c,%1\n"	3
reg:	MULI2(reg, am)		"?\tw mov %c,%0\n\tw mul  %c,%1\n"	3
reg:	MULU2(reg, am)		"?\tw mov %c,%0\n\tw mul  %c,%1\n"	3
reg:	MULI4(reg, am)		"?\td mov %c,%0\n\td mul  %c,%1\n"	3
reg:	MULU4(reg, am)		"?\td mov %c,%0\n\td mul  %c,%1\n"	3
reg:	MULI8(reg, am)		"?\tq mov %c,%0\n\tq mul  %c,%1\n"	3
reg:	MULU8(reg, am)		"?\tq mov %c,%0\n\tq mul  %c,%1\n"	3
reg:	MULF4(reg, am)		"?\td mov %c,%0\n\td mulf %c,%1\n"	3
reg:	MULF8(reg, am)		"?\tq mov %c,%0\n\tq mulf %c,%1\n"	3

reg:	MULI4(reg,conv2)	"?\td mov %c,%0\n\td lsh %c,=1\n"	2
reg:	MULI4(reg,conv4)	"?\td mov %c,%0\n\td lsh %c,=2\n"	2
reg:	MULI4(reg,conv8)	"?\td mov %c,%0\n\td lsh %c,=3\n"	2
reg:	MULI4(reg,conv16)	"?\td mov %c,%0\n\td lsh %c,=4\n"	2
reg:	MULI4(reg,conv32)	"?\td mov %c,%0\n\td lsh %c,=5\n"	2
reg:	MULI4(reg,conv64)	"?\td mov %c,%0\n\td lsh %c,=6\n"	2
reg:	MULI4(reg,conv128)	"?\td mov %c,%0\n\td lsh %c,=7\n"	2
reg:	MULI4(reg,conv256)	"?\td mov %c,%0\n\td lsh %c,=8\n"	2
reg:	MULU4(reg,conv2)	"?\td mov %c,%0\n\td lsh %c,=1\n"	2
reg:	MULU4(reg,conv4)	"?\td mov %c,%0\n\td lsh %c,=2\n"	2
reg:	MULU4(reg,conv8)	"?\td mov %c,%0\n\td lsh %c,=3\n"	2
reg:	MULU4(reg,conv16)	"?\td mov %c,%0\n\td lsh %c,=4\n"	2
reg:	MULU4(reg,conv32)	"?\td mov %c,%0\n\td lsh %c,=5\n"	2
reg:	MULU4(reg,conv64)	"?\td mov %c,%0\n\td lsh %c,=6\n"	2
reg:	MULU4(reg,conv128)	"?\td mov %c,%0\n\td lsh %c,=7\n"	2
reg:	MULU4(reg,conv256)	"?\td mov %c,%0\n\td lsh %c,=8\n"	2

reg:	DIVI1(reg, am)		"?\tb mov %c,%0\n\tb divi %c,%1\n"	3
reg:	DIVU1(reg, am)		"?\tb mov %c,%0\n\tb divu %c,%1\n"	3
reg:	DIVI2(reg, am)		"?\tw mov %c,%0\n\tw divi %c,%1\n"	3
reg:	DIVU2(reg, am)		"?\tw mov %c,%0\n\tw divu %c,%1\n"	3
reg:	DIVI4(reg, am)		"?\td mov %c,%0\n\td divi %c,%1\n"	3
reg:	DIVU4(reg, am)		"?\td mov %c,%0\n\td divu %c,%1\n"	3
reg:	DIVI8(reg, am)		"?\tq mov %c,%0\n\tq divi %c,%1\n"	3
reg:	DIVU8(reg, am)		"?\tq mov %c,%0\n\tq divu %c,%1\n"	3
reg:	DIVF4(reg, am)		"?\td mov %c,%0\n\td divf %c,%1\n"	3
reg:	DIVF8(reg, am)		"?\tq mov %c,%0\n\tq divf %c,%1\n"	3

reg:	DIVI4(reg,conv2)	"?\td mov %c,%0\n\td rshi %c,=1\n"	2
reg:	DIVI4(reg,conv4)	"?\td mov %c,%0\n\td rshi %c,=2\n"	2
reg:	DIVI4(reg,conv8)	"?\td mov %c,%0\n\td rshi %c,=3\n"	2
reg:	DIVI4(reg,conv16)	"?\td mov %c,%0\n\td rshi %c,=4\n"	2
reg:	DIVI4(reg,conv32)	"?\td mov %c,%0\n\td rshi %c,=5\n"	2
reg:	DIVI4(reg,conv64)	"?\td mov %c,%0\n\td rshi %c,=6\n"	2
reg:	DIVI4(reg,conv128)	"?\td mov %c,%0\n\td rshi %c,=7\n"	2
reg:	DIVI4(reg,conv256)	"?\td mov %c,%0\n\td rshi %c,=8\n"	2
reg:	DIVU4(reg,conv2)	"?\td mov %c,%0\n\td rshu %c,=1\n"	2
reg:	DIVU4(reg,conv4)	"?\td mov %c,%0\n\td rshu %c,=2\n"	2
reg:	DIVU4(reg,conv8)	"?\td mov %c,%0\n\td rshu %c,=3\n"	2
reg:	DIVU4(reg,conv16)	"?\td mov %c,%0\n\td rshu %c,=4\n"	2
reg:	DIVU4(reg,conv32)	"?\td mov %c,%0\n\td rshu %c,=5\n"	2
reg:	DIVU4(reg,conv64)	"?\td mov %c,%0\n\td rshu %c,=6\n"	2
reg:	DIVU4(reg,conv128)	"?\td mov %c,%0\n\td rshu %c,=7\n"	2
reg:	DIVU4(reg,conv256)	"?\td mov %c,%0\n\td rshu %c,=8\n"	2

reg:	MODI1(reg, am)		"?\tb mov %c,%0\n\tb modi %c,%1\n"	2
reg:	MODU1(reg, am)		"?\tb mov %c,%0\n\tb modu %c,%1\n"	2
reg:	MODI2(reg, am)		"?\tw mov %c,%0\n\tw modi %c,%1\n"	2
reg:	MODU2(reg, am)		"?\tw mov %c,%0\n\tw modu %c,%1\n"	2
reg:	MODI4(reg, am)		"?\td mov %c,%0\n\td modi %c,%1\n"	2
reg:	MODU4(reg, am)		"?\td mov %c,%0\n\td modu %c,%1\n"	2
reg:	MODI8(reg, am)		"?\tq mov %c,%0\n\tq modi %c,%1\n"	2
reg:	MODU8(reg, am)		"?\tq mov %c,%0\n\tq modu %c,%1\n"	2

reg:	RSHI1(reg, am)		"?\tb mov %c,%0\n\tb rshi %c,%1\n"	2
reg:	RSHU1(reg, am)		"?\tb mov %c,%0\n\tb rshu %c,%1\n"	2
reg:	RSHI2(reg, am)		"?\tw mov %c,%0\n\tw rshi %c,%1\n"	2
reg:	RSHU2(reg, am)		"?\tw mov %c,%0\n\tw rshu %c,%1\n"	2
reg:	RSHI4(reg, am)		"?\td mov %c,%0\n\td rshi %c,%1\n"	2
reg:	RSHU4(reg, am)		"?\td mov %c,%0\n\td rshu %c,%1\n"	2
reg:	RSHI8(reg, am)		"?\tq mov %c,%0\n\tq rshi %c,%1\n"	2
reg:	RSHU8(reg, am)		"?\tq mov %c,%0\n\tq rshu %c,%1\n"	2

reg:	LSHI1(reg, am)		"?\tb mov %c,%0\n\tb lsh %c,%1\n"	1
reg:	LSHU1(reg, am)		"?\tb mov %c,%0\n\tb lsh %c,%1\n"	1
reg:	LSHI2(reg, am)		"?\tw mov %c,%0\n\tw lsh %c,%1\n"	1
reg:	LSHU2(reg, am)		"?\tw mov %c,%0\n\tw lsh %c,%1\n"	1
reg:	LSHI4(reg, am)		"?\td mov %c,%0\n\td lsh %c,%1\n"	1
reg:	LSHU4(reg, am)		"?\td mov %c,%0\n\td lsh %c,%1\n"	1
reg:	LSHI8(reg, am)		"?\tq mov %c,%0\n\tq lsh %c,%1\n"	1
reg:	LSHU8(reg, am)		"?\tq mov %c,%0\n\tq lsh %c,%1\n"	1

reg:	BANDI1(reg, am)		"?\tb mov %c,%0\n\tb and %c,%1\n"	2
reg:	BANDU1(reg, am)		"?\tb mov %c,%0\n\tb and %c,%1\n"	2
reg:	BANDI2(reg, am)		"?\tw mov %c,%0\n\tw and %c,%1\n"	2
reg:	BANDU2(reg, am)		"?\tw mov %c,%0\n\tw and %c,%1\n"	2
reg:	BANDI4(reg, am)		"?\td mov %c,%0\n\td and %c,%1\n"	2
reg:	BANDU4(reg, am)		"?\td mov %c,%0\n\td and %c,%1\n"	2
reg:	BANDI8(reg, am)		"?\tq mov %c,%0\n\tq and %c,%1\n"	2
reg:	BANDU8(reg, am)		"?\tq mov %c,%0\n\tq and %c,%1\n"	2

reg:	BORI1(reg, am)		"?\tb mov %c,%0\n\tb or %c,%1\n"	2
reg:	BORU1(reg, am)		"?\tb mov %c,%0\n\tb or %c,%1\n"	2
reg:	BORI2(reg, am)		"?\tw mov %c,%0\n\tw or %c,%1\n"	2
reg:	BORU2(reg, am)		"?\tw mov %c,%0\n\tw or %c,%1\n"	2
reg:	BORI4(reg, am)		"?\td mov %c,%0\n\td or %c,%1\n"	2
reg:	BORU4(reg, am)		"?\td mov %c,%0\n\td or %c,%1\n"	2
reg:	BORI8(reg, am)		"?\tq mov %c,%0\n\tq or %c,%1\n"	2
reg:	BORU8(reg, am)		"?\tq mov %c,%0\n\tq or %c,%1\n"	2

reg:	BXORI1(reg, am)		"?\tb mov %c,%0\n\tb xor %c,%1\n"	2
reg:	BXORU1(reg, am)		"?\tb mov %c,%0\n\tb xor %c,%1\n"	2
reg:	BXORI2(reg, am)		"?\tw mov %c,%0\n\tw xor %c,%1\n"	2
reg:	BXORU2(reg, am)		"?\tw mov %c,%0\n\tw xor %c,%1\n"	2
reg:	BXORI4(reg, am)		"?\td mov %c,%0\n\td xor %c,%1\n"	2
reg:	BXORU4(reg, am)		"?\td mov %c,%0\n\td xor %c,%1\n"	2
reg:	BXORI8(reg, am)		"?\tq mov %c,%0\n\tq xor %c,%1\n"	2
reg:	BXORU8(reg, am)		"?\tq mov %c,%0\n\tq xor %c,%1\n"	2

reg:	BCOMI1(reg)			"?\tb mov %c,%0\n\tb not %c\n"	1
reg:	BCOMU1(reg)			"?\tb mov %c,%0\n\tb not %c\n"	1
reg:	BCOMI2(reg)			"?\tw mov %c,%0\n\tw not %c\n"	1
reg:	BCOMU2(reg)			"?\tw mov %c,%0\n\tw not %c\n"	1
reg:	BCOMI4(reg)			"?\td mov %c,%0\n\td not %c\n"	1
reg:	BCOMU4(reg)			"?\td mov %c,%0\n\td not %c\n"	1
reg:	BCOMI8(reg)			"?\tq mov %c,%0\n\tq not %c\n"	1
reg:	BCOMU8(reg)			"?\tq mov %c,%0\n\tq not %c\n"	1

stmt:	BCOMI1(am)			"\tb not %0\n"	2
stmt:	BCOMU1(am)			"\tb not %0\n"	2
stmt:	BCOMI2(am)			"\tw not %0\n"	2
stmt:	BCOMU2(am)			"\tw not %0\n"	2
stmt:	BCOMI4(am)			"\td not %0\n"	2
stmt:	BCOMU4(am)			"\td not %0\n"	2
stmt:	BCOMI8(am)			"\tq not %0\n"	2
stmt:	BCOMU8(am)			"\tq not %0\n"	2

reg:	NEGI1(reg)			"?\tb mov %c,%0\n\tb neg  %c\n"	2
reg:	NEGI2(reg)			"?\tw mov %c,%0\n\tw neg  %c\n"	2
reg:	NEGI4(reg)			"?\td mov %c,%0\n\td neg  %c\n"	2
reg:	NEGI8(reg)			"?\tq mov %c,%0\n\tq neg  %c\n"	2
reg:	NEGF4(reg)			"?\td mov %c,%0\n\td negf %c\n"	2
reg:	NEGF8(reg)			"?\tq mov %c,%0\n\tq negf %c\n"	2

stmt:	EQI1(dm, conv0)			"\tb tljeq %0,=%a\n"	1
stmt:	EQU1(dm, conv0)			"\tb tljeq %0,=%a\n"	1
stmt:	EQI2(dm, conv0)			"\tw tljeq %0,=%a\n"	1
stmt:	EQU2(dm, conv0)			"\tw tljeq %0,=%a\n"	1
stmt:	EQI4(dm, conv0)			"\td tljeq %0,=%a\n"	1
stmt:	EQU4(dm, conv0)			"\td tljeq %0,=%a\n"	1
stmt:	EQI8(dm, conv0)			"\tq tljeq %0,=%a\n"	1
stmt:	EQU8(dm, conv0)			"\tq tljeq %0,=%a\n"	1
stmt:	EQF4(dm, conv0)			"\td tljeq %0,=%a\n"	1
stmt:	EQF8(dm, conv0)			"\tq tljeq %0,=%a\n"	1

stmt:	EQI1(BANDI1(dm,conv1), conv0)	"\tb tljev %0,=%a\n"	1
stmt:	EQU1(BANDU1(dm,conv1), conv0)	"\tb tljev %0,=%a\n"	1
stmt:	EQI2(BANDI2(dm,conv1), conv0)	"\tw tljev %0,=%a\n"	1
stmt:	EQU2(BANDU2(dm,conv1), conv0)	"\tw tljev %0,=%a\n"	1
stmt:	EQI4(BANDI4(dm,conv1), conv0)	"\td tljev %0,=%a\n"	1
stmt:	EQU4(BANDU4(dm,conv1), conv0)	"\td tljev %0,=%a\n"	1
stmt:	EQI8(BANDI8(dm,conv1), conv0)	"\tq tljev %0,=%a\n"	1
stmt:	EQU8(BANDU8(dm,conv1), conv0)	"\tq tljev %0,=%a\n"	1

stmt:	EQI1(BANDI1(dm,conv1), conv1)	"\tb tljod %0,=%a\n" 1
stmt:	EQU1(BANDU1(dm,conv1), conv1)	"\tb tljod %0,=%a\n" 1
stmt:	EQI2(BANDI2(dm,conv1), conv1)	"\tw tljod %0,=%a\n" 1
stmt:	EQU2(BANDU2(dm,conv1), conv1)	"\tw tljod %0,=%a\n" 1
stmt:	EQI4(BANDI4(dm,conv1), conv1)	"\td tljod %0,=%a\n" 1
stmt:	EQU4(BANDU4(dm,conv1), conv1)	"\td tljod %0,=%a\n" 1
stmt:	EQI8(BANDI8(dm,conv1), conv1)	"\tq tljod %0,=%a\n" 1
stmt:	EQU8(BANDU8(dm,conv1), conv1)	"\tq tljod %0,=%a\n" 1

stmt:	EQI1(dm, am)			"\tb cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQU1(dm, am)			"\tb cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQI2(dm, am)			"\tw cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQU2(dm, am)			"\tw cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQI4(dm, am)			"\td cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQU4(dm, am)			"\td cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQI8(dm, am)			"\tq cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQU8(dm, am)			"\tq cmp  %0,%1\n\tljeq =%a\n" 1
stmt:	EQF4(dm, am)			"\td cmpf %0,%1\n\tljeq =%a\n" 1
stmt:	EQF8(dm, am)			"\tq cmpf %0,%1\n\tljeq =%a\n" 1

stmt:	NEI1(dm, conv0)			"\tb tljne %0,=%a\n" 1
stmt:	NEU1(dm, conv0)			"\tb tljne %0,=%a\n" 1
stmt:	NEI2(dm, conv0)			"\tw tljne %0,=%a\n" 1
stmt:	NEU2(dm, conv0)			"\tw tljne %0,=%a\n" 1
stmt:	NEI4(dm, conv0)			"\td tljne %0,=%a\n" 1
stmt:	NEU4(dm, conv0)			"\td tljne %0,=%a\n" 1
stmt:	NEI8(dm, conv0)			"\tq tljne %0,=%a\n" 1
stmt:	NEU8(dm, conv0)			"\tq tljne %0,=%a\n" 1
stmt:	NEF4(dm, conv0)			"\td tljne %0,=%a\n" 1
stmt:	NEF8(dm, conv0)			"\tq tljne %0,=%a\n" 1

stmt:	NEI1(BANDI1(dm,conv1), conv0)	"\tb tljod %0,=%a\n" 1
stmt:	NEU1(BANDU1(dm,conv1), conv0)	"\tb tljod %0,=%a\n" 1
stmt:	NEI2(BANDI2(dm,conv1), conv0)	"\tw tljod %0,=%a\n" 1
stmt:	NEU2(BANDU2(dm,conv1), conv0)	"\tw tljod %0,=%a\n" 1
stmt:	NEI4(BANDI4(dm,conv1), conv0)	"\td tljod %0,=%a\n" 1
stmt:	NEU4(BANDU4(dm,conv1), conv0)	"\td tljod %0,=%a\n" 1
stmt:	NEI8(BANDI8(dm,conv1), conv0)	"\tq tljod %0,=%a\n" 1
stmt:	NEU8(BANDU8(dm,conv1), conv0)	"\tq tljod %0,=%a\n" 1

stmt:	NEI1(BANDI1(dm,conv1), conv1)	"\tb tljev %0,=%a\n" 1
stmt:	NEU1(BANDU1(dm,conv1), conv1)	"\tb tljev %0,=%a\n" 1
stmt:	NEI2(BANDI2(dm,conv1), conv1)	"\tw tljev %0,=%a\n" 1
stmt:	NEU2(BANDU2(dm,conv1), conv1)	"\tw tljev %0,=%a\n" 1
stmt:	NEI4(BANDI4(dm,conv1), conv1)	"\td tljev %0,=%a\n" 1
stmt:	NEU4(BANDU4(dm,conv1), conv1)	"\td tljev %0,=%a\n" 1
stmt:	NEI8(BANDI8(dm,conv1), conv1)	"\tq tljev %0,=%a\n" 1
stmt:	NEU8(BANDU8(dm,conv1), conv1)	"\tq tljev %0,=%a\n" 1

stmt:	NEI1(dm, am)			"\tb cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEU1(dm, am)			"\tb cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEI2(dm, am)			"\tw cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEU2(dm, am)			"\tw cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEI4(dm, am)			"\td cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEU4(dm, am)			"\td cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEI8(dm, am)			"\tq cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEU8(dm, am)			"\tq cmp  %0,%1\n\tljne =%a\n" 1
stmt:	NEF4(dm, am)			"\td cmpf %0,%1\n\tljne =%a\n" 1
stmt:	NEF8(dm, am)			"\tq cmpf %0,%1\n\tljne =%a\n" 1

stmt:	GEI1(dm, am)			"\tb cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEU1(dm, am)			"\tb cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEI2(dm, am)			"\tw cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEU2(dm, am)			"\tw cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEI4(dm, am)			"\td cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEU4(dm, am)			"\td cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEI8(dm, am)			"\tq cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEU8(dm, am)			"\tq cmp  %0,%1\n\tljge =%a\n" 1
stmt:	GEF4(dm, am)			"\td cmpf %0,%1\n\tljge =%a\n" 1
stmt:	GEF8(dm, am)			"\tq cmpf %0,%1\n\tljge =%a\n" 1

stmt:	GTI1(dm, am)			"\tb cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTU1(dm, am)			"\tb cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTI2(dm, am)			"\tw cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTU2(dm, am)			"\tw cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTI4(dm, am)			"\td cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTU4(dm, am)			"\td cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTI8(dm, am)			"\tq cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTU8(dm, am)			"\tq cmp  %0,%1\n\tljgt =%a\n" 1
stmt:	GTF4(dm, am)			"\td cmpf %0,%1\n\tljgt =%a\n" 1
stmt:	GTF8(dm, am)			"\tq cmpf %0,%1\n\tljgt =%a\n" 1

stmt:	LEI1(dm, am)			"\tb cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEU1(dm, am)			"\tb cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEI2(dm, am)			"\tw cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEU2(dm, am)			"\tw cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEI4(dm, am)			"\td cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEU4(dm, am)			"\td cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEI8(dm, am)			"\tq cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEU8(dm, am)			"\tq cmp  %0,%1\n\tljle =%a\n" 1
stmt:	LEF4(dm, am)			"\td cmpf %0,%1\n\tljle =%a\n" 1
stmt:	LEF8(dm, am)			"\tq cmpf %0,%1\n\tljle =%a\n" 1

stmt:	LTI1(dm, am)			"\tb cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU1(dm, am)			"\tb cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTI2(dm, am)			"\tw cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU2(dm, am)			"\tw cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTI4(dm, am)			"\td cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU4(dm, am)			"\td cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTI8(dm, am)			"\tq cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU8(dm, am)			"\tq cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTF4(dm, am)			"\td cmpf %0,%1\n\tljlt =%a\n" 1
stmt:	LTF8(dm, am)			"\tq cmpf %0,%1\n\tljlt =%a\n" 1

stmt:	LTI1(eadr,eadr)			"\tb cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU1(eadr,eadr)			"\tb cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTI2(eadr,eadr)			"\tw cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU2(eadr,eadr)			"\tw cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTI4(eadr,eadr)			"\td cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU4(eadr,eadr)			"\td cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTI8(eadr,eadr)			"\tq cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTU8(eadr,eadr)			"\tq cmp  %0,%1\n\tljlt =%a\n" 1
stmt:	LTF4(eadr,eadr)			"\td cmpf %0,%1\n\tljlt =%a\n" 1
stmt:	LTF8(eadr,eadr)			"\tq cmpf %0,%1\n\tljlt =%a\n" 1

reg:	CVPU4(reg)			"\td mov %c,%0\n"  move(a)
reg:	CVUP4(reg)			"\td mov %c,%0\n"  move(a)

reg:	CVII1(reg)			"#\n"	1
reg:	CVIU1(reg)			"#\n"	1
reg:	CVII2(reg)			"#\n"	1
reg:	CVIU2(reg)			"#\n"	1
reg:	CVII4(reg)			"#\n"	1
reg:	CVIU4(reg)			"#\n"	1
reg:	CVII8(reg)			"#\n"	1
reg:	CVIU8(reg)			"#\n"	1
reg:	CVIF4(reg)			"\td mov %c,%0\n\tcvsf %c\n"	1
reg:	CVIF8(reg)			"\tq mov %c,%0\n\tcvsd %c\n"	1

reg:	CVUI1(reg)			"#\n"	1
reg:	CVUU1(reg)			"#\n"	1
reg:	CVUI2(reg)			"#\n"	1
reg:	CVUU2(reg)			"#\n"	1
reg:	CVUI4(reg)			"#\n"	1
reg:	CVUU4(reg)			"#\n"	1
reg:	CVUI8(reg)			"#\n"	1
reg:	CVUU8(reg)			"#\n"	1
reg:	CVUF4(reg)			"\td mov %c,%0\n\tcvuf %c\n"	1
reg:	CVUF8(reg)			"\tq mov %c,%0\n\tcvud %c\n"	1

reg:	CVFF4(reg)			"\tq mov %c,%0\n\tcvdf %c\n"	1
reg:	CVFF8(reg)			"?\td mov %c,%0\n\tcvfd %c\n"	1
reg:	CVFI2(reg)			"#\tw mov %c,%0\n\tcvfs %c\n"	1
reg:	CVFI4(reg)			"#\td mov %c,%0\n\tcvfs %c\n"	1
reg:	CVFI8(reg)			"#\tq mov %c,%0\n\tcvfs %c\n"	1

reg:	CVFF8(INDIRF4(ADDP4(dm, am)))	"\td mov %c,%0\n\tadd %c,%1\n\tmov %c,[%c]\n\tcvfd %c\n" 1
reg:	CVFI4(MULF8(iadr, reg))			"?\tq mov %c,%0\n\tq mulf %c,%1\n\tcvds %c\n" 1


stmt:	ARGI2(LOADI4(am))		"\tw push %0\n"	1
stmt:	ARGU2(LOADU4(am))		"\tw push %0\n"	1
stmt:	ARGI4(LOADI4(am))		"\td push %0\n"	1
stmt:	ARGU4(LOADU4(ad))		"\td pea %0\n" 1
stmt:	ARGU4(LOADU4(reg))		"\td push %0\n"	1
stmt:	ARGU4(LOADU4(con))		"\td push %0\n"	1
stmt:	ARGU4(LOADU4(iadr))		"\td push %0\n"	1
stmt:	ARGI4(LOADI8(am))		"\tq push %0\n"	1
stmt:	ARGU4(LOADU8(am))		"\tq push %0\n"	1

stmt:	ARGI2(am)			"\tw push %0\n"	1
stmt:	ARGU2(am)			"\tw push %0\n"	1
stmt:	ARGI4(am)			"\td push %0\n"	1
stmt:	ARGU4(am)			"\td push %0\n"	1
stmt:	ARGI8(am)			"\tq push %0\n"	1
stmt:	ARGU8(am)			"\tq push %0\n"	1
stmt:	ARGF4(am)			"\td push %0\n"	1
stmt:	ARGF8(am)			"\tq push %0\n"	1

stmt:	ARGP4(eadr)			"\td push %0\n"	1
stmt:	ARGP4(ad)			"\td pea  %0\n"	1
stmt:	ARGP4(reg)			"\td push %0\n"	1

reg:	CALLI2(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLU2(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLI4(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLU4(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLP4(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLI8(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLU8(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLF4(ca)			"\tjsrfr %a,%0\n"	1
reg:	CALLF8(ca)			"\tjsrfr %a,%0\n"	1
stmt:	CALLV(ca)			"\tjsrfr %a,%0\n"	1

reg:	RETI1(reg)			"?\tb mov %c,%0\n"	1
reg:	RETU1(reg)			"?\tb mov %c,%0\n"	1
reg:	RETI2(reg)			"?\tw mov %c,%0\n"	1
reg:	RETU2(reg)			"?\tw mov %c,%0\n"	1
reg:	RETI4(reg)			"?\td mov %c,%0\n"	1
reg:	RETU4(reg)			"?\td mov %c,%0\n"	1
reg:	RETP4(reg)			"?\td mov %c,%0\n"	1
reg:	RETI8(reg)			"?\tq mov %c,%0\n"	1
reg:	RETU8(reg)			"?\tq mov %c,%0\n"	1

reg:	RETI1(con)			"\tb mov %c,=%0\n"	1
reg:	RETU1(con)			"\tb mov %c,=%0\n"	1
reg:	RETI2(con)			"\tw mov %c,=%0\n"	1
reg:	RETU2(con)			"\tw mov %c,=%0\n"	1
reg:	RETI4(con)			"\td mov %c,=%0\n"	1
reg:	RETU4(con)			"\td mov %c,=%0\n"	1
reg:	RETP4(con)			"\td mov %c,=%0\n"	1
reg:	RETI8(con)			"\tq mov %c,=%0\n"	1
reg:	RETU8(con)			"\tq mov %c,=%0\n"	1

stmt:	RETI4(ad)			"\td mov r6,%0\n"	1
stmt:	RETU4(ad)			"\td mov r6,%0\n"	1
stmt:	RETP4(ad)			"\td mov r6,%0\n"	1

stmt:	RETI1(reg)			"*\tb mov r6,%0\n"
stmt:	RETU1(reg)			"*\tb mov r6,%0\n"
stmt:	RETI2(reg)			"*\tw mov r6,%0\n"
stmt:	RETU2(reg)			"*\tw mov r6,%0\n"
stmt:	RETI4(reg)			"*\td mov r6,%0\n"
stmt:	RETU4(reg)			"*\td mov r6,%0\n"
stmt:	RETP4(reg)			"*\td mov r6,%0\n"
stmt:	RETI8(reg)			"*\tq mov r6,%0\n"
stmt:	RETU8(reg)			"*\tq mov r6,%0\n"
stmt:	RETF4(reg)			"*\td mov r6,%0\n"
stmt:	RETF8(reg)			"*\tq mov r6,%0\n"

stmt:	RETI1(con)			"\tb mov r6,=%0\n"
stmt:	RETU1(con)			"\tb mov r6,=%0\n"
stmt:	RETI2(con)			"\tw mov r6,=%0\n"
stmt:	RETU2(con)			"\tw mov r6,=%0\n"
stmt:	RETI4(con)			"\td mov r6,=%0\n"
stmt:	RETU4(con)			"\td mov r6,=%0\n"
stmt:	RETP4(con)			"\td mov r6,=%0\n"
stmt:	RETI8(con)			"\tq mov r6,=%0\n"
stmt:	RETU8(con)			"\tq mov r6,=%0\n"
stmt:	RETF4(con)			"\td mov r6,=%0\n"
stmt:	RETF8(con)			"\tq mov r6,=%0\n"

stmt:	JUMPV(am)			"\tjmp %0\n"			1
stmt:	JUMPV(adr)			"\tjmp %0\n"			1

stmt:	LABELV				"#:%a\n"

reg:	ASMRETU4			"#\n" 1
reg:	ASMRETU8			"#\n" 1
stmt:	ASMRETU4			"#\n" 1
stmt:	ASMRETU8			"#\n" 1

%%

//*********************************
static void progbeg(long argc, char *argv[]) {

	{
		union {
			char c;
			long i;
		} u;
		u.i = 0;
		u.c = 1;
		swap = (long)((unsigned)(u.i == 1)) != IR->little_endian;
	}
	parseflags(argc, argv);

	fltreg[R6] = mkreg("R6", R6, 1, IREG);
	fltreg[R0] = mkreg("R0", R0, 1, IREG);
	fltreg[R1] = mkreg("R1", R1, 1, IREG);
	fltreg[R2] = mkreg("R2", R2, 1, IREG);
	fltreg[R3] = mkreg("R3", R3, 1, IREG);
	fltreg[R4] = mkreg("R4", R4, 1, IREG);
	fltreg[R5] = mkreg("R5", R5, 1, IREG);
	
	longlongreg[R6] = mkreg("R6", R6, 1, IREG);
	longlongreg[R0] = mkreg("R0", R0, 1, IREG);
	longlongreg[R1] = mkreg("R1", R1, 1, IREG);
	longlongreg[R2] = mkreg("R2", R2, 1, IREG);
	longlongreg[R3] = mkreg("R3", R3, 1, IREG);
	longlongreg[R4] = mkreg("R4", R4, 1, IREG);
	longlongreg[R5] = mkreg("R5", R5, 1, IREG);
	
	longreg[R6] = mkreg("R6", R6, 1, IREG);
	longreg[R0] = mkreg("R0", R0, 1, IREG);
	longreg[R1] = mkreg("R1", R1, 1, IREG);
	longreg[R2] = mkreg("R2", R2, 1, IREG);
	longreg[R3] = mkreg("R3", R3, 1, IREG);
	longreg[R4] = mkreg("R4", R4, 1, IREG);
	longreg[R5] = mkreg("R5", R5, 1, IREG);
	
	shortreg[R6] = mkreg("R6", R6, 1, IREG);
	shortreg[R0] = mkreg("R0", R0, 1, IREG);
	shortreg[R1] = mkreg("R1", R1, 1, IREG);
	shortreg[R2] = mkreg("R2", R2, 1, IREG);
	shortreg[R3] = mkreg("R3", R3, 1, IREG);
	shortreg[R4] = mkreg("R4", R4, 1, IREG);
	shortreg[R5] = mkreg("R5", R5, 1, IREG);
	
	charreg[R6] = mkreg("R6", R6, 1, IREG);
	charreg[R0] = mkreg("R0", R0, 1, IREG);
	charreg[R1] = mkreg("R1", R1, 1, IREG);
	charreg[R2] = mkreg("R2", R2, 1, IREG);
	charreg[R3] = mkreg("R3", R3, 1, IREG);
	charreg[R4] = mkreg("R4", R4, 1, IREG);
	charreg[R5] = mkreg("R5", R5, 1, IREG);
	
	charregw = mkwildcard(charreg);
	shortregw = mkwildcard(shortreg);
	longregw = mkwildcard(longreg);
	longlongregw = mkwildcard(longlongreg);
	fltregw = mkwildcard(fltreg);

	tmask[IREG] = (1<<R6) | (1<<R0) | (1<<R1) | (1<<R2) | (1<<R3) | (1<<R4) | (1<<R5);
	vmask[IREG] = 0;

	tmask[FREG] = (1<<R6) | (1<<R0) | (1<<R1) | (1<<R2) | (1<<R3) | (1<<R4) | (1<<R5);
	vmask[FREG] = 0;

	cseg = 0;

	asmresult = mkreg("R6", R6, 1, IREG);
	// output UTF-8 BOM (UTF8)
	// Nope, on second thought, don't do it.
	// UTF-8, according to the standard, shouldn't
	// have the UTF-16 BOM at the start of it,
	// even though many do. So, chew it up if it's
	// there, but don't put it out yourself.
	// print("\357\273\277");
}

//*********************************
static Symbol rmap(long opk) {

	switch (optype(opk)) {
	 case B:	// struct
	 case P:	// pointer
		return longregw;
	 case I:
	 case U:
		if( opsize(opk)==1 ) return charregw;
		else if( opsize(opk)==2 ) return shortregw;
		else if( opsize(opk)==4 ) return longregw;
		else return longlongregw;
	 case F:
		return fltregw;
	 default:
		return 0;
	}
}

//*********************************
static void segment(long n) {

	if( ((cseg==CODE || cseg==LIT || cseg==DATA) && (n==0 || n==BSS))
	 || (cseg==BSS && n!=BSS) ) print("\t@seg_end\n\t; *************************\n");	// end of segment
	if( (n==CODE || n==LIT || n==DATA) && (cseg==0 || cseg==BSS) ) print("\t@code\n");
	if( n==BSS && cseg!=BSS ) print("\t@zero\n");

	cseg = n;
}

//*********************************
static void progend(void) {
	segment(0);
	print("\t@end\n");
}

//*********************************
static void target(Node p) {

	assert(p);
	switch (specific(p->op)) {
	 case ASMRET+U:
	 case ASM0+U:
	 case ASM1+U:
	 case ASM2+U:
		setreg(p, asmresult);
		break;
	 case CALL+I:
	 case CALL+U:
	 case CALL+P:
		setreg(p, longreg[R6]);
		break;
	 case RET+I:
	 case RET+U:
	 case RET+P:
	 case RET+F:
		rtarget(p, 0, longreg[R6]);
		break;
	}
}

//****************************************
static void clobber(Node p) {

	assert(p);
	switch (specific(p->op)) {
	 case EQ+F:
	 case LE+F:
	 case GE+F:
	 case LT+F:
	 case GT+F:
	 case NE+F:
		if( specific(p->op) == EQ+F ) p->syms[1] = findlabel(genlabel(1));
		break;

	}
}

//*******************************************
static int sametree(Node p, Node q) {

	return (p == NULL && q == NULL)
		|| (p && q && p->op == q->op && p->syms[0] == q->syms[0]
				&& sametree(p->kids[0], q->kids[0])
				&& sametree(p->kids[1], q->kids[1]));
}

//*******************************************
static int memop(Node p) {

	assert(p);
	assert(generic(p->op) == ASGN);
	assert(p->kids[0]);
	assert(p->kids[1]);
	return (generic(p->kids[1]->kids[0]->op) == INDIR && sametree(p->kids[0], p->kids[1]->kids[0]->kids[0])) ? 1 : LBURG_MAX;
}

//*******************************************
// return 1 if handled, else return 0
static int emit2(Node p) {

	long sop = specific(p->op);
	long gop = generic(p->op);
	long ops = opsize(p->op);
	long opsk= 0;

	if( p->kids[0] ) opsk = opsize(p->kids[0]->op);

#define preg(f) ((f)[getregnum(p->x.kids[0])]->x.name)

//print("EMIT2 sop=%d p->op=%d\r\n", sop, p->op);

	// GEEKDO: check: shouldn't the following opsize(p->x.kids[0]->op) be opsk (above)???
	
	// GEEKDO: these conversions should check to see if the SRC & DST are the same and omit the 'mov'
	
// convert CHAR to signed
//	if (sop == CVI+I && opsize(p->x.kids[0]->op) == 1) {
	if (sop == CVI+I && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb cvss %s\n", p->syms[RX]->x.name);
// convert SHORT to signed
//	} else if (sop == CVI+I && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVI+I && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvss %s\n", p->syms[RX]->x.name);
// convert LONG to signed
//	} else if (sop == CVI+I && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVI+I && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
		print("\td cvss %s\n", p->syms[RX]->x.name);

// convert CHAR to unsigned
//	} else if (sop == CVI+U && opsize(p->x.kids[0]->op) == 1) {
	} else if (sop == CVI+U && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb ncvuu %s\n", p->syms[RX]->x.name);
// convert SHORT to unsigned
//	} else if (sop == CVI+U && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVI+U && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvuu %s\n", p->syms[RX]->x.name);
// convert LONG to unsigned
//	} else if (sop == CVI+U && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVI+U && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
		print("\td cvuu %s\n", p->syms[RX]->x.name);

// convert BYTE to signed
//	} else if (sop == CVU+I && opsize(p->x.kids[0]->op) == 1) {
	} else if (sop == CVU+I && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb cvuu %s\n", p->syms[RX]->x.name);
// convert WORD to signed
//	} else if (sop == CVU+I && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVU+I && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvuu %s\n", p->syms[RX]->x.name);
// convert WORD to signed
//	} else if (sop == CVU+I && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVU+I && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\td cvuu %s\n", p->syms[RX]->x.name);
		
// convert BYTE to unsigned
//	} else if (sop == CVU+U && opsize(p->x.kids[0]->op) == 1) {
	} else if (sop == CVU+U && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb cvuu %s\n", p->syms[RX]->x.name);
// convert WORD to unsigned
//	} else if (sop == CVU+U && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVU+U && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvuu %s\n", p->syms[RX]->x.name);
// convert DWORD to unsigned
//	} else if (sop == CVU+U && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVU+U && opsk == 4) {
		if( ops!=8 || p->kids[0]->op!=4870 ) {	// don't convert ASMRETU4 if output is QWORD
			if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
			print("\td cvuu %s\n", p->syms[RX]->x.name);
		}
		
// convert FLOAT to signed
	} else if( sop == CVF+I && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
		print("\tcvfs %s\n", p->syms[RX]->x.name);
// convert DOUBLE to signed
	} else if( sop == CVF+I && opsk == 8) {
		if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\tq mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
		print("\tcvds %s\n", p->syms[RX]->x.name);
		
// convert CHAR to 


// GEEKDO: need to finish these!

	} else if( sop == CVI+F && opsk==1 ) {
		opsk = 1;
	} else if( sop == CVI+F && opsk==2 ) {
		opsk = 2;
	} else if( sop == CVI+F && opsk==4 ) {
		opsk = 4;
	} else if( sop == CVI+F && opsk==8 ) {
		opsk = 8;
	} else if( sop == CVI+F && opsk==1 ) {
	
	
	
	
		if( opsk == 1 ) {
			if( opsize(p->op) == 4 ) print("\tb mov %s,%s\n\tcvss %sb\n\tcvsf %s", p->syms[RX]->x.name, preg(longreg), p->syms[RX]->x.name, p->syms[RX]->x.name);
		} else if( opsk==2 ) {
		
		} else if( opsk==4 ) {
			if( ops==4 ) ;
			else if( ops==8 ) ;
			else print("Oh, crap. emit2(CVI+F) opsk=4, ops=%d\n", ops);
		} else if( opsk==8 ) {
		} else print("Oh, crap. emit2(CVI+F) opsk=%d\n", opsk);
	} else if( gop==CVI || gop==CVU || gop==LOAD ) {

		char	*dst;
		char	*src;
		long	srcopsize, dstopsize;
		long	srcreg,	dstreg;

/*		if( gop==LOAD && generic(p->kids[0]->op)==INDIR && p->kids[0]->kids[0]->op==VREG+P ) {
			return 1;
		}
*/
		if( p->x.kids[0] ) {
			srcopsize = opsize(p->x.kids[0]->op);
			srcreg = getregnum(p->x.kids[0]);
		} else {
			srcopsize = opsize(p->kids[0]->op);
			srcreg = getregnum(p->kids[0]);
		}
		dstopsize = opsize(p->op);
		dstreg = getregnum(p);

		if( generic(p->kids[0]->op)==ASMRET ) {
			return 1;
		}
		if( dstopsize==1 )		dst = charreg[dstreg]->x.name;
		else if( dstopsize==2 )	dst = shortreg[dstreg]->x.name;
		else if( dstopsize==4 )	dst = longreg[dstreg]->x.name;
		else					dst = longlongreg[dstreg]->x.name;

		if( srcopsize==1 ) 		src = charreg[srcreg]->x.name;
		else if( srcopsize==2 )	src = shortreg[srcreg]->x.name;
		else if( srcopsize==4 )	src = longreg[srcreg]->x.name;
		else					src = longlongreg[srcreg]->x.name;

//		assert(dstopsize <= srcopsize);
		if (dst != src) {
				 if( dstopsize==1 ) print("\tb mov %s,%s\n", dst, src);
			else if( dstopsize==2 ) print("\tw mov %s,%s\n", dst, src);
			else if( dstopsize==4 ) print("\td mov %s,%s\n", dst, src);
			else if( dstopsize==8 ) print("\tq mov %s,%s\n", dst, src);
		}
	} else if( gop==RET+I || gop==RET+U ) {
		// check if %0 is 'r6' (return causing "mov r6,r6"). if so, return 1, else return 0
//		if( stricmp("r6", p->syms[0]->x.name) ) return 0;
		return 1;
	} else if( gop==LABEL ) {
		print(":%s\n", p->syms[0]->x.name);
	} else if( gop==ASM0 ) {
		if( p->syms[0]->name[0]==':' ) print("%s", p->syms[0]->name);
		else print("\t%s", p->syms[0]->name);
	} else if( gop==ASMRET || gop==ASM2 ) {
		if( p->kids[0] ) emit2(p->kids[0]);
		if( p->kids[1] ) emit2(p->kids[1]);
	}
	return 1;
}

//*********************************
static void doarg(Node p) {

	assert(p && p->syms[0]);
	mkactual(4, p->syms[0]->u.c.v.ill);
}

//*********************************
static void blkfetch(long k, long off, long reg, long tmp) {}

//*********************************
static void blkstore(long k, long off, long reg, long tmp) {}

//*********************************
static void blkloop(long dreg, long doff, long sreg, long soff, long size, long tmps[]) {}

//*********************************
static void local(Symbol p) {

	if( isfloat(p->type) ) p->sclass = AUTO;
	if( askregvar(p, (*IR->x.rmap)(ttob(p->type))) == 0 ) {
		assert(p->sclass == AUTO);
		offset = roundup(offset + p->type->size, p->type->align);// < 4 ? 4 : p->type->align);
		p->x.offset = -offset;
		p->x.name = stringd(-offset);
	}
}

//***************************************************************************
static void function(Symbol f, Symbol caller[], Symbol callee[], long n) {

	long 			i, argsize;
	unsigned long	mask;

	usedmask[0] = usedmask[1] = 0;
	freemask[0] = freemask[1] = ~(unsigned)0;
	offset = maxoffset = maxargoffset = 12;	// for retaddr & saved FP & saved OldSP

	argsize = 0;
	for(i=0; callee[i]; i++) {
			Symbol p = callee[i];
			Symbol q = caller[i];
			assert(q);
			p->x.offset = q->x.offset = offset;
			p->x.name = q->x.name = stringf("%d", p->x.offset);
			p->sclass = q->sclass = AUTO;
			offset += roundup(q->type->size, 4);
			argsize += q->type->size;
	}
	assert(caller[i]==0);

	offset = maxoffset = 0;
	
	gencode(caller, callee);

	framesize = roundup(maxoffset, 4);

	mask = usedmask[IREG] & ~1;	// never save R6, return value register, only 0-5
	
	print("\t@align 2\n");
	print(":%s\n", f->x.name);
	
// if framesize==0, no local vars
// if mask==0, no registers to save
	
	if( framesize ) {
		if( mask ) print("\tSFVR %u,%u\n", framesize, mask);	// start frame with both LOCAL VARS and REGS saved
		else print("\tSFV %u\n", framesize);					// start frame with LOCAL VARS but no regs saved
	} else {
		if( mask ) print("\tSFR %u\n", mask);					// start frame with no local vars but REGS saved
		else print("\tSF\n");									// start frame with no local vars and no regs saved
	}
			
	emitcode();

	if( mask ) print("\tEFR %u\n", mask);		// end frame with saved REGS to restore
	else print("\tEF\n");						// end frame with no saved regs to restore
	print("\t; =========================\n");
}

//***********************************
static void defsymbol(Symbol p) {

	if (p->scope >= LOCAL && p->sclass == STATIC) p->x.name = stringf("%c%d", 'A'+Kfilenum, genlabel(1));
	else if (p->generated) p->x.name = stringf("%c%s", 'A'+Kfilenum, p->name);
	else if (p->scope == GLOBAL || p->sclass == EXTERN) p->x.name = stringf("_%s", p->name);
	else if (p->scope == CONSTANTS && (isint(p->type) || isptr(p->type))
		  && p->name[0] == '0' && p->name[1] == 'x') p->x.name = stringf("0x%s", &p->name[2]);
	else p->x.name = p->name;
}

//***********************************************
static void address(Symbol q, Symbol p, long n) {

	if( p->scope == GLOBAL || p->sclass == STATIC || p->sclass == EXTERN ) {
		q->x.name = stringf("%s%s%d", p->x.name, n >= 0 ? "+" : "", n);
	} else {
		assert(n <= INT_MAX && n >= INT_MIN);
		q->x.offset = p->x.offset + n;
		q->x.name = stringd(q->x.offset);
	}
}

//******************************************************
static void defconst(long suffix, long size, Value v) {

	assert(sizeof(float)==4 && sizeof(double)==8);
		 if (suffix == I && size == 1) print("\tb %d\n", (unsigned long)v.ull);
	else if (suffix == I && size == 2) print("\tw %d\n", (unsigned long)v.ill);
	else if (suffix == I && size == 4) print("\td %d\n",(unsigned long)v.ill);
	else if (suffix == I && size == 8) print("\tq 0x%X\n",v.ill);
	else if (suffix == U && size == 1) print("\tb 0x%x\n", (unsigned)((unsigned char)v.ull));
	else if (suffix == U && size == 2) print("\tw 0x%x\n", (unsigned)((unsigned short)v.ull));
	else if (suffix == U && size == 4) print("\td 0x%x\n", (unsigned)v.ull);
	else if (suffix == U && size == 8) print("\tq 0x%X\n",v.ull);
	else if (suffix == P && size == 4) print("\td 0x%x\n", (unsigned)v.p);
	else if (suffix == F && size == 4) {
		float f = (float)v.d;
		print("\td 0x%x\n", *(unsigned *)&f);
	} else if (suffix == F && size == 8) print("\tq 0x%X\n", *(unsigned long long *)&v.d);
	else assert(0);
}

//*********************************
static void defaddress(Symbol p) {

	print("\t& %s\n", p->x.name);
}

//*********************************
static void defstring(long n, char *str) {

	char 	*s;
	long	a, b, last;

	for(s=str, a=last=0; s<str+n; s++) if( *s>=' ' && *s<128 ) {++a; last = s-str;}
	if( 100*a/n > 90 || (a==n-1 && (*(s-1)<32 || *(s-1)>127)) ) {
		print("\tb \"");
		for(s=str; s<str+n; s++) {
			if( *s=='"' || *s=='\\' ) print("\\%c", *s);
			else if( *s>=32 && *s<128 ) print("%c", *s);
			else if( *s=='\t' ) print("\\t");
			else if( *s=='\r' ) print("\\r");
			else if( *s=='\n' ) print("\\n");
			else print("\\0%o", *s);
		}
		print("\"\n");
	} else {
		for(s=str, b=0; s<str+n; b++) {
			print("\tb ");
			for(a=0; s<str+n && a<8; ++s, ++a) {
				if( a<7 && s+1<str+n ) print("%d,", *s);
				else print("%d", *s);
			}
			print("\n");
		}
	}
}

//*********************************
static void export(Symbol p) {

// KASM assembles all files in a project at the
// same time, so in that environment, 'public'
// and 'extern' aren't needed and just clog up
// the assembly language source code needlessly.
//	print("public %s\n", p->x.name);
}

//*********************************
static void import(Symbol p) {

// See comment above in export()
//	if( p->ref > 0 ) print("extern %s\n", p->x.name);
}

//*********************************
static void global(Symbol p) {

	Type	t;

	print("\t@align %d\n", p->type->align > 4 ? 4 : p->type->align);
	print(":%s\n", p->x.name);
	for(t=p->type; isarray(t); t=t->type) ;
	if( t->op==STRUCT ) print("\t@align 1\n");
}

//*********************************
static void space(long n) {

	if( cseg==BSS ) print("\tb [%d]\n", n);
	else print("\tb [%d] 0\n", n);
}

//****************************************************
// returns 0 if cp doesn't point to a register name, or len of register name if it does
static long ASMisreg() {

	return (toupper(*cp)!='R' || *(cp+1)<'0' || *(cp+1)>'7' || isalnum(*(cp+2)) || *(cp+2)=='_') ? 0 : 2;
}

//*********************************
Interface kint6IR = {
	{1, 1, 0},	/* char */
	{2, 2, 0},	/* short */
	{4, 4, 0},	/* int */
	{4, 4, 0},	/* long */
	{8, 4, 0},	/* long long */
	{4, 4, 1},	/* float */
	{8, 4, 1},	/* double (size, alignment, outofline) */
	{8, 4, 1},	/* long double */
	{4, 4, 0},	/* T * */
	{0, 1, 0},	/* struct */
	
	1,			/* little_endian */
	0,			/* mulops_calls */
	0,			/* wants_callb */
	1,			/* wants_argb */
	0,			/* left_to_right */
	0,			/* wants_dag */
	0,			/* unsigned_char */
	
	address,
	blockbeg,
	blockend,
	defaddress,
	defconst,
	defstring,
	defsymbol,
	emit,
	export,
	function,
	gen,
	global,
	import,
	local,
	progbeg,
	progend,
	segment,
	space,
	0, 0, 0, 0, 0, 0, 0,
	{
		1, rmap,
		blkfetch, blkstore, blkloop,
		_label,
		_rule,
		_nts,
		_kids,
		_string,
		_templates,
		_isinstruction,
		_ntname,
		emit2,
		doarg,
		target,
		clobber,
		ASMisreg,
	}
};
