
/*
41   ARGB
73   INDIRB
584  JUMPV
711  VREGP

4117 CNSTI4
4118 CNSTU4
4133 ARGI4
4134 ARGU4
4149 ASGNI4
4150 ASGNU4
4161 INDIRF4
4165 INDIRI4
4166 INDIRU4
4167 INDIRP4

4213 CVFI4

4310 CALLU4
4326 LOADU4
4359 ADDRGP4
4375 ADDRFP4
4391 ADDRLP4

4405 ADDI4
4407 ADDP4
4421 SUBI4
4437 LSHI4
4470 RSHU4
4486 BANDU4

4501 BCOMI4
4582 EQU4

8257 INDIRF8

8305 CVFF8

8657 MULF8
*/

/*
NOTE: Within the various blocks of an .MD file, you can
use "%{\n" to start a literal block (just as THIS block
is started in that way above), and terminate the literal
block with "%}\n" (without the quotes and where '\n'
represents the EOL newline char. These blocks are not
interpreted by LBURG, but rather are copied as-is. So,
if you wish to COMMENT out something that will appear
as a COMMENTED BLOCK in the .c output code, then you'll
also need to nest it between SLASH-STAR and STAR-SLASH
within the "%{\n" "%}\n" block (as is this comment).

So, there's nothing special about THIS block being first
in the file, it's just a block of code/comments that are
included in the output file "as-is". You can use this
method to include comments within the 'rules' section
as well, although those comments will appear "up-front"
in the .c output file, since the rules themselves get
turned into several arrays and switch statements, which
makes it pretty much impossible to include the comment
"in place" where it appears in the .md file.

With the modified KLBURG.EXE, you can also comment out
single lines within the 'rules' section using the
standard '//' C "single-line comment". A line thus
commented-out will NOT appear, even as a comment, in
the .c output file, whereas the above-mentioned comment
BLOCK method WILL appear in the output file, albeit
"out of place."
*/

/*
NOTE: Book pgs 418-419:
ralloc frees the input registers before allocating the output register,
which allows it to reuse an input register as the output register. This
econonmy is always safe when the node is implemented by a single instruction,
but it can be unsafe if a node is implemented by a sequence of instructions:
If the output register is also one of the input registers, and if the
sequence changes the output register before reading the corresponding input
register, then the read fetches a corrupt value. We take care that all rules
that emit instruction sequences set their output register ONLY after they
finish reading all input registers. Most templates emit just one instruction,
so this assumption is a good default, but it does require considerable care
with multi-instruction sequences.
*/
/*
	R15 is used as a scratch register
	R14 is used to hold return values from funcions
	R13-R0 are general purpose registers
*/


enum {	R14=0, R0=1, R1=2, R2=3, R3=4, R4=5, R5=6, R6=7, R7=8, R8=9, R9=10, R10=11, R11=12, R12=13, R13=14};

#include <io.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/stat.h>

#include "c.h"
#define NODEPTR_TYPE Node
#define OP_LABEL(p) ((p)->op)
#define LEFT_CHILD(p) ((p)->kids[0])
#define RIGHT_CHILD(p) ((p)->kids[1])
#define STATE_LABEL(p) ((p)->x.state)
static void address(Symbol, Symbol, long);
static void blkfetch(long, long, long, long);
static void blkloop(long, long, long, long, long, long[]);
static void blkstore(long, long, long, long);
static void defaddress(Symbol);
static void defconst(long, long, Value);
static void defstring(long, char *);
static void defsymbol(Symbol);
static void doarg(Node);
static int emit2(Node);
static void export(Symbol);
static void clobber(Node);
static void function(Symbol, Symbol [], Symbol [], long);
static void global(Symbol);
static void import(Symbol);
static void local(Symbol);
static void progbeg(long, char **);
static void progend(void);
static void segment(long);
static void space(long);
static void target(Node);
static Symbol charreg[32], shortreg[32], longreg[32], longlongreg[32];
static Symbol fltreg[32];
static int memop(Node p);
static int sametree(Node p, Node q);

static Symbol charregw, shortregw, longregw, longlongregw, fltregw, asmresult;

static long cseg;

//static Symbol quo, rem;

/*
generated at Thu May 12 10:27:14 2022
by $Id$
*/
static void _kids(NODEPTR_TYPE, int, NODEPTR_TYPE[]);
static void _label(NODEPTR_TYPE);
static int _rule(void*, int);

#define _stmt_NT 1
#define _con1_NT 2
#define _con2_NT 3
#define _con4_NT 4
#define _con8_NT 5
#define _con_NT 6
#define _conv0_NT 7
#define _conv1_NT 8
#define _conv2_NT 9
#define _conv4_NT 10
#define _conv8_NT 11
#define _conv16_NT 12
#define _conv32_NT 13
#define _conv64_NT 14
#define _conv128_NT 15
#define _conv256_NT 16
#define _reg_NT 17
#define _adr_NT 18
#define _ad_NT 19
#define _ca_NT 20
#define _iadr1_NT 21
#define _iadr2_NT 22
#define _iadr4_NT 23
#define _iadr8_NT 24
#define _iadr_NT 25
#define _eadr1_NT 26
#define _eadr2_NT 27
#define _eadr4_NT 28
#define _eadr8_NT 29
#define _eadr_NT 30
#define _dm_NT 31
#define _am_NT 32

static char *_ntname[] = {
	0,
	"stmt",
	"con1",
	"con2",
	"con4",
	"con8",
	"con",
	"conv0",
	"conv1",
	"conv2",
	"conv4",
	"conv8",
	"conv16",
	"conv32",
	"conv64",
	"conv128",
	"conv256",
	"reg",
	"adr",
	"ad",
	"ca",
	"iadr1",
	"iadr2",
	"iadr4",
	"iadr8",
	"iadr",
	"eadr1",
	"eadr2",
	"eadr4",
	"eadr8",
	"eadr",
	"dm",
	"am",
	0
};

struct _state {
	short cost[33];
	struct {
		unsigned int _stmt:8;
		unsigned int _con1:2;
		unsigned int _con2:2;
		unsigned int _con4:3;
		unsigned int _con8:2;
		unsigned int _con:3;
		unsigned int _conv0:4;
		unsigned int _conv1:4;
		unsigned int _conv2:2;
		unsigned int _conv4:2;
		unsigned int _conv8:2;
		unsigned int _conv16:2;
		unsigned int _conv32:2;
		unsigned int _conv64:2;
		unsigned int _conv128:2;
		unsigned int _conv256:2;
		unsigned int _reg:8;
		unsigned int _adr:2;
		unsigned int _ad:2;
		unsigned int _ca:2;
		unsigned int _iadr1:2;
		unsigned int _iadr2:2;
		unsigned int _iadr4:3;
		unsigned int _iadr8:2;
		unsigned int _iadr:3;
		unsigned int _eadr1:2;
		unsigned int _eadr2:2;
		unsigned int _eadr4:3;
		unsigned int _eadr8:2;
		unsigned int _eadr:3;
		unsigned int _dm:3;
		unsigned int _am:3;
	} rule;
};

static short _nts_0[] = { 0 };
static short _nts_1[] = { _con1_NT, 0 };
static short _nts_2[] = { _con2_NT, 0 };
static short _nts_3[] = { _con4_NT, 0 };
static short _nts_4[] = { _con8_NT, 0 };
static short _nts_5[] = { _reg_NT, 0 };
static short _nts_6[] = { _adr_NT, 0 };
static short _nts_7[] = { _ad_NT, 0 };
static short _nts_8[] = { _iadr1_NT, 0 };
static short _nts_9[] = { _iadr2_NT, 0 };
static short _nts_10[] = { _iadr4_NT, 0 };
static short _nts_11[] = { _iadr8_NT, 0 };
static short _nts_12[] = { _eadr1_NT, 0 };
static short _nts_13[] = { _eadr2_NT, 0 };
static short _nts_14[] = { _eadr4_NT, 0 };
static short _nts_15[] = { _eadr8_NT, 0 };
static short _nts_16[] = { _iadr_NT, 0 };
static short _nts_17[] = { _con_NT, 0 };
static short _nts_18[] = { _adr_NT, _am_NT, 0 };
static short _nts_19[] = { _adr_NT, _ad_NT, 0 };
static short _nts_20[] = { _adr_NT, _reg_NT, 0 };
static short _nts_21[] = { _adr_NT, _con_NT, 0 };
static short _nts_22[] = { _adr_NT, _iadr_NT, 0 };
static short _nts_23[] = { _adr_NT, _adr_NT, _am_NT, 0 };
static short _nts_24[] = { _reg_NT, _am_NT, 0 };
static short _nts_25[] = { _reg_NT, _ca_NT, 0 };
static short _nts_26[] = { _con_NT, _eadr_NT, 0 };
static short _nts_27[] = { _reg_NT, _conv2_NT, 0 };
static short _nts_28[] = { _reg_NT, _conv4_NT, 0 };
static short _nts_29[] = { _reg_NT, _conv8_NT, 0 };
static short _nts_30[] = { _reg_NT, _conv16_NT, 0 };
static short _nts_31[] = { _reg_NT, _conv32_NT, 0 };
static short _nts_32[] = { _reg_NT, _conv64_NT, 0 };
static short _nts_33[] = { _reg_NT, _conv128_NT, 0 };
static short _nts_34[] = { _reg_NT, _conv256_NT, 0 };
static short _nts_35[] = { _am_NT, 0 };
static short _nts_36[] = { _dm_NT, _conv0_NT, 0 };
static short _nts_37[] = { _dm_NT, _conv1_NT, _conv0_NT, 0 };
static short _nts_38[] = { _dm_NT, _conv1_NT, _conv1_NT, 0 };
static short _nts_39[] = { _dm_NT, _am_NT, 0 };
static short _nts_40[] = { _eadr_NT, _eadr_NT, 0 };
static short _nts_41[] = { _eadr_NT, 0 };
static short _nts_42[] = { _ca_NT, 0 };

static short *_nts[] = {
	0,	/* 0 */
	_nts_0,	/* 1 */
	_nts_0,	/* 2 */
	_nts_0,	/* 3 */
	_nts_0,	/* 4 */
	_nts_0,	/* 5 */
	_nts_0,	/* 6 */
	_nts_0,	/* 7 */
	_nts_0,	/* 8 */
	_nts_0,	/* 9 */
	_nts_0,	/* 10 */
	_nts_0,	/* 11 */
	_nts_1,	/* 12 */
	_nts_2,	/* 13 */
	_nts_3,	/* 14 */
	_nts_4,	/* 15 */
	_nts_0,	/* 16 */
	_nts_0,	/* 17 */
	_nts_0,	/* 18 */
	_nts_0,	/* 19 */
	_nts_0,	/* 20 */
	_nts_0,	/* 21 */
	_nts_0,	/* 22 */
	_nts_0,	/* 23 */
	_nts_0,	/* 24 */
	_nts_0,	/* 25 */
	_nts_0,	/* 26 */
	_nts_0,	/* 27 */
	_nts_0,	/* 28 */
	_nts_0,	/* 29 */
	_nts_0,	/* 30 */
	_nts_0,	/* 31 */
	_nts_0,	/* 32 */
	_nts_0,	/* 33 */
	_nts_0,	/* 34 */
	_nts_0,	/* 35 */
	_nts_0,	/* 36 */
	_nts_0,	/* 37 */
	_nts_0,	/* 38 */
	_nts_0,	/* 39 */
	_nts_0,	/* 40 */
	_nts_0,	/* 41 */
	_nts_0,	/* 42 */
	_nts_0,	/* 43 */
	_nts_0,	/* 44 */
	_nts_0,	/* 45 */
	_nts_0,	/* 46 */
	_nts_0,	/* 47 */
	_nts_0,	/* 48 */
	_nts_0,	/* 49 */
	_nts_0,	/* 50 */
	_nts_0,	/* 51 */
	_nts_0,	/* 52 */
	_nts_0,	/* 53 */
	_nts_0,	/* 54 */
	_nts_0,	/* 55 */
	_nts_0,	/* 56 */
	_nts_0,	/* 57 */
	_nts_0,	/* 58 */
	_nts_5,	/* 59 */
	_nts_5,	/* 60 */
	_nts_5,	/* 61 */
	_nts_5,	/* 62 */
	_nts_5,	/* 63 */
	_nts_5,	/* 64 */
	_nts_5,	/* 65 */
	_nts_5,	/* 66 */
	_nts_5,	/* 67 */
	_nts_6,	/* 68 */
	_nts_7,	/* 69 */
	_nts_7,	/* 70 */
	_nts_5,	/* 71 */
	_nts_1,	/* 72 */
	_nts_2,	/* 73 */
	_nts_3,	/* 74 */
	_nts_4,	/* 75 */
	_nts_5,	/* 76 */
	_nts_0,	/* 77 */
	_nts_5,	/* 78 */
	_nts_0,	/* 79 */
	_nts_0,	/* 80 */
	_nts_0,	/* 81 */
	_nts_5,	/* 82 */
	_nts_7,	/* 83 */
	_nts_6,	/* 84 */
	_nts_6,	/* 85 */
	_nts_6,	/* 86 */
	_nts_6,	/* 87 */
	_nts_6,	/* 88 */
	_nts_6,	/* 89 */
	_nts_6,	/* 90 */
	_nts_6,	/* 91 */
	_nts_6,	/* 92 */
	_nts_6,	/* 93 */
	_nts_6,	/* 94 */
	_nts_8,	/* 95 */
	_nts_9,	/* 96 */
	_nts_10,	/* 97 */
	_nts_11,	/* 98 */
	_nts_6,	/* 99 */
	_nts_7,	/* 100 */
	_nts_7,	/* 101 */
	_nts_7,	/* 102 */
	_nts_7,	/* 103 */
	_nts_7,	/* 104 */
	_nts_7,	/* 105 */
	_nts_7,	/* 106 */
	_nts_7,	/* 107 */
	_nts_7,	/* 108 */
	_nts_7,	/* 109 */
	_nts_7,	/* 110 */
	_nts_12,	/* 111 */
	_nts_13,	/* 112 */
	_nts_14,	/* 113 */
	_nts_15,	/* 114 */
	_nts_7,	/* 115 */
	_nts_8,	/* 116 */
	_nts_9,	/* 117 */
	_nts_10,	/* 118 */
	_nts_11,	/* 119 */
	_nts_7,	/* 120 */
	_nts_5,	/* 121 */
	_nts_16,	/* 122 */
	_nts_0,	/* 123 */
	_nts_0,	/* 124 */
	_nts_0,	/* 125 */
	_nts_5,	/* 126 */
	_nts_17,	/* 127 */
	_nts_16,	/* 128 */
	_nts_0,	/* 129 */
	_nts_0,	/* 130 */
	_nts_0,	/* 131 */
	_nts_5,	/* 132 */
	_nts_5,	/* 133 */
	_nts_5,	/* 134 */
	_nts_5,	/* 135 */
	_nts_5,	/* 136 */
	_nts_5,	/* 137 */
	_nts_5,	/* 138 */
	_nts_5,	/* 139 */
	_nts_5,	/* 140 */
	_nts_5,	/* 141 */
	_nts_5,	/* 142 */
	_nts_0,	/* 143 */
	_nts_18,	/* 144 */
	_nts_18,	/* 145 */
	_nts_18,	/* 146 */
	_nts_18,	/* 147 */
	_nts_18,	/* 148 */
	_nts_18,	/* 149 */
	_nts_5,	/* 150 */
	_nts_5,	/* 151 */
	_nts_19,	/* 152 */
	_nts_19,	/* 153 */
	_nts_19,	/* 154 */
	_nts_19,	/* 155 */
	_nts_20,	/* 156 */
	_nts_21,	/* 157 */
	_nts_22,	/* 158 */
	_nts_19,	/* 159 */
	_nts_18,	/* 160 */
	_nts_18,	/* 161 */
	_nts_18,	/* 162 */
	_nts_18,	/* 163 */
	_nts_23,	/* 164 */
	_nts_23,	/* 165 */
	_nts_23,	/* 166 */
	_nts_23,	/* 167 */
	_nts_23,	/* 168 */
	_nts_23,	/* 169 */
	_nts_23,	/* 170 */
	_nts_23,	/* 171 */
	_nts_23,	/* 172 */
	_nts_23,	/* 173 */
	_nts_23,	/* 174 */
	_nts_23,	/* 175 */
	_nts_23,	/* 176 */
	_nts_23,	/* 177 */
	_nts_23,	/* 178 */
	_nts_23,	/* 179 */
	_nts_23,	/* 180 */
	_nts_23,	/* 181 */
	_nts_23,	/* 182 */
	_nts_23,	/* 183 */
	_nts_22,	/* 184 */
	_nts_16,	/* 185 */
	_nts_24,	/* 186 */
	_nts_24,	/* 187 */
	_nts_24,	/* 188 */
	_nts_24,	/* 189 */
	_nts_24,	/* 190 */
	_nts_24,	/* 191 */
	_nts_24,	/* 192 */
	_nts_24,	/* 193 */
	_nts_24,	/* 194 */
	_nts_24,	/* 195 */
	_nts_25,	/* 196 */
	_nts_24,	/* 197 */
	_nts_24,	/* 198 */
	_nts_24,	/* 199 */
	_nts_24,	/* 200 */
	_nts_24,	/* 201 */
	_nts_24,	/* 202 */
	_nts_24,	/* 203 */
	_nts_24,	/* 204 */
	_nts_24,	/* 205 */
	_nts_24,	/* 206 */
	_nts_24,	/* 207 */
	_nts_26,	/* 208 */
	_nts_24,	/* 209 */
	_nts_24,	/* 210 */
	_nts_24,	/* 211 */
	_nts_24,	/* 212 */
	_nts_24,	/* 213 */
	_nts_24,	/* 214 */
	_nts_24,	/* 215 */
	_nts_24,	/* 216 */
	_nts_24,	/* 217 */
	_nts_24,	/* 218 */
	_nts_27,	/* 219 */
	_nts_28,	/* 220 */
	_nts_29,	/* 221 */
	_nts_30,	/* 222 */
	_nts_31,	/* 223 */
	_nts_32,	/* 224 */
	_nts_33,	/* 225 */
	_nts_34,	/* 226 */
	_nts_27,	/* 227 */
	_nts_28,	/* 228 */
	_nts_29,	/* 229 */
	_nts_30,	/* 230 */
	_nts_31,	/* 231 */
	_nts_32,	/* 232 */
	_nts_33,	/* 233 */
	_nts_34,	/* 234 */
	_nts_24,	/* 235 */
	_nts_24,	/* 236 */
	_nts_24,	/* 237 */
	_nts_24,	/* 238 */
	_nts_24,	/* 239 */
	_nts_24,	/* 240 */
	_nts_24,	/* 241 */
	_nts_24,	/* 242 */
	_nts_24,	/* 243 */
	_nts_24,	/* 244 */
	_nts_27,	/* 245 */
	_nts_28,	/* 246 */
	_nts_29,	/* 247 */
	_nts_30,	/* 248 */
	_nts_31,	/* 249 */
	_nts_32,	/* 250 */
	_nts_33,	/* 251 */
	_nts_34,	/* 252 */
	_nts_27,	/* 253 */
	_nts_28,	/* 254 */
	_nts_29,	/* 255 */
	_nts_30,	/* 256 */
	_nts_31,	/* 257 */
	_nts_32,	/* 258 */
	_nts_33,	/* 259 */
	_nts_34,	/* 260 */
	_nts_24,	/* 261 */
	_nts_24,	/* 262 */
	_nts_24,	/* 263 */
	_nts_24,	/* 264 */
	_nts_24,	/* 265 */
	_nts_24,	/* 266 */
	_nts_24,	/* 267 */
	_nts_24,	/* 268 */
	_nts_24,	/* 269 */
	_nts_24,	/* 270 */
	_nts_24,	/* 271 */
	_nts_24,	/* 272 */
	_nts_24,	/* 273 */
	_nts_24,	/* 274 */
	_nts_24,	/* 275 */
	_nts_24,	/* 276 */
	_nts_24,	/* 277 */
	_nts_24,	/* 278 */
	_nts_24,	/* 279 */
	_nts_24,	/* 280 */
	_nts_24,	/* 281 */
	_nts_24,	/* 282 */
	_nts_24,	/* 283 */
	_nts_24,	/* 284 */
	_nts_24,	/* 285 */
	_nts_24,	/* 286 */
	_nts_24,	/* 287 */
	_nts_24,	/* 288 */
	_nts_24,	/* 289 */
	_nts_24,	/* 290 */
	_nts_24,	/* 291 */
	_nts_24,	/* 292 */
	_nts_24,	/* 293 */
	_nts_24,	/* 294 */
	_nts_24,	/* 295 */
	_nts_24,	/* 296 */
	_nts_24,	/* 297 */
	_nts_24,	/* 298 */
	_nts_24,	/* 299 */
	_nts_24,	/* 300 */
	_nts_24,	/* 301 */
	_nts_24,	/* 302 */
	_nts_24,	/* 303 */
	_nts_24,	/* 304 */
	_nts_24,	/* 305 */
	_nts_24,	/* 306 */
	_nts_24,	/* 307 */
	_nts_24,	/* 308 */
	_nts_5,	/* 309 */
	_nts_5,	/* 310 */
	_nts_5,	/* 311 */
	_nts_5,	/* 312 */
	_nts_5,	/* 313 */
	_nts_5,	/* 314 */
	_nts_5,	/* 315 */
	_nts_5,	/* 316 */
	_nts_35,	/* 317 */
	_nts_35,	/* 318 */
	_nts_35,	/* 319 */
	_nts_35,	/* 320 */
	_nts_35,	/* 321 */
	_nts_35,	/* 322 */
	_nts_35,	/* 323 */
	_nts_35,	/* 324 */
	_nts_5,	/* 325 */
	_nts_5,	/* 326 */
	_nts_5,	/* 327 */
	_nts_5,	/* 328 */
	_nts_5,	/* 329 */
	_nts_5,	/* 330 */
	_nts_36,	/* 331 */
	_nts_36,	/* 332 */
	_nts_36,	/* 333 */
	_nts_36,	/* 334 */
	_nts_36,	/* 335 */
	_nts_36,	/* 336 */
	_nts_36,	/* 337 */
	_nts_36,	/* 338 */
	_nts_36,	/* 339 */
	_nts_36,	/* 340 */
	_nts_37,	/* 341 */
	_nts_37,	/* 342 */
	_nts_37,	/* 343 */
	_nts_37,	/* 344 */
	_nts_37,	/* 345 */
	_nts_37,	/* 346 */
	_nts_37,	/* 347 */
	_nts_37,	/* 348 */
	_nts_38,	/* 349 */
	_nts_38,	/* 350 */
	_nts_38,	/* 351 */
	_nts_38,	/* 352 */
	_nts_38,	/* 353 */
	_nts_38,	/* 354 */
	_nts_38,	/* 355 */
	_nts_38,	/* 356 */
	_nts_39,	/* 357 */
	_nts_39,	/* 358 */
	_nts_39,	/* 359 */
	_nts_39,	/* 360 */
	_nts_39,	/* 361 */
	_nts_39,	/* 362 */
	_nts_39,	/* 363 */
	_nts_39,	/* 364 */
	_nts_39,	/* 365 */
	_nts_39,	/* 366 */
	_nts_36,	/* 367 */
	_nts_36,	/* 368 */
	_nts_36,	/* 369 */
	_nts_36,	/* 370 */
	_nts_36,	/* 371 */
	_nts_36,	/* 372 */
	_nts_36,	/* 373 */
	_nts_36,	/* 374 */
	_nts_36,	/* 375 */
	_nts_36,	/* 376 */
	_nts_37,	/* 377 */
	_nts_37,	/* 378 */
	_nts_37,	/* 379 */
	_nts_37,	/* 380 */
	_nts_37,	/* 381 */
	_nts_37,	/* 382 */
	_nts_37,	/* 383 */
	_nts_37,	/* 384 */
	_nts_38,	/* 385 */
	_nts_38,	/* 386 */
	_nts_38,	/* 387 */
	_nts_38,	/* 388 */
	_nts_38,	/* 389 */
	_nts_38,	/* 390 */
	_nts_38,	/* 391 */
	_nts_38,	/* 392 */
	_nts_39,	/* 393 */
	_nts_39,	/* 394 */
	_nts_39,	/* 395 */
	_nts_39,	/* 396 */
	_nts_39,	/* 397 */
	_nts_39,	/* 398 */
	_nts_39,	/* 399 */
	_nts_39,	/* 400 */
	_nts_39,	/* 401 */
	_nts_39,	/* 402 */
	_nts_39,	/* 403 */
	_nts_39,	/* 404 */
	_nts_39,	/* 405 */
	_nts_39,	/* 406 */
	_nts_39,	/* 407 */
	_nts_39,	/* 408 */
	_nts_39,	/* 409 */
	_nts_39,	/* 410 */
	_nts_39,	/* 411 */
	_nts_39,	/* 412 */
	_nts_39,	/* 413 */
	_nts_39,	/* 414 */
	_nts_39,	/* 415 */
	_nts_39,	/* 416 */
	_nts_39,	/* 417 */
	_nts_39,	/* 418 */
	_nts_39,	/* 419 */
	_nts_39,	/* 420 */
	_nts_39,	/* 421 */
	_nts_39,	/* 422 */
	_nts_39,	/* 423 */
	_nts_39,	/* 424 */
	_nts_39,	/* 425 */
	_nts_39,	/* 426 */
	_nts_39,	/* 427 */
	_nts_39,	/* 428 */
	_nts_39,	/* 429 */
	_nts_39,	/* 430 */
	_nts_39,	/* 431 */
	_nts_39,	/* 432 */
	_nts_39,	/* 433 */
	_nts_39,	/* 434 */
	_nts_39,	/* 435 */
	_nts_39,	/* 436 */
	_nts_39,	/* 437 */
	_nts_39,	/* 438 */
	_nts_39,	/* 439 */
	_nts_39,	/* 440 */
	_nts_39,	/* 441 */
	_nts_39,	/* 442 */
	_nts_40,	/* 443 */
	_nts_40,	/* 444 */
	_nts_40,	/* 445 */
	_nts_40,	/* 446 */
	_nts_40,	/* 447 */
	_nts_40,	/* 448 */
	_nts_40,	/* 449 */
	_nts_40,	/* 450 */
	_nts_40,	/* 451 */
	_nts_40,	/* 452 */
	_nts_5,	/* 453 */
	_nts_5,	/* 454 */
	_nts_5,	/* 455 */
	_nts_5,	/* 456 */
	_nts_5,	/* 457 */
	_nts_5,	/* 458 */
	_nts_5,	/* 459 */
	_nts_5,	/* 460 */
	_nts_5,	/* 461 */
	_nts_5,	/* 462 */
	_nts_5,	/* 463 */
	_nts_5,	/* 464 */
	_nts_5,	/* 465 */
	_nts_5,	/* 466 */
	_nts_5,	/* 467 */
	_nts_5,	/* 468 */
	_nts_5,	/* 469 */
	_nts_5,	/* 470 */
	_nts_5,	/* 471 */
	_nts_5,	/* 472 */
	_nts_5,	/* 473 */
	_nts_5,	/* 474 */
	_nts_5,	/* 475 */
	_nts_5,	/* 476 */
	_nts_5,	/* 477 */
	_nts_5,	/* 478 */
	_nts_5,	/* 479 */
	_nts_39,	/* 480 */
	_nts_35,	/* 481 */
	_nts_35,	/* 482 */
	_nts_35,	/* 483 */
	_nts_7,	/* 484 */
	_nts_5,	/* 485 */
	_nts_17,	/* 486 */
	_nts_16,	/* 487 */
	_nts_35,	/* 488 */
	_nts_35,	/* 489 */
	_nts_35,	/* 490 */
	_nts_35,	/* 491 */
	_nts_35,	/* 492 */
	_nts_35,	/* 493 */
	_nts_35,	/* 494 */
	_nts_35,	/* 495 */
	_nts_35,	/* 496 */
	_nts_35,	/* 497 */
	_nts_41,	/* 498 */
	_nts_7,	/* 499 */
	_nts_5,	/* 500 */
	_nts_42,	/* 501 */
	_nts_42,	/* 502 */
	_nts_42,	/* 503 */
	_nts_42,	/* 504 */
	_nts_42,	/* 505 */
	_nts_42,	/* 506 */
	_nts_42,	/* 507 */
	_nts_42,	/* 508 */
	_nts_42,	/* 509 */
	_nts_42,	/* 510 */
	_nts_5,	/* 511 */
	_nts_5,	/* 512 */
	_nts_5,	/* 513 */
	_nts_5,	/* 514 */
	_nts_5,	/* 515 */
	_nts_5,	/* 516 */
	_nts_5,	/* 517 */
	_nts_5,	/* 518 */
	_nts_5,	/* 519 */
	_nts_17,	/* 520 */
	_nts_17,	/* 521 */
	_nts_17,	/* 522 */
	_nts_17,	/* 523 */
	_nts_17,	/* 524 */
	_nts_17,	/* 525 */
	_nts_17,	/* 526 */
	_nts_17,	/* 527 */
	_nts_17,	/* 528 */
	_nts_7,	/* 529 */
	_nts_7,	/* 530 */
	_nts_7,	/* 531 */
	_nts_5,	/* 532 */
	_nts_5,	/* 533 */
	_nts_5,	/* 534 */
	_nts_5,	/* 535 */
	_nts_5,	/* 536 */
	_nts_5,	/* 537 */
	_nts_5,	/* 538 */
	_nts_5,	/* 539 */
	_nts_5,	/* 540 */
	_nts_5,	/* 541 */
	_nts_5,	/* 542 */
	_nts_17,	/* 543 */
	_nts_17,	/* 544 */
	_nts_17,	/* 545 */
	_nts_17,	/* 546 */
	_nts_17,	/* 547 */
	_nts_17,	/* 548 */
	_nts_17,	/* 549 */
	_nts_17,	/* 550 */
	_nts_17,	/* 551 */
	_nts_17,	/* 552 */
	_nts_17,	/* 553 */
	_nts_35,	/* 554 */
	_nts_6,	/* 555 */
	_nts_0,	/* 556 */
	_nts_0,	/* 557 */
	_nts_0,	/* 558 */
	_nts_0,	/* 559 */
	_nts_0,	/* 560 */
};

static char *_templates[] = {
/* 0 */	0,
/* 1 */	"%a",	/* con1: CNSTI1 */
/* 2 */	"%a",	/* con1: CNSTU1 */
/* 3 */	"%a",	/* con2: CNSTI2 */
/* 4 */	"%a",	/* con2: CNSTU2 */
/* 5 */	"%a",	/* con4: CNSTI4 */
/* 6 */	"%a",	/* con4: CNSTU4 */
/* 7 */	"%a",	/* con4: CNSTP4 */
/* 8 */	"%a",	/* con8: CNSTI8 */
/* 9 */	"%a",	/* con8: CNSTU8 */
/* 10 */	"%a",	/* con4: CNSTF4 */
/* 11 */	"%a",	/* con8: CNSTF8 */
/* 12 */	"%a",	/* con: con1 */
/* 13 */	"%a",	/* con: con2 */
/* 14 */	"%a",	/* con: con4 */
/* 15 */	"%a",	/* con: con8 */
/* 16 */	"0",	/* conv0: CNSTI1 */
/* 17 */	"0",	/* conv0: CNSTI2 */
/* 18 */	"0",	/* conv0: CNSTI4 */
/* 19 */	"0",	/* conv0: CNSTI8 */
/* 20 */	"0",	/* conv0: CNSTU1 */
/* 21 */	"0",	/* conv0: CNSTU2 */
/* 22 */	"0",	/* conv0: CNSTU4 */
/* 23 */	"0",	/* conv0: CNSTU8 */
/* 24 */	"1",	/* conv1: CNSTI1 */
/* 25 */	"1",	/* conv1: CNSTI2 */
/* 26 */	"1",	/* conv1: CNSTI4 */
/* 27 */	"1",	/* conv1: CNSTI8 */
/* 28 */	"1",	/* conv1: CNSTU1 */
/* 29 */	"1",	/* conv1: CNSTU2 */
/* 30 */	"1",	/* conv1: CNSTU4 */
/* 31 */	"1",	/* conv1: CNSTU8 */
/* 32 */	"2",	/* conv2: CNSTI4 */
/* 33 */	"2",	/* conv2: CNSTU4 */
/* 34 */	"4",	/* conv4: CNSTI4 */
/* 35 */	"4",	/* conv4: CNSTU4 */
/* 36 */	"8",	/* conv8: CNSTI4 */
/* 37 */	"8",	/* conv8: CNSTU4 */
/* 38 */	"16",	/* conv16: CNSTI4 */
/* 39 */	"16",	/* conv16: CNSTU4 */
/* 40 */	"32",	/* conv32: CNSTI4 */
/* 41 */	"32",	/* conv32: CNSTU4 */
/* 42 */	"64",	/* conv64: CNSTI4 */
/* 43 */	"64",	/* conv64: CNSTU4 */
/* 44 */	"128",	/* conv128: CNSTI4 */
/* 45 */	"128",	/* conv128: CNSTU4 */
/* 46 */	"256",	/* conv256: CNSTI4 */
/* 47 */	"256",	/* conv256: CNSTU4 */
/* 48 */	"#\n",	/* reg: INDIRI1(VREGP) */
/* 49 */	"#\n",	/* reg: INDIRU1(VREGP) */
/* 50 */	"#\n",	/* reg: INDIRI2(VREGP) */
/* 51 */	"#\n",	/* reg: INDIRU2(VREGP) */
/* 52 */	"#\n",	/* reg: INDIRI4(VREGP) */
/* 53 */	"#\n",	/* reg: INDIRU4(VREGP) */
/* 54 */	"#\n",	/* reg: INDIRP4(VREGP) */
/* 55 */	"#\n",	/* reg: INDIRI8(VREGP) */
/* 56 */	"#\n",	/* reg: INDIRU8(VREGP) */
/* 57 */	"#\n",	/* reg: INDIRF4(VREGP) */
/* 58 */	"#\n",	/* reg: INDIRF8(VREGP) */
/* 59 */	"#\n",	/* reg: LOADI1(reg) */
/* 60 */	"#\n",	/* reg: LOADU1(reg) */
/* 61 */	"#\n",	/* reg: LOADI2(reg) */
/* 62 */	"#\n",	/* reg: LOADU2(reg) */
/* 63 */	"#\n",	/* reg: LOADI4(reg) */
/* 64 */	"#\n",	/* reg: LOADU4(reg) */
/* 65 */	"#\n",	/* reg: LOADP4(reg) */
/* 66 */	"#\n",	/* reg: LOADI8(reg) */
/* 67 */	"#\n",	/* reg: LOADU8(reg) */
/* 68 */	"\td mov %c,%0\n",	/* reg: LOADI4(adr) */
/* 69 */	"\td lea %c,%0\n",	/* reg: LOADU4(ad) */
/* 70 */	"\td lea %c,%0\n",	/* reg: LOADU4(INDIRU8(ad)) */
/* 71 */	"#\n",	/* reg: LOADU2(LOADU4(reg)) */
/* 72 */	"\tb mov %c,=%0\n",	/* reg: con1 */
/* 73 */	"\tw mov %c,=%0\n",	/* reg: con2 */
/* 74 */	"\td mov %c,=%0\n",	/* reg: con4 */
/* 75 */	"\tq mov %c,=%0\n",	/* reg: con8 */
/* 76 */	"",	/* stmt: reg */
/* 77 */	"=%a",	/* ca: ADDRGP4 */
/* 78 */	"%0",	/* ca: reg */
/* 79 */	"[%a]",	/* ad: ADDRGP4 */
/* 80 */	"[FP][%a]",	/* ad: ADDRFP4 */
/* 81 */	"[FP][%a]",	/* ad: ADDRLP4 */
/* 82 */	"[%0]",	/* adr: reg */
/* 83 */	"%0",	/* adr: ad */
/* 84 */	"%0",	/* iadr1: INDIRI1(adr) */
/* 85 */	"%0",	/* iadr1: INDIRU1(adr) */
/* 86 */	"%0",	/* iadr2: INDIRI2(adr) */
/* 87 */	"%0",	/* iadr2: INDIRU2(adr) */
/* 88 */	"%0",	/* iadr4: INDIRI4(adr) */
/* 89 */	"%0",	/* iadr4: INDIRU4(adr) */
/* 90 */	"%0",	/* iadr4: INDIRP4(adr) */
/* 91 */	"%0",	/* iadr8: INDIRI8(adr) */
/* 92 */	"%0",	/* iadr8: INDIRU8(adr) */
/* 93 */	"%0",	/* iadr4: INDIRF4(adr) */
/* 94 */	"%0",	/* iadr8: INDIRF8(adr) */
/* 95 */	"%0",	/* iadr: iadr1 */
/* 96 */	"%0",	/* iadr: iadr2 */
/* 97 */	"%0",	/* iadr: iadr4 */
/* 98 */	"%0",	/* iadr: iadr8 */
/* 99 */	"%0",	/* iadr: INDIRB(adr) */
/* 100 */	"%0",	/* eadr1: INDIRI1(ad) */
/* 101 */	"%0",	/* eadr1: INDIRU1(ad) */
/* 102 */	"%0",	/* eadr2: INDIRI2(ad) */
/* 103 */	"%0",	/* eadr2: INDIRU2(ad) */
/* 104 */	"%0",	/* eadr4: INDIRI4(ad) */
/* 105 */	"%0",	/* eadr4: INDIRU4(ad) */
/* 106 */	"%0",	/* eadr4: INDIRP4(ad) */
/* 107 */	"%0",	/* eadr8: INDIRI8(ad) */
/* 108 */	"%0",	/* eadr8: INDIRU8(ad) */
/* 109 */	"%0",	/* eadr4: INDIRF4(ad) */
/* 110 */	"%0",	/* eadr8: INDIRF8(ad) */
/* 111 */	"%0",	/* eadr: eadr1 */
/* 112 */	"%0",	/* eadr: eadr2 */
/* 113 */	"%0",	/* eadr: eadr4 */
/* 114 */	"%0",	/* eadr: eadr8 */
/* 115 */	"%0",	/* eadr: INDIRB(ad) */
/* 116 */	"\tb mov %c,%0\n",	/* reg: iadr1 */
/* 117 */	"\tw mov %c,%0\n",	/* reg: iadr2 */
/* 118 */	"\td mov %c,%0\n",	/* reg: iadr4 */
/* 119 */	"\tq mov %c,%0\n",	/* reg: iadr8 */
/* 120 */	"\td lea %c,%0\n",	/* reg: ad */
/* 121 */	"%0",	/* dm: reg */
/* 122 */	"%0",	/* dm: iadr */
/* 123 */	"%a",	/* dm: ADDRGP4 */
/* 124 */	"[FP][%a]",	/* dm: ADDRFP4 */
/* 125 */	"[FP][%a]",	/* dm: ADDRLP4 */
/* 126 */	"%0",	/* am: reg */
/* 127 */	"=%a",	/* am: con */
/* 128 */	"%0",	/* am: iadr */
/* 129 */	"=%a",	/* am: ADDRGP4 */
/* 130 */	"[FP][%a]",	/* am: ADDRFP4 */
/* 131 */	"[FP][%a]",	/* am: ADDRLP4 */
/* 132 */	"# write register\n",	/* stmt: ASGNI1(VREGP,reg) */
/* 133 */	"# write register\n",	/* stmt: ASGNU1(VREGP,reg) */
/* 134 */	"# write register\n",	/* stmt: ASGNI2(VREGP,reg) */
/* 135 */	"# write register\n",	/* stmt: ASGNU2(VREGP,reg) */
/* 136 */	"# write register\n",	/* stmt: ASGNI4(VREGP,reg) */
/* 137 */	"# write register\n",	/* stmt: ASGNU4(VREGP,reg) */
/* 138 */	"# write register\n",	/* stmt: ASGNP4(VREGP,reg) */
/* 139 */	"# write register\n",	/* stmt: ASGNF4(VREGP,reg) */
/* 140 */	"# write register\n",	/* stmt: ASGNI8(VREGP,reg) */
/* 141 */	"# write register\n",	/* stmt: ASGNU8(VREGP,reg) */
/* 142 */	"# write register\n",	/* stmt: ASGNF8(VREGP,reg) */
/* 143 */	"\td mov %c,%1\n",	/* stmt: ASGNP4(VREGP,ADDRGP4) */
/* 144 */	"\tb mov %0,%1\n",	/* stmt: ASGNI1(adr,am) */
/* 145 */	"\tb mov %0,%1\n",	/* stmt: ASGNU1(adr,am) */
/* 146 */	"\tw mov %0,%1\n",	/* stmt: ASGNI2(adr,am) */
/* 147 */	"\tw mov %0,%1\n",	/* stmt: ASGNU2(adr,am) */
/* 148 */	"\td mov %0,%1\n",	/* stmt: ASGNI4(adr,am) */
/* 149 */	"\td mov %0,%1\n",	/* stmt: ASGNU4(adr,am) */
/* 150 */	"\tw mov [%a],%0\n",	/* stmt: ASGNU2(INDIRP4(VREGP),LOADU2(reg)) */
/* 151 */	"\tw mov %a,%0\n",	/* reg: ASGNU2(INDIRP4(VREGP),LOADU2(reg)) */
/* 152 */	"\tw mov %0,%1\n",	/* stmt: ASGNU1(adr,LOADU1(INDIRU4(ad))) */
/* 153 */	"\tw mov %0,%1\n",	/* stmt: ASGNU2(adr,LOADU2(INDIRU4(ad))) */
/* 154 */	"\tw mov %0,%1\n",	/* stmt: ASGNI1(adr,LOADI1(INDIRI4(ad))) */
/* 155 */	"\tw mov %0,%1\n",	/* stmt: ASGNI2(adr,LOADI2(INDIRI4(ad))) */
/* 156 */	"\td mov %0,%1\n",	/* stmt: ASGNP4(adr,reg) */
/* 157 */	"\td mov %0,=%1\n",	/* stmt: ASGNP4(adr,con) */
/* 158 */	"\td mov %0,%1\n",	/* stmt: ASGNP4(adr,iadr) */
/* 159 */	"\td lea %0,%1\n",	/* stmt: ASGNP4(adr,ad) */
/* 160 */	"\tq mov %0,%1\n",	/* stmt: ASGNI8(adr,am) */
/* 161 */	"\tq mov %0,%1\n",	/* stmt: ASGNU8(adr,am) */
/* 162 */	"\td mov %0,%1\n",	/* stmt: ASGNF4(adr,am) */
/* 163 */	"\tq mov %0,%1\n",	/* stmt: ASGNF8(adr,am) */
/* 164 */	"\tb add %0,%2\n",	/* stmt: ASGNI1(adr,ADDI1(INDIRI1(adr),am)) */
/* 165 */	"\tb add %0,%2\n",	/* stmt: ASGNU1(adr,ADDU1(INDIRU1(adr),am)) */
/* 166 */	"\tw add %0,%2\n",	/* stmt: ASGNI2(adr,ADDI2(INDIRI2(adr),am)) */
/* 167 */	"\tw add %0,%2\n",	/* stmt: ASGNU2(adr,ADDU2(INDIRU2(adr),am)) */
/* 168 */	"\td add %0,%2\n",	/* stmt: ASGNI4(adr,ADDI4(INDIRI4(adr),am)) */
/* 169 */	"\td add %0,%2\n",	/* stmt: ASGNU4(adr,ADDU4(INDIRU4(adr),am)) */
/* 170 */	"\tq add %0,%2\n",	/* stmt: ASGNI8(adr,ADDI8(INDIRI8(adr),am)) */
/* 171 */	"\tq add %0,%2\n",	/* stmt: ASGNU8(adr,ADDU8(INDIRU8(adr),am)) */
/* 172 */	"\td sub %0,%2\n",	/* stmt: ASGNI4(adr,SUBI4(INDIRI4(adr),am)) */
/* 173 */	"\td sub %0,%2\n",	/* stmt: ASGNU4(adr,SUBU4(INDIRU4(adr),am)) */
/* 174 */	"\td rshi %0,%2\n",	/* stmt: ASGNI4(adr,RSHI4(INDIRI4(adr),am)) */
/* 175 */	"\td rshu %0,%2\n",	/* stmt: ASGNU4(adr,RSHU4(INDIRU4(adr),am)) */
/* 176 */	"\td lsh %0,%2\n",	/* stmt: ASGNI4(adr,LSHI4(INDIRI4(adr),am)) */
/* 177 */	"\td lsh %0,%2\n",	/* stmt: ASGNU4(adr,LSHU4(INDIRU4(adr),am)) */
/* 178 */	"\td and %0,%2\n",	/* stmt: ASGNI4(adr,BANDI4(INDIRI4(adr),am)) */
/* 179 */	"\td and %0,%2\n",	/* stmt: ASGNU4(adr,BANDU4(INDIRU4(adr),am)) */
/* 180 */	"\td or %0,%2\n",	/* stmt: ASGNI4(adr,BORI4(INDIRI4(adr),am)) */
/* 181 */	"\td or %0,%2\n",	/* stmt: ASGNU4(adr,BORU4(INDIRU4(adr),am)) */
/* 182 */	"\td xor %0,%2\n",	/* stmt: ASGNI4(adr,BXORI4(INDIRI4(adr),am)) */
/* 183 */	"\td xor %0,%2\n",	/* stmt: ASGNU4(adr,BXORU4(INDIRU4(adr),am)) */
/* 184 */	"\td mov [0xB8],=%a\n\tb memmov %0,%1\n",	/* stmt: ASGNB(adr,iadr) */
/* 185 */	"\tpumem =%a,%0\n",	/* stmt: ARGB(iadr) */
/* 186 */	"?\tb mov %c,%0\n\tb add  %c,%1\n",	/* reg: ADDI1(reg,am) */
/* 187 */	"?\tb mov %c,%0\n\tb add  %c,%1\n",	/* reg: ADDU1(reg,am) */
/* 188 */	"?\tw mov %c,%0\n\tw add  %c,%1\n",	/* reg: ADDI2(reg,am) */
/* 189 */	"?\tw mov %c,%0\n\tw add  %c,%1\n",	/* reg: ADDU2(reg,am) */
/* 190 */	"?\td mov %c,%0\n\td add  %c,%1\n",	/* reg: ADDI4(reg,am) */
/* 191 */	"?\td mov %c,%0\n\td add  %c,%1\n",	/* reg: ADDU4(reg,am) */
/* 192 */	"?\tq mov %c,%0\n\tq add  %c,%1\n",	/* reg: ADDI8(reg,am) */
/* 193 */	"?\tq mov %c,%0\n\tq add  %c,%1\n",	/* reg: ADDU8(reg,am) */
/* 194 */	"?\td mov %c,%0\n\td addf %c,%1\n",	/* reg: ADDF4(reg,am) */
/* 195 */	"?\tq mov %c,%0\n\tq addf %c,%1\n",	/* reg: ADDF8(reg,am) */
/* 196 */	"?\td mov %c,%0\n\td add  %c,%1\n",	/* reg: ADDP4(reg,ca) */
/* 197 */	"?\tb mov %c,%0\n\tb sub  %c,%1\n",	/* reg: SUBI1(reg,am) */
/* 198 */	"?\tb mov %c,%0\n\tb sub  %c,%1\n",	/* reg: SUBU1(reg,am) */
/* 199 */	"?\tw mov %c,%0\n\tw sub  %c,%1\n",	/* reg: SUBI2(reg,am) */
/* 200 */	"?\tw mov %c,%0\n\tw sub  %c,%1\n",	/* reg: SUBU2(reg,am) */
/* 201 */	"?\td mov %c,%0\n\td sub  %c,%1\n",	/* reg: SUBI4(reg,am) */
/* 202 */	"?\td mov %c,%0\n\td sub  %c,%1\n",	/* reg: SUBU4(reg,am) */
/* 203 */	"?\td mov %c,%0\n\td sub  %c,%1\n",	/* reg: SUBP4(reg,am) */
/* 204 */	"?\tq mov %c,%0\n\tq sub  %c,%1\n",	/* reg: SUBI8(reg,am) */
/* 205 */	"?\tq mov %c,%0\n\tq sub  %c,%1\n",	/* reg: SUBU8(reg,am) */
/* 206 */	"?\td mov %c,%0\n\td subf %c,%1\n",	/* reg: SUBF4(reg,am) */
/* 207 */	"?\tq mov %c,%0\n\tq subf %c,%1\n",	/* reg: SUBF8(reg,am) */
/* 208 */	"\td mov %c,=%0\n\td mul %c,%1\n",	/* reg: MULU4(con,LOADU4(eadr)) */
/* 209 */	"?\tb mov %c,%0\n\tb mul  %c,%1\n",	/* reg: MULI1(reg,am) */
/* 210 */	"?\tb mov %c,%0\n\tb mul  %c,%1\n",	/* reg: MULU1(reg,am) */
/* 211 */	"?\tw mov %c,%0\n\tw mul  %c,%1\n",	/* reg: MULI2(reg,am) */
/* 212 */	"?\tw mov %c,%0\n\tw mul  %c,%1\n",	/* reg: MULU2(reg,am) */
/* 213 */	"?\td mov %c,%0\n\td mul  %c,%1\n",	/* reg: MULI4(reg,am) */
/* 214 */	"?\td mov %c,%0\n\td mul  %c,%1\n",	/* reg: MULU4(reg,am) */
/* 215 */	"?\tq mov %c,%0\n\tq mul  %c,%1\n",	/* reg: MULI8(reg,am) */
/* 216 */	"?\tq mov %c,%0\n\tq mul  %c,%1\n",	/* reg: MULU8(reg,am) */
/* 217 */	"?\td mov %c,%0\n\td mulf %c,%1\n",	/* reg: MULF4(reg,am) */
/* 218 */	"?\tq mov %c,%0\n\tq mulf %c,%1\n",	/* reg: MULF8(reg,am) */
/* 219 */	"?\td mov %c,%0\n\td lsh %c,=1\n",	/* reg: MULI4(reg,conv2) */
/* 220 */	"?\td mov %c,%0\n\td lsh %c,=2\n",	/* reg: MULI4(reg,conv4) */
/* 221 */	"?\td mov %c,%0\n\td lsh %c,=3\n",	/* reg: MULI4(reg,conv8) */
/* 222 */	"?\td mov %c,%0\n\td lsh %c,=4\n",	/* reg: MULI4(reg,conv16) */
/* 223 */	"?\td mov %c,%0\n\td lsh %c,=5\n",	/* reg: MULI4(reg,conv32) */
/* 224 */	"?\td mov %c,%0\n\td lsh %c,=6\n",	/* reg: MULI4(reg,conv64) */
/* 225 */	"?\td mov %c,%0\n\td lsh %c,=7\n",	/* reg: MULI4(reg,conv128) */
/* 226 */	"?\td mov %c,%0\n\td lsh %c,=8\n",	/* reg: MULI4(reg,conv256) */
/* 227 */	"?\td mov %c,%0\n\td lsh %c,=1\n",	/* reg: MULU4(reg,conv2) */
/* 228 */	"?\td mov %c,%0\n\td lsh %c,=2\n",	/* reg: MULU4(reg,conv4) */
/* 229 */	"?\td mov %c,%0\n\td lsh %c,=3\n",	/* reg: MULU4(reg,conv8) */
/* 230 */	"?\td mov %c,%0\n\td lsh %c,=4\n",	/* reg: MULU4(reg,conv16) */
/* 231 */	"?\td mov %c,%0\n\td lsh %c,=5\n",	/* reg: MULU4(reg,conv32) */
/* 232 */	"?\td mov %c,%0\n\td lsh %c,=6\n",	/* reg: MULU4(reg,conv64) */
/* 233 */	"?\td mov %c,%0\n\td lsh %c,=7\n",	/* reg: MULU4(reg,conv128) */
/* 234 */	"?\td mov %c,%0\n\td lsh %c,=8\n",	/* reg: MULU4(reg,conv256) */
/* 235 */	"?\tb mov %c,%0\n\tb divi %c,%1\n",	/* reg: DIVI1(reg,am) */
/* 236 */	"?\tb mov %c,%0\n\tb divu %c,%1\n",	/* reg: DIVU1(reg,am) */
/* 237 */	"?\tw mov %c,%0\n\tw divi %c,%1\n",	/* reg: DIVI2(reg,am) */
/* 238 */	"?\tw mov %c,%0\n\tw divu %c,%1\n",	/* reg: DIVU2(reg,am) */
/* 239 */	"?\td mov %c,%0\n\td divi %c,%1\n",	/* reg: DIVI4(reg,am) */
/* 240 */	"?\td mov %c,%0\n\td divu %c,%1\n",	/* reg: DIVU4(reg,am) */
/* 241 */	"?\tq mov %c,%0\n\tq divi %c,%1\n",	/* reg: DIVI8(reg,am) */
/* 242 */	"?\tq mov %c,%0\n\tq divu %c,%1\n",	/* reg: DIVU8(reg,am) */
/* 243 */	"?\td mov %c,%0\n\td divf %c,%1\n",	/* reg: DIVF4(reg,am) */
/* 244 */	"?\tq mov %c,%0\n\tq divf %c,%1\n",	/* reg: DIVF8(reg,am) */
/* 245 */	"?\td mov %c,%0\n\td rshi %c,=1\n",	/* reg: DIVI4(reg,conv2) */
/* 246 */	"?\td mov %c,%0\n\td rshi %c,=2\n",	/* reg: DIVI4(reg,conv4) */
/* 247 */	"?\td mov %c,%0\n\td rshi %c,=3\n",	/* reg: DIVI4(reg,conv8) */
/* 248 */	"?\td mov %c,%0\n\td rshi %c,=4\n",	/* reg: DIVI4(reg,conv16) */
/* 249 */	"?\td mov %c,%0\n\td rshi %c,=5\n",	/* reg: DIVI4(reg,conv32) */
/* 250 */	"?\td mov %c,%0\n\td rshi %c,=6\n",	/* reg: DIVI4(reg,conv64) */
/* 251 */	"?\td mov %c,%0\n\td rshi %c,=7\n",	/* reg: DIVI4(reg,conv128) */
/* 252 */	"?\td mov %c,%0\n\td rshi %c,=8\n",	/* reg: DIVI4(reg,conv256) */
/* 253 */	"?\td mov %c,%0\n\td rshu %c,=1\n",	/* reg: DIVU4(reg,conv2) */
/* 254 */	"?\td mov %c,%0\n\td rshu %c,=2\n",	/* reg: DIVU4(reg,conv4) */
/* 255 */	"?\td mov %c,%0\n\td rshu %c,=3\n",	/* reg: DIVU4(reg,conv8) */
/* 256 */	"?\td mov %c,%0\n\td rshu %c,=4\n",	/* reg: DIVU4(reg,conv16) */
/* 257 */	"?\td mov %c,%0\n\td rshu %c,=5\n",	/* reg: DIVU4(reg,conv32) */
/* 258 */	"?\td mov %c,%0\n\td rshu %c,=6\n",	/* reg: DIVU4(reg,conv64) */
/* 259 */	"?\td mov %c,%0\n\td rshu %c,=7\n",	/* reg: DIVU4(reg,conv128) */
/* 260 */	"?\td mov %c,%0\n\td rshu %c,=8\n",	/* reg: DIVU4(reg,conv256) */
/* 261 */	"?\tb mov %c,%0\n\tb modi %c,%1\n",	/* reg: MODI1(reg,am) */
/* 262 */	"?\tb mov %c,%0\n\tb modu %c,%1\n",	/* reg: MODU1(reg,am) */
/* 263 */	"?\tw mov %c,%0\n\tw modi %c,%1\n",	/* reg: MODI2(reg,am) */
/* 264 */	"?\tw mov %c,%0\n\tw modu %c,%1\n",	/* reg: MODU2(reg,am) */
/* 265 */	"?\td mov %c,%0\n\td modi %c,%1\n",	/* reg: MODI4(reg,am) */
/* 266 */	"?\td mov %c,%0\n\td modu %c,%1\n",	/* reg: MODU4(reg,am) */
/* 267 */	"?\tq mov %c,%0\n\tq modi %c,%1\n",	/* reg: MODI8(reg,am) */
/* 268 */	"?\tq mov %c,%0\n\tq modu %c,%1\n",	/* reg: MODU8(reg,am) */
/* 269 */	"?\tb mov %c,%0\n\tb rshi %c,%1\n",	/* reg: RSHI1(reg,am) */
/* 270 */	"?\tb mov %c,%0\n\tb rshu %c,%1\n",	/* reg: RSHU1(reg,am) */
/* 271 */	"?\tw mov %c,%0\n\tw rshi %c,%1\n",	/* reg: RSHI2(reg,am) */
/* 272 */	"?\tw mov %c,%0\n\tw rshu %c,%1\n",	/* reg: RSHU2(reg,am) */
/* 273 */	"?\td mov %c,%0\n\td rshi %c,%1\n",	/* reg: RSHI4(reg,am) */
/* 274 */	"?\td mov %c,%0\n\td rshu %c,%1\n",	/* reg: RSHU4(reg,am) */
/* 275 */	"?\tq mov %c,%0\n\tq rshi %c,%1\n",	/* reg: RSHI8(reg,am) */
/* 276 */	"?\tq mov %c,%0\n\tq rshu %c,%1\n",	/* reg: RSHU8(reg,am) */
/* 277 */	"?\tb mov %c,%0\n\tb lsh %c,%1\n",	/* reg: LSHI1(reg,am) */
/* 278 */	"?\tb mov %c,%0\n\tb lsh %c,%1\n",	/* reg: LSHU1(reg,am) */
/* 279 */	"?\tw mov %c,%0\n\tw lsh %c,%1\n",	/* reg: LSHI2(reg,am) */
/* 280 */	"?\tw mov %c,%0\n\tw lsh %c,%1\n",	/* reg: LSHU2(reg,am) */
/* 281 */	"?\td mov %c,%0\n\td lsh %c,%1\n",	/* reg: LSHI4(reg,am) */
/* 282 */	"?\td mov %c,%0\n\td lsh %c,%1\n",	/* reg: LSHU4(reg,am) */
/* 283 */	"?\tq mov %c,%0\n\tq lsh %c,%1\n",	/* reg: LSHI8(reg,am) */
/* 284 */	"?\tq mov %c,%0\n\tq lsh %c,%1\n",	/* reg: LSHU8(reg,am) */
/* 285 */	"?\tb mov %c,%0\n\tb and %c,%1\n",	/* reg: BANDI1(reg,am) */
/* 286 */	"?\tb mov %c,%0\n\tb and %c,%1\n",	/* reg: BANDU1(reg,am) */
/* 287 */	"?\tw mov %c,%0\n\tw and %c,%1\n",	/* reg: BANDI2(reg,am) */
/* 288 */	"?\tw mov %c,%0\n\tw and %c,%1\n",	/* reg: BANDU2(reg,am) */
/* 289 */	"?\td mov %c,%0\n\td and %c,%1\n",	/* reg: BANDI4(reg,am) */
/* 290 */	"?\td mov %c,%0\n\td and %c,%1\n",	/* reg: BANDU4(reg,am) */
/* 291 */	"?\tq mov %c,%0\n\tq and %c,%1\n",	/* reg: BANDI8(reg,am) */
/* 292 */	"?\tq mov %c,%0\n\tq and %c,%1\n",	/* reg: BANDU8(reg,am) */
/* 293 */	"?\tb mov %c,%0\n\tb or %c,%1\n",	/* reg: BORI1(reg,am) */
/* 294 */	"?\tb mov %c,%0\n\tb or %c,%1\n",	/* reg: BORU1(reg,am) */
/* 295 */	"?\tw mov %c,%0\n\tw or %c,%1\n",	/* reg: BORI2(reg,am) */
/* 296 */	"?\tw mov %c,%0\n\tw or %c,%1\n",	/* reg: BORU2(reg,am) */
/* 297 */	"?\td mov %c,%0\n\td or %c,%1\n",	/* reg: BORI4(reg,am) */
/* 298 */	"?\td mov %c,%0\n\td or %c,%1\n",	/* reg: BORU4(reg,am) */
/* 299 */	"?\tq mov %c,%0\n\tq or %c,%1\n",	/* reg: BORI8(reg,am) */
/* 300 */	"?\tq mov %c,%0\n\tq or %c,%1\n",	/* reg: BORU8(reg,am) */
/* 301 */	"?\tb mov %c,%0\n\tb xor %c,%1\n",	/* reg: BXORI1(reg,am) */
/* 302 */	"?\tb mov %c,%0\n\tb xor %c,%1\n",	/* reg: BXORU1(reg,am) */
/* 303 */	"?\tw mov %c,%0\n\tw xor %c,%1\n",	/* reg: BXORI2(reg,am) */
/* 304 */	"?\tw mov %c,%0\n\tw xor %c,%1\n",	/* reg: BXORU2(reg,am) */
/* 305 */	"?\td mov %c,%0\n\td xor %c,%1\n",	/* reg: BXORI4(reg,am) */
/* 306 */	"?\td mov %c,%0\n\td xor %c,%1\n",	/* reg: BXORU4(reg,am) */
/* 307 */	"?\tq mov %c,%0\n\tq xor %c,%1\n",	/* reg: BXORI8(reg,am) */
/* 308 */	"?\tq mov %c,%0\n\tq xor %c,%1\n",	/* reg: BXORU8(reg,am) */
/* 309 */	"?\tb mov %c,%0\n\tb not %c\n",	/* reg: BCOMI1(reg) */
/* 310 */	"?\tb mov %c,%0\n\tb not %c\n",	/* reg: BCOMU1(reg) */
/* 311 */	"?\tw mov %c,%0\n\tw not %c\n",	/* reg: BCOMI2(reg) */
/* 312 */	"?\tw mov %c,%0\n\tw not %c\n",	/* reg: BCOMU2(reg) */
/* 313 */	"?\td mov %c,%0\n\td not %c\n",	/* reg: BCOMI4(reg) */
/* 314 */	"?\td mov %c,%0\n\td not %c\n",	/* reg: BCOMU4(reg) */
/* 315 */	"?\tq mov %c,%0\n\tq not %c\n",	/* reg: BCOMI8(reg) */
/* 316 */	"?\tq mov %c,%0\n\tq not %c\n",	/* reg: BCOMU8(reg) */
/* 317 */	"\tb not %0\n",	/* stmt: BCOMI1(am) */
/* 318 */	"\tb not %0\n",	/* stmt: BCOMU1(am) */
/* 319 */	"\tw not %0\n",	/* stmt: BCOMI2(am) */
/* 320 */	"\tw not %0\n",	/* stmt: BCOMU2(am) */
/* 321 */	"\td not %0\n",	/* stmt: BCOMI4(am) */
/* 322 */	"\td not %0\n",	/* stmt: BCOMU4(am) */
/* 323 */	"\tq not %0\n",	/* stmt: BCOMI8(am) */
/* 324 */	"\tq not %0\n",	/* stmt: BCOMU8(am) */
/* 325 */	"?\tb mov %c,%0\n\tb neg  %c\n",	/* reg: NEGI1(reg) */
/* 326 */	"?\tw mov %c,%0\n\tw neg  %c\n",	/* reg: NEGI2(reg) */
/* 327 */	"?\td mov %c,%0\n\td neg  %c\n",	/* reg: NEGI4(reg) */
/* 328 */	"?\tq mov %c,%0\n\tq neg  %c\n",	/* reg: NEGI8(reg) */
/* 329 */	"?\td mov %c,%0\n\td negf %c\n",	/* reg: NEGF4(reg) */
/* 330 */	"?\tq mov %c,%0\n\tq negf %c\n",	/* reg: NEGF8(reg) */
/* 331 */	"\tb tljeq %0,=%a\n",	/* stmt: EQI1(dm,conv0) */
/* 332 */	"\tb tljeq %0,=%a\n",	/* stmt: EQU1(dm,conv0) */
/* 333 */	"\tw tljeq %0,=%a\n",	/* stmt: EQI2(dm,conv0) */
/* 334 */	"\tw tljeq %0,=%a\n",	/* stmt: EQU2(dm,conv0) */
/* 335 */	"\td tljeq %0,=%a\n",	/* stmt: EQI4(dm,conv0) */
/* 336 */	"\td tljeq %0,=%a\n",	/* stmt: EQU4(dm,conv0) */
/* 337 */	"\tq tljeq %0,=%a\n",	/* stmt: EQI8(dm,conv0) */
/* 338 */	"\tq tljeq %0,=%a\n",	/* stmt: EQU8(dm,conv0) */
/* 339 */	"\td tljeq %0,=%a\n",	/* stmt: EQF4(dm,conv0) */
/* 340 */	"\tq tljeq %0,=%a\n",	/* stmt: EQF8(dm,conv0) */
/* 341 */	"\tb tljev %0,=%a\n",	/* stmt: EQI1(BANDI1(dm,conv1),conv0) */
/* 342 */	"\tb tljev %0,=%a\n",	/* stmt: EQU1(BANDU1(dm,conv1),conv0) */
/* 343 */	"\tw tljev %0,=%a\n",	/* stmt: EQI2(BANDI2(dm,conv1),conv0) */
/* 344 */	"\tw tljev %0,=%a\n",	/* stmt: EQU2(BANDU2(dm,conv1),conv0) */
/* 345 */	"\td tljev %0,=%a\n",	/* stmt: EQI4(BANDI4(dm,conv1),conv0) */
/* 346 */	"\td tljev %0,=%a\n",	/* stmt: EQU4(BANDU4(dm,conv1),conv0) */
/* 347 */	"\tq tljev %0,=%a\n",	/* stmt: EQI8(BANDI8(dm,conv1),conv0) */
/* 348 */	"\tq tljev %0,=%a\n",	/* stmt: EQU8(BANDU8(dm,conv1),conv0) */
/* 349 */	"\tb tljod %0,=%a\n",	/* stmt: EQI1(BANDI1(dm,conv1),conv1) */
/* 350 */	"\tb tljod %0,=%a\n",	/* stmt: EQU1(BANDU1(dm,conv1),conv1) */
/* 351 */	"\tw tljod %0,=%a\n",	/* stmt: EQI2(BANDI2(dm,conv1),conv1) */
/* 352 */	"\tw tljod %0,=%a\n",	/* stmt: EQU2(BANDU2(dm,conv1),conv1) */
/* 353 */	"\td tljod %0,=%a\n",	/* stmt: EQI4(BANDI4(dm,conv1),conv1) */
/* 354 */	"\td tljod %0,=%a\n",	/* stmt: EQU4(BANDU4(dm,conv1),conv1) */
/* 355 */	"\tq tljod %0,=%a\n",	/* stmt: EQI8(BANDI8(dm,conv1),conv1) */
/* 356 */	"\tq tljod %0,=%a\n",	/* stmt: EQU8(BANDU8(dm,conv1),conv1) */
/* 357 */	"\tb cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQI1(dm,am) */
/* 358 */	"\tb cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQU1(dm,am) */
/* 359 */	"\tw cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQI2(dm,am) */
/* 360 */	"\tw cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQU2(dm,am) */
/* 361 */	"\td cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQI4(dm,am) */
/* 362 */	"\td cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQU4(dm,am) */
/* 363 */	"\tq cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQI8(dm,am) */
/* 364 */	"\tq cmp  %0,%1\n\tljeq =%a\n",	/* stmt: EQU8(dm,am) */
/* 365 */	"\td cmpf %0,%1\n\tljeq =%a\n",	/* stmt: EQF4(dm,am) */
/* 366 */	"\tq cmpf %0,%1\n\tljeq =%a\n",	/* stmt: EQF8(dm,am) */
/* 367 */	"\tb tljne %0,=%a\n",	/* stmt: NEI1(dm,conv0) */
/* 368 */	"\tb tljne %0,=%a\n",	/* stmt: NEU1(dm,conv0) */
/* 369 */	"\tw tljne %0,=%a\n",	/* stmt: NEI2(dm,conv0) */
/* 370 */	"\tw tljne %0,=%a\n",	/* stmt: NEU2(dm,conv0) */
/* 371 */	"\td tljne %0,=%a\n",	/* stmt: NEI4(dm,conv0) */
/* 372 */	"\td tljne %0,=%a\n",	/* stmt: NEU4(dm,conv0) */
/* 373 */	"\tq tljne %0,=%a\n",	/* stmt: NEI8(dm,conv0) */
/* 374 */	"\tq tljne %0,=%a\n",	/* stmt: NEU8(dm,conv0) */
/* 375 */	"\td tljne %0,=%a\n",	/* stmt: NEF4(dm,conv0) */
/* 376 */	"\tq tljne %0,=%a\n",	/* stmt: NEF8(dm,conv0) */
/* 377 */	"\tb tljod %0,=%a\n",	/* stmt: NEI1(BANDI1(dm,conv1),conv0) */
/* 378 */	"\tb tljod %0,=%a\n",	/* stmt: NEU1(BANDU1(dm,conv1),conv0) */
/* 379 */	"\tw tljod %0,=%a\n",	/* stmt: NEI2(BANDI2(dm,conv1),conv0) */
/* 380 */	"\tw tljod %0,=%a\n",	/* stmt: NEU2(BANDU2(dm,conv1),conv0) */
/* 381 */	"\td tljod %0,=%a\n",	/* stmt: NEI4(BANDI4(dm,conv1),conv0) */
/* 382 */	"\td tljod %0,=%a\n",	/* stmt: NEU4(BANDU4(dm,conv1),conv0) */
/* 383 */	"\tq tljod %0,=%a\n",	/* stmt: NEI8(BANDI8(dm,conv1),conv0) */
/* 384 */	"\tq tljod %0,=%a\n",	/* stmt: NEU8(BANDU8(dm,conv1),conv0) */
/* 385 */	"\tb tljev %0,=%a\n",	/* stmt: NEI1(BANDI1(dm,conv1),conv1) */
/* 386 */	"\tb tljev %0,=%a\n",	/* stmt: NEU1(BANDU1(dm,conv1),conv1) */
/* 387 */	"\tw tljev %0,=%a\n",	/* stmt: NEI2(BANDI2(dm,conv1),conv1) */
/* 388 */	"\tw tljev %0,=%a\n",	/* stmt: NEU2(BANDU2(dm,conv1),conv1) */
/* 389 */	"\td tljev %0,=%a\n",	/* stmt: NEI4(BANDI4(dm,conv1),conv1) */
/* 390 */	"\td tljev %0,=%a\n",	/* stmt: NEU4(BANDU4(dm,conv1),conv1) */
/* 391 */	"\tq tljev %0,=%a\n",	/* stmt: NEI8(BANDI8(dm,conv1),conv1) */
/* 392 */	"\tq tljev %0,=%a\n",	/* stmt: NEU8(BANDU8(dm,conv1),conv1) */
/* 393 */	"\tb cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEI1(dm,am) */
/* 394 */	"\tb cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEU1(dm,am) */
/* 395 */	"\tw cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEI2(dm,am) */
/* 396 */	"\tw cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEU2(dm,am) */
/* 397 */	"\td cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEI4(dm,am) */
/* 398 */	"\td cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEU4(dm,am) */
/* 399 */	"\tq cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEI8(dm,am) */
/* 400 */	"\tq cmp  %0,%1\n\tljne =%a\n",	/* stmt: NEU8(dm,am) */
/* 401 */	"\td cmpf %0,%1\n\tljne =%a\n",	/* stmt: NEF4(dm,am) */
/* 402 */	"\tq cmpf %0,%1\n\tljne =%a\n",	/* stmt: NEF8(dm,am) */
/* 403 */	"\tb cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEI1(dm,am) */
/* 404 */	"\tb cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEU1(dm,am) */
/* 405 */	"\tw cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEI2(dm,am) */
/* 406 */	"\tw cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEU2(dm,am) */
/* 407 */	"\td cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEI4(dm,am) */
/* 408 */	"\td cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEU4(dm,am) */
/* 409 */	"\tq cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEI8(dm,am) */
/* 410 */	"\tq cmp  %0,%1\n\tljge =%a\n",	/* stmt: GEU8(dm,am) */
/* 411 */	"\td cmpf %0,%1\n\tljge =%a\n",	/* stmt: GEF4(dm,am) */
/* 412 */	"\tq cmpf %0,%1\n\tljge =%a\n",	/* stmt: GEF8(dm,am) */
/* 413 */	"\tb cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTI1(dm,am) */
/* 414 */	"\tb cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTU1(dm,am) */
/* 415 */	"\tw cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTI2(dm,am) */
/* 416 */	"\tw cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTU2(dm,am) */
/* 417 */	"\td cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTI4(dm,am) */
/* 418 */	"\td cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTU4(dm,am) */
/* 419 */	"\tq cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTI8(dm,am) */
/* 420 */	"\tq cmp  %0,%1\n\tljgt =%a\n",	/* stmt: GTU8(dm,am) */
/* 421 */	"\td cmpf %0,%1\n\tljgt =%a\n",	/* stmt: GTF4(dm,am) */
/* 422 */	"\tq cmpf %0,%1\n\tljgt =%a\n",	/* stmt: GTF8(dm,am) */
/* 423 */	"\tb cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEI1(dm,am) */
/* 424 */	"\tb cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEU1(dm,am) */
/* 425 */	"\tw cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEI2(dm,am) */
/* 426 */	"\tw cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEU2(dm,am) */
/* 427 */	"\td cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEI4(dm,am) */
/* 428 */	"\td cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEU4(dm,am) */
/* 429 */	"\tq cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEI8(dm,am) */
/* 430 */	"\tq cmp  %0,%1\n\tljle =%a\n",	/* stmt: LEU8(dm,am) */
/* 431 */	"\td cmpf %0,%1\n\tljle =%a\n",	/* stmt: LEF4(dm,am) */
/* 432 */	"\tq cmpf %0,%1\n\tljle =%a\n",	/* stmt: LEF8(dm,am) */
/* 433 */	"\tb cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI1(dm,am) */
/* 434 */	"\tb cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU1(dm,am) */
/* 435 */	"\tw cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI2(dm,am) */
/* 436 */	"\tw cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU2(dm,am) */
/* 437 */	"\td cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI4(dm,am) */
/* 438 */	"\td cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU4(dm,am) */
/* 439 */	"\tq cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI8(dm,am) */
/* 440 */	"\tq cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU8(dm,am) */
/* 441 */	"\td cmpf %0,%1\n\tljlt =%a\n",	/* stmt: LTF4(dm,am) */
/* 442 */	"\tq cmpf %0,%1\n\tljlt =%a\n",	/* stmt: LTF8(dm,am) */
/* 443 */	"\tb cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI1(eadr,eadr) */
/* 444 */	"\tb cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU1(eadr,eadr) */
/* 445 */	"\tw cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI2(eadr,eadr) */
/* 446 */	"\tw cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU2(eadr,eadr) */
/* 447 */	"\td cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI4(eadr,eadr) */
/* 448 */	"\td cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU4(eadr,eadr) */
/* 449 */	"\tq cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTI8(eadr,eadr) */
/* 450 */	"\tq cmp  %0,%1\n\tljlt =%a\n",	/* stmt: LTU8(eadr,eadr) */
/* 451 */	"\td cmpf %0,%1\n\tljlt =%a\n",	/* stmt: LTF4(eadr,eadr) */
/* 452 */	"\tq cmpf %0,%1\n\tljlt =%a\n",	/* stmt: LTF8(eadr,eadr) */
/* 453 */	"\td mov %c,%0\n",	/* reg: CVPU4(reg) */
/* 454 */	"\td mov %c,%0\n",	/* reg: CVUP4(reg) */
/* 455 */	"#\n",	/* reg: CVII1(reg) */
/* 456 */	"#\n",	/* reg: CVIU1(reg) */
/* 457 */	"#\n",	/* reg: CVII2(reg) */
/* 458 */	"#\n",	/* reg: CVIU2(reg) */
/* 459 */	"#\n",	/* reg: CVII4(reg) */
/* 460 */	"#\n",	/* reg: CVIU4(reg) */
/* 461 */	"#\n",	/* reg: CVII8(reg) */
/* 462 */	"#\n",	/* reg: CVIU8(reg) */
/* 463 */	"#\td mov %c,%0\n\tcvsf %c\n",	/* reg: CVIF4(reg) */
/* 464 */	"#\tq mov %c,%0\n\tcvsd %c\n",	/* reg: CVIF8(reg) */
/* 465 */	"#\n",	/* reg: CVUI1(reg) */
/* 466 */	"#\n",	/* reg: CVUU1(reg) */
/* 467 */	"#\n",	/* reg: CVUI2(reg) */
/* 468 */	"#\n",	/* reg: CVUU2(reg) */
/* 469 */	"#\n",	/* reg: CVUI4(reg) */
/* 470 */	"#\n",	/* reg: CVUU4(reg) */
/* 471 */	"#\n",	/* reg: CVUI8(reg) */
/* 472 */	"#\n",	/* reg: CVUU8(reg) */
/* 473 */	"#\td mov %c,%0\n\tcvuf %c\n",	/* reg: CVUF4(reg) */
/* 474 */	"#\tq mov %c,%0\n\tcvud %c\n",	/* reg: CVUF8(reg) */
/* 475 */	"\tq mov %c,%0\n\tcvdf %c\n",	/* reg: CVFF4(reg) */
/* 476 */	"?\td mov %c,%0\n\tcvfd %c\n",	/* reg: CVFF8(reg) */
/* 477 */	"#\tw mov %c,%0\n\tcvfs %c\n",	/* reg: CVFI2(reg) */
/* 478 */	"#\td mov %c,%0\n\tcvfs %c\n",	/* reg: CVFI4(reg) */
/* 479 */	"#\tq mov %c,%0\n\tcvfs %c\n",	/* reg: CVFI8(reg) */
/* 480 */	"\td mov %c,%0\n\tadd %c,%1\n\tmov %c,[%c]\n\tcvfd %c\n",	/* reg: CVFF8(INDIRF4(ADDP4(dm,am))) */
/* 481 */	"\tw push %0\n",	/* stmt: ARGI2(LOADI4(am)) */
/* 482 */	"\tw push %0\n",	/* stmt: ARGU2(LOADU4(am)) */
/* 483 */	"\td push %0\n",	/* stmt: ARGI4(LOADI4(am)) */
/* 484 */	"\td pea %0\n",	/* stmt: ARGU4(LOADU4(ad)) */
/* 485 */	"\td push %0\n",	/* stmt: ARGU4(LOADU4(reg)) */
/* 486 */	"\td push %0\n",	/* stmt: ARGU4(LOADU4(con)) */
/* 487 */	"\td push %0\n",	/* stmt: ARGU4(LOADU4(iadr)) */
/* 488 */	"\tq push %0\n",	/* stmt: ARGI4(LOADI8(am)) */
/* 489 */	"\tq push %0\n",	/* stmt: ARGU4(LOADU8(am)) */
/* 490 */	"\tw push %0\n",	/* stmt: ARGI2(am) */
/* 491 */	"\tw push %0\n",	/* stmt: ARGU2(am) */
/* 492 */	"\td push %0\n",	/* stmt: ARGI4(am) */
/* 493 */	"\td push %0\n",	/* stmt: ARGU4(am) */
/* 494 */	"\tq push %0\n",	/* stmt: ARGI8(am) */
/* 495 */	"\tq push %0\n",	/* stmt: ARGU8(am) */
/* 496 */	"\td push %0\n",	/* stmt: ARGF4(am) */
/* 497 */	"\tq push %0\n",	/* stmt: ARGF8(am) */
/* 498 */	"\td push %0\n",	/* stmt: ARGP4(eadr) */
/* 499 */	"\td pea  %0\n",	/* stmt: ARGP4(ad) */
/* 500 */	"\td push %0\n",	/* stmt: ARGP4(reg) */
/* 501 */	"\tjsrfr %a,%0\n",	/* reg: CALLI2(ca) */
/* 502 */	"\tjsrfr %a,%0\n",	/* reg: CALLU2(ca) */
/* 503 */	"\tjsrfr %a,%0\n",	/* reg: CALLI4(ca) */
/* 504 */	"\tjsrfr %a,%0\n",	/* reg: CALLU4(ca) */
/* 505 */	"\tjsrfr %a,%0\n",	/* reg: CALLP4(ca) */
/* 506 */	"\tjsrfr %a,%0\n",	/* reg: CALLI8(ca) */
/* 507 */	"\tjsrfr %a,%0\n",	/* reg: CALLU8(ca) */
/* 508 */	"\tjsrfr %a,%0\n",	/* reg: CALLF4(ca) */
/* 509 */	"\tjsrfr %a,%0\n",	/* reg: CALLF8(ca) */
/* 510 */	"\tjsrfr %a,%0\n",	/* stmt: CALLV(ca) */
/* 511 */	"?\tb mov %c,%0\n",	/* reg: RETI1(reg) */
/* 512 */	"?\tb mov %c,%0\n",	/* reg: RETU1(reg) */
/* 513 */	"?\tw mov %c,%0\n",	/* reg: RETI2(reg) */
/* 514 */	"?\tw mov %c,%0\n",	/* reg: RETU2(reg) */
/* 515 */	"?\td mov %c,%0\n",	/* reg: RETI4(reg) */
/* 516 */	"?\td mov %c,%0\n",	/* reg: RETU4(reg) */
/* 517 */	"?\td mov %c,%0\n",	/* reg: RETP4(reg) */
/* 518 */	"?\tq mov %c,%0\n",	/* reg: RETI8(reg) */
/* 519 */	"?\tq mov %c,%0\n",	/* reg: RETU8(reg) */
/* 520 */	"\tb mov %c,=%0\n",	/* reg: RETI1(con) */
/* 521 */	"\tb mov %c,=%0\n",	/* reg: RETU1(con) */
/* 522 */	"\tw mov %c,=%0\n",	/* reg: RETI2(con) */
/* 523 */	"\tw mov %c,=%0\n",	/* reg: RETU2(con) */
/* 524 */	"\td mov %c,=%0\n",	/* reg: RETI4(con) */
/* 525 */	"\td mov %c,=%0\n",	/* reg: RETU4(con) */
/* 526 */	"\td mov %c,=%0\n",	/* reg: RETP4(con) */
/* 527 */	"\tq mov %c,=%0\n",	/* reg: RETI8(con) */
/* 528 */	"\tq mov %c,=%0\n",	/* reg: RETU8(con) */
/* 529 */	"\td mov $14,%0\n",	/* stmt: RETI4(ad) */
/* 530 */	"\td mov $14,%0\n",	/* stmt: RETU4(ad) */
/* 531 */	"\td mov $14,%0\n",	/* stmt: RETP4(ad) */
/* 532 */	"*\tb mov $14,%0\n",	/* stmt: RETI1(reg) */
/* 533 */	"*\tb mov $14,%0\n",	/* stmt: RETU1(reg) */
/* 534 */	"*\tw mov $14,%0\n",	/* stmt: RETI2(reg) */
/* 535 */	"*\tw mov $14,%0\n",	/* stmt: RETU2(reg) */
/* 536 */	"*\td mov $14,%0\n",	/* stmt: RETI4(reg) */
/* 537 */	"*\td mov $14,%0\n",	/* stmt: RETU4(reg) */
/* 538 */	"*\td mov $14,%0\n",	/* stmt: RETP4(reg) */
/* 539 */	"*\tq mov $14,%0\n",	/* stmt: RETI8(reg) */
/* 540 */	"*\tq mov $14,%0\n",	/* stmt: RETU8(reg) */
/* 541 */	"*\td mov $14,%0\n",	/* stmt: RETF4(reg) */
/* 542 */	"*\tq mov $14,%0\n",	/* stmt: RETF8(reg) */
/* 543 */	"\tb mov $14,=%0\n",	/* stmt: RETI1(con) */
/* 544 */	"\tb mov $14,=%0\n",	/* stmt: RETU1(con) */
/* 545 */	"\tw mov $14,=%0\n",	/* stmt: RETI2(con) */
/* 546 */	"\tw mov $14,=%0\n",	/* stmt: RETU2(con) */
/* 547 */	"\td mov $14,=%0\n",	/* stmt: RETI4(con) */
/* 548 */	"\td mov $14,=%0\n",	/* stmt: RETU4(con) */
/* 549 */	"\td mov $14,=%0\n",	/* stmt: RETP4(con) */
/* 550 */	"\tq mov $14,=%0\n",	/* stmt: RETI8(con) */
/* 551 */	"\tq mov $14,=%0\n",	/* stmt: RETU8(con) */
/* 552 */	"\td mov $14,=%0\n",	/* stmt: RETF4(con) */
/* 553 */	"\tq mov $14,=%0\n",	/* stmt: RETF8(con) */
/* 554 */	"\tjmp %0\n",	/* stmt: JUMPV(am) */
/* 555 */	"\tjmp %0\n",	/* stmt: JUMPV(adr) */
/* 556 */	"#:%a\n",	/* stmt: LABELV */
/* 557 */	"#\n",	/* reg: ASMRETU4 */
/* 558 */	"#\n",	/* reg: ASMRETU8 */
/* 559 */	"#\n",	/* stmt: ASMRETU4 */
/* 560 */	"#\n",	/* stmt: ASMRETU8 */
};

static char _isinstruction[] = {
/* 0 */	0,
/* 1 */	0,	/* %a */
/* 2 */	0,	/* %a */
/* 3 */	0,	/* %a */
/* 4 */	0,	/* %a */
/* 5 */	0,	/* %a */
/* 6 */	0,	/* %a */
/* 7 */	0,	/* %a */
/* 8 */	0,	/* %a */
/* 9 */	0,	/* %a */
/* 10 */	0,	/* %a */
/* 11 */	0,	/* %a */
/* 12 */	0,	/* %a */
/* 13 */	0,	/* %a */
/* 14 */	0,	/* %a */
/* 15 */	0,	/* %a */
/* 16 */	0,	/* 0 */
/* 17 */	0,	/* 0 */
/* 18 */	0,	/* 0 */
/* 19 */	0,	/* 0 */
/* 20 */	0,	/* 0 */
/* 21 */	0,	/* 0 */
/* 22 */	0,	/* 0 */
/* 23 */	0,	/* 0 */
/* 24 */	0,	/* 1 */
/* 25 */	0,	/* 1 */
/* 26 */	0,	/* 1 */
/* 27 */	0,	/* 1 */
/* 28 */	0,	/* 1 */
/* 29 */	0,	/* 1 */
/* 30 */	0,	/* 1 */
/* 31 */	0,	/* 1 */
/* 32 */	0,	/* 2 */
/* 33 */	0,	/* 2 */
/* 34 */	0,	/* 4 */
/* 35 */	0,	/* 4 */
/* 36 */	0,	/* 8 */
/* 37 */	0,	/* 8 */
/* 38 */	0,	/* 16 */
/* 39 */	0,	/* 16 */
/* 40 */	0,	/* 32 */
/* 41 */	0,	/* 32 */
/* 42 */	0,	/* 64 */
/* 43 */	0,	/* 64 */
/* 44 */	0,	/* 128 */
/* 45 */	0,	/* 128 */
/* 46 */	0,	/* 256 */
/* 47 */	0,	/* 256 */
/* 48 */	1,	/* #\n */
/* 49 */	1,	/* #\n */
/* 50 */	1,	/* #\n */
/* 51 */	1,	/* #\n */
/* 52 */	1,	/* #\n */
/* 53 */	1,	/* #\n */
/* 54 */	1,	/* #\n */
/* 55 */	1,	/* #\n */
/* 56 */	1,	/* #\n */
/* 57 */	1,	/* #\n */
/* 58 */	1,	/* #\n */
/* 59 */	1,	/* #\n */
/* 60 */	1,	/* #\n */
/* 61 */	1,	/* #\n */
/* 62 */	1,	/* #\n */
/* 63 */	1,	/* #\n */
/* 64 */	1,	/* #\n */
/* 65 */	1,	/* #\n */
/* 66 */	1,	/* #\n */
/* 67 */	1,	/* #\n */
/* 68 */	1,	/* \td mov %c,%0\n */
/* 69 */	1,	/* \td lea %c,%0\n */
/* 70 */	1,	/* \td lea %c,%0\n */
/* 71 */	1,	/* #\n */
/* 72 */	1,	/* \tb mov %c,=%0\n */
/* 73 */	1,	/* \tw mov %c,=%0\n */
/* 74 */	1,	/* \td mov %c,=%0\n */
/* 75 */	1,	/* \tq mov %c,=%0\n */
/* 76 */	0,	/*  */
/* 77 */	0,	/* =%a */
/* 78 */	0,	/* %0 */
/* 79 */	0,	/* [%a] */
/* 80 */	0,	/* [FP][%a] */
/* 81 */	0,	/* [FP][%a] */
/* 82 */	0,	/* [%0] */
/* 83 */	0,	/* %0 */
/* 84 */	0,	/* %0 */
/* 85 */	0,	/* %0 */
/* 86 */	0,	/* %0 */
/* 87 */	0,	/* %0 */
/* 88 */	0,	/* %0 */
/* 89 */	0,	/* %0 */
/* 90 */	0,	/* %0 */
/* 91 */	0,	/* %0 */
/* 92 */	0,	/* %0 */
/* 93 */	0,	/* %0 */
/* 94 */	0,	/* %0 */
/* 95 */	0,	/* %0 */
/* 96 */	0,	/* %0 */
/* 97 */	0,	/* %0 */
/* 98 */	0,	/* %0 */
/* 99 */	0,	/* %0 */
/* 100 */	0,	/* %0 */
/* 101 */	0,	/* %0 */
/* 102 */	0,	/* %0 */
/* 103 */	0,	/* %0 */
/* 104 */	0,	/* %0 */
/* 105 */	0,	/* %0 */
/* 106 */	0,	/* %0 */
/* 107 */	0,	/* %0 */
/* 108 */	0,	/* %0 */
/* 109 */	0,	/* %0 */
/* 110 */	0,	/* %0 */
/* 111 */	0,	/* %0 */
/* 112 */	0,	/* %0 */
/* 113 */	0,	/* %0 */
/* 114 */	0,	/* %0 */
/* 115 */	0,	/* %0 */
/* 116 */	1,	/* \tb mov %c,%0\n */
/* 117 */	1,	/* \tw mov %c,%0\n */
/* 118 */	1,	/* \td mov %c,%0\n */
/* 119 */	1,	/* \tq mov %c,%0\n */
/* 120 */	1,	/* \td lea %c,%0\n */
/* 121 */	0,	/* %0 */
/* 122 */	0,	/* %0 */
/* 123 */	0,	/* %a */
/* 124 */	0,	/* [FP][%a] */
/* 125 */	0,	/* [FP][%a] */
/* 126 */	0,	/* %0 */
/* 127 */	0,	/* =%a */
/* 128 */	0,	/* %0 */
/* 129 */	0,	/* =%a */
/* 130 */	0,	/* [FP][%a] */
/* 131 */	0,	/* [FP][%a] */
/* 132 */	1,	/* # write register\n */
/* 133 */	1,	/* # write register\n */
/* 134 */	1,	/* # write register\n */
/* 135 */	1,	/* # write register\n */
/* 136 */	1,	/* # write register\n */
/* 137 */	1,	/* # write register\n */
/* 138 */	1,	/* # write register\n */
/* 139 */	1,	/* # write register\n */
/* 140 */	1,	/* # write register\n */
/* 141 */	1,	/* # write register\n */
/* 142 */	1,	/* # write register\n */
/* 143 */	1,	/* \td mov %c,%1\n */
/* 144 */	1,	/* \tb mov %0,%1\n */
/* 145 */	1,	/* \tb mov %0,%1\n */
/* 146 */	1,	/* \tw mov %0,%1\n */
/* 147 */	1,	/* \tw mov %0,%1\n */
/* 148 */	1,	/* \td mov %0,%1\n */
/* 149 */	1,	/* \td mov %0,%1\n */
/* 150 */	1,	/* \tw mov [%a],%0\n */
/* 151 */	1,	/* \tw mov %a,%0\n */
/* 152 */	1,	/* \tw mov %0,%1\n */
/* 153 */	1,	/* \tw mov %0,%1\n */
/* 154 */	1,	/* \tw mov %0,%1\n */
/* 155 */	1,	/* \tw mov %0,%1\n */
/* 156 */	1,	/* \td mov %0,%1\n */
/* 157 */	1,	/* \td mov %0,=%1\n */
/* 158 */	1,	/* \td mov %0,%1\n */
/* 159 */	1,	/* \td lea %0,%1\n */
/* 160 */	1,	/* \tq mov %0,%1\n */
/* 161 */	1,	/* \tq mov %0,%1\n */
/* 162 */	1,	/* \td mov %0,%1\n */
/* 163 */	1,	/* \tq mov %0,%1\n */
/* 164 */	1,	/* \tb add %0,%2\n */
/* 165 */	1,	/* \tb add %0,%2\n */
/* 166 */	1,	/* \tw add %0,%2\n */
/* 167 */	1,	/* \tw add %0,%2\n */
/* 168 */	1,	/* \td add %0,%2\n */
/* 169 */	1,	/* \td add %0,%2\n */
/* 170 */	1,	/* \tq add %0,%2\n */
/* 171 */	1,	/* \tq add %0,%2\n */
/* 172 */	1,	/* \td sub %0,%2\n */
/* 173 */	1,	/* \td sub %0,%2\n */
/* 174 */	1,	/* \td rshi %0,%2\n */
/* 175 */	1,	/* \td rshu %0,%2\n */
/* 176 */	1,	/* \td lsh %0,%2\n */
/* 177 */	1,	/* \td lsh %0,%2\n */
/* 178 */	1,	/* \td and %0,%2\n */
/* 179 */	1,	/* \td and %0,%2\n */
/* 180 */	1,	/* \td or %0,%2\n */
/* 181 */	1,	/* \td or %0,%2\n */
/* 182 */	1,	/* \td xor %0,%2\n */
/* 183 */	1,	/* \td xor %0,%2\n */
/* 184 */	1,	/* \td mov [0xB8],=%a\n\tb memmov %0,%1\n */
/* 185 */	1,	/* \tpumem =%a,%0\n */
/* 186 */	1,	/* ?\tb mov %c,%0\n\tb add  %c,%1\n */
/* 187 */	1,	/* ?\tb mov %c,%0\n\tb add  %c,%1\n */
/* 188 */	1,	/* ?\tw mov %c,%0\n\tw add  %c,%1\n */
/* 189 */	1,	/* ?\tw mov %c,%0\n\tw add  %c,%1\n */
/* 190 */	1,	/* ?\td mov %c,%0\n\td add  %c,%1\n */
/* 191 */	1,	/* ?\td mov %c,%0\n\td add  %c,%1\n */
/* 192 */	1,	/* ?\tq mov %c,%0\n\tq add  %c,%1\n */
/* 193 */	1,	/* ?\tq mov %c,%0\n\tq add  %c,%1\n */
/* 194 */	1,	/* ?\td mov %c,%0\n\td addf %c,%1\n */
/* 195 */	1,	/* ?\tq mov %c,%0\n\tq addf %c,%1\n */
/* 196 */	1,	/* ?\td mov %c,%0\n\td add  %c,%1\n */
/* 197 */	1,	/* ?\tb mov %c,%0\n\tb sub  %c,%1\n */
/* 198 */	1,	/* ?\tb mov %c,%0\n\tb sub  %c,%1\n */
/* 199 */	1,	/* ?\tw mov %c,%0\n\tw sub  %c,%1\n */
/* 200 */	1,	/* ?\tw mov %c,%0\n\tw sub  %c,%1\n */
/* 201 */	1,	/* ?\td mov %c,%0\n\td sub  %c,%1\n */
/* 202 */	1,	/* ?\td mov %c,%0\n\td sub  %c,%1\n */
/* 203 */	1,	/* ?\td mov %c,%0\n\td sub  %c,%1\n */
/* 204 */	1,	/* ?\tq mov %c,%0\n\tq sub  %c,%1\n */
/* 205 */	1,	/* ?\tq mov %c,%0\n\tq sub  %c,%1\n */
/* 206 */	1,	/* ?\td mov %c,%0\n\td subf %c,%1\n */
/* 207 */	1,	/* ?\tq mov %c,%0\n\tq subf %c,%1\n */
/* 208 */	1,	/* \td mov %c,=%0\n\td mul %c,%1\n */
/* 209 */	1,	/* ?\tb mov %c,%0\n\tb mul  %c,%1\n */
/* 210 */	1,	/* ?\tb mov %c,%0\n\tb mul  %c,%1\n */
/* 211 */	1,	/* ?\tw mov %c,%0\n\tw mul  %c,%1\n */
/* 212 */	1,	/* ?\tw mov %c,%0\n\tw mul  %c,%1\n */
/* 213 */	1,	/* ?\td mov %c,%0\n\td mul  %c,%1\n */
/* 214 */	1,	/* ?\td mov %c,%0\n\td mul  %c,%1\n */
/* 215 */	1,	/* ?\tq mov %c,%0\n\tq mul  %c,%1\n */
/* 216 */	1,	/* ?\tq mov %c,%0\n\tq mul  %c,%1\n */
/* 217 */	1,	/* ?\td mov %c,%0\n\td mulf %c,%1\n */
/* 218 */	1,	/* ?\tq mov %c,%0\n\tq mulf %c,%1\n */
/* 219 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=1\n */
/* 220 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=2\n */
/* 221 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=3\n */
/* 222 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=4\n */
/* 223 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=5\n */
/* 224 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=6\n */
/* 225 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=7\n */
/* 226 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=8\n */
/* 227 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=1\n */
/* 228 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=2\n */
/* 229 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=3\n */
/* 230 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=4\n */
/* 231 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=5\n */
/* 232 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=6\n */
/* 233 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=7\n */
/* 234 */	1,	/* ?\td mov %c,%0\n\td lsh %c,=8\n */
/* 235 */	1,	/* ?\tb mov %c,%0\n\tb divi %c,%1\n */
/* 236 */	1,	/* ?\tb mov %c,%0\n\tb divu %c,%1\n */
/* 237 */	1,	/* ?\tw mov %c,%0\n\tw divi %c,%1\n */
/* 238 */	1,	/* ?\tw mov %c,%0\n\tw divu %c,%1\n */
/* 239 */	1,	/* ?\td mov %c,%0\n\td divi %c,%1\n */
/* 240 */	1,	/* ?\td mov %c,%0\n\td divu %c,%1\n */
/* 241 */	1,	/* ?\tq mov %c,%0\n\tq divi %c,%1\n */
/* 242 */	1,	/* ?\tq mov %c,%0\n\tq divu %c,%1\n */
/* 243 */	1,	/* ?\td mov %c,%0\n\td divf %c,%1\n */
/* 244 */	1,	/* ?\tq mov %c,%0\n\tq divf %c,%1\n */
/* 245 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=1\n */
/* 246 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=2\n */
/* 247 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=3\n */
/* 248 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=4\n */
/* 249 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=5\n */
/* 250 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=6\n */
/* 251 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=7\n */
/* 252 */	1,	/* ?\td mov %c,%0\n\td rshi %c,=8\n */
/* 253 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=1\n */
/* 254 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=2\n */
/* 255 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=3\n */
/* 256 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=4\n */
/* 257 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=5\n */
/* 258 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=6\n */
/* 259 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=7\n */
/* 260 */	1,	/* ?\td mov %c,%0\n\td rshu %c,=8\n */
/* 261 */	1,	/* ?\tb mov %c,%0\n\tb modi %c,%1\n */
/* 262 */	1,	/* ?\tb mov %c,%0\n\tb modu %c,%1\n */
/* 263 */	1,	/* ?\tw mov %c,%0\n\tw modi %c,%1\n */
/* 264 */	1,	/* ?\tw mov %c,%0\n\tw modu %c,%1\n */
/* 265 */	1,	/* ?\td mov %c,%0\n\td modi %c,%1\n */
/* 266 */	1,	/* ?\td mov %c,%0\n\td modu %c,%1\n */
/* 267 */	1,	/* ?\tq mov %c,%0\n\tq modi %c,%1\n */
/* 268 */	1,	/* ?\tq mov %c,%0\n\tq modu %c,%1\n */
/* 269 */	1,	/* ?\tb mov %c,%0\n\tb rshi %c,%1\n */
/* 270 */	1,	/* ?\tb mov %c,%0\n\tb rshu %c,%1\n */
/* 271 */	1,	/* ?\tw mov %c,%0\n\tw rshi %c,%1\n */
/* 272 */	1,	/* ?\tw mov %c,%0\n\tw rshu %c,%1\n */
/* 273 */	1,	/* ?\td mov %c,%0\n\td rshi %c,%1\n */
/* 274 */	1,	/* ?\td mov %c,%0\n\td rshu %c,%1\n */
/* 275 */	1,	/* ?\tq mov %c,%0\n\tq rshi %c,%1\n */
/* 276 */	1,	/* ?\tq mov %c,%0\n\tq rshu %c,%1\n */
/* 277 */	1,	/* ?\tb mov %c,%0\n\tb lsh %c,%1\n */
/* 278 */	1,	/* ?\tb mov %c,%0\n\tb lsh %c,%1\n */
/* 279 */	1,	/* ?\tw mov %c,%0\n\tw lsh %c,%1\n */
/* 280 */	1,	/* ?\tw mov %c,%0\n\tw lsh %c,%1\n */
/* 281 */	1,	/* ?\td mov %c,%0\n\td lsh %c,%1\n */
/* 282 */	1,	/* ?\td mov %c,%0\n\td lsh %c,%1\n */
/* 283 */	1,	/* ?\tq mov %c,%0\n\tq lsh %c,%1\n */
/* 284 */	1,	/* ?\tq mov %c,%0\n\tq lsh %c,%1\n */
/* 285 */	1,	/* ?\tb mov %c,%0\n\tb and %c,%1\n */
/* 286 */	1,	/* ?\tb mov %c,%0\n\tb and %c,%1\n */
/* 287 */	1,	/* ?\tw mov %c,%0\n\tw and %c,%1\n */
/* 288 */	1,	/* ?\tw mov %c,%0\n\tw and %c,%1\n */
/* 289 */	1,	/* ?\td mov %c,%0\n\td and %c,%1\n */
/* 290 */	1,	/* ?\td mov %c,%0\n\td and %c,%1\n */
/* 291 */	1,	/* ?\tq mov %c,%0\n\tq and %c,%1\n */
/* 292 */	1,	/* ?\tq mov %c,%0\n\tq and %c,%1\n */
/* 293 */	1,	/* ?\tb mov %c,%0\n\tb or %c,%1\n */
/* 294 */	1,	/* ?\tb mov %c,%0\n\tb or %c,%1\n */
/* 295 */	1,	/* ?\tw mov %c,%0\n\tw or %c,%1\n */
/* 296 */	1,	/* ?\tw mov %c,%0\n\tw or %c,%1\n */
/* 297 */	1,	/* ?\td mov %c,%0\n\td or %c,%1\n */
/* 298 */	1,	/* ?\td mov %c,%0\n\td or %c,%1\n */
/* 299 */	1,	/* ?\tq mov %c,%0\n\tq or %c,%1\n */
/* 300 */	1,	/* ?\tq mov %c,%0\n\tq or %c,%1\n */
/* 301 */	1,	/* ?\tb mov %c,%0\n\tb xor %c,%1\n */
/* 302 */	1,	/* ?\tb mov %c,%0\n\tb xor %c,%1\n */
/* 303 */	1,	/* ?\tw mov %c,%0\n\tw xor %c,%1\n */
/* 304 */	1,	/* ?\tw mov %c,%0\n\tw xor %c,%1\n */
/* 305 */	1,	/* ?\td mov %c,%0\n\td xor %c,%1\n */
/* 306 */	1,	/* ?\td mov %c,%0\n\td xor %c,%1\n */
/* 307 */	1,	/* ?\tq mov %c,%0\n\tq xor %c,%1\n */
/* 308 */	1,	/* ?\tq mov %c,%0\n\tq xor %c,%1\n */
/* 309 */	1,	/* ?\tb mov %c,%0\n\tb not %c\n */
/* 310 */	1,	/* ?\tb mov %c,%0\n\tb not %c\n */
/* 311 */	1,	/* ?\tw mov %c,%0\n\tw not %c\n */
/* 312 */	1,	/* ?\tw mov %c,%0\n\tw not %c\n */
/* 313 */	1,	/* ?\td mov %c,%0\n\td not %c\n */
/* 314 */	1,	/* ?\td mov %c,%0\n\td not %c\n */
/* 315 */	1,	/* ?\tq mov %c,%0\n\tq not %c\n */
/* 316 */	1,	/* ?\tq mov %c,%0\n\tq not %c\n */
/* 317 */	1,	/* \tb not %0\n */
/* 318 */	1,	/* \tb not %0\n */
/* 319 */	1,	/* \tw not %0\n */
/* 320 */	1,	/* \tw not %0\n */
/* 321 */	1,	/* \td not %0\n */
/* 322 */	1,	/* \td not %0\n */
/* 323 */	1,	/* \tq not %0\n */
/* 324 */	1,	/* \tq not %0\n */
/* 325 */	1,	/* ?\tb mov %c,%0\n\tb neg  %c\n */
/* 326 */	1,	/* ?\tw mov %c,%0\n\tw neg  %c\n */
/* 327 */	1,	/* ?\td mov %c,%0\n\td neg  %c\n */
/* 328 */	1,	/* ?\tq mov %c,%0\n\tq neg  %c\n */
/* 329 */	1,	/* ?\td mov %c,%0\n\td negf %c\n */
/* 330 */	1,	/* ?\tq mov %c,%0\n\tq negf %c\n */
/* 331 */	1,	/* \tb tljeq %0,=%a\n */
/* 332 */	1,	/* \tb tljeq %0,=%a\n */
/* 333 */	1,	/* \tw tljeq %0,=%a\n */
/* 334 */	1,	/* \tw tljeq %0,=%a\n */
/* 335 */	1,	/* \td tljeq %0,=%a\n */
/* 336 */	1,	/* \td tljeq %0,=%a\n */
/* 337 */	1,	/* \tq tljeq %0,=%a\n */
/* 338 */	1,	/* \tq tljeq %0,=%a\n */
/* 339 */	1,	/* \td tljeq %0,=%a\n */
/* 340 */	1,	/* \tq tljeq %0,=%a\n */
/* 341 */	1,	/* \tb tljev %0,=%a\n */
/* 342 */	1,	/* \tb tljev %0,=%a\n */
/* 343 */	1,	/* \tw tljev %0,=%a\n */
/* 344 */	1,	/* \tw tljev %0,=%a\n */
/* 345 */	1,	/* \td tljev %0,=%a\n */
/* 346 */	1,	/* \td tljev %0,=%a\n */
/* 347 */	1,	/* \tq tljev %0,=%a\n */
/* 348 */	1,	/* \tq tljev %0,=%a\n */
/* 349 */	1,	/* \tb tljod %0,=%a\n */
/* 350 */	1,	/* \tb tljod %0,=%a\n */
/* 351 */	1,	/* \tw tljod %0,=%a\n */
/* 352 */	1,	/* \tw tljod %0,=%a\n */
/* 353 */	1,	/* \td tljod %0,=%a\n */
/* 354 */	1,	/* \td tljod %0,=%a\n */
/* 355 */	1,	/* \tq tljod %0,=%a\n */
/* 356 */	1,	/* \tq tljod %0,=%a\n */
/* 357 */	1,	/* \tb cmp  %0,%1\n\tljeq =%a\n */
/* 358 */	1,	/* \tb cmp  %0,%1\n\tljeq =%a\n */
/* 359 */	1,	/* \tw cmp  %0,%1\n\tljeq =%a\n */
/* 360 */	1,	/* \tw cmp  %0,%1\n\tljeq =%a\n */
/* 361 */	1,	/* \td cmp  %0,%1\n\tljeq =%a\n */
/* 362 */	1,	/* \td cmp  %0,%1\n\tljeq =%a\n */
/* 363 */	1,	/* \tq cmp  %0,%1\n\tljeq =%a\n */
/* 364 */	1,	/* \tq cmp  %0,%1\n\tljeq =%a\n */
/* 365 */	1,	/* \td cmpf %0,%1\n\tljeq =%a\n */
/* 366 */	1,	/* \tq cmpf %0,%1\n\tljeq =%a\n */
/* 367 */	1,	/* \tb tljne %0,=%a\n */
/* 368 */	1,	/* \tb tljne %0,=%a\n */
/* 369 */	1,	/* \tw tljne %0,=%a\n */
/* 370 */	1,	/* \tw tljne %0,=%a\n */
/* 371 */	1,	/* \td tljne %0,=%a\n */
/* 372 */	1,	/* \td tljne %0,=%a\n */
/* 373 */	1,	/* \tq tljne %0,=%a\n */
/* 374 */	1,	/* \tq tljne %0,=%a\n */
/* 375 */	1,	/* \td tljne %0,=%a\n */
/* 376 */	1,	/* \tq tljne %0,=%a\n */
/* 377 */	1,	/* \tb tljod %0,=%a\n */
/* 378 */	1,	/* \tb tljod %0,=%a\n */
/* 379 */	1,	/* \tw tljod %0,=%a\n */
/* 380 */	1,	/* \tw tljod %0,=%a\n */
/* 381 */	1,	/* \td tljod %0,=%a\n */
/* 382 */	1,	/* \td tljod %0,=%a\n */
/* 383 */	1,	/* \tq tljod %0,=%a\n */
/* 384 */	1,	/* \tq tljod %0,=%a\n */
/* 385 */	1,	/* \tb tljev %0,=%a\n */
/* 386 */	1,	/* \tb tljev %0,=%a\n */
/* 387 */	1,	/* \tw tljev %0,=%a\n */
/* 388 */	1,	/* \tw tljev %0,=%a\n */
/* 389 */	1,	/* \td tljev %0,=%a\n */
/* 390 */	1,	/* \td tljev %0,=%a\n */
/* 391 */	1,	/* \tq tljev %0,=%a\n */
/* 392 */	1,	/* \tq tljev %0,=%a\n */
/* 393 */	1,	/* \tb cmp  %0,%1\n\tljne =%a\n */
/* 394 */	1,	/* \tb cmp  %0,%1\n\tljne =%a\n */
/* 395 */	1,	/* \tw cmp  %0,%1\n\tljne =%a\n */
/* 396 */	1,	/* \tw cmp  %0,%1\n\tljne =%a\n */
/* 397 */	1,	/* \td cmp  %0,%1\n\tljne =%a\n */
/* 398 */	1,	/* \td cmp  %0,%1\n\tljne =%a\n */
/* 399 */	1,	/* \tq cmp  %0,%1\n\tljne =%a\n */
/* 400 */	1,	/* \tq cmp  %0,%1\n\tljne =%a\n */
/* 401 */	1,	/* \td cmpf %0,%1\n\tljne =%a\n */
/* 402 */	1,	/* \tq cmpf %0,%1\n\tljne =%a\n */
/* 403 */	1,	/* \tb cmp  %0,%1\n\tljge =%a\n */
/* 404 */	1,	/* \tb cmp  %0,%1\n\tljge =%a\n */
/* 405 */	1,	/* \tw cmp  %0,%1\n\tljge =%a\n */
/* 406 */	1,	/* \tw cmp  %0,%1\n\tljge =%a\n */
/* 407 */	1,	/* \td cmp  %0,%1\n\tljge =%a\n */
/* 408 */	1,	/* \td cmp  %0,%1\n\tljge =%a\n */
/* 409 */	1,	/* \tq cmp  %0,%1\n\tljge =%a\n */
/* 410 */	1,	/* \tq cmp  %0,%1\n\tljge =%a\n */
/* 411 */	1,	/* \td cmpf %0,%1\n\tljge =%a\n */
/* 412 */	1,	/* \tq cmpf %0,%1\n\tljge =%a\n */
/* 413 */	1,	/* \tb cmp  %0,%1\n\tljgt =%a\n */
/* 414 */	1,	/* \tb cmp  %0,%1\n\tljgt =%a\n */
/* 415 */	1,	/* \tw cmp  %0,%1\n\tljgt =%a\n */
/* 416 */	1,	/* \tw cmp  %0,%1\n\tljgt =%a\n */
/* 417 */	1,	/* \td cmp  %0,%1\n\tljgt =%a\n */
/* 418 */	1,	/* \td cmp  %0,%1\n\tljgt =%a\n */
/* 419 */	1,	/* \tq cmp  %0,%1\n\tljgt =%a\n */
/* 420 */	1,	/* \tq cmp  %0,%1\n\tljgt =%a\n */
/* 421 */	1,	/* \td cmpf %0,%1\n\tljgt =%a\n */
/* 422 */	1,	/* \tq cmpf %0,%1\n\tljgt =%a\n */
/* 423 */	1,	/* \tb cmp  %0,%1\n\tljle =%a\n */
/* 424 */	1,	/* \tb cmp  %0,%1\n\tljle =%a\n */
/* 425 */	1,	/* \tw cmp  %0,%1\n\tljle =%a\n */
/* 426 */	1,	/* \tw cmp  %0,%1\n\tljle =%a\n */
/* 427 */	1,	/* \td cmp  %0,%1\n\tljle =%a\n */
/* 428 */	1,	/* \td cmp  %0,%1\n\tljle =%a\n */
/* 429 */	1,	/* \tq cmp  %0,%1\n\tljle =%a\n */
/* 430 */	1,	/* \tq cmp  %0,%1\n\tljle =%a\n */
/* 431 */	1,	/* \td cmpf %0,%1\n\tljle =%a\n */
/* 432 */	1,	/* \tq cmpf %0,%1\n\tljle =%a\n */
/* 433 */	1,	/* \tb cmp  %0,%1\n\tljlt =%a\n */
/* 434 */	1,	/* \tb cmp  %0,%1\n\tljlt =%a\n */
/* 435 */	1,	/* \tw cmp  %0,%1\n\tljlt =%a\n */
/* 436 */	1,	/* \tw cmp  %0,%1\n\tljlt =%a\n */
/* 437 */	1,	/* \td cmp  %0,%1\n\tljlt =%a\n */
/* 438 */	1,	/* \td cmp  %0,%1\n\tljlt =%a\n */
/* 439 */	1,	/* \tq cmp  %0,%1\n\tljlt =%a\n */
/* 440 */	1,	/* \tq cmp  %0,%1\n\tljlt =%a\n */
/* 441 */	1,	/* \td cmpf %0,%1\n\tljlt =%a\n */
/* 442 */	1,	/* \tq cmpf %0,%1\n\tljlt =%a\n */
/* 443 */	1,	/* \tb cmp  %0,%1\n\tljlt =%a\n */
/* 444 */	1,	/* \tb cmp  %0,%1\n\tljlt =%a\n */
/* 445 */	1,	/* \tw cmp  %0,%1\n\tljlt =%a\n */
/* 446 */	1,	/* \tw cmp  %0,%1\n\tljlt =%a\n */
/* 447 */	1,	/* \td cmp  %0,%1\n\tljlt =%a\n */
/* 448 */	1,	/* \td cmp  %0,%1\n\tljlt =%a\n */
/* 449 */	1,	/* \tq cmp  %0,%1\n\tljlt =%a\n */
/* 450 */	1,	/* \tq cmp  %0,%1\n\tljlt =%a\n */
/* 451 */	1,	/* \td cmpf %0,%1\n\tljlt =%a\n */
/* 452 */	1,	/* \tq cmpf %0,%1\n\tljlt =%a\n */
/* 453 */	1,	/* \td mov %c,%0\n */
/* 454 */	1,	/* \td mov %c,%0\n */
/* 455 */	1,	/* #\n */
/* 456 */	1,	/* #\n */
/* 457 */	1,	/* #\n */
/* 458 */	1,	/* #\n */
/* 459 */	1,	/* #\n */
/* 460 */	1,	/* #\n */
/* 461 */	1,	/* #\n */
/* 462 */	1,	/* #\n */
/* 463 */	1,	/* #\td mov %c,%0\n\tcvsf %c\n */
/* 464 */	1,	/* #\tq mov %c,%0\n\tcvsd %c\n */
/* 465 */	1,	/* #\n */
/* 466 */	1,	/* #\n */
/* 467 */	1,	/* #\n */
/* 468 */	1,	/* #\n */
/* 469 */	1,	/* #\n */
/* 470 */	1,	/* #\n */
/* 471 */	1,	/* #\n */
/* 472 */	1,	/* #\n */
/* 473 */	1,	/* #\td mov %c,%0\n\tcvuf %c\n */
/* 474 */	1,	/* #\tq mov %c,%0\n\tcvud %c\n */
/* 475 */	1,	/* \tq mov %c,%0\n\tcvdf %c\n */
/* 476 */	1,	/* ?\td mov %c,%0\n\tcvfd %c\n */
/* 477 */	1,	/* #\tw mov %c,%0\n\tcvfs %c\n */
/* 478 */	1,	/* #\td mov %c,%0\n\tcvfs %c\n */
/* 479 */	1,	/* #\tq mov %c,%0\n\tcvfs %c\n */
/* 480 */	1,	/* \td mov %c,%0\n\tadd %c,%1\n\tmov %c,[%c]\n\tcvfd %c\n */
/* 481 */	1,	/* \tw push %0\n */
/* 482 */	1,	/* \tw push %0\n */
/* 483 */	1,	/* \td push %0\n */
/* 484 */	1,	/* \td pea %0\n */
/* 485 */	1,	/* \td push %0\n */
/* 486 */	1,	/* \td push %0\n */
/* 487 */	1,	/* \td push %0\n */
/* 488 */	1,	/* \tq push %0\n */
/* 489 */	1,	/* \tq push %0\n */
/* 490 */	1,	/* \tw push %0\n */
/* 491 */	1,	/* \tw push %0\n */
/* 492 */	1,	/* \td push %0\n */
/* 493 */	1,	/* \td push %0\n */
/* 494 */	1,	/* \tq push %0\n */
/* 495 */	1,	/* \tq push %0\n */
/* 496 */	1,	/* \td push %0\n */
/* 497 */	1,	/* \tq push %0\n */
/* 498 */	1,	/* \td push %0\n */
/* 499 */	1,	/* \td pea  %0\n */
/* 500 */	1,	/* \td push %0\n */
/* 501 */	1,	/* \tjsrfr %a,%0\n */
/* 502 */	1,	/* \tjsrfr %a,%0\n */
/* 503 */	1,	/* \tjsrfr %a,%0\n */
/* 504 */	1,	/* \tjsrfr %a,%0\n */
/* 505 */	1,	/* \tjsrfr %a,%0\n */
/* 506 */	1,	/* \tjsrfr %a,%0\n */
/* 507 */	1,	/* \tjsrfr %a,%0\n */
/* 508 */	1,	/* \tjsrfr %a,%0\n */
/* 509 */	1,	/* \tjsrfr %a,%0\n */
/* 510 */	1,	/* \tjsrfr %a,%0\n */
/* 511 */	1,	/* ?\tb mov %c,%0\n */
/* 512 */	1,	/* ?\tb mov %c,%0\n */
/* 513 */	1,	/* ?\tw mov %c,%0\n */
/* 514 */	1,	/* ?\tw mov %c,%0\n */
/* 515 */	1,	/* ?\td mov %c,%0\n */
/* 516 */	1,	/* ?\td mov %c,%0\n */
/* 517 */	1,	/* ?\td mov %c,%0\n */
/* 518 */	1,	/* ?\tq mov %c,%0\n */
/* 519 */	1,	/* ?\tq mov %c,%0\n */
/* 520 */	1,	/* \tb mov %c,=%0\n */
/* 521 */	1,	/* \tb mov %c,=%0\n */
/* 522 */	1,	/* \tw mov %c,=%0\n */
/* 523 */	1,	/* \tw mov %c,=%0\n */
/* 524 */	1,	/* \td mov %c,=%0\n */
/* 525 */	1,	/* \td mov %c,=%0\n */
/* 526 */	1,	/* \td mov %c,=%0\n */
/* 527 */	1,	/* \tq mov %c,=%0\n */
/* 528 */	1,	/* \tq mov %c,=%0\n */
/* 529 */	1,	/* \td mov $14,%0\n */
/* 530 */	1,	/* \td mov $14,%0\n */
/* 531 */	1,	/* \td mov $14,%0\n */
/* 532 */	1,	/* *\tb mov $14,%0\n */
/* 533 */	1,	/* *\tb mov $14,%0\n */
/* 534 */	1,	/* *\tw mov $14,%0\n */
/* 535 */	1,	/* *\tw mov $14,%0\n */
/* 536 */	1,	/* *\td mov $14,%0\n */
/* 537 */	1,	/* *\td mov $14,%0\n */
/* 538 */	1,	/* *\td mov $14,%0\n */
/* 539 */	1,	/* *\tq mov $14,%0\n */
/* 540 */	1,	/* *\tq mov $14,%0\n */
/* 541 */	1,	/* *\td mov $14,%0\n */
/* 542 */	1,	/* *\tq mov $14,%0\n */
/* 543 */	1,	/* \tb mov $14,=%0\n */
/* 544 */	1,	/* \tb mov $14,=%0\n */
/* 545 */	1,	/* \tw mov $14,=%0\n */
/* 546 */	1,	/* \tw mov $14,=%0\n */
/* 547 */	1,	/* \td mov $14,=%0\n */
/* 548 */	1,	/* \td mov $14,=%0\n */
/* 549 */	1,	/* \td mov $14,=%0\n */
/* 550 */	1,	/* \tq mov $14,=%0\n */
/* 551 */	1,	/* \tq mov $14,=%0\n */
/* 552 */	1,	/* \td mov $14,=%0\n */
/* 553 */	1,	/* \tq mov $14,=%0\n */
/* 554 */	1,	/* \tjmp %0\n */
/* 555 */	1,	/* \tjmp %0\n */
/* 556 */	1,	/* #:%a\n */
/* 557 */	1,	/* #\n */
/* 558 */	1,	/* #\n */
/* 559 */	1,	/* #\n */
/* 560 */	1,	/* #\n */
};

static char *_string[] = {
/* 0 */	0,
/* 1 */	"con1: CNSTI1",
/* 2 */	"con1: CNSTU1",
/* 3 */	"con2: CNSTI2",
/* 4 */	"con2: CNSTU2",
/* 5 */	"con4: CNSTI4",
/* 6 */	"con4: CNSTU4",
/* 7 */	"con4: CNSTP4",
/* 8 */	"con8: CNSTI8",
/* 9 */	"con8: CNSTU8",
/* 10 */	"con4: CNSTF4",
/* 11 */	"con8: CNSTF8",
/* 12 */	"con: con1",
/* 13 */	"con: con2",
/* 14 */	"con: con4",
/* 15 */	"con: con8",
/* 16 */	"conv0: CNSTI1",
/* 17 */	"conv0: CNSTI2",
/* 18 */	"conv0: CNSTI4",
/* 19 */	"conv0: CNSTI8",
/* 20 */	"conv0: CNSTU1",
/* 21 */	"conv0: CNSTU2",
/* 22 */	"conv0: CNSTU4",
/* 23 */	"conv0: CNSTU8",
/* 24 */	"conv1: CNSTI1",
/* 25 */	"conv1: CNSTI2",
/* 26 */	"conv1: CNSTI4",
/* 27 */	"conv1: CNSTI8",
/* 28 */	"conv1: CNSTU1",
/* 29 */	"conv1: CNSTU2",
/* 30 */	"conv1: CNSTU4",
/* 31 */	"conv1: CNSTU8",
/* 32 */	"conv2: CNSTI4",
/* 33 */	"conv2: CNSTU4",
/* 34 */	"conv4: CNSTI4",
/* 35 */	"conv4: CNSTU4",
/* 36 */	"conv8: CNSTI4",
/* 37 */	"conv8: CNSTU4",
/* 38 */	"conv16: CNSTI4",
/* 39 */	"conv16: CNSTU4",
/* 40 */	"conv32: CNSTI4",
/* 41 */	"conv32: CNSTU4",
/* 42 */	"conv64: CNSTI4",
/* 43 */	"conv64: CNSTU4",
/* 44 */	"conv128: CNSTI4",
/* 45 */	"conv128: CNSTU4",
/* 46 */	"conv256: CNSTI4",
/* 47 */	"conv256: CNSTU4",
/* 48 */	"reg: INDIRI1(VREGP)",
/* 49 */	"reg: INDIRU1(VREGP)",
/* 50 */	"reg: INDIRI2(VREGP)",
/* 51 */	"reg: INDIRU2(VREGP)",
/* 52 */	"reg: INDIRI4(VREGP)",
/* 53 */	"reg: INDIRU4(VREGP)",
/* 54 */	"reg: INDIRP4(VREGP)",
/* 55 */	"reg: INDIRI8(VREGP)",
/* 56 */	"reg: INDIRU8(VREGP)",
/* 57 */	"reg: INDIRF4(VREGP)",
/* 58 */	"reg: INDIRF8(VREGP)",
/* 59 */	"reg: LOADI1(reg)",
/* 60 */	"reg: LOADU1(reg)",
/* 61 */	"reg: LOADI2(reg)",
/* 62 */	"reg: LOADU2(reg)",
/* 63 */	"reg: LOADI4(reg)",
/* 64 */	"reg: LOADU4(reg)",
/* 65 */	"reg: LOADP4(reg)",
/* 66 */	"reg: LOADI8(reg)",
/* 67 */	"reg: LOADU8(reg)",
/* 68 */	"reg: LOADI4(adr)",
/* 69 */	"reg: LOADU4(ad)",
/* 70 */	"reg: LOADU4(INDIRU8(ad))",
/* 71 */	"reg: LOADU2(LOADU4(reg))",
/* 72 */	"reg: con1",
/* 73 */	"reg: con2",
/* 74 */	"reg: con4",
/* 75 */	"reg: con8",
/* 76 */	"stmt: reg",
/* 77 */	"ca: ADDRGP4",
/* 78 */	"ca: reg",
/* 79 */	"ad: ADDRGP4",
/* 80 */	"ad: ADDRFP4",
/* 81 */	"ad: ADDRLP4",
/* 82 */	"adr: reg",
/* 83 */	"adr: ad",
/* 84 */	"iadr1: INDIRI1(adr)",
/* 85 */	"iadr1: INDIRU1(adr)",
/* 86 */	"iadr2: INDIRI2(adr)",
/* 87 */	"iadr2: INDIRU2(adr)",
/* 88 */	"iadr4: INDIRI4(adr)",
/* 89 */	"iadr4: INDIRU4(adr)",
/* 90 */	"iadr4: INDIRP4(adr)",
/* 91 */	"iadr8: INDIRI8(adr)",
/* 92 */	"iadr8: INDIRU8(adr)",
/* 93 */	"iadr4: INDIRF4(adr)",
/* 94 */	"iadr8: INDIRF8(adr)",
/* 95 */	"iadr: iadr1",
/* 96 */	"iadr: iadr2",
/* 97 */	"iadr: iadr4",
/* 98 */	"iadr: iadr8",
/* 99 */	"iadr: INDIRB(adr)",
/* 100 */	"eadr1: INDIRI1(ad)",
/* 101 */	"eadr1: INDIRU1(ad)",
/* 102 */	"eadr2: INDIRI2(ad)",
/* 103 */	"eadr2: INDIRU2(ad)",
/* 104 */	"eadr4: INDIRI4(ad)",
/* 105 */	"eadr4: INDIRU4(ad)",
/* 106 */	"eadr4: INDIRP4(ad)",
/* 107 */	"eadr8: INDIRI8(ad)",
/* 108 */	"eadr8: INDIRU8(ad)",
/* 109 */	"eadr4: INDIRF4(ad)",
/* 110 */	"eadr8: INDIRF8(ad)",
/* 111 */	"eadr: eadr1",
/* 112 */	"eadr: eadr2",
/* 113 */	"eadr: eadr4",
/* 114 */	"eadr: eadr8",
/* 115 */	"eadr: INDIRB(ad)",
/* 116 */	"reg: iadr1",
/* 117 */	"reg: iadr2",
/* 118 */	"reg: iadr4",
/* 119 */	"reg: iadr8",
/* 120 */	"reg: ad",
/* 121 */	"dm: reg",
/* 122 */	"dm: iadr",
/* 123 */	"dm: ADDRGP4",
/* 124 */	"dm: ADDRFP4",
/* 125 */	"dm: ADDRLP4",
/* 126 */	"am: reg",
/* 127 */	"am: con",
/* 128 */	"am: iadr",
/* 129 */	"am: ADDRGP4",
/* 130 */	"am: ADDRFP4",
/* 131 */	"am: ADDRLP4",
/* 132 */	"stmt: ASGNI1(VREGP,reg)",
/* 133 */	"stmt: ASGNU1(VREGP,reg)",
/* 134 */	"stmt: ASGNI2(VREGP,reg)",
/* 135 */	"stmt: ASGNU2(VREGP,reg)",
/* 136 */	"stmt: ASGNI4(VREGP,reg)",
/* 137 */	"stmt: ASGNU4(VREGP,reg)",
/* 138 */	"stmt: ASGNP4(VREGP,reg)",
/* 139 */	"stmt: ASGNF4(VREGP,reg)",
/* 140 */	"stmt: ASGNI8(VREGP,reg)",
/* 141 */	"stmt: ASGNU8(VREGP,reg)",
/* 142 */	"stmt: ASGNF8(VREGP,reg)",
/* 143 */	"stmt: ASGNP4(VREGP,ADDRGP4)",
/* 144 */	"stmt: ASGNI1(adr,am)",
/* 145 */	"stmt: ASGNU1(adr,am)",
/* 146 */	"stmt: ASGNI2(adr,am)",
/* 147 */	"stmt: ASGNU2(adr,am)",
/* 148 */	"stmt: ASGNI4(adr,am)",
/* 149 */	"stmt: ASGNU4(adr,am)",
/* 150 */	"stmt: ASGNU2(INDIRP4(VREGP),LOADU2(reg))",
/* 151 */	"reg: ASGNU2(INDIRP4(VREGP),LOADU2(reg))",
/* 152 */	"stmt: ASGNU1(adr,LOADU1(INDIRU4(ad)))",
/* 153 */	"stmt: ASGNU2(adr,LOADU2(INDIRU4(ad)))",
/* 154 */	"stmt: ASGNI1(adr,LOADI1(INDIRI4(ad)))",
/* 155 */	"stmt: ASGNI2(adr,LOADI2(INDIRI4(ad)))",
/* 156 */	"stmt: ASGNP4(adr,reg)",
/* 157 */	"stmt: ASGNP4(adr,con)",
/* 158 */	"stmt: ASGNP4(adr,iadr)",
/* 159 */	"stmt: ASGNP4(adr,ad)",
/* 160 */	"stmt: ASGNI8(adr,am)",
/* 161 */	"stmt: ASGNU8(adr,am)",
/* 162 */	"stmt: ASGNF4(adr,am)",
/* 163 */	"stmt: ASGNF8(adr,am)",
/* 164 */	"stmt: ASGNI1(adr,ADDI1(INDIRI1(adr),am))",
/* 165 */	"stmt: ASGNU1(adr,ADDU1(INDIRU1(adr),am))",
/* 166 */	"stmt: ASGNI2(adr,ADDI2(INDIRI2(adr),am))",
/* 167 */	"stmt: ASGNU2(adr,ADDU2(INDIRU2(adr),am))",
/* 168 */	"stmt: ASGNI4(adr,ADDI4(INDIRI4(adr),am))",
/* 169 */	"stmt: ASGNU4(adr,ADDU4(INDIRU4(adr),am))",
/* 170 */	"stmt: ASGNI8(adr,ADDI8(INDIRI8(adr),am))",
/* 171 */	"stmt: ASGNU8(adr,ADDU8(INDIRU8(adr),am))",
/* 172 */	"stmt: ASGNI4(adr,SUBI4(INDIRI4(adr),am))",
/* 173 */	"stmt: ASGNU4(adr,SUBU4(INDIRU4(adr),am))",
/* 174 */	"stmt: ASGNI4(adr,RSHI4(INDIRI4(adr),am))",
/* 175 */	"stmt: ASGNU4(adr,RSHU4(INDIRU4(adr),am))",
/* 176 */	"stmt: ASGNI4(adr,LSHI4(INDIRI4(adr),am))",
/* 177 */	"stmt: ASGNU4(adr,LSHU4(INDIRU4(adr),am))",
/* 178 */	"stmt: ASGNI4(adr,BANDI4(INDIRI4(adr),am))",
/* 179 */	"stmt: ASGNU4(adr,BANDU4(INDIRU4(adr),am))",
/* 180 */	"stmt: ASGNI4(adr,BORI4(INDIRI4(adr),am))",
/* 181 */	"stmt: ASGNU4(adr,BORU4(INDIRU4(adr),am))",
/* 182 */	"stmt: ASGNI4(adr,BXORI4(INDIRI4(adr),am))",
/* 183 */	"stmt: ASGNU4(adr,BXORU4(INDIRU4(adr),am))",
/* 184 */	"stmt: ASGNB(adr,iadr)",
/* 185 */	"stmt: ARGB(iadr)",
/* 186 */	"reg: ADDI1(reg,am)",
/* 187 */	"reg: ADDU1(reg,am)",
/* 188 */	"reg: ADDI2(reg,am)",
/* 189 */	"reg: ADDU2(reg,am)",
/* 190 */	"reg: ADDI4(reg,am)",
/* 191 */	"reg: ADDU4(reg,am)",
/* 192 */	"reg: ADDI8(reg,am)",
/* 193 */	"reg: ADDU8(reg,am)",
/* 194 */	"reg: ADDF4(reg,am)",
/* 195 */	"reg: ADDF8(reg,am)",
/* 196 */	"reg: ADDP4(reg,ca)",
/* 197 */	"reg: SUBI1(reg,am)",
/* 198 */	"reg: SUBU1(reg,am)",
/* 199 */	"reg: SUBI2(reg,am)",
/* 200 */	"reg: SUBU2(reg,am)",
/* 201 */	"reg: SUBI4(reg,am)",
/* 202 */	"reg: SUBU4(reg,am)",
/* 203 */	"reg: SUBP4(reg,am)",
/* 204 */	"reg: SUBI8(reg,am)",
/* 205 */	"reg: SUBU8(reg,am)",
/* 206 */	"reg: SUBF4(reg,am)",
/* 207 */	"reg: SUBF8(reg,am)",
/* 208 */	"reg: MULU4(con,LOADU4(eadr))",
/* 209 */	"reg: MULI1(reg,am)",
/* 210 */	"reg: MULU1(reg,am)",
/* 211 */	"reg: MULI2(reg,am)",
/* 212 */	"reg: MULU2(reg,am)",
/* 213 */	"reg: MULI4(reg,am)",
/* 214 */	"reg: MULU4(reg,am)",
/* 215 */	"reg: MULI8(reg,am)",
/* 216 */	"reg: MULU8(reg,am)",
/* 217 */	"reg: MULF4(reg,am)",
/* 218 */	"reg: MULF8(reg,am)",
/* 219 */	"reg: MULI4(reg,conv2)",
/* 220 */	"reg: MULI4(reg,conv4)",
/* 221 */	"reg: MULI4(reg,conv8)",
/* 222 */	"reg: MULI4(reg,conv16)",
/* 223 */	"reg: MULI4(reg,conv32)",
/* 224 */	"reg: MULI4(reg,conv64)",
/* 225 */	"reg: MULI4(reg,conv128)",
/* 226 */	"reg: MULI4(reg,conv256)",
/* 227 */	"reg: MULU4(reg,conv2)",
/* 228 */	"reg: MULU4(reg,conv4)",
/* 229 */	"reg: MULU4(reg,conv8)",
/* 230 */	"reg: MULU4(reg,conv16)",
/* 231 */	"reg: MULU4(reg,conv32)",
/* 232 */	"reg: MULU4(reg,conv64)",
/* 233 */	"reg: MULU4(reg,conv128)",
/* 234 */	"reg: MULU4(reg,conv256)",
/* 235 */	"reg: DIVI1(reg,am)",
/* 236 */	"reg: DIVU1(reg,am)",
/* 237 */	"reg: DIVI2(reg,am)",
/* 238 */	"reg: DIVU2(reg,am)",
/* 239 */	"reg: DIVI4(reg,am)",
/* 240 */	"reg: DIVU4(reg,am)",
/* 241 */	"reg: DIVI8(reg,am)",
/* 242 */	"reg: DIVU8(reg,am)",
/* 243 */	"reg: DIVF4(reg,am)",
/* 244 */	"reg: DIVF8(reg,am)",
/* 245 */	"reg: DIVI4(reg,conv2)",
/* 246 */	"reg: DIVI4(reg,conv4)",
/* 247 */	"reg: DIVI4(reg,conv8)",
/* 248 */	"reg: DIVI4(reg,conv16)",
/* 249 */	"reg: DIVI4(reg,conv32)",
/* 250 */	"reg: DIVI4(reg,conv64)",
/* 251 */	"reg: DIVI4(reg,conv128)",
/* 252 */	"reg: DIVI4(reg,conv256)",
/* 253 */	"reg: DIVU4(reg,conv2)",
/* 254 */	"reg: DIVU4(reg,conv4)",
/* 255 */	"reg: DIVU4(reg,conv8)",
/* 256 */	"reg: DIVU4(reg,conv16)",
/* 257 */	"reg: DIVU4(reg,conv32)",
/* 258 */	"reg: DIVU4(reg,conv64)",
/* 259 */	"reg: DIVU4(reg,conv128)",
/* 260 */	"reg: DIVU4(reg,conv256)",
/* 261 */	"reg: MODI1(reg,am)",
/* 262 */	"reg: MODU1(reg,am)",
/* 263 */	"reg: MODI2(reg,am)",
/* 264 */	"reg: MODU2(reg,am)",
/* 265 */	"reg: MODI4(reg,am)",
/* 266 */	"reg: MODU4(reg,am)",
/* 267 */	"reg: MODI8(reg,am)",
/* 268 */	"reg: MODU8(reg,am)",
/* 269 */	"reg: RSHI1(reg,am)",
/* 270 */	"reg: RSHU1(reg,am)",
/* 271 */	"reg: RSHI2(reg,am)",
/* 272 */	"reg: RSHU2(reg,am)",
/* 273 */	"reg: RSHI4(reg,am)",
/* 274 */	"reg: RSHU4(reg,am)",
/* 275 */	"reg: RSHI8(reg,am)",
/* 276 */	"reg: RSHU8(reg,am)",
/* 277 */	"reg: LSHI1(reg,am)",
/* 278 */	"reg: LSHU1(reg,am)",
/* 279 */	"reg: LSHI2(reg,am)",
/* 280 */	"reg: LSHU2(reg,am)",
/* 281 */	"reg: LSHI4(reg,am)",
/* 282 */	"reg: LSHU4(reg,am)",
/* 283 */	"reg: LSHI8(reg,am)",
/* 284 */	"reg: LSHU8(reg,am)",
/* 285 */	"reg: BANDI1(reg,am)",
/* 286 */	"reg: BANDU1(reg,am)",
/* 287 */	"reg: BANDI2(reg,am)",
/* 288 */	"reg: BANDU2(reg,am)",
/* 289 */	"reg: BANDI4(reg,am)",
/* 290 */	"reg: BANDU4(reg,am)",
/* 291 */	"reg: BANDI8(reg,am)",
/* 292 */	"reg: BANDU8(reg,am)",
/* 293 */	"reg: BORI1(reg,am)",
/* 294 */	"reg: BORU1(reg,am)",
/* 295 */	"reg: BORI2(reg,am)",
/* 296 */	"reg: BORU2(reg,am)",
/* 297 */	"reg: BORI4(reg,am)",
/* 298 */	"reg: BORU4(reg,am)",
/* 299 */	"reg: BORI8(reg,am)",
/* 300 */	"reg: BORU8(reg,am)",
/* 301 */	"reg: BXORI1(reg,am)",
/* 302 */	"reg: BXORU1(reg,am)",
/* 303 */	"reg: BXORI2(reg,am)",
/* 304 */	"reg: BXORU2(reg,am)",
/* 305 */	"reg: BXORI4(reg,am)",
/* 306 */	"reg: BXORU4(reg,am)",
/* 307 */	"reg: BXORI8(reg,am)",
/* 308 */	"reg: BXORU8(reg,am)",
/* 309 */	"reg: BCOMI1(reg)",
/* 310 */	"reg: BCOMU1(reg)",
/* 311 */	"reg: BCOMI2(reg)",
/* 312 */	"reg: BCOMU2(reg)",
/* 313 */	"reg: BCOMI4(reg)",
/* 314 */	"reg: BCOMU4(reg)",
/* 315 */	"reg: BCOMI8(reg)",
/* 316 */	"reg: BCOMU8(reg)",
/* 317 */	"stmt: BCOMI1(am)",
/* 318 */	"stmt: BCOMU1(am)",
/* 319 */	"stmt: BCOMI2(am)",
/* 320 */	"stmt: BCOMU2(am)",
/* 321 */	"stmt: BCOMI4(am)",
/* 322 */	"stmt: BCOMU4(am)",
/* 323 */	"stmt: BCOMI8(am)",
/* 324 */	"stmt: BCOMU8(am)",
/* 325 */	"reg: NEGI1(reg)",
/* 326 */	"reg: NEGI2(reg)",
/* 327 */	"reg: NEGI4(reg)",
/* 328 */	"reg: NEGI8(reg)",
/* 329 */	"reg: NEGF4(reg)",
/* 330 */	"reg: NEGF8(reg)",
/* 331 */	"stmt: EQI1(dm,conv0)",
/* 332 */	"stmt: EQU1(dm,conv0)",
/* 333 */	"stmt: EQI2(dm,conv0)",
/* 334 */	"stmt: EQU2(dm,conv0)",
/* 335 */	"stmt: EQI4(dm,conv0)",
/* 336 */	"stmt: EQU4(dm,conv0)",
/* 337 */	"stmt: EQI8(dm,conv0)",
/* 338 */	"stmt: EQU8(dm,conv0)",
/* 339 */	"stmt: EQF4(dm,conv0)",
/* 340 */	"stmt: EQF8(dm,conv0)",
/* 341 */	"stmt: EQI1(BANDI1(dm,conv1),conv0)",
/* 342 */	"stmt: EQU1(BANDU1(dm,conv1),conv0)",
/* 343 */	"stmt: EQI2(BANDI2(dm,conv1),conv0)",
/* 344 */	"stmt: EQU2(BANDU2(dm,conv1),conv0)",
/* 345 */	"stmt: EQI4(BANDI4(dm,conv1),conv0)",
/* 346 */	"stmt: EQU4(BANDU4(dm,conv1),conv0)",
/* 347 */	"stmt: EQI8(BANDI8(dm,conv1),conv0)",
/* 348 */	"stmt: EQU8(BANDU8(dm,conv1),conv0)",
/* 349 */	"stmt: EQI1(BANDI1(dm,conv1),conv1)",
/* 350 */	"stmt: EQU1(BANDU1(dm,conv1),conv1)",
/* 351 */	"stmt: EQI2(BANDI2(dm,conv1),conv1)",
/* 352 */	"stmt: EQU2(BANDU2(dm,conv1),conv1)",
/* 353 */	"stmt: EQI4(BANDI4(dm,conv1),conv1)",
/* 354 */	"stmt: EQU4(BANDU4(dm,conv1),conv1)",
/* 355 */	"stmt: EQI8(BANDI8(dm,conv1),conv1)",
/* 356 */	"stmt: EQU8(BANDU8(dm,conv1),conv1)",
/* 357 */	"stmt: EQI1(dm,am)",
/* 358 */	"stmt: EQU1(dm,am)",
/* 359 */	"stmt: EQI2(dm,am)",
/* 360 */	"stmt: EQU2(dm,am)",
/* 361 */	"stmt: EQI4(dm,am)",
/* 362 */	"stmt: EQU4(dm,am)",
/* 363 */	"stmt: EQI8(dm,am)",
/* 364 */	"stmt: EQU8(dm,am)",
/* 365 */	"stmt: EQF4(dm,am)",
/* 366 */	"stmt: EQF8(dm,am)",
/* 367 */	"stmt: NEI1(dm,conv0)",
/* 368 */	"stmt: NEU1(dm,conv0)",
/* 369 */	"stmt: NEI2(dm,conv0)",
/* 370 */	"stmt: NEU2(dm,conv0)",
/* 371 */	"stmt: NEI4(dm,conv0)",
/* 372 */	"stmt: NEU4(dm,conv0)",
/* 373 */	"stmt: NEI8(dm,conv0)",
/* 374 */	"stmt: NEU8(dm,conv0)",
/* 375 */	"stmt: NEF4(dm,conv0)",
/* 376 */	"stmt: NEF8(dm,conv0)",
/* 377 */	"stmt: NEI1(BANDI1(dm,conv1),conv0)",
/* 378 */	"stmt: NEU1(BANDU1(dm,conv1),conv0)",
/* 379 */	"stmt: NEI2(BANDI2(dm,conv1),conv0)",
/* 380 */	"stmt: NEU2(BANDU2(dm,conv1),conv0)",
/* 381 */	"stmt: NEI4(BANDI4(dm,conv1),conv0)",
/* 382 */	"stmt: NEU4(BANDU4(dm,conv1),conv0)",
/* 383 */	"stmt: NEI8(BANDI8(dm,conv1),conv0)",
/* 384 */	"stmt: NEU8(BANDU8(dm,conv1),conv0)",
/* 385 */	"stmt: NEI1(BANDI1(dm,conv1),conv1)",
/* 386 */	"stmt: NEU1(BANDU1(dm,conv1),conv1)",
/* 387 */	"stmt: NEI2(BANDI2(dm,conv1),conv1)",
/* 388 */	"stmt: NEU2(BANDU2(dm,conv1),conv1)",
/* 389 */	"stmt: NEI4(BANDI4(dm,conv1),conv1)",
/* 390 */	"stmt: NEU4(BANDU4(dm,conv1),conv1)",
/* 391 */	"stmt: NEI8(BANDI8(dm,conv1),conv1)",
/* 392 */	"stmt: NEU8(BANDU8(dm,conv1),conv1)",
/* 393 */	"stmt: NEI1(dm,am)",
/* 394 */	"stmt: NEU1(dm,am)",
/* 395 */	"stmt: NEI2(dm,am)",
/* 396 */	"stmt: NEU2(dm,am)",
/* 397 */	"stmt: NEI4(dm,am)",
/* 398 */	"stmt: NEU4(dm,am)",
/* 399 */	"stmt: NEI8(dm,am)",
/* 400 */	"stmt: NEU8(dm,am)",
/* 401 */	"stmt: NEF4(dm,am)",
/* 402 */	"stmt: NEF8(dm,am)",
/* 403 */	"stmt: GEI1(dm,am)",
/* 404 */	"stmt: GEU1(dm,am)",
/* 405 */	"stmt: GEI2(dm,am)",
/* 406 */	"stmt: GEU2(dm,am)",
/* 407 */	"stmt: GEI4(dm,am)",
/* 408 */	"stmt: GEU4(dm,am)",
/* 409 */	"stmt: GEI8(dm,am)",
/* 410 */	"stmt: GEU8(dm,am)",
/* 411 */	"stmt: GEF4(dm,am)",
/* 412 */	"stmt: GEF8(dm,am)",
/* 413 */	"stmt: GTI1(dm,am)",
/* 414 */	"stmt: GTU1(dm,am)",
/* 415 */	"stmt: GTI2(dm,am)",
/* 416 */	"stmt: GTU2(dm,am)",
/* 417 */	"stmt: GTI4(dm,am)",
/* 418 */	"stmt: GTU4(dm,am)",
/* 419 */	"stmt: GTI8(dm,am)",
/* 420 */	"stmt: GTU8(dm,am)",
/* 421 */	"stmt: GTF4(dm,am)",
/* 422 */	"stmt: GTF8(dm,am)",
/* 423 */	"stmt: LEI1(dm,am)",
/* 424 */	"stmt: LEU1(dm,am)",
/* 425 */	"stmt: LEI2(dm,am)",
/* 426 */	"stmt: LEU2(dm,am)",
/* 427 */	"stmt: LEI4(dm,am)",
/* 428 */	"stmt: LEU4(dm,am)",
/* 429 */	"stmt: LEI8(dm,am)",
/* 430 */	"stmt: LEU8(dm,am)",
/* 431 */	"stmt: LEF4(dm,am)",
/* 432 */	"stmt: LEF8(dm,am)",
/* 433 */	"stmt: LTI1(dm,am)",
/* 434 */	"stmt: LTU1(dm,am)",
/* 435 */	"stmt: LTI2(dm,am)",
/* 436 */	"stmt: LTU2(dm,am)",
/* 437 */	"stmt: LTI4(dm,am)",
/* 438 */	"stmt: LTU4(dm,am)",
/* 439 */	"stmt: LTI8(dm,am)",
/* 440 */	"stmt: LTU8(dm,am)",
/* 441 */	"stmt: LTF4(dm,am)",
/* 442 */	"stmt: LTF8(dm,am)",
/* 443 */	"stmt: LTI1(eadr,eadr)",
/* 444 */	"stmt: LTU1(eadr,eadr)",
/* 445 */	"stmt: LTI2(eadr,eadr)",
/* 446 */	"stmt: LTU2(eadr,eadr)",
/* 447 */	"stmt: LTI4(eadr,eadr)",
/* 448 */	"stmt: LTU4(eadr,eadr)",
/* 449 */	"stmt: LTI8(eadr,eadr)",
/* 450 */	"stmt: LTU8(eadr,eadr)",
/* 451 */	"stmt: LTF4(eadr,eadr)",
/* 452 */	"stmt: LTF8(eadr,eadr)",
/* 453 */	"reg: CVPU4(reg)",
/* 454 */	"reg: CVUP4(reg)",
/* 455 */	"reg: CVII1(reg)",
/* 456 */	"reg: CVIU1(reg)",
/* 457 */	"reg: CVII2(reg)",
/* 458 */	"reg: CVIU2(reg)",
/* 459 */	"reg: CVII4(reg)",
/* 460 */	"reg: CVIU4(reg)",
/* 461 */	"reg: CVII8(reg)",
/* 462 */	"reg: CVIU8(reg)",
/* 463 */	"reg: CVIF4(reg)",
/* 464 */	"reg: CVIF8(reg)",
/* 465 */	"reg: CVUI1(reg)",
/* 466 */	"reg: CVUU1(reg)",
/* 467 */	"reg: CVUI2(reg)",
/* 468 */	"reg: CVUU2(reg)",
/* 469 */	"reg: CVUI4(reg)",
/* 470 */	"reg: CVUU4(reg)",
/* 471 */	"reg: CVUI8(reg)",
/* 472 */	"reg: CVUU8(reg)",
/* 473 */	"reg: CVUF4(reg)",
/* 474 */	"reg: CVUF8(reg)",
/* 475 */	"reg: CVFF4(reg)",
/* 476 */	"reg: CVFF8(reg)",
/* 477 */	"reg: CVFI2(reg)",
/* 478 */	"reg: CVFI4(reg)",
/* 479 */	"reg: CVFI8(reg)",
/* 480 */	"reg: CVFF8(INDIRF4(ADDP4(dm,am)))",
/* 481 */	"stmt: ARGI2(LOADI4(am))",
/* 482 */	"stmt: ARGU2(LOADU4(am))",
/* 483 */	"stmt: ARGI4(LOADI4(am))",
/* 484 */	"stmt: ARGU4(LOADU4(ad))",
/* 485 */	"stmt: ARGU4(LOADU4(reg))",
/* 486 */	"stmt: ARGU4(LOADU4(con))",
/* 487 */	"stmt: ARGU4(LOADU4(iadr))",
/* 488 */	"stmt: ARGI4(LOADI8(am))",
/* 489 */	"stmt: ARGU4(LOADU8(am))",
/* 490 */	"stmt: ARGI2(am)",
/* 491 */	"stmt: ARGU2(am)",
/* 492 */	"stmt: ARGI4(am)",
/* 493 */	"stmt: ARGU4(am)",
/* 494 */	"stmt: ARGI8(am)",
/* 495 */	"stmt: ARGU8(am)",
/* 496 */	"stmt: ARGF4(am)",
/* 497 */	"stmt: ARGF8(am)",
/* 498 */	"stmt: ARGP4(eadr)",
/* 499 */	"stmt: ARGP4(ad)",
/* 500 */	"stmt: ARGP4(reg)",
/* 501 */	"reg: CALLI2(ca)",
/* 502 */	"reg: CALLU2(ca)",
/* 503 */	"reg: CALLI4(ca)",
/* 504 */	"reg: CALLU4(ca)",
/* 505 */	"reg: CALLP4(ca)",
/* 506 */	"reg: CALLI8(ca)",
/* 507 */	"reg: CALLU8(ca)",
/* 508 */	"reg: CALLF4(ca)",
/* 509 */	"reg: CALLF8(ca)",
/* 510 */	"stmt: CALLV(ca)",
/* 511 */	"reg: RETI1(reg)",
/* 512 */	"reg: RETU1(reg)",
/* 513 */	"reg: RETI2(reg)",
/* 514 */	"reg: RETU2(reg)",
/* 515 */	"reg: RETI4(reg)",
/* 516 */	"reg: RETU4(reg)",
/* 517 */	"reg: RETP4(reg)",
/* 518 */	"reg: RETI8(reg)",
/* 519 */	"reg: RETU8(reg)",
/* 520 */	"reg: RETI1(con)",
/* 521 */	"reg: RETU1(con)",
/* 522 */	"reg: RETI2(con)",
/* 523 */	"reg: RETU2(con)",
/* 524 */	"reg: RETI4(con)",
/* 525 */	"reg: RETU4(con)",
/* 526 */	"reg: RETP4(con)",
/* 527 */	"reg: RETI8(con)",
/* 528 */	"reg: RETU8(con)",
/* 529 */	"stmt: RETI4(ad)",
/* 530 */	"stmt: RETU4(ad)",
/* 531 */	"stmt: RETP4(ad)",
/* 532 */	"stmt: RETI1(reg)",
/* 533 */	"stmt: RETU1(reg)",
/* 534 */	"stmt: RETI2(reg)",
/* 535 */	"stmt: RETU2(reg)",
/* 536 */	"stmt: RETI4(reg)",
/* 537 */	"stmt: RETU4(reg)",
/* 538 */	"stmt: RETP4(reg)",
/* 539 */	"stmt: RETI8(reg)",
/* 540 */	"stmt: RETU8(reg)",
/* 541 */	"stmt: RETF4(reg)",
/* 542 */	"stmt: RETF8(reg)",
/* 543 */	"stmt: RETI1(con)",
/* 544 */	"stmt: RETU1(con)",
/* 545 */	"stmt: RETI2(con)",
/* 546 */	"stmt: RETU2(con)",
/* 547 */	"stmt: RETI4(con)",
/* 548 */	"stmt: RETU4(con)",
/* 549 */	"stmt: RETP4(con)",
/* 550 */	"stmt: RETI8(con)",
/* 551 */	"stmt: RETU8(con)",
/* 552 */	"stmt: RETF4(con)",
/* 553 */	"stmt: RETF8(con)",
/* 554 */	"stmt: JUMPV(am)",
/* 555 */	"stmt: JUMPV(adr)",
/* 556 */	"stmt: LABELV",
/* 557 */	"reg: ASMRETU4",
/* 558 */	"reg: ASMRETU8",
/* 559 */	"stmt: ASMRETU4",
/* 560 */	"stmt: ASMRETU8",
};

static short _decode_stmt[] = {
	0,
	76,
	132,
	133,
	134,
	135,
	136,
	137,
	138,
	139,
	140,
	141,
	142,
	143,
	144,
	145,
	146,
	147,
	148,
	149,
	150,
	152,
	153,
	154,
	155,
	156,
	157,
	158,
	159,
	160,
	161,
	162,
	163,
	164,
	165,
	166,
	167,
	168,
	169,
	170,
	171,
	172,
	173,
	174,
	175,
	176,
	177,
	178,
	179,
	180,
	181,
	182,
	183,
	184,
	185,
	317,
	318,
	319,
	320,
	321,
	322,
	323,
	324,
	331,
	332,
	333,
	334,
	335,
	336,
	337,
	338,
	339,
	340,
	341,
	342,
	343,
	344,
	345,
	346,
	347,
	348,
	349,
	350,
	351,
	352,
	353,
	354,
	355,
	356,
	357,
	358,
	359,
	360,
	361,
	362,
	363,
	364,
	365,
	366,
	367,
	368,
	369,
	370,
	371,
	372,
	373,
	374,
	375,
	376,
	377,
	378,
	379,
	380,
	381,
	382,
	383,
	384,
	385,
	386,
	387,
	388,
	389,
	390,
	391,
	392,
	393,
	394,
	395,
	396,
	397,
	398,
	399,
	400,
	401,
	402,
	403,
	404,
	405,
	406,
	407,
	408,
	409,
	410,
	411,
	412,
	413,
	414,
	415,
	416,
	417,
	418,
	419,
	420,
	421,
	422,
	423,
	424,
	425,
	426,
	427,
	428,
	429,
	430,
	431,
	432,
	433,
	434,
	435,
	436,
	437,
	438,
	439,
	440,
	441,
	442,
	443,
	444,
	445,
	446,
	447,
	448,
	449,
	450,
	451,
	452,
	481,
	482,
	483,
	484,
	485,
	486,
	487,
	488,
	489,
	490,
	491,
	492,
	493,
	494,
	495,
	496,
	497,
	498,
	499,
	500,
	510,
	529,
	530,
	531,
	532,
	533,
	534,
	535,
	536,
	537,
	538,
	539,
	540,
	541,
	542,
	543,
	544,
	545,
	546,
	547,
	548,
	549,
	550,
	551,
	552,
	553,
	554,
	555,
	556,
	559,
	560,
};

static short _decode_con1[] = {
	0,
	1,
	2,
};

static short _decode_con2[] = {
	0,
	3,
	4,
};

static short _decode_con4[] = {
	0,
	5,
	6,
	7,
	10,
};

static short _decode_con8[] = {
	0,
	8,
	9,
	11,
};

static short _decode_con[] = {
	0,
	12,
	13,
	14,
	15,
};

static short _decode_conv0[] = {
	0,
	16,
	17,
	18,
	19,
	20,
	21,
	22,
	23,
};

static short _decode_conv1[] = {
	0,
	24,
	25,
	26,
	27,
	28,
	29,
	30,
	31,
};

static short _decode_conv2[] = {
	0,
	32,
	33,
};

static short _decode_conv4[] = {
	0,
	34,
	35,
};

static short _decode_conv8[] = {
	0,
	36,
	37,
};

static short _decode_conv16[] = {
	0,
	38,
	39,
};

static short _decode_conv32[] = {
	0,
	40,
	41,
};

static short _decode_conv64[] = {
	0,
	42,
	43,
};

static short _decode_conv128[] = {
	0,
	44,
	45,
};

static short _decode_conv256[] = {
	0,
	46,
	47,
};

static short _decode_reg[] = {
	0,
	48,
	49,
	50,
	51,
	52,
	53,
	54,
	55,
	56,
	57,
	58,
	59,
	60,
	61,
	62,
	63,
	64,
	65,
	66,
	67,
	68,
	69,
	70,
	71,
	72,
	73,
	74,
	75,
	116,
	117,
	118,
	119,
	120,
	151,
	186,
	187,
	188,
	189,
	190,
	191,
	192,
	193,
	194,
	195,
	196,
	197,
	198,
	199,
	200,
	201,
	202,
	203,
	204,
	205,
	206,
	207,
	208,
	209,
	210,
	211,
	212,
	213,
	214,
	215,
	216,
	217,
	218,
	219,
	220,
	221,
	222,
	223,
	224,
	225,
	226,
	227,
	228,
	229,
	230,
	231,
	232,
	233,
	234,
	235,
	236,
	237,
	238,
	239,
	240,
	241,
	242,
	243,
	244,
	245,
	246,
	247,
	248,
	249,
	250,
	251,
	252,
	253,
	254,
	255,
	256,
	257,
	258,
	259,
	260,
	261,
	262,
	263,
	264,
	265,
	266,
	267,
	268,
	269,
	270,
	271,
	272,
	273,
	274,
	275,
	276,
	277,
	278,
	279,
	280,
	281,
	282,
	283,
	284,
	285,
	286,
	287,
	288,
	289,
	290,
	291,
	292,
	293,
	294,
	295,
	296,
	297,
	298,
	299,
	300,
	301,
	302,
	303,
	304,
	305,
	306,
	307,
	308,
	309,
	310,
	311,
	312,
	313,
	314,
	315,
	316,
	325,
	326,
	327,
	328,
	329,
	330,
	453,
	454,
	455,
	456,
	457,
	458,
	459,
	460,
	461,
	462,
	463,
	464,
	465,
	466,
	467,
	468,
	469,
	470,
	471,
	472,
	473,
	474,
	475,
	476,
	477,
	478,
	479,
	480,
	501,
	502,
	503,
	504,
	505,
	506,
	507,
	508,
	509,
	511,
	512,
	513,
	514,
	515,
	516,
	517,
	518,
	519,
	520,
	521,
	522,
	523,
	524,
	525,
	526,
	527,
	528,
	557,
	558,
};

static short _decode_adr[] = {
	0,
	82,
	83,
};

static short _decode_ad[] = {
	0,
	79,
	80,
	81,
};

static short _decode_ca[] = {
	0,
	77,
	78,
};

static short _decode_iadr1[] = {
	0,
	84,
	85,
};

static short _decode_iadr2[] = {
	0,
	86,
	87,
};

static short _decode_iadr4[] = {
	0,
	88,
	89,
	90,
	93,
};

static short _decode_iadr8[] = {
	0,
	91,
	92,
	94,
};

static short _decode_iadr[] = {
	0,
	95,
	96,
	97,
	98,
	99,
};

static short _decode_eadr1[] = {
	0,
	100,
	101,
};

static short _decode_eadr2[] = {
	0,
	102,
	103,
};

static short _decode_eadr4[] = {
	0,
	104,
	105,
	106,
	109,
};

static short _decode_eadr8[] = {
	0,
	107,
	108,
	110,
};

static short _decode_eadr[] = {
	0,
	111,
	112,
	113,
	114,
	115,
};

static short _decode_dm[] = {
	0,
	121,
	122,
	123,
	124,
	125,
};

static short _decode_am[] = {
	0,
	126,
	127,
	128,
	129,
	130,
	131,
};

static int _rule(void *state, int goalnt) {
	if (goalnt < 1 || goalnt > 32)
		fatal("_rule", "Bad goal nonterminal %d\n", goalnt);
	if (!state)
		return 0;
	switch (goalnt) {
	case _stmt_NT:	return _decode_stmt[((struct _state *)state)->rule._stmt];
	case _con1_NT:	return _decode_con1[((struct _state *)state)->rule._con1];
	case _con2_NT:	return _decode_con2[((struct _state *)state)->rule._con2];
	case _con4_NT:	return _decode_con4[((struct _state *)state)->rule._con4];
	case _con8_NT:	return _decode_con8[((struct _state *)state)->rule._con8];
	case _con_NT:	return _decode_con[((struct _state *)state)->rule._con];
	case _conv0_NT:	return _decode_conv0[((struct _state *)state)->rule._conv0];
	case _conv1_NT:	return _decode_conv1[((struct _state *)state)->rule._conv1];
	case _conv2_NT:	return _decode_conv2[((struct _state *)state)->rule._conv2];
	case _conv4_NT:	return _decode_conv4[((struct _state *)state)->rule._conv4];
	case _conv8_NT:	return _decode_conv8[((struct _state *)state)->rule._conv8];
	case _conv16_NT:	return _decode_conv16[((struct _state *)state)->rule._conv16];
	case _conv32_NT:	return _decode_conv32[((struct _state *)state)->rule._conv32];
	case _conv64_NT:	return _decode_conv64[((struct _state *)state)->rule._conv64];
	case _conv128_NT:	return _decode_conv128[((struct _state *)state)->rule._conv128];
	case _conv256_NT:	return _decode_conv256[((struct _state *)state)->rule._conv256];
	case _reg_NT:	return _decode_reg[((struct _state *)state)->rule._reg];
	case _adr_NT:	return _decode_adr[((struct _state *)state)->rule._adr];
	case _ad_NT:	return _decode_ad[((struct _state *)state)->rule._ad];
	case _ca_NT:	return _decode_ca[((struct _state *)state)->rule._ca];
	case _iadr1_NT:	return _decode_iadr1[((struct _state *)state)->rule._iadr1];
	case _iadr2_NT:	return _decode_iadr2[((struct _state *)state)->rule._iadr2];
	case _iadr4_NT:	return _decode_iadr4[((struct _state *)state)->rule._iadr4];
	case _iadr8_NT:	return _decode_iadr8[((struct _state *)state)->rule._iadr8];
	case _iadr_NT:	return _decode_iadr[((struct _state *)state)->rule._iadr];
	case _eadr1_NT:	return _decode_eadr1[((struct _state *)state)->rule._eadr1];
	case _eadr2_NT:	return _decode_eadr2[((struct _state *)state)->rule._eadr2];
	case _eadr4_NT:	return _decode_eadr4[((struct _state *)state)->rule._eadr4];
	case _eadr8_NT:	return _decode_eadr8[((struct _state *)state)->rule._eadr8];
	case _eadr_NT:	return _decode_eadr[((struct _state *)state)->rule._eadr];
	case _dm_NT:	return _decode_dm[((struct _state *)state)->rule._dm];
	case _am_NT:	return _decode_am[((struct _state *)state)->rule._am];
	default:
		fatal("_rule", "Bad goal nonterminal %d\n", goalnt);
		return 0;
	}
}

static void _closure_con1(NODEPTR_TYPE, int);
static void _closure_con2(NODEPTR_TYPE, int);
static void _closure_con4(NODEPTR_TYPE, int);
static void _closure_con8(NODEPTR_TYPE, int);
static void _closure_con(NODEPTR_TYPE, int);
static void _closure_reg(NODEPTR_TYPE, int);
static void _closure_ad(NODEPTR_TYPE, int);
static void _closure_iadr1(NODEPTR_TYPE, int);
static void _closure_iadr2(NODEPTR_TYPE, int);
static void _closure_iadr4(NODEPTR_TYPE, int);
static void _closure_iadr8(NODEPTR_TYPE, int);
static void _closure_iadr(NODEPTR_TYPE, int);
static void _closure_eadr1(NODEPTR_TYPE, int);
static void _closure_eadr2(NODEPTR_TYPE, int);
static void _closure_eadr4(NODEPTR_TYPE, int);
static void _closure_eadr8(NODEPTR_TYPE, int);

static void _closure_con1(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 25;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_con_NT]) {
		p->cost[_con_NT] = c + 0;
		p->rule._con = 1;
		_closure_con(a, c + 0);
	}
}

static void _closure_con2(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 26;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_con_NT]) {
		p->cost[_con_NT] = c + 0;
		p->rule._con = 2;
		_closure_con(a, c + 0);
	}
}

static void _closure_con4(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 27;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_con_NT]) {
		p->cost[_con_NT] = c + 0;
		p->rule._con = 3;
		_closure_con(a, c + 0);
	}
}

static void _closure_con8(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 28;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_con_NT]) {
		p->cost[_con_NT] = c + 0;
		p->rule._con = 4;
		_closure_con(a, c + 0);
	}
}

static void _closure_con(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_am_NT]) {
		p->cost[_am_NT] = c + 0;
		p->rule._am = 2;
	}
}

static void _closure_reg(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_am_NT]) {
		p->cost[_am_NT] = c + 0;
		p->rule._am = 1;
	}
	if (c + 0 < p->cost[_dm_NT]) {
		p->cost[_dm_NT] = c + 0;
		p->rule._dm = 1;
	}
	if (c + 0 < p->cost[_adr_NT]) {
		p->cost[_adr_NT] = c + 0;
		p->rule._adr = 1;
	}
	if (c + 0 < p->cost[_ca_NT]) {
		p->cost[_ca_NT] = c + 0;
		p->rule._ca = 2;
	}
	if (c + 0 < p->cost[_stmt_NT]) {
		p->cost[_stmt_NT] = c + 0;
		p->rule._stmt = 1;
	}
}

static void _closure_ad(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 33;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_adr_NT]) {
		p->cost[_adr_NT] = c + 0;
		p->rule._adr = 2;
	}
}

static void _closure_iadr1(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 29;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_iadr_NT]) {
		p->cost[_iadr_NT] = c + 0;
		p->rule._iadr = 1;
		_closure_iadr(a, c + 0);
	}
}

static void _closure_iadr2(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 30;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_iadr_NT]) {
		p->cost[_iadr_NT] = c + 0;
		p->rule._iadr = 2;
		_closure_iadr(a, c + 0);
	}
}

static void _closure_iadr4(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 31;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_iadr_NT]) {
		p->cost[_iadr_NT] = c + 0;
		p->rule._iadr = 3;
		_closure_iadr(a, c + 0);
	}
}

static void _closure_iadr8(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg_NT]) {
		p->cost[_reg_NT] = c + 1;
		p->rule._reg = 32;
		_closure_reg(a, c + 1);
	}
	if (c + 0 < p->cost[_iadr_NT]) {
		p->cost[_iadr_NT] = c + 0;
		p->rule._iadr = 4;
		_closure_iadr(a, c + 0);
	}
}

static void _closure_iadr(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_am_NT]) {
		p->cost[_am_NT] = c + 0;
		p->rule._am = 3;
	}
	if (c + 0 < p->cost[_dm_NT]) {
		p->cost[_dm_NT] = c + 0;
		p->rule._dm = 2;
	}
}

static void _closure_eadr1(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_eadr_NT]) {
		p->cost[_eadr_NT] = c + 0;
		p->rule._eadr = 1;
	}
}

static void _closure_eadr2(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_eadr_NT]) {
		p->cost[_eadr_NT] = c + 0;
		p->rule._eadr = 2;
	}
}

static void _closure_eadr4(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_eadr_NT]) {
		p->cost[_eadr_NT] = c + 0;
		p->rule._eadr = 3;
	}
}

static void _closure_eadr8(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_eadr_NT]) {
		p->cost[_eadr_NT] = c + 0;
		p->rule._eadr = 4;
	}
}

static void _label(NODEPTR_TYPE a) {
	int c;
	struct _state *p;

	if (!a)
		fatal("_label", "Null tree\n", 0);
	STATE_LABEL(a) = p = allocate(sizeof *p, FUNC);
	p->rule._stmt = 0;
	p->cost[1] =
	p->cost[2] =
	p->cost[3] =
	p->cost[4] =
	p->cost[5] =
	p->cost[6] =
	p->cost[7] =
	p->cost[8] =
	p->cost[9] =
	p->cost[10] =
	p->cost[11] =
	p->cost[12] =
	p->cost[13] =
	p->cost[14] =
	p->cost[15] =
	p->cost[16] =
	p->cost[17] =
	p->cost[18] =
	p->cost[19] =
	p->cost[20] =
	p->cost[21] =
	p->cost[22] =
	p->cost[23] =
	p->cost[24] =
	p->cost[25] =
	p->cost[26] =
	p->cost[27] =
	p->cost[28] =
	p->cost[29] =
	p->cost[30] =
	p->cost[31] =
	p->cost[32] =
		0x7fff;
	switch (OP_LABEL(a)) {
	case 41: /* ARGB */
		_label(LEFT_CHILD(a));
		/* stmt: ARGB(iadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_iadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 54;
		}
		break;
	case 57: /* ASGNB */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: ASGNB(adr,iadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_iadr_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 53;
		}
		break;
	case 73: /* INDIRB */
		_label(LEFT_CHILD(a));
		/* iadr: INDIRB(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr_NT]) {
			p->cost[_iadr_NT] = c + 0;
			p->rule._iadr = 5;
			_closure_iadr(a, c + 0);
		}
		/* eadr: INDIRB(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr_NT]) {
			p->cost[_eadr_NT] = c + 0;
			p->rule._eadr = 5;
		}
		break;
	case 216: /* CALLV */
		_label(LEFT_CHILD(a));
		/* stmt: CALLV(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 205;
		}
		break;
	case 217: /* CALLB */
		break;
	case 233: /* LOADB */
		break;
	case 248: /* RETV */
		break;
	case 584: /* JUMPV */
		_label(LEFT_CHILD(a));
		/* stmt: JUMPV(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 231;
		}
		/* stmt: JUMPV(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 232;
		}
		break;
	case 600: /* LABELV */
		/* stmt: LABELV */
		if (0 + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = 0 + 0;
			p->rule._stmt = 233;
		}
		break;
	case 711: /* VREGP */
		break;
	case 720: /* ASM0 */
		break;
	case 736: /* ASM1 */
		break;
	case 752: /* ASM2 */
		break;
	case 1045: /* CNSTI1 */
		/* con1: CNSTI1 */
		if (0 + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = 0 + 0;
			p->rule._con1 = 1;
			_closure_con1(a, 0 + 0);
		}
		/* conv0: CNSTI1 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 1;
		}
		/* conv1: CNSTI1 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 1;
		}
		break;
	case 1046: /* CNSTU1 */
		/* con1: CNSTU1 */
		if (0 + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = 0 + 0;
			p->rule._con1 = 2;
			_closure_con1(a, 0 + 0);
		}
		/* conv0: CNSTU1 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 5;
		}
		/* conv1: CNSTU1 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 5;
		}
		break;
	case 1077: /* ASGNI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI1(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 2;
			}
		}
		/* stmt: ASGNI1(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 14;
		}
		if (	/* stmt: ASGNI1(adr,LOADI1(INDIRI4(ad))) */
			RIGHT_CHILD(a)->op == 1253 && /* LOADI1 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_ad_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 23;
			}
		}
		if (	/* stmt: ASGNI1(adr,ADDI1(INDIRI1(adr),am)) */
			RIGHT_CHILD(a)->op == 1333 && /* ADDI1 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 33;
			}
		}
		break;
	case 1078: /* ASGNU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU1(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 3;
			}
		}
		/* stmt: ASGNU1(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 15;
		}
		if (	/* stmt: ASGNU1(adr,LOADU1(INDIRU4(ad))) */
			RIGHT_CHILD(a)->op == 1254 && /* LOADU1 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_ad_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 21;
			}
		}
		if (	/* stmt: ASGNU1(adr,ADDU1(INDIRU1(adr),am)) */
			RIGHT_CHILD(a)->op == 1334 && /* ADDU1 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 1094 /* INDIRU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 34;
			}
		}
		break;
	case 1093: /* INDIRI1 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRI1(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 1;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr1: INDIRI1(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr1_NT]) {
			p->cost[_iadr1_NT] = c + 0;
			p->rule._iadr1 = 1;
			_closure_iadr1(a, c + 0);
		}
		/* eadr1: INDIRI1(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr1_NT]) {
			p->cost[_eadr1_NT] = c + 0;
			p->rule._eadr1 = 1;
			_closure_eadr1(a, c + 0);
		}
		break;
	case 1094: /* INDIRU1 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRU1(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 2;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr1: INDIRU1(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr1_NT]) {
			p->cost[_iadr1_NT] = c + 0;
			p->rule._iadr1 = 2;
			_closure_iadr1(a, c + 0);
		}
		/* eadr1: INDIRU1(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr1_NT]) {
			p->cost[_eadr1_NT] = c + 0;
			p->rule._eadr1 = 2;
			_closure_eadr1(a, c + 0);
		}
		break;
	case 1157: /* CVII1 */
		_label(LEFT_CHILD(a));
		/* reg: CVII1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 174;
			_closure_reg(a, c + 0);
		}
		break;
	case 1158: /* CVIU1 */
		_label(LEFT_CHILD(a));
		/* reg: CVIU1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 175;
			_closure_reg(a, c + 0);
		}
		break;
	case 1205: /* CVUI1 */
		_label(LEFT_CHILD(a));
		/* reg: CVUI1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 184;
			_closure_reg(a, c + 0);
		}
		break;
	case 1206: /* CVUU1 */
		_label(LEFT_CHILD(a));
		/* reg: CVUU1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 185;
			_closure_reg(a, c + 0);
		}
		break;
	case 1221: /* NEGI1 */
		_label(LEFT_CHILD(a));
		/* reg: NEGI1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 166;
			_closure_reg(a, c + 0);
		}
		break;
	case 1253: /* LOADI1 */
		_label(LEFT_CHILD(a));
		/* reg: LOADI1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 12;
			_closure_reg(a, c + 0);
		}
		break;
	case 1254: /* LOADU1 */
		_label(LEFT_CHILD(a));
		/* reg: LOADU1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 13;
			_closure_reg(a, c + 0);
		}
		break;
	case 1269: /* RETI1 */
		_label(LEFT_CHILD(a));
		/* reg: RETI1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 209;
			_closure_reg(a, c + 0);
		}
		/* reg: RETI1(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 218;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETI1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 209;
		}
		/* stmt: RETI1(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 220;
		}
		break;
	case 1270: /* RETU1 */
		_label(LEFT_CHILD(a));
		/* reg: RETU1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 210;
			_closure_reg(a, c + 0);
		}
		/* reg: RETU1(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 219;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETU1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 210;
		}
		/* stmt: RETU1(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 221;
		}
		break;
	case 1333: /* ADDI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 35;
			_closure_reg(a, c + 0);
		}
		break;
	case 1334: /* ADDU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 36;
			_closure_reg(a, c + 0);
		}
		break;
	case 1349: /* SUBI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 46;
			_closure_reg(a, c + 0);
		}
		break;
	case 1350: /* SUBU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 47;
			_closure_reg(a, c + 0);
		}
		break;
	case 1365: /* LSHI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 126;
			_closure_reg(a, c + 0);
		}
		break;
	case 1366: /* LSHU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 127;
			_closure_reg(a, c + 0);
		}
		break;
	case 1381: /* MODI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 110;
			_closure_reg(a, c + 0);
		}
		break;
	case 1382: /* MODU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 111;
			_closure_reg(a, c + 0);
		}
		break;
	case 1397: /* RSHI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 118;
			_closure_reg(a, c + 0);
		}
		break;
	case 1398: /* RSHU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 119;
			_closure_reg(a, c + 0);
		}
		break;
	case 1413: /* BANDI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 134;
			_closure_reg(a, c + 0);
		}
		break;
	case 1414: /* BANDU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 135;
			_closure_reg(a, c + 0);
		}
		break;
	case 1429: /* BCOMI1 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMI1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 158;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMI1(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 55;
		}
		break;
	case 1430: /* BCOMU1 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMU1(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 159;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMU1(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 56;
		}
		break;
	case 1445: /* BORI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 142;
			_closure_reg(a, c + 0);
		}
		break;
	case 1446: /* BORU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 143;
			_closure_reg(a, c + 0);
		}
		break;
	case 1461: /* BXORI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 150;
			_closure_reg(a, c + 0);
		}
		break;
	case 1462: /* BXORU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 151;
			_closure_reg(a, c + 0);
		}
		break;
	case 1477: /* DIVI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 84;
			_closure_reg(a, c + 0);
		}
		break;
	case 1478: /* DIVU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 85;
			_closure_reg(a, c + 0);
		}
		break;
	case 1493: /* MULI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULI1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 58;
			_closure_reg(a, c + 0);
		}
		break;
	case 1494: /* MULU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULU1(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 59;
			_closure_reg(a, c + 0);
		}
		break;
	case 1509: /* EQI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI1(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 63;
		}
		if (	/* stmt: EQI1(BANDI1(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 1413 /* BANDI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 73;
			}
		}
		if (	/* stmt: EQI1(BANDI1(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 1413 /* BANDI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 81;
			}
		}
		/* stmt: EQI1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 89;
		}
		break;
	case 1510: /* EQU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU1(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 64;
		}
		if (	/* stmt: EQU1(BANDU1(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 1414 /* BANDU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 74;
			}
		}
		if (	/* stmt: EQU1(BANDU1(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 1414 /* BANDU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 82;
			}
		}
		/* stmt: EQU1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 90;
		}
		break;
	case 1525: /* GEI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 135;
		}
		break;
	case 1526: /* GEU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 136;
		}
		break;
	case 1541: /* GTI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 145;
		}
		break;
	case 1542: /* GTU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 146;
		}
		break;
	case 1557: /* LEI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 155;
		}
		break;
	case 1558: /* LEU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 156;
		}
		break;
	case 1573: /* LTI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 165;
		}
		/* stmt: LTI1(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 175;
		}
		break;
	case 1574: /* LTU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 166;
		}
		/* stmt: LTU1(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 176;
		}
		break;
	case 1589: /* NEI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI1(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 99;
		}
		if (	/* stmt: NEI1(BANDI1(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 1413 /* BANDI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 109;
			}
		}
		if (	/* stmt: NEI1(BANDI1(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 1413 /* BANDI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 117;
			}
		}
		/* stmt: NEI1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 125;
		}
		break;
	case 1590: /* NEU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU1(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 100;
		}
		if (	/* stmt: NEU1(BANDU1(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 1414 /* BANDU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 110;
			}
		}
		if (	/* stmt: NEU1(BANDU1(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 1414 /* BANDU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 118;
			}
		}
		/* stmt: NEU1(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 126;
		}
		break;
	case 2069: /* CNSTI2 */
		/* con2: CNSTI2 */
		if (0 + 0 < p->cost[_con2_NT]) {
			p->cost[_con2_NT] = 0 + 0;
			p->rule._con2 = 1;
			_closure_con2(a, 0 + 0);
		}
		/* conv0: CNSTI2 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 2;
		}
		/* conv1: CNSTI2 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 2;
		}
		break;
	case 2070: /* CNSTU2 */
		/* con2: CNSTU2 */
		if (0 + 0 < p->cost[_con2_NT]) {
			p->cost[_con2_NT] = 0 + 0;
			p->rule._con2 = 2;
			_closure_con2(a, 0 + 0);
		}
		/* conv0: CNSTU2 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 6;
		}
		/* conv1: CNSTU2 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 6;
		}
		break;
	case 2085: /* ARGI2 */
		_label(LEFT_CHILD(a));
		if (	/* stmt: ARGI2(LOADI4(am)) */
			LEFT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_am_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 185;
			}
		}
		/* stmt: ARGI2(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 194;
		}
		break;
	case 2086: /* ARGU2 */
		_label(LEFT_CHILD(a));
		if (	/* stmt: ARGU2(LOADU4(am)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_am_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 186;
			}
		}
		/* stmt: ARGU2(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 195;
		}
		break;
	case 2101: /* ASGNI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI2(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 4;
			}
		}
		/* stmt: ASGNI2(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 16;
		}
		if (	/* stmt: ASGNI2(adr,LOADI2(INDIRI4(ad))) */
			RIGHT_CHILD(a)->op == 2277 && /* LOADI2 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_ad_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 24;
			}
		}
		if (	/* stmt: ASGNI2(adr,ADDI2(INDIRI2(adr),am)) */
			RIGHT_CHILD(a)->op == 2357 && /* ADDI2 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 35;
			}
		}
		break;
	case 2102: /* ASGNU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU2(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 5;
			}
		}
		/* stmt: ASGNU2(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 17;
		}
		if (	/* stmt: ASGNU2(INDIRP4(VREGP),LOADU2(reg)) */
			LEFT_CHILD(a)->op == 4167 && /* INDIRP4 */
			LEFT_CHILD(LEFT_CHILD(a))->op == 711 && /* VREGP */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 20;
			}
		}
		if (	/* reg: ASGNU2(INDIRP4(VREGP),LOADU2(reg)) */
			LEFT_CHILD(a)->op == 4167 && /* INDIRP4 */
			LEFT_CHILD(LEFT_CHILD(a))->op == 711 && /* VREGP */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 34;
				_closure_reg(a, c + 0);
			}
		}
		if (	/* stmt: ASGNU2(adr,LOADU2(INDIRU4(ad))) */
			RIGHT_CHILD(a)->op == 2278 && /* LOADU2 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_ad_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 22;
			}
		}
		if (	/* stmt: ASGNU2(adr,ADDU2(INDIRU2(adr),am)) */
			RIGHT_CHILD(a)->op == 2358 && /* ADDU2 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 2118 /* INDIRU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 36;
			}
		}
		break;
	case 2117: /* INDIRI2 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRI2(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 3;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr2: INDIRI2(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr2_NT]) {
			p->cost[_iadr2_NT] = c + 0;
			p->rule._iadr2 = 1;
			_closure_iadr2(a, c + 0);
		}
		/* eadr2: INDIRI2(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr2_NT]) {
			p->cost[_eadr2_NT] = c + 0;
			p->rule._eadr2 = 1;
			_closure_eadr2(a, c + 0);
		}
		break;
	case 2118: /* INDIRU2 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRU2(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 4;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr2: INDIRU2(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr2_NT]) {
			p->cost[_iadr2_NT] = c + 0;
			p->rule._iadr2 = 2;
			_closure_iadr2(a, c + 0);
		}
		/* eadr2: INDIRU2(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr2_NT]) {
			p->cost[_eadr2_NT] = c + 0;
			p->rule._eadr2 = 2;
			_closure_eadr2(a, c + 0);
		}
		break;
	case 2165: /* CVFI2 */
		_label(LEFT_CHILD(a));
		/* reg: CVFI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 196;
			_closure_reg(a, c + 0);
		}
		break;
	case 2181: /* CVII2 */
		_label(LEFT_CHILD(a));
		/* reg: CVII2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 176;
			_closure_reg(a, c + 0);
		}
		break;
	case 2182: /* CVIU2 */
		_label(LEFT_CHILD(a));
		/* reg: CVIU2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 177;
			_closure_reg(a, c + 0);
		}
		break;
	case 2229: /* CVUI2 */
		_label(LEFT_CHILD(a));
		/* reg: CVUI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 186;
			_closure_reg(a, c + 0);
		}
		break;
	case 2230: /* CVUU2 */
		_label(LEFT_CHILD(a));
		/* reg: CVUU2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 187;
			_closure_reg(a, c + 0);
		}
		break;
	case 2245: /* NEGI2 */
		_label(LEFT_CHILD(a));
		/* reg: NEGI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 167;
			_closure_reg(a, c + 0);
		}
		break;
	case 2261: /* CALLI2 */
		_label(LEFT_CHILD(a));
		/* reg: CALLI2(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 200;
			_closure_reg(a, c + 0);
		}
		break;
	case 2262: /* CALLU2 */
		_label(LEFT_CHILD(a));
		/* reg: CALLU2(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 201;
			_closure_reg(a, c + 0);
		}
		break;
	case 2277: /* LOADI2 */
		_label(LEFT_CHILD(a));
		/* reg: LOADI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 14;
			_closure_reg(a, c + 0);
		}
		break;
	case 2278: /* LOADU2 */
		_label(LEFT_CHILD(a));
		/* reg: LOADU2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 15;
			_closure_reg(a, c + 0);
		}
		if (	/* reg: LOADU2(LOADU4(reg)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 24;
				_closure_reg(a, c + 0);
			}
		}
		break;
	case 2293: /* RETI2 */
		_label(LEFT_CHILD(a));
		/* reg: RETI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 211;
			_closure_reg(a, c + 0);
		}
		/* reg: RETI2(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 220;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 211;
		}
		/* stmt: RETI2(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 222;
		}
		break;
	case 2294: /* RETU2 */
		_label(LEFT_CHILD(a));
		/* reg: RETU2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 212;
			_closure_reg(a, c + 0);
		}
		/* reg: RETU2(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 221;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETU2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 212;
		}
		/* stmt: RETU2(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 223;
		}
		break;
	case 2357: /* ADDI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 37;
			_closure_reg(a, c + 0);
		}
		break;
	case 2358: /* ADDU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 38;
			_closure_reg(a, c + 0);
		}
		break;
	case 2373: /* SUBI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 48;
			_closure_reg(a, c + 0);
		}
		break;
	case 2374: /* SUBU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 49;
			_closure_reg(a, c + 0);
		}
		break;
	case 2389: /* LSHI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 128;
			_closure_reg(a, c + 0);
		}
		break;
	case 2390: /* LSHU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 129;
			_closure_reg(a, c + 0);
		}
		break;
	case 2405: /* MODI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 112;
			_closure_reg(a, c + 0);
		}
		break;
	case 2406: /* MODU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 113;
			_closure_reg(a, c + 0);
		}
		break;
	case 2421: /* RSHI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 120;
			_closure_reg(a, c + 0);
		}
		break;
	case 2422: /* RSHU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 121;
			_closure_reg(a, c + 0);
		}
		break;
	case 2437: /* BANDI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 136;
			_closure_reg(a, c + 0);
		}
		break;
	case 2438: /* BANDU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 137;
			_closure_reg(a, c + 0);
		}
		break;
	case 2453: /* BCOMI2 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMI2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 160;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMI2(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 57;
		}
		break;
	case 2454: /* BCOMU2 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMU2(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 161;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMU2(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 58;
		}
		break;
	case 2469: /* BORI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 144;
			_closure_reg(a, c + 0);
		}
		break;
	case 2470: /* BORU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 145;
			_closure_reg(a, c + 0);
		}
		break;
	case 2485: /* BXORI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 152;
			_closure_reg(a, c + 0);
		}
		break;
	case 2486: /* BXORU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 153;
			_closure_reg(a, c + 0);
		}
		break;
	case 2501: /* DIVI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 86;
			_closure_reg(a, c + 0);
		}
		break;
	case 2502: /* DIVU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 87;
			_closure_reg(a, c + 0);
		}
		break;
	case 2517: /* MULI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULI2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 60;
			_closure_reg(a, c + 0);
		}
		break;
	case 2518: /* MULU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULU2(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 61;
			_closure_reg(a, c + 0);
		}
		break;
	case 2533: /* EQI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI2(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 65;
		}
		if (	/* stmt: EQI2(BANDI2(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 2437 /* BANDI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 75;
			}
		}
		if (	/* stmt: EQI2(BANDI2(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 2437 /* BANDI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 83;
			}
		}
		/* stmt: EQI2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 91;
		}
		break;
	case 2534: /* EQU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU2(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 66;
		}
		if (	/* stmt: EQU2(BANDU2(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 2438 /* BANDU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 76;
			}
		}
		if (	/* stmt: EQU2(BANDU2(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 2438 /* BANDU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 84;
			}
		}
		/* stmt: EQU2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 92;
		}
		break;
	case 2549: /* GEI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 137;
		}
		break;
	case 2550: /* GEU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 138;
		}
		break;
	case 2565: /* GTI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 147;
		}
		break;
	case 2566: /* GTU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 148;
		}
		break;
	case 2581: /* LEI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 157;
		}
		break;
	case 2582: /* LEU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 158;
		}
		break;
	case 2597: /* LTI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 167;
		}
		/* stmt: LTI2(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 177;
		}
		break;
	case 2598: /* LTU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 168;
		}
		/* stmt: LTU2(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 178;
		}
		break;
	case 2613: /* NEI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI2(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 101;
		}
		if (	/* stmt: NEI2(BANDI2(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 2437 /* BANDI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 111;
			}
		}
		if (	/* stmt: NEI2(BANDI2(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 2437 /* BANDI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 119;
			}
		}
		/* stmt: NEI2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 127;
		}
		break;
	case 2614: /* NEU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU2(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 102;
		}
		if (	/* stmt: NEU2(BANDU2(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 2438 /* BANDU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 112;
			}
		}
		if (	/* stmt: NEU2(BANDU2(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 2438 /* BANDU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 120;
			}
		}
		/* stmt: NEU2(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 128;
		}
		break;
	case 4113: /* CNSTF4 */
		/* con4: CNSTF4 */
		if (0 + 0 < p->cost[_con4_NT]) {
			p->cost[_con4_NT] = 0 + 0;
			p->rule._con4 = 4;
			_closure_con4(a, 0 + 0);
		}
		break;
	case 4117: /* CNSTI4 */
		/* con4: CNSTI4 */
		if (0 + 0 < p->cost[_con4_NT]) {
			p->cost[_con4_NT] = 0 + 0;
			p->rule._con4 = 1;
			_closure_con4(a, 0 + 0);
		}
		/* conv0: CNSTI4 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 3;
		}
		/* conv1: CNSTI4 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 3;
		}
		/* conv2: CNSTI4 */
		c = (range(a, 2, 2));
		if (c + 0 < p->cost[_conv2_NT]) {
			p->cost[_conv2_NT] = c + 0;
			p->rule._conv2 = 1;
		}
		/* conv4: CNSTI4 */
		c = (range(a, 4, 4));
		if (c + 0 < p->cost[_conv4_NT]) {
			p->cost[_conv4_NT] = c + 0;
			p->rule._conv4 = 1;
		}
		/* conv8: CNSTI4 */
		c = (range(a, 8, 8));
		if (c + 0 < p->cost[_conv8_NT]) {
			p->cost[_conv8_NT] = c + 0;
			p->rule._conv8 = 1;
		}
		/* conv16: CNSTI4 */
		c = (range(a, 16, 16));
		if (c + 0 < p->cost[_conv16_NT]) {
			p->cost[_conv16_NT] = c + 0;
			p->rule._conv16 = 1;
		}
		/* conv32: CNSTI4 */
		c = (range(a, 32, 32));
		if (c + 0 < p->cost[_conv32_NT]) {
			p->cost[_conv32_NT] = c + 0;
			p->rule._conv32 = 1;
		}
		/* conv64: CNSTI4 */
		c = (range(a, 64, 64));
		if (c + 0 < p->cost[_conv64_NT]) {
			p->cost[_conv64_NT] = c + 0;
			p->rule._conv64 = 1;
		}
		/* conv128: CNSTI4 */
		c = (range(a, 128, 128));
		if (c + 0 < p->cost[_conv128_NT]) {
			p->cost[_conv128_NT] = c + 0;
			p->rule._conv128 = 1;
		}
		/* conv256: CNSTI4 */
		c = (range(a, 256, 256));
		if (c + 0 < p->cost[_conv256_NT]) {
			p->cost[_conv256_NT] = c + 0;
			p->rule._conv256 = 1;
		}
		break;
	case 4118: /* CNSTU4 */
		/* con4: CNSTU4 */
		if (0 + 0 < p->cost[_con4_NT]) {
			p->cost[_con4_NT] = 0 + 0;
			p->rule._con4 = 2;
			_closure_con4(a, 0 + 0);
		}
		/* conv0: CNSTU4 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 7;
		}
		/* conv1: CNSTU4 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 7;
		}
		/* conv2: CNSTU4 */
		c = (range(a, 2, 2));
		if (c + 0 < p->cost[_conv2_NT]) {
			p->cost[_conv2_NT] = c + 0;
			p->rule._conv2 = 2;
		}
		/* conv4: CNSTU4 */
		c = (range(a, 4, 4));
		if (c + 0 < p->cost[_conv4_NT]) {
			p->cost[_conv4_NT] = c + 0;
			p->rule._conv4 = 2;
		}
		/* conv8: CNSTU4 */
		c = (range(a, 8, 8));
		if (c + 0 < p->cost[_conv8_NT]) {
			p->cost[_conv8_NT] = c + 0;
			p->rule._conv8 = 2;
		}
		/* conv16: CNSTU4 */
		c = (range(a, 16, 16));
		if (c + 0 < p->cost[_conv16_NT]) {
			p->cost[_conv16_NT] = c + 0;
			p->rule._conv16 = 2;
		}
		/* conv32: CNSTU4 */
		c = (range(a, 32, 32));
		if (c + 0 < p->cost[_conv32_NT]) {
			p->cost[_conv32_NT] = c + 0;
			p->rule._conv32 = 2;
		}
		/* conv64: CNSTU4 */
		c = (range(a, 64, 64));
		if (c + 0 < p->cost[_conv64_NT]) {
			p->cost[_conv64_NT] = c + 0;
			p->rule._conv64 = 2;
		}
		/* conv128: CNSTU4 */
		c = (range(a, 128, 128));
		if (c + 0 < p->cost[_conv128_NT]) {
			p->cost[_conv128_NT] = c + 0;
			p->rule._conv128 = 2;
		}
		/* conv256: CNSTU4 */
		c = (range(a, 256, 256));
		if (c + 0 < p->cost[_conv256_NT]) {
			p->cost[_conv256_NT] = c + 0;
			p->rule._conv256 = 2;
		}
		break;
	case 4119: /* CNSTP4 */
		/* con4: CNSTP4 */
		if (0 + 0 < p->cost[_con4_NT]) {
			p->cost[_con4_NT] = 0 + 0;
			p->rule._con4 = 3;
			_closure_con4(a, 0 + 0);
		}
		break;
	case 4129: /* ARGF4 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGF4(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 200;
		}
		break;
	case 4133: /* ARGI4 */
		_label(LEFT_CHILD(a));
		if (	/* stmt: ARGI4(LOADI4(am)) */
			LEFT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_am_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 187;
			}
		}
		if (	/* stmt: ARGI4(LOADI8(am)) */
			LEFT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_am_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 192;
			}
		}
		/* stmt: ARGI4(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 196;
		}
		break;
	case 4134: /* ARGU4 */
		_label(LEFT_CHILD(a));
		if (	/* stmt: ARGU4(LOADU4(ad)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_ad_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 188;
			}
		}
		if (	/* stmt: ARGU4(LOADU4(reg)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 189;
			}
		}
		if (	/* stmt: ARGU4(LOADU4(con)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_con_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 190;
			}
		}
		if (	/* stmt: ARGU4(LOADU4(iadr)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_iadr_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 191;
			}
		}
		if (	/* stmt: ARGU4(LOADU8(am)) */
			LEFT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_am_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 193;
			}
		}
		/* stmt: ARGU4(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 197;
		}
		break;
	case 4135: /* ARGP4 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGP4(eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 202;
		}
		/* stmt: ARGP4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 203;
		}
		/* stmt: ARGP4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 204;
		}
		break;
	case 4145: /* ASGNF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNF4(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 9;
			}
		}
		/* stmt: ASGNF4(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 31;
		}
		break;
	case 4149: /* ASGNI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI4(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 6;
			}
		}
		/* stmt: ASGNI4(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 18;
		}
		if (	/* stmt: ASGNI4(adr,ADDI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4405 && /* ADDI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 37;
			}
		}
		if (	/* stmt: ASGNI4(adr,SUBI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4421 && /* SUBI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 41;
			}
		}
		if (	/* stmt: ASGNI4(adr,RSHI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4469 && /* RSHI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 43;
			}
		}
		if (	/* stmt: ASGNI4(adr,LSHI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4437 && /* LSHI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 45;
			}
		}
		if (	/* stmt: ASGNI4(adr,BANDI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4485 && /* BANDI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 47;
			}
		}
		if (	/* stmt: ASGNI4(adr,BORI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4517 && /* BORI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 49;
			}
		}
		if (	/* stmt: ASGNI4(adr,BXORI4(INDIRI4(adr),am)) */
			RIGHT_CHILD(a)->op == 4533 && /* BXORI4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 51;
			}
		}
		break;
	case 4150: /* ASGNU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU4(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 7;
			}
		}
		/* stmt: ASGNU4(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 19;
		}
		if (	/* stmt: ASGNU4(adr,ADDU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4406 && /* ADDU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 38;
			}
		}
		if (	/* stmt: ASGNU4(adr,SUBU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4422 && /* SUBU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 42;
			}
		}
		if (	/* stmt: ASGNU4(adr,RSHU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4470 && /* RSHU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 44;
			}
		}
		if (	/* stmt: ASGNU4(adr,LSHU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4438 && /* LSHU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 46;
			}
		}
		if (	/* stmt: ASGNU4(adr,BANDU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4486 && /* BANDU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 48;
			}
		}
		if (	/* stmt: ASGNU4(adr,BORU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4518 && /* BORU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 50;
			}
		}
		if (	/* stmt: ASGNU4(adr,BXORU4(INDIRU4(adr),am)) */
			RIGHT_CHILD(a)->op == 4534 && /* BXORU4 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 4166 /* INDIRU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 52;
			}
		}
		break;
	case 4151: /* ASGNP4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNP4(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 8;
			}
		}
		if (	/* stmt: ASGNP4(VREGP,ADDRGP4) */
			LEFT_CHILD(a)->op == 711 && /* VREGP */
			RIGHT_CHILD(a)->op == 4359 /* ADDRGP4 */
		) {
			c = 20;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 13;
			}
		}
		/* stmt: ASGNP4(adr,reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 25;
		}
		/* stmt: ASGNP4(adr,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 26;
		}
		/* stmt: ASGNP4(adr,iadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_iadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 27;
		}
		/* stmt: ASGNP4(adr,ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_ad_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 28;
		}
		break;
	case 4161: /* INDIRF4 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRF4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 10;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr4: INDIRF4(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr4_NT]) {
			p->cost[_iadr4_NT] = c + 0;
			p->rule._iadr4 = 4;
			_closure_iadr4(a, c + 0);
		}
		/* eadr4: INDIRF4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr4_NT]) {
			p->cost[_eadr4_NT] = c + 0;
			p->rule._eadr4 = 4;
			_closure_eadr4(a, c + 0);
		}
		break;
	case 4165: /* INDIRI4 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRI4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 5;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr4: INDIRI4(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr4_NT]) {
			p->cost[_iadr4_NT] = c + 0;
			p->rule._iadr4 = 1;
			_closure_iadr4(a, c + 0);
		}
		/* eadr4: INDIRI4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr4_NT]) {
			p->cost[_eadr4_NT] = c + 0;
			p->rule._eadr4 = 1;
			_closure_eadr4(a, c + 0);
		}
		break;
	case 4166: /* INDIRU4 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRU4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 6;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr4: INDIRU4(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr4_NT]) {
			p->cost[_iadr4_NT] = c + 0;
			p->rule._iadr4 = 2;
			_closure_iadr4(a, c + 0);
		}
		/* eadr4: INDIRU4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr4_NT]) {
			p->cost[_eadr4_NT] = c + 0;
			p->rule._eadr4 = 2;
			_closure_eadr4(a, c + 0);
		}
		break;
	case 4167: /* INDIRP4 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRP4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 7;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr4: INDIRP4(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr4_NT]) {
			p->cost[_iadr4_NT] = c + 0;
			p->rule._iadr4 = 3;
			_closure_iadr4(a, c + 0);
		}
		/* eadr4: INDIRP4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr4_NT]) {
			p->cost[_eadr4_NT] = c + 0;
			p->rule._eadr4 = 3;
			_closure_eadr4(a, c + 0);
		}
		break;
	case 4209: /* CVFF4 */
		_label(LEFT_CHILD(a));
		/* reg: CVFF4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 194;
			_closure_reg(a, c + 0);
		}
		break;
	case 4213: /* CVFI4 */
		_label(LEFT_CHILD(a));
		/* reg: CVFI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 197;
			_closure_reg(a, c + 0);
		}
		break;
	case 4225: /* CVIF4 */
		_label(LEFT_CHILD(a));
		/* reg: CVIF4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 182;
			_closure_reg(a, c + 0);
		}
		break;
	case 4229: /* CVII4 */
		_label(LEFT_CHILD(a));
		/* reg: CVII4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 178;
			_closure_reg(a, c + 0);
		}
		break;
	case 4230: /* CVIU4 */
		_label(LEFT_CHILD(a));
		/* reg: CVIU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 179;
			_closure_reg(a, c + 0);
		}
		break;
	case 4246: /* CVPU4 */
		_label(LEFT_CHILD(a));
		/* reg: CVPU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + (move(a));
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 172;
			_closure_reg(a, c + 0);
		}
		break;
	case 4247: /* CVPP4 */
		break;
	case 4273: /* CVUF4 */
		_label(LEFT_CHILD(a));
		/* reg: CVUF4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 192;
			_closure_reg(a, c + 0);
		}
		break;
	case 4277: /* CVUI4 */
		_label(LEFT_CHILD(a));
		/* reg: CVUI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 188;
			_closure_reg(a, c + 0);
		}
		break;
	case 4278: /* CVUU4 */
		_label(LEFT_CHILD(a));
		/* reg: CVUU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 189;
			_closure_reg(a, c + 0);
		}
		break;
	case 4279: /* CVUP4 */
		_label(LEFT_CHILD(a));
		/* reg: CVUP4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + (move(a));
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 173;
			_closure_reg(a, c + 0);
		}
		break;
	case 4289: /* NEGF4 */
		_label(LEFT_CHILD(a));
		/* reg: NEGF4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 170;
			_closure_reg(a, c + 0);
		}
		break;
	case 4293: /* NEGI4 */
		_label(LEFT_CHILD(a));
		/* reg: NEGI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 168;
			_closure_reg(a, c + 0);
		}
		break;
	case 4305: /* CALLF4 */
		_label(LEFT_CHILD(a));
		/* reg: CALLF4(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 207;
			_closure_reg(a, c + 0);
		}
		break;
	case 4309: /* CALLI4 */
		_label(LEFT_CHILD(a));
		/* reg: CALLI4(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 202;
			_closure_reg(a, c + 0);
		}
		break;
	case 4310: /* CALLU4 */
		_label(LEFT_CHILD(a));
		/* reg: CALLU4(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 203;
			_closure_reg(a, c + 0);
		}
		break;
	case 4311: /* CALLP4 */
		_label(LEFT_CHILD(a));
		/* reg: CALLP4(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 204;
			_closure_reg(a, c + 0);
		}
		break;
	case 4321: /* LOADF4 */
		break;
	case 4325: /* LOADI4 */
		_label(LEFT_CHILD(a));
		/* reg: LOADI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 16;
			_closure_reg(a, c + 0);
		}
		/* reg: LOADI4(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 21;
			_closure_reg(a, c + 0);
		}
		break;
	case 4326: /* LOADU4 */
		_label(LEFT_CHILD(a));
		/* reg: LOADU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 17;
			_closure_reg(a, c + 0);
		}
		/* reg: LOADU4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 22;
			_closure_reg(a, c + 0);
		}
		if (	/* reg: LOADU4(INDIRU8(ad)) */
			LEFT_CHILD(a)->op == 8262 /* INDIRU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_ad_NT] + 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 23;
				_closure_reg(a, c + 0);
			}
		}
		break;
	case 4327: /* LOADP4 */
		_label(LEFT_CHILD(a));
		/* reg: LOADP4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 18;
			_closure_reg(a, c + 0);
		}
		break;
	case 4337: /* RETF4 */
		_label(LEFT_CHILD(a));
		/* stmt: RETF4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 218;
		}
		/* stmt: RETF4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 229;
		}
		break;
	case 4341: /* RETI4 */
		_label(LEFT_CHILD(a));
		/* reg: RETI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 213;
			_closure_reg(a, c + 0);
		}
		/* reg: RETI4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 222;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETI4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 206;
		}
		/* stmt: RETI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 213;
		}
		/* stmt: RETI4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 224;
		}
		break;
	case 4342: /* RETU4 */
		_label(LEFT_CHILD(a));
		/* reg: RETU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 214;
			_closure_reg(a, c + 0);
		}
		/* reg: RETU4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 223;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETU4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 207;
		}
		/* stmt: RETU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 214;
		}
		/* stmt: RETU4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 225;
		}
		break;
	case 4343: /* RETP4 */
		_label(LEFT_CHILD(a));
		/* reg: RETP4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 215;
			_closure_reg(a, c + 0);
		}
		/* reg: RETP4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 224;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETP4(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 208;
		}
		/* stmt: RETP4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 215;
		}
		/* stmt: RETP4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 226;
		}
		break;
	case 4359: /* ADDRGP4 */
		/* ca: ADDRGP4 */
		if (0 + 0 < p->cost[_ca_NT]) {
			p->cost[_ca_NT] = 0 + 0;
			p->rule._ca = 1;
		}
		/* ad: ADDRGP4 */
		if (0 + 0 < p->cost[_ad_NT]) {
			p->cost[_ad_NT] = 0 + 0;
			p->rule._ad = 1;
			_closure_ad(a, 0 + 0);
		}
		/* dm: ADDRGP4 */
		if (0 + 0 < p->cost[_dm_NT]) {
			p->cost[_dm_NT] = 0 + 0;
			p->rule._dm = 3;
		}
		/* am: ADDRGP4 */
		if (0 + 0 < p->cost[_am_NT]) {
			p->cost[_am_NT] = 0 + 0;
			p->rule._am = 4;
		}
		break;
	case 4375: /* ADDRFP4 */
		/* ad: ADDRFP4 */
		if (0 + 0 < p->cost[_ad_NT]) {
			p->cost[_ad_NT] = 0 + 0;
			p->rule._ad = 2;
			_closure_ad(a, 0 + 0);
		}
		/* dm: ADDRFP4 */
		if (0 + 0 < p->cost[_dm_NT]) {
			p->cost[_dm_NT] = 0 + 0;
			p->rule._dm = 4;
		}
		/* am: ADDRFP4 */
		if (0 + 0 < p->cost[_am_NT]) {
			p->cost[_am_NT] = 0 + 0;
			p->rule._am = 5;
		}
		break;
	case 4391: /* ADDRLP4 */
		/* ad: ADDRLP4 */
		if (0 + 0 < p->cost[_ad_NT]) {
			p->cost[_ad_NT] = 0 + 0;
			p->rule._ad = 3;
			_closure_ad(a, 0 + 0);
		}
		/* dm: ADDRLP4 */
		if (0 + 0 < p->cost[_dm_NT]) {
			p->cost[_dm_NT] = 0 + 0;
			p->rule._dm = 5;
		}
		/* am: ADDRLP4 */
		if (0 + 0 < p->cost[_am_NT]) {
			p->cost[_am_NT] = 0 + 0;
			p->rule._am = 6;
		}
		break;
	case 4401: /* ADDF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDF4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 43;
			_closure_reg(a, c + 0);
		}
		break;
	case 4405: /* ADDI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 39;
			_closure_reg(a, c + 0);
		}
		break;
	case 4406: /* ADDU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 40;
			_closure_reg(a, c + 0);
		}
		break;
	case 4407: /* ADDP4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDP4(reg,ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_ca_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 45;
			_closure_reg(a, c + 0);
		}
		break;
	case 4417: /* SUBF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBF4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 55;
			_closure_reg(a, c + 0);
		}
		break;
	case 4421: /* SUBI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 50;
			_closure_reg(a, c + 0);
		}
		break;
	case 4422: /* SUBU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 51;
			_closure_reg(a, c + 0);
		}
		break;
	case 4423: /* SUBP4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBP4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 52;
			_closure_reg(a, c + 0);
		}
		break;
	case 4437: /* LSHI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 130;
			_closure_reg(a, c + 0);
		}
		break;
	case 4438: /* LSHU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 131;
			_closure_reg(a, c + 0);
		}
		break;
	case 4453: /* MODI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 114;
			_closure_reg(a, c + 0);
		}
		break;
	case 4454: /* MODU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 115;
			_closure_reg(a, c + 0);
		}
		break;
	case 4469: /* RSHI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 122;
			_closure_reg(a, c + 0);
		}
		break;
	case 4470: /* RSHU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 123;
			_closure_reg(a, c + 0);
		}
		break;
	case 4485: /* BANDI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 138;
			_closure_reg(a, c + 0);
		}
		break;
	case 4486: /* BANDU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 139;
			_closure_reg(a, c + 0);
		}
		break;
	case 4501: /* BCOMI4 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMI4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 162;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMI4(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 59;
		}
		break;
	case 4502: /* BCOMU4 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMU4(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 163;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMU4(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 60;
		}
		break;
	case 4517: /* BORI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 146;
			_closure_reg(a, c + 0);
		}
		break;
	case 4518: /* BORU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 147;
			_closure_reg(a, c + 0);
		}
		break;
	case 4533: /* BXORI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 154;
			_closure_reg(a, c + 0);
		}
		break;
	case 4534: /* BXORU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 155;
			_closure_reg(a, c + 0);
		}
		break;
	case 4545: /* DIVF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVF4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 92;
			_closure_reg(a, c + 0);
		}
		break;
	case 4549: /* DIVI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 88;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv2_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 94;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv4_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 95;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv8_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 96;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv16_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 97;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv32_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 98;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv64_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 99;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv128_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 100;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVI4(reg,conv256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv256_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 101;
			_closure_reg(a, c + 0);
		}
		break;
	case 4550: /* DIVU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 89;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv2_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 102;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv4_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 103;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv8_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 104;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv16_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 105;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv32_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 106;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv64_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 107;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv128_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 108;
			_closure_reg(a, c + 0);
		}
		/* reg: DIVU4(reg,conv256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv256_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 109;
			_closure_reg(a, c + 0);
		}
		break;
	case 4561: /* MULF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULF4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 66;
			_closure_reg(a, c + 0);
		}
		break;
	case 4565: /* MULI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULI4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 62;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv2_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 68;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv4_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 69;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv8_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 70;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv16_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 71;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv32_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 72;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv64_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 73;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv128_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 74;
			_closure_reg(a, c + 0);
		}
		/* reg: MULI4(reg,conv256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv256_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 75;
			_closure_reg(a, c + 0);
		}
		break;
	case 4566: /* MULU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* reg: MULU4(con,LOADU4(eadr)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_eadr_NT] + 2;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 57;
				_closure_reg(a, c + 0);
			}
		}
		/* reg: MULU4(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 63;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv2_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 76;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv4_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 77;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv8_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 78;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv16_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 79;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv32_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 80;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv64_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 81;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv128_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 82;
			_closure_reg(a, c + 0);
		}
		/* reg: MULU4(reg,conv256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv256_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 83;
			_closure_reg(a, c + 0);
		}
		break;
	case 4577: /* EQF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQF4(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 71;
		}
		/* stmt: EQF4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 97;
		}
		break;
	case 4581: /* EQI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI4(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 67;
		}
		if (	/* stmt: EQI4(BANDI4(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 4485 /* BANDI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 77;
			}
		}
		if (	/* stmt: EQI4(BANDI4(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 4485 /* BANDI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 85;
			}
		}
		/* stmt: EQI4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 93;
		}
		break;
	case 4582: /* EQU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU4(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 68;
		}
		if (	/* stmt: EQU4(BANDU4(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 4486 /* BANDU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 78;
			}
		}
		if (	/* stmt: EQU4(BANDU4(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 4486 /* BANDU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 86;
			}
		}
		/* stmt: EQU4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 94;
		}
		break;
	case 4593: /* GEF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEF4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 143;
		}
		break;
	case 4597: /* GEI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 139;
		}
		break;
	case 4598: /* GEU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 140;
		}
		break;
	case 4609: /* GTF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTF4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 153;
		}
		break;
	case 4613: /* GTI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 149;
		}
		break;
	case 4614: /* GTU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 150;
		}
		break;
	case 4625: /* LEF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEF4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 163;
		}
		break;
	case 4629: /* LEI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 159;
		}
		break;
	case 4630: /* LEU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 160;
		}
		break;
	case 4641: /* LTF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTF4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 173;
		}
		/* stmt: LTF4(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 183;
		}
		break;
	case 4645: /* LTI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 169;
		}
		/* stmt: LTI4(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 179;
		}
		break;
	case 4646: /* LTU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 170;
		}
		/* stmt: LTU4(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 180;
		}
		break;
	case 4657: /* NEF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEF4(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 107;
		}
		/* stmt: NEF4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 133;
		}
		break;
	case 4661: /* NEI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI4(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 103;
		}
		if (	/* stmt: NEI4(BANDI4(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 4485 /* BANDI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 113;
			}
		}
		if (	/* stmt: NEI4(BANDI4(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 4485 /* BANDI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 121;
			}
		}
		/* stmt: NEI4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 129;
		}
		break;
	case 4662: /* NEU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU4(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 104;
		}
		if (	/* stmt: NEU4(BANDU4(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 4486 /* BANDU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 114;
			}
		}
		if (	/* stmt: NEU4(BANDU4(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 4486 /* BANDU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 122;
			}
		}
		/* stmt: NEU4(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 130;
		}
		break;
	case 4870: /* ASMRETU4 */
		/* reg: ASMRETU4 */
		if (1 + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = 1 + 0;
			p->rule._reg = 227;
			_closure_reg(a, 1 + 0);
		}
		/* stmt: ASMRETU4 */
		if (1 + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = 1 + 0;
			p->rule._stmt = 234;
		}
		break;
	case 8209: /* CNSTF8 */
		/* con8: CNSTF8 */
		if (0 + 0 < p->cost[_con8_NT]) {
			p->cost[_con8_NT] = 0 + 0;
			p->rule._con8 = 3;
			_closure_con8(a, 0 + 0);
		}
		break;
	case 8213: /* CNSTI8 */
		/* con8: CNSTI8 */
		if (0 + 0 < p->cost[_con8_NT]) {
			p->cost[_con8_NT] = 0 + 0;
			p->rule._con8 = 1;
			_closure_con8(a, 0 + 0);
		}
		/* conv0: CNSTI8 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 4;
		}
		/* conv1: CNSTI8 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 4;
		}
		break;
	case 8214: /* CNSTU8 */
		/* con8: CNSTU8 */
		if (0 + 0 < p->cost[_con8_NT]) {
			p->cost[_con8_NT] = 0 + 0;
			p->rule._con8 = 2;
			_closure_con8(a, 0 + 0);
		}
		/* conv0: CNSTU8 */
		c = (range(a,0,0));
		if (c + 0 < p->cost[_conv0_NT]) {
			p->cost[_conv0_NT] = c + 0;
			p->rule._conv0 = 8;
		}
		/* conv1: CNSTU8 */
		c = (range(a,1,1));
		if (c + 0 < p->cost[_conv1_NT]) {
			p->cost[_conv1_NT] = c + 0;
			p->rule._conv1 = 8;
		}
		break;
	case 8225: /* ARGF8 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGF8(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 201;
		}
		break;
	case 8229: /* ARGI8 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGI8(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 198;
		}
		break;
	case 8230: /* ARGU8 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGU8(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 199;
		}
		break;
	case 8241: /* ASGNF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNF8(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 12;
			}
		}
		/* stmt: ASGNF8(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 32;
		}
		break;
	case 8245: /* ASGNI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI8(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 10;
			}
		}
		/* stmt: ASGNI8(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 29;
		}
		if (	/* stmt: ASGNI8(adr,ADDI8(INDIRI8(adr),am)) */
			RIGHT_CHILD(a)->op == 8501 && /* ADDI8 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 39;
			}
		}
		break;
	case 8246: /* ASGNU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU8(VREGP,reg) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 11;
			}
		}
		/* stmt: ASGNU8(adr,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 30;
		}
		if (	/* stmt: ASGNU8(adr,ADDU8(INDIRU8(adr),am)) */
			RIGHT_CHILD(a)->op == 8502 && /* ADDU8 */
			LEFT_CHILD(RIGHT_CHILD(a))->op == 8262 /* INDIRU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + ((struct _state *)(LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(a)))->x.state))->cost[_adr_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_am_NT] + (memop(a));
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 40;
			}
		}
		break;
	case 8257: /* INDIRF8 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRF8(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 11;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr8: INDIRF8(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr8_NT]) {
			p->cost[_iadr8_NT] = c + 0;
			p->rule._iadr8 = 3;
			_closure_iadr8(a, c + 0);
		}
		/* eadr8: INDIRF8(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr8_NT]) {
			p->cost[_eadr8_NT] = c + 0;
			p->rule._eadr8 = 3;
			_closure_eadr8(a, c + 0);
		}
		break;
	case 8261: /* INDIRI8 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRI8(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 8;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr8: INDIRI8(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr8_NT]) {
			p->cost[_iadr8_NT] = c + 0;
			p->rule._iadr8 = 1;
			_closure_iadr8(a, c + 0);
		}
		/* eadr8: INDIRI8(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr8_NT]) {
			p->cost[_eadr8_NT] = c + 0;
			p->rule._eadr8 = 1;
			_closure_eadr8(a, c + 0);
		}
		break;
	case 8262: /* INDIRU8 */
		_label(LEFT_CHILD(a));
		if (	/* reg: INDIRU8(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conv0_NT] == 0) {
					p->cost[_conv0_NT] = 0;
					p->rule._conv0 = q->rule._conv0;
				}
				if (q->cost[_conv1_NT] == 0) {
					p->cost[_conv1_NT] = 0;
					p->rule._conv1 = q->rule._conv1;
				}
				if (q->cost[_conv2_NT] == 0) {
					p->cost[_conv2_NT] = 0;
					p->rule._conv2 = q->rule._conv2;
				}
				if (q->cost[_conv4_NT] == 0) {
					p->cost[_conv4_NT] = 0;
					p->rule._conv4 = q->rule._conv4;
				}
				if (q->cost[_conv8_NT] == 0) {
					p->cost[_conv8_NT] = 0;
					p->rule._conv8 = q->rule._conv8;
				}
				if (q->cost[_conv16_NT] == 0) {
					p->cost[_conv16_NT] = 0;
					p->rule._conv16 = q->rule._conv16;
				}
				if (q->cost[_conv32_NT] == 0) {
					p->cost[_conv32_NT] = 0;
					p->rule._conv32 = q->rule._conv32;
				}
				if (q->cost[_conv64_NT] == 0) {
					p->cost[_conv64_NT] = 0;
					p->rule._conv64 = q->rule._conv64;
				}
				if (q->cost[_conv128_NT] == 0) {
					p->cost[_conv128_NT] = 0;
					p->rule._conv128 = q->rule._conv128;
				}
				if (q->cost[_conv256_NT] == 0) {
					p->cost[_conv256_NT] = 0;
					p->rule._conv256 = q->rule._conv256;
				}
				if (q->cost[_reg_NT] == 0) {
					p->cost[_reg_NT] = 0;
					p->rule._reg = q->rule._reg;
				}
				if (q->cost[_adr_NT] == 0) {
					p->cost[_adr_NT] = 0;
					p->rule._adr = q->rule._adr;
				}
				if (q->cost[_ad_NT] == 0) {
					p->cost[_ad_NT] = 0;
					p->rule._ad = q->rule._ad;
				}
				if (q->cost[_ca_NT] == 0) {
					p->cost[_ca_NT] = 0;
					p->rule._ca = q->rule._ca;
				}
				if (q->cost[_iadr1_NT] == 0) {
					p->cost[_iadr1_NT] = 0;
					p->rule._iadr1 = q->rule._iadr1;
				}
				if (q->cost[_iadr2_NT] == 0) {
					p->cost[_iadr2_NT] = 0;
					p->rule._iadr2 = q->rule._iadr2;
				}
				if (q->cost[_iadr4_NT] == 0) {
					p->cost[_iadr4_NT] = 0;
					p->rule._iadr4 = q->rule._iadr4;
				}
				if (q->cost[_iadr8_NT] == 0) {
					p->cost[_iadr8_NT] = 0;
					p->rule._iadr8 = q->rule._iadr8;
				}
				if (q->cost[_iadr_NT] == 0) {
					p->cost[_iadr_NT] = 0;
					p->rule._iadr = q->rule._iadr;
				}
				if (q->cost[_eadr1_NT] == 0) {
					p->cost[_eadr1_NT] = 0;
					p->rule._eadr1 = q->rule._eadr1;
				}
				if (q->cost[_eadr2_NT] == 0) {
					p->cost[_eadr2_NT] = 0;
					p->rule._eadr2 = q->rule._eadr2;
				}
				if (q->cost[_eadr4_NT] == 0) {
					p->cost[_eadr4_NT] = 0;
					p->rule._eadr4 = q->rule._eadr4;
				}
				if (q->cost[_eadr8_NT] == 0) {
					p->cost[_eadr8_NT] = 0;
					p->rule._eadr8 = q->rule._eadr8;
				}
				if (q->cost[_eadr_NT] == 0) {
					p->cost[_eadr_NT] = 0;
					p->rule._eadr = q->rule._eadr;
				}
				if (q->cost[_dm_NT] == 0) {
					p->cost[_dm_NT] = 0;
					p->rule._dm = q->rule._dm;
				}
				if (q->cost[_am_NT] == 0) {
					p->cost[_am_NT] = 0;
					p->rule._am = q->rule._am;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 9;
				_closure_reg(a, c + 0);
			}
		}
		/* iadr8: INDIRU8(adr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_adr_NT] + 0;
		if (c + 0 < p->cost[_iadr8_NT]) {
			p->cost[_iadr8_NT] = c + 0;
			p->rule._iadr8 = 2;
			_closure_iadr8(a, c + 0);
		}
		/* eadr8: INDIRU8(ad) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ad_NT] + 0;
		if (c + 0 < p->cost[_eadr8_NT]) {
			p->cost[_eadr8_NT] = c + 0;
			p->rule._eadr8 = 2;
			_closure_eadr8(a, c + 0);
		}
		break;
	case 8305: /* CVFF8 */
		_label(LEFT_CHILD(a));
		/* reg: CVFF8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 195;
			_closure_reg(a, c + 0);
		}
		if (	/* reg: CVFF8(INDIRF4(ADDP4(dm,am))) */
			LEFT_CHILD(a)->op == 4161 && /* INDIRF4 */
			LEFT_CHILD(LEFT_CHILD(a))->op == 4407 /* ADDP4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(LEFT_CHILD(a)))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(LEFT_CHILD(a)))->x.state))->cost[_am_NT] + 1;
			if (c + 0 < p->cost[_reg_NT]) {
				p->cost[_reg_NT] = c + 0;
				p->rule._reg = 199;
				_closure_reg(a, c + 0);
			}
		}
		break;
	case 8309: /* CVFI8 */
		_label(LEFT_CHILD(a));
		/* reg: CVFI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 198;
			_closure_reg(a, c + 0);
		}
		break;
	case 8321: /* CVIF8 */
		_label(LEFT_CHILD(a));
		/* reg: CVIF8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 183;
			_closure_reg(a, c + 0);
		}
		break;
	case 8325: /* CVII8 */
		_label(LEFT_CHILD(a));
		/* reg: CVII8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 180;
			_closure_reg(a, c + 0);
		}
		break;
	case 8326: /* CVIU8 */
		_label(LEFT_CHILD(a));
		/* reg: CVIU8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 181;
			_closure_reg(a, c + 0);
		}
		break;
	case 8342: /* CVPU8 */
		break;
	case 8369: /* CVUF8 */
		_label(LEFT_CHILD(a));
		/* reg: CVUF8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 193;
			_closure_reg(a, c + 0);
		}
		break;
	case 8373: /* CVUI8 */
		_label(LEFT_CHILD(a));
		/* reg: CVUI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 190;
			_closure_reg(a, c + 0);
		}
		break;
	case 8374: /* CVUU8 */
		_label(LEFT_CHILD(a));
		/* reg: CVUU8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 191;
			_closure_reg(a, c + 0);
		}
		break;
	case 8385: /* NEGF8 */
		_label(LEFT_CHILD(a));
		/* reg: NEGF8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 171;
			_closure_reg(a, c + 0);
		}
		break;
	case 8389: /* NEGI8 */
		_label(LEFT_CHILD(a));
		/* reg: NEGI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 169;
			_closure_reg(a, c + 0);
		}
		break;
	case 8401: /* CALLF8 */
		_label(LEFT_CHILD(a));
		/* reg: CALLF8(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 208;
			_closure_reg(a, c + 0);
		}
		break;
	case 8405: /* CALLI8 */
		_label(LEFT_CHILD(a));
		/* reg: CALLI8(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 205;
			_closure_reg(a, c + 0);
		}
		break;
	case 8406: /* CALLU8 */
		_label(LEFT_CHILD(a));
		/* reg: CALLU8(ca) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_ca_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 206;
			_closure_reg(a, c + 0);
		}
		break;
	case 8417: /* LOADF8 */
		break;
	case 8421: /* LOADI8 */
		_label(LEFT_CHILD(a));
		/* reg: LOADI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 19;
			_closure_reg(a, c + 0);
		}
		break;
	case 8422: /* LOADU8 */
		_label(LEFT_CHILD(a));
		/* reg: LOADU8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 20;
			_closure_reg(a, c + 0);
		}
		break;
	case 8433: /* RETF8 */
		_label(LEFT_CHILD(a));
		/* stmt: RETF8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 219;
		}
		/* stmt: RETF8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 230;
		}
		break;
	case 8437: /* RETI8 */
		_label(LEFT_CHILD(a));
		/* reg: RETI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 216;
			_closure_reg(a, c + 0);
		}
		/* reg: RETI8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 225;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 216;
		}
		/* stmt: RETI8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 227;
		}
		break;
	case 8438: /* RETU8 */
		_label(LEFT_CHILD(a));
		/* reg: RETU8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 217;
			_closure_reg(a, c + 0);
		}
		/* reg: RETU8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 226;
			_closure_reg(a, c + 0);
		}
		/* stmt: RETU8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 217;
		}
		/* stmt: RETU8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 228;
		}
		break;
	case 8497: /* ADDF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDF8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 44;
			_closure_reg(a, c + 0);
		}
		break;
	case 8501: /* ADDI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 41;
			_closure_reg(a, c + 0);
		}
		break;
	case 8502: /* ADDU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: ADDU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 42;
			_closure_reg(a, c + 0);
		}
		break;
	case 8513: /* SUBF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBF8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 56;
			_closure_reg(a, c + 0);
		}
		break;
	case 8517: /* SUBI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 53;
			_closure_reg(a, c + 0);
		}
		break;
	case 8518: /* SUBU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: SUBU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 54;
			_closure_reg(a, c + 0);
		}
		break;
	case 8533: /* LSHI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 132;
			_closure_reg(a, c + 0);
		}
		break;
	case 8534: /* LSHU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: LSHU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 133;
			_closure_reg(a, c + 0);
		}
		break;
	case 8549: /* MODI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 116;
			_closure_reg(a, c + 0);
		}
		break;
	case 8550: /* MODU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MODU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 117;
			_closure_reg(a, c + 0);
		}
		break;
	case 8565: /* RSHI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 124;
			_closure_reg(a, c + 0);
		}
		break;
	case 8566: /* RSHU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: RSHU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 125;
			_closure_reg(a, c + 0);
		}
		break;
	case 8581: /* BANDI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 140;
			_closure_reg(a, c + 0);
		}
		break;
	case 8582: /* BANDU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BANDU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 141;
			_closure_reg(a, c + 0);
		}
		break;
	case 8597: /* BCOMI8 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMI8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 164;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMI8(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 61;
		}
		break;
	case 8598: /* BCOMU8 */
		_label(LEFT_CHILD(a));
		/* reg: BCOMU8(reg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + 1;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 165;
			_closure_reg(a, c + 0);
		}
		/* stmt: BCOMU8(am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 62;
		}
		break;
	case 8613: /* BORI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 148;
			_closure_reg(a, c + 0);
		}
		break;
	case 8614: /* BORU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BORU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 149;
			_closure_reg(a, c + 0);
		}
		break;
	case 8629: /* BXORI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 156;
			_closure_reg(a, c + 0);
		}
		break;
	case 8630: /* BXORU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: BXORU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 2;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 157;
			_closure_reg(a, c + 0);
		}
		break;
	case 8641: /* DIVF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVF8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 93;
			_closure_reg(a, c + 0);
		}
		break;
	case 8645: /* DIVI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 90;
			_closure_reg(a, c + 0);
		}
		break;
	case 8646: /* DIVU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: DIVU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 91;
			_closure_reg(a, c + 0);
		}
		break;
	case 8657: /* MULF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULF8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 67;
			_closure_reg(a, c + 0);
		}
		break;
	case 8661: /* MULI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULI8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 64;
			_closure_reg(a, c + 0);
		}
		break;
	case 8662: /* MULU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg: MULU8(reg,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 3;
		if (c + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = c + 0;
			p->rule._reg = 65;
			_closure_reg(a, c + 0);
		}
		break;
	case 8673: /* EQF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQF8(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 72;
		}
		/* stmt: EQF8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 98;
		}
		break;
	case 8677: /* EQI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI8(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 69;
		}
		if (	/* stmt: EQI8(BANDI8(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 8581 /* BANDI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 79;
			}
		}
		if (	/* stmt: EQI8(BANDI8(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 8581 /* BANDI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 87;
			}
		}
		/* stmt: EQI8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 95;
		}
		break;
	case 8678: /* EQU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU8(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 70;
		}
		if (	/* stmt: EQU8(BANDU8(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 8582 /* BANDU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 80;
			}
		}
		if (	/* stmt: EQU8(BANDU8(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 8582 /* BANDU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 88;
			}
		}
		/* stmt: EQU8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 96;
		}
		break;
	case 8689: /* GEF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEF8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 144;
		}
		break;
	case 8693: /* GEI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 141;
		}
		break;
	case 8694: /* GEU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 142;
		}
		break;
	case 8705: /* GTF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTF8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 154;
		}
		break;
	case 8709: /* GTI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 151;
		}
		break;
	case 8710: /* GTU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 152;
		}
		break;
	case 8721: /* LEF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEF8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 164;
		}
		break;
	case 8725: /* LEI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 161;
		}
		break;
	case 8726: /* LEU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 162;
		}
		break;
	case 8737: /* LTF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTF8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 174;
		}
		/* stmt: LTF8(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 184;
		}
		break;
	case 8741: /* LTI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 171;
		}
		/* stmt: LTI8(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 181;
		}
		break;
	case 8742: /* LTU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 172;
		}
		/* stmt: LTU8(eadr,eadr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_eadr_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_eadr_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 182;
		}
		break;
	case 8753: /* NEF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEF8(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 108;
		}
		/* stmt: NEF8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 134;
		}
		break;
	case 8757: /* NEI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI8(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 105;
		}
		if (	/* stmt: NEI8(BANDI8(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 8581 /* BANDI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 115;
			}
		}
		if (	/* stmt: NEI8(BANDI8(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 8581 /* BANDI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 123;
			}
		}
		/* stmt: NEI8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 131;
		}
		break;
	case 8758: /* NEU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU8(dm,conv0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 106;
		}
		if (	/* stmt: NEU8(BANDU8(dm,conv1),conv0) */
			LEFT_CHILD(a)->op == 8582 /* BANDU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv0_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 116;
			}
		}
		if (	/* stmt: NEU8(BANDU8(dm,conv1),conv1) */
			LEFT_CHILD(a)->op == 8582 /* BANDU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(a))->x.state))->cost[_conv1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conv1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 124;
			}
		}
		/* stmt: NEU8(dm,am) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dm_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_am_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 132;
		}
		break;
	case 8966: /* ASMRETU8 */
		/* reg: ASMRETU8 */
		if (1 + 0 < p->cost[_reg_NT]) {
			p->cost[_reg_NT] = 1 + 0;
			p->rule._reg = 228;
			_closure_reg(a, 1 + 0);
		}
		/* stmt: ASMRETU8 */
		if (1 + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = 1 + 0;
			p->rule._stmt = 235;
		}
		break;
	default:
		fatal("_label", "Bad terminal %d\n", OP_LABEL(a));
	}
}

static void _kids(NODEPTR_TYPE p, int eruleno, NODEPTR_TYPE kids[]) {
	if (!p)
		fatal("_kids", "Null tree\n", 0);
	if (!kids)
		fatal("_kids", "Null kids\n", 0);
	switch (eruleno) {
	case 560: /* stmt: ASMRETU8 */
	case 559: /* stmt: ASMRETU4 */
	case 558: /* reg: ASMRETU8 */
	case 557: /* reg: ASMRETU4 */
	case 556: /* stmt: LABELV */
	case 143: /* stmt: ASGNP4(VREGP,ADDRGP4) */
	case 131: /* am: ADDRLP4 */
	case 130: /* am: ADDRFP4 */
	case 129: /* am: ADDRGP4 */
	case 125: /* dm: ADDRLP4 */
	case 124: /* dm: ADDRFP4 */
	case 123: /* dm: ADDRGP4 */
	case 81: /* ad: ADDRLP4 */
	case 80: /* ad: ADDRFP4 */
	case 79: /* ad: ADDRGP4 */
	case 77: /* ca: ADDRGP4 */
	case 58: /* reg: INDIRF8(VREGP) */
	case 57: /* reg: INDIRF4(VREGP) */
	case 56: /* reg: INDIRU8(VREGP) */
	case 55: /* reg: INDIRI8(VREGP) */
	case 54: /* reg: INDIRP4(VREGP) */
	case 53: /* reg: INDIRU4(VREGP) */
	case 52: /* reg: INDIRI4(VREGP) */
	case 51: /* reg: INDIRU2(VREGP) */
	case 50: /* reg: INDIRI2(VREGP) */
	case 49: /* reg: INDIRU1(VREGP) */
	case 48: /* reg: INDIRI1(VREGP) */
	case 47: /* conv256: CNSTU4 */
	case 46: /* conv256: CNSTI4 */
	case 45: /* conv128: CNSTU4 */
	case 44: /* conv128: CNSTI4 */
	case 43: /* conv64: CNSTU4 */
	case 42: /* conv64: CNSTI4 */
	case 41: /* conv32: CNSTU4 */
	case 40: /* conv32: CNSTI4 */
	case 39: /* conv16: CNSTU4 */
	case 38: /* conv16: CNSTI4 */
	case 37: /* conv8: CNSTU4 */
	case 36: /* conv8: CNSTI4 */
	case 35: /* conv4: CNSTU4 */
	case 34: /* conv4: CNSTI4 */
	case 33: /* conv2: CNSTU4 */
	case 32: /* conv2: CNSTI4 */
	case 31: /* conv1: CNSTU8 */
	case 30: /* conv1: CNSTU4 */
	case 29: /* conv1: CNSTU2 */
	case 28: /* conv1: CNSTU1 */
	case 27: /* conv1: CNSTI8 */
	case 26: /* conv1: CNSTI4 */
	case 25: /* conv1: CNSTI2 */
	case 24: /* conv1: CNSTI1 */
	case 23: /* conv0: CNSTU8 */
	case 22: /* conv0: CNSTU4 */
	case 21: /* conv0: CNSTU2 */
	case 20: /* conv0: CNSTU1 */
	case 19: /* conv0: CNSTI8 */
	case 18: /* conv0: CNSTI4 */
	case 17: /* conv0: CNSTI2 */
	case 16: /* conv0: CNSTI1 */
	case 11: /* con8: CNSTF8 */
	case 10: /* con4: CNSTF4 */
	case 9: /* con8: CNSTU8 */
	case 8: /* con8: CNSTI8 */
	case 7: /* con4: CNSTP4 */
	case 6: /* con4: CNSTU4 */
	case 5: /* con4: CNSTI4 */
	case 4: /* con2: CNSTU2 */
	case 3: /* con2: CNSTI2 */
	case 2: /* con1: CNSTU1 */
	case 1: /* con1: CNSTI1 */
		break;
	case 128: /* am: iadr */
	case 127: /* am: con */
	case 126: /* am: reg */
	case 122: /* dm: iadr */
	case 121: /* dm: reg */
	case 120: /* reg: ad */
	case 119: /* reg: iadr8 */
	case 118: /* reg: iadr4 */
	case 117: /* reg: iadr2 */
	case 116: /* reg: iadr1 */
	case 114: /* eadr: eadr8 */
	case 113: /* eadr: eadr4 */
	case 112: /* eadr: eadr2 */
	case 111: /* eadr: eadr1 */
	case 98: /* iadr: iadr8 */
	case 97: /* iadr: iadr4 */
	case 96: /* iadr: iadr2 */
	case 95: /* iadr: iadr1 */
	case 83: /* adr: ad */
	case 82: /* adr: reg */
	case 78: /* ca: reg */
	case 76: /* stmt: reg */
	case 75: /* reg: con8 */
	case 74: /* reg: con4 */
	case 73: /* reg: con2 */
	case 72: /* reg: con1 */
	case 15: /* con: con8 */
	case 14: /* con: con4 */
	case 13: /* con: con2 */
	case 12: /* con: con1 */
		kids[0] = p;
		break;
	case 555: /* stmt: JUMPV(adr) */
	case 554: /* stmt: JUMPV(am) */
	case 553: /* stmt: RETF8(con) */
	case 552: /* stmt: RETF4(con) */
	case 551: /* stmt: RETU8(con) */
	case 550: /* stmt: RETI8(con) */
	case 549: /* stmt: RETP4(con) */
	case 548: /* stmt: RETU4(con) */
	case 547: /* stmt: RETI4(con) */
	case 546: /* stmt: RETU2(con) */
	case 545: /* stmt: RETI2(con) */
	case 544: /* stmt: RETU1(con) */
	case 543: /* stmt: RETI1(con) */
	case 542: /* stmt: RETF8(reg) */
	case 541: /* stmt: RETF4(reg) */
	case 540: /* stmt: RETU8(reg) */
	case 539: /* stmt: RETI8(reg) */
	case 538: /* stmt: RETP4(reg) */
	case 537: /* stmt: RETU4(reg) */
	case 536: /* stmt: RETI4(reg) */
	case 535: /* stmt: RETU2(reg) */
	case 534: /* stmt: RETI2(reg) */
	case 533: /* stmt: RETU1(reg) */
	case 532: /* stmt: RETI1(reg) */
	case 531: /* stmt: RETP4(ad) */
	case 530: /* stmt: RETU4(ad) */
	case 529: /* stmt: RETI4(ad) */
	case 528: /* reg: RETU8(con) */
	case 527: /* reg: RETI8(con) */
	case 526: /* reg: RETP4(con) */
	case 525: /* reg: RETU4(con) */
	case 524: /* reg: RETI4(con) */
	case 523: /* reg: RETU2(con) */
	case 522: /* reg: RETI2(con) */
	case 521: /* reg: RETU1(con) */
	case 520: /* reg: RETI1(con) */
	case 519: /* reg: RETU8(reg) */
	case 518: /* reg: RETI8(reg) */
	case 517: /* reg: RETP4(reg) */
	case 516: /* reg: RETU4(reg) */
	case 515: /* reg: RETI4(reg) */
	case 514: /* reg: RETU2(reg) */
	case 513: /* reg: RETI2(reg) */
	case 512: /* reg: RETU1(reg) */
	case 511: /* reg: RETI1(reg) */
	case 510: /* stmt: CALLV(ca) */
	case 509: /* reg: CALLF8(ca) */
	case 508: /* reg: CALLF4(ca) */
	case 507: /* reg: CALLU8(ca) */
	case 506: /* reg: CALLI8(ca) */
	case 505: /* reg: CALLP4(ca) */
	case 504: /* reg: CALLU4(ca) */
	case 503: /* reg: CALLI4(ca) */
	case 502: /* reg: CALLU2(ca) */
	case 501: /* reg: CALLI2(ca) */
	case 500: /* stmt: ARGP4(reg) */
	case 499: /* stmt: ARGP4(ad) */
	case 498: /* stmt: ARGP4(eadr) */
	case 497: /* stmt: ARGF8(am) */
	case 496: /* stmt: ARGF4(am) */
	case 495: /* stmt: ARGU8(am) */
	case 494: /* stmt: ARGI8(am) */
	case 493: /* stmt: ARGU4(am) */
	case 492: /* stmt: ARGI4(am) */
	case 491: /* stmt: ARGU2(am) */
	case 490: /* stmt: ARGI2(am) */
	case 479: /* reg: CVFI8(reg) */
	case 478: /* reg: CVFI4(reg) */
	case 477: /* reg: CVFI2(reg) */
	case 476: /* reg: CVFF8(reg) */
	case 475: /* reg: CVFF4(reg) */
	case 474: /* reg: CVUF8(reg) */
	case 473: /* reg: CVUF4(reg) */
	case 472: /* reg: CVUU8(reg) */
	case 471: /* reg: CVUI8(reg) */
	case 470: /* reg: CVUU4(reg) */
	case 469: /* reg: CVUI4(reg) */
	case 468: /* reg: CVUU2(reg) */
	case 467: /* reg: CVUI2(reg) */
	case 466: /* reg: CVUU1(reg) */
	case 465: /* reg: CVUI1(reg) */
	case 464: /* reg: CVIF8(reg) */
	case 463: /* reg: CVIF4(reg) */
	case 462: /* reg: CVIU8(reg) */
	case 461: /* reg: CVII8(reg) */
	case 460: /* reg: CVIU4(reg) */
	case 459: /* reg: CVII4(reg) */
	case 458: /* reg: CVIU2(reg) */
	case 457: /* reg: CVII2(reg) */
	case 456: /* reg: CVIU1(reg) */
	case 455: /* reg: CVII1(reg) */
	case 454: /* reg: CVUP4(reg) */
	case 453: /* reg: CVPU4(reg) */
	case 330: /* reg: NEGF8(reg) */
	case 329: /* reg: NEGF4(reg) */
	case 328: /* reg: NEGI8(reg) */
	case 327: /* reg: NEGI4(reg) */
	case 326: /* reg: NEGI2(reg) */
	case 325: /* reg: NEGI1(reg) */
	case 324: /* stmt: BCOMU8(am) */
	case 323: /* stmt: BCOMI8(am) */
	case 322: /* stmt: BCOMU4(am) */
	case 321: /* stmt: BCOMI4(am) */
	case 320: /* stmt: BCOMU2(am) */
	case 319: /* stmt: BCOMI2(am) */
	case 318: /* stmt: BCOMU1(am) */
	case 317: /* stmt: BCOMI1(am) */
	case 316: /* reg: BCOMU8(reg) */
	case 315: /* reg: BCOMI8(reg) */
	case 314: /* reg: BCOMU4(reg) */
	case 313: /* reg: BCOMI4(reg) */
	case 312: /* reg: BCOMU2(reg) */
	case 311: /* reg: BCOMI2(reg) */
	case 310: /* reg: BCOMU1(reg) */
	case 309: /* reg: BCOMI1(reg) */
	case 185: /* stmt: ARGB(iadr) */
	case 115: /* eadr: INDIRB(ad) */
	case 110: /* eadr8: INDIRF8(ad) */
	case 109: /* eadr4: INDIRF4(ad) */
	case 108: /* eadr8: INDIRU8(ad) */
	case 107: /* eadr8: INDIRI8(ad) */
	case 106: /* eadr4: INDIRP4(ad) */
	case 105: /* eadr4: INDIRU4(ad) */
	case 104: /* eadr4: INDIRI4(ad) */
	case 103: /* eadr2: INDIRU2(ad) */
	case 102: /* eadr2: INDIRI2(ad) */
	case 101: /* eadr1: INDIRU1(ad) */
	case 100: /* eadr1: INDIRI1(ad) */
	case 99: /* iadr: INDIRB(adr) */
	case 94: /* iadr8: INDIRF8(adr) */
	case 93: /* iadr4: INDIRF4(adr) */
	case 92: /* iadr8: INDIRU8(adr) */
	case 91: /* iadr8: INDIRI8(adr) */
	case 90: /* iadr4: INDIRP4(adr) */
	case 89: /* iadr4: INDIRU4(adr) */
	case 88: /* iadr4: INDIRI4(adr) */
	case 87: /* iadr2: INDIRU2(adr) */
	case 86: /* iadr2: INDIRI2(adr) */
	case 85: /* iadr1: INDIRU1(adr) */
	case 84: /* iadr1: INDIRI1(adr) */
	case 69: /* reg: LOADU4(ad) */
	case 68: /* reg: LOADI4(adr) */
	case 67: /* reg: LOADU8(reg) */
	case 66: /* reg: LOADI8(reg) */
	case 65: /* reg: LOADP4(reg) */
	case 64: /* reg: LOADU4(reg) */
	case 63: /* reg: LOADI4(reg) */
	case 62: /* reg: LOADU2(reg) */
	case 61: /* reg: LOADI2(reg) */
	case 60: /* reg: LOADU1(reg) */
	case 59: /* reg: LOADI1(reg) */
		kids[0] = LEFT_CHILD(p);
		break;
	case 489: /* stmt: ARGU4(LOADU8(am)) */
	case 488: /* stmt: ARGI4(LOADI8(am)) */
	case 487: /* stmt: ARGU4(LOADU4(iadr)) */
	case 486: /* stmt: ARGU4(LOADU4(con)) */
	case 485: /* stmt: ARGU4(LOADU4(reg)) */
	case 484: /* stmt: ARGU4(LOADU4(ad)) */
	case 483: /* stmt: ARGI4(LOADI4(am)) */
	case 482: /* stmt: ARGU2(LOADU4(am)) */
	case 481: /* stmt: ARGI2(LOADI4(am)) */
	case 71: /* reg: LOADU2(LOADU4(reg)) */
	case 70: /* reg: LOADU4(INDIRU8(ad)) */
		kids[0] = LEFT_CHILD(LEFT_CHILD(p));
		break;
	case 142: /* stmt: ASGNF8(VREGP,reg) */
	case 141: /* stmt: ASGNU8(VREGP,reg) */
	case 140: /* stmt: ASGNI8(VREGP,reg) */
	case 139: /* stmt: ASGNF4(VREGP,reg) */
	case 138: /* stmt: ASGNP4(VREGP,reg) */
	case 137: /* stmt: ASGNU4(VREGP,reg) */
	case 136: /* stmt: ASGNI4(VREGP,reg) */
	case 135: /* stmt: ASGNU2(VREGP,reg) */
	case 134: /* stmt: ASGNI2(VREGP,reg) */
	case 133: /* stmt: ASGNU1(VREGP,reg) */
	case 132: /* stmt: ASGNI1(VREGP,reg) */
		kids[0] = RIGHT_CHILD(p);
		break;
	case 452: /* stmt: LTF8(eadr,eadr) */
	case 451: /* stmt: LTF4(eadr,eadr) */
	case 450: /* stmt: LTU8(eadr,eadr) */
	case 449: /* stmt: LTI8(eadr,eadr) */
	case 448: /* stmt: LTU4(eadr,eadr) */
	case 447: /* stmt: LTI4(eadr,eadr) */
	case 446: /* stmt: LTU2(eadr,eadr) */
	case 445: /* stmt: LTI2(eadr,eadr) */
	case 444: /* stmt: LTU1(eadr,eadr) */
	case 443: /* stmt: LTI1(eadr,eadr) */
	case 442: /* stmt: LTF8(dm,am) */
	case 441: /* stmt: LTF4(dm,am) */
	case 440: /* stmt: LTU8(dm,am) */
	case 439: /* stmt: LTI8(dm,am) */
	case 438: /* stmt: LTU4(dm,am) */
	case 437: /* stmt: LTI4(dm,am) */
	case 436: /* stmt: LTU2(dm,am) */
	case 435: /* stmt: LTI2(dm,am) */
	case 434: /* stmt: LTU1(dm,am) */
	case 433: /* stmt: LTI1(dm,am) */
	case 432: /* stmt: LEF8(dm,am) */
	case 431: /* stmt: LEF4(dm,am) */
	case 430: /* stmt: LEU8(dm,am) */
	case 429: /* stmt: LEI8(dm,am) */
	case 428: /* stmt: LEU4(dm,am) */
	case 427: /* stmt: LEI4(dm,am) */
	case 426: /* stmt: LEU2(dm,am) */
	case 425: /* stmt: LEI2(dm,am) */
	case 424: /* stmt: LEU1(dm,am) */
	case 423: /* stmt: LEI1(dm,am) */
	case 422: /* stmt: GTF8(dm,am) */
	case 421: /* stmt: GTF4(dm,am) */
	case 420: /* stmt: GTU8(dm,am) */
	case 419: /* stmt: GTI8(dm,am) */
	case 418: /* stmt: GTU4(dm,am) */
	case 417: /* stmt: GTI4(dm,am) */
	case 416: /* stmt: GTU2(dm,am) */
	case 415: /* stmt: GTI2(dm,am) */
	case 414: /* stmt: GTU1(dm,am) */
	case 413: /* stmt: GTI1(dm,am) */
	case 412: /* stmt: GEF8(dm,am) */
	case 411: /* stmt: GEF4(dm,am) */
	case 410: /* stmt: GEU8(dm,am) */
	case 409: /* stmt: GEI8(dm,am) */
	case 408: /* stmt: GEU4(dm,am) */
	case 407: /* stmt: GEI4(dm,am) */
	case 406: /* stmt: GEU2(dm,am) */
	case 405: /* stmt: GEI2(dm,am) */
	case 404: /* stmt: GEU1(dm,am) */
	case 403: /* stmt: GEI1(dm,am) */
	case 402: /* stmt: NEF8(dm,am) */
	case 401: /* stmt: NEF4(dm,am) */
	case 400: /* stmt: NEU8(dm,am) */
	case 399: /* stmt: NEI8(dm,am) */
	case 398: /* stmt: NEU4(dm,am) */
	case 397: /* stmt: NEI4(dm,am) */
	case 396: /* stmt: NEU2(dm,am) */
	case 395: /* stmt: NEI2(dm,am) */
	case 394: /* stmt: NEU1(dm,am) */
	case 393: /* stmt: NEI1(dm,am) */
	case 376: /* stmt: NEF8(dm,conv0) */
	case 375: /* stmt: NEF4(dm,conv0) */
	case 374: /* stmt: NEU8(dm,conv0) */
	case 373: /* stmt: NEI8(dm,conv0) */
	case 372: /* stmt: NEU4(dm,conv0) */
	case 371: /* stmt: NEI4(dm,conv0) */
	case 370: /* stmt: NEU2(dm,conv0) */
	case 369: /* stmt: NEI2(dm,conv0) */
	case 368: /* stmt: NEU1(dm,conv0) */
	case 367: /* stmt: NEI1(dm,conv0) */
	case 366: /* stmt: EQF8(dm,am) */
	case 365: /* stmt: EQF4(dm,am) */
	case 364: /* stmt: EQU8(dm,am) */
	case 363: /* stmt: EQI8(dm,am) */
	case 362: /* stmt: EQU4(dm,am) */
	case 361: /* stmt: EQI4(dm,am) */
	case 360: /* stmt: EQU2(dm,am) */
	case 359: /* stmt: EQI2(dm,am) */
	case 358: /* stmt: EQU1(dm,am) */
	case 357: /* stmt: EQI1(dm,am) */
	case 340: /* stmt: EQF8(dm,conv0) */
	case 339: /* stmt: EQF4(dm,conv0) */
	case 338: /* stmt: EQU8(dm,conv0) */
	case 337: /* stmt: EQI8(dm,conv0) */
	case 336: /* stmt: EQU4(dm,conv0) */
	case 335: /* stmt: EQI4(dm,conv0) */
	case 334: /* stmt: EQU2(dm,conv0) */
	case 333: /* stmt: EQI2(dm,conv0) */
	case 332: /* stmt: EQU1(dm,conv0) */
	case 331: /* stmt: EQI1(dm,conv0) */
	case 308: /* reg: BXORU8(reg,am) */
	case 307: /* reg: BXORI8(reg,am) */
	case 306: /* reg: BXORU4(reg,am) */
	case 305: /* reg: BXORI4(reg,am) */
	case 304: /* reg: BXORU2(reg,am) */
	case 303: /* reg: BXORI2(reg,am) */
	case 302: /* reg: BXORU1(reg,am) */
	case 301: /* reg: BXORI1(reg,am) */
	case 300: /* reg: BORU8(reg,am) */
	case 299: /* reg: BORI8(reg,am) */
	case 298: /* reg: BORU4(reg,am) */
	case 297: /* reg: BORI4(reg,am) */
	case 296: /* reg: BORU2(reg,am) */
	case 295: /* reg: BORI2(reg,am) */
	case 294: /* reg: BORU1(reg,am) */
	case 293: /* reg: BORI1(reg,am) */
	case 292: /* reg: BANDU8(reg,am) */
	case 291: /* reg: BANDI8(reg,am) */
	case 290: /* reg: BANDU4(reg,am) */
	case 289: /* reg: BANDI4(reg,am) */
	case 288: /* reg: BANDU2(reg,am) */
	case 287: /* reg: BANDI2(reg,am) */
	case 286: /* reg: BANDU1(reg,am) */
	case 285: /* reg: BANDI1(reg,am) */
	case 284: /* reg: LSHU8(reg,am) */
	case 283: /* reg: LSHI8(reg,am) */
	case 282: /* reg: LSHU4(reg,am) */
	case 281: /* reg: LSHI4(reg,am) */
	case 280: /* reg: LSHU2(reg,am) */
	case 279: /* reg: LSHI2(reg,am) */
	case 278: /* reg: LSHU1(reg,am) */
	case 277: /* reg: LSHI1(reg,am) */
	case 276: /* reg: RSHU8(reg,am) */
	case 275: /* reg: RSHI8(reg,am) */
	case 274: /* reg: RSHU4(reg,am) */
	case 273: /* reg: RSHI4(reg,am) */
	case 272: /* reg: RSHU2(reg,am) */
	case 271: /* reg: RSHI2(reg,am) */
	case 270: /* reg: RSHU1(reg,am) */
	case 269: /* reg: RSHI1(reg,am) */
	case 268: /* reg: MODU8(reg,am) */
	case 267: /* reg: MODI8(reg,am) */
	case 266: /* reg: MODU4(reg,am) */
	case 265: /* reg: MODI4(reg,am) */
	case 264: /* reg: MODU2(reg,am) */
	case 263: /* reg: MODI2(reg,am) */
	case 262: /* reg: MODU1(reg,am) */
	case 261: /* reg: MODI1(reg,am) */
	case 260: /* reg: DIVU4(reg,conv256) */
	case 259: /* reg: DIVU4(reg,conv128) */
	case 258: /* reg: DIVU4(reg,conv64) */
	case 257: /* reg: DIVU4(reg,conv32) */
	case 256: /* reg: DIVU4(reg,conv16) */
	case 255: /* reg: DIVU4(reg,conv8) */
	case 254: /* reg: DIVU4(reg,conv4) */
	case 253: /* reg: DIVU4(reg,conv2) */
	case 252: /* reg: DIVI4(reg,conv256) */
	case 251: /* reg: DIVI4(reg,conv128) */
	case 250: /* reg: DIVI4(reg,conv64) */
	case 249: /* reg: DIVI4(reg,conv32) */
	case 248: /* reg: DIVI4(reg,conv16) */
	case 247: /* reg: DIVI4(reg,conv8) */
	case 246: /* reg: DIVI4(reg,conv4) */
	case 245: /* reg: DIVI4(reg,conv2) */
	case 244: /* reg: DIVF8(reg,am) */
	case 243: /* reg: DIVF4(reg,am) */
	case 242: /* reg: DIVU8(reg,am) */
	case 241: /* reg: DIVI8(reg,am) */
	case 240: /* reg: DIVU4(reg,am) */
	case 239: /* reg: DIVI4(reg,am) */
	case 238: /* reg: DIVU2(reg,am) */
	case 237: /* reg: DIVI2(reg,am) */
	case 236: /* reg: DIVU1(reg,am) */
	case 235: /* reg: DIVI1(reg,am) */
	case 234: /* reg: MULU4(reg,conv256) */
	case 233: /* reg: MULU4(reg,conv128) */
	case 232: /* reg: MULU4(reg,conv64) */
	case 231: /* reg: MULU4(reg,conv32) */
	case 230: /* reg: MULU4(reg,conv16) */
	case 229: /* reg: MULU4(reg,conv8) */
	case 228: /* reg: MULU4(reg,conv4) */
	case 227: /* reg: MULU4(reg,conv2) */
	case 226: /* reg: MULI4(reg,conv256) */
	case 225: /* reg: MULI4(reg,conv128) */
	case 224: /* reg: MULI4(reg,conv64) */
	case 223: /* reg: MULI4(reg,conv32) */
	case 222: /* reg: MULI4(reg,conv16) */
	case 221: /* reg: MULI4(reg,conv8) */
	case 220: /* reg: MULI4(reg,conv4) */
	case 219: /* reg: MULI4(reg,conv2) */
	case 218: /* reg: MULF8(reg,am) */
	case 217: /* reg: MULF4(reg,am) */
	case 216: /* reg: MULU8(reg,am) */
	case 215: /* reg: MULI8(reg,am) */
	case 214: /* reg: MULU4(reg,am) */
	case 213: /* reg: MULI4(reg,am) */
	case 212: /* reg: MULU2(reg,am) */
	case 211: /* reg: MULI2(reg,am) */
	case 210: /* reg: MULU1(reg,am) */
	case 209: /* reg: MULI1(reg,am) */
	case 207: /* reg: SUBF8(reg,am) */
	case 206: /* reg: SUBF4(reg,am) */
	case 205: /* reg: SUBU8(reg,am) */
	case 204: /* reg: SUBI8(reg,am) */
	case 203: /* reg: SUBP4(reg,am) */
	case 202: /* reg: SUBU4(reg,am) */
	case 201: /* reg: SUBI4(reg,am) */
	case 200: /* reg: SUBU2(reg,am) */
	case 199: /* reg: SUBI2(reg,am) */
	case 198: /* reg: SUBU1(reg,am) */
	case 197: /* reg: SUBI1(reg,am) */
	case 196: /* reg: ADDP4(reg,ca) */
	case 195: /* reg: ADDF8(reg,am) */
	case 194: /* reg: ADDF4(reg,am) */
	case 193: /* reg: ADDU8(reg,am) */
	case 192: /* reg: ADDI8(reg,am) */
	case 191: /* reg: ADDU4(reg,am) */
	case 190: /* reg: ADDI4(reg,am) */
	case 189: /* reg: ADDU2(reg,am) */
	case 188: /* reg: ADDI2(reg,am) */
	case 187: /* reg: ADDU1(reg,am) */
	case 186: /* reg: ADDI1(reg,am) */
	case 184: /* stmt: ASGNB(adr,iadr) */
	case 163: /* stmt: ASGNF8(adr,am) */
	case 162: /* stmt: ASGNF4(adr,am) */
	case 161: /* stmt: ASGNU8(adr,am) */
	case 160: /* stmt: ASGNI8(adr,am) */
	case 159: /* stmt: ASGNP4(adr,ad) */
	case 158: /* stmt: ASGNP4(adr,iadr) */
	case 157: /* stmt: ASGNP4(adr,con) */
	case 156: /* stmt: ASGNP4(adr,reg) */
	case 149: /* stmt: ASGNU4(adr,am) */
	case 148: /* stmt: ASGNI4(adr,am) */
	case 147: /* stmt: ASGNU2(adr,am) */
	case 146: /* stmt: ASGNI2(adr,am) */
	case 145: /* stmt: ASGNU1(adr,am) */
	case 144: /* stmt: ASGNI1(adr,am) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = RIGHT_CHILD(p);
		break;
	case 151: /* reg: ASGNU2(INDIRP4(VREGP),LOADU2(reg)) */
	case 150: /* stmt: ASGNU2(INDIRP4(VREGP),LOADU2(reg)) */
		kids[0] = LEFT_CHILD(RIGHT_CHILD(p));
		break;
	case 155: /* stmt: ASGNI2(adr,LOADI2(INDIRI4(ad))) */
	case 154: /* stmt: ASGNI1(adr,LOADI1(INDIRI4(ad))) */
	case 153: /* stmt: ASGNU2(adr,LOADU2(INDIRU4(ad))) */
	case 152: /* stmt: ASGNU1(adr,LOADU1(INDIRU4(ad))) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(p)));
		break;
	case 183: /* stmt: ASGNU4(adr,BXORU4(INDIRU4(adr),am)) */
	case 182: /* stmt: ASGNI4(adr,BXORI4(INDIRI4(adr),am)) */
	case 181: /* stmt: ASGNU4(adr,BORU4(INDIRU4(adr),am)) */
	case 180: /* stmt: ASGNI4(adr,BORI4(INDIRI4(adr),am)) */
	case 179: /* stmt: ASGNU4(adr,BANDU4(INDIRU4(adr),am)) */
	case 178: /* stmt: ASGNI4(adr,BANDI4(INDIRI4(adr),am)) */
	case 177: /* stmt: ASGNU4(adr,LSHU4(INDIRU4(adr),am)) */
	case 176: /* stmt: ASGNI4(adr,LSHI4(INDIRI4(adr),am)) */
	case 175: /* stmt: ASGNU4(adr,RSHU4(INDIRU4(adr),am)) */
	case 174: /* stmt: ASGNI4(adr,RSHI4(INDIRI4(adr),am)) */
	case 173: /* stmt: ASGNU4(adr,SUBU4(INDIRU4(adr),am)) */
	case 172: /* stmt: ASGNI4(adr,SUBI4(INDIRI4(adr),am)) */
	case 171: /* stmt: ASGNU8(adr,ADDU8(INDIRU8(adr),am)) */
	case 170: /* stmt: ASGNI8(adr,ADDI8(INDIRI8(adr),am)) */
	case 169: /* stmt: ASGNU4(adr,ADDU4(INDIRU4(adr),am)) */
	case 168: /* stmt: ASGNI4(adr,ADDI4(INDIRI4(adr),am)) */
	case 167: /* stmt: ASGNU2(adr,ADDU2(INDIRU2(adr),am)) */
	case 166: /* stmt: ASGNI2(adr,ADDI2(INDIRI2(adr),am)) */
	case 165: /* stmt: ASGNU1(adr,ADDU1(INDIRU1(adr),am)) */
	case 164: /* stmt: ASGNI1(adr,ADDI1(INDIRI1(adr),am)) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = LEFT_CHILD(LEFT_CHILD(RIGHT_CHILD(p)));
		kids[2] = RIGHT_CHILD(RIGHT_CHILD(p));
		break;
	case 208: /* reg: MULU4(con,LOADU4(eadr)) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = LEFT_CHILD(RIGHT_CHILD(p));
		break;
	case 392: /* stmt: NEU8(BANDU8(dm,conv1),conv1) */
	case 391: /* stmt: NEI8(BANDI8(dm,conv1),conv1) */
	case 390: /* stmt: NEU4(BANDU4(dm,conv1),conv1) */
	case 389: /* stmt: NEI4(BANDI4(dm,conv1),conv1) */
	case 388: /* stmt: NEU2(BANDU2(dm,conv1),conv1) */
	case 387: /* stmt: NEI2(BANDI2(dm,conv1),conv1) */
	case 386: /* stmt: NEU1(BANDU1(dm,conv1),conv1) */
	case 385: /* stmt: NEI1(BANDI1(dm,conv1),conv1) */
	case 384: /* stmt: NEU8(BANDU8(dm,conv1),conv0) */
	case 383: /* stmt: NEI8(BANDI8(dm,conv1),conv0) */
	case 382: /* stmt: NEU4(BANDU4(dm,conv1),conv0) */
	case 381: /* stmt: NEI4(BANDI4(dm,conv1),conv0) */
	case 380: /* stmt: NEU2(BANDU2(dm,conv1),conv0) */
	case 379: /* stmt: NEI2(BANDI2(dm,conv1),conv0) */
	case 378: /* stmt: NEU1(BANDU1(dm,conv1),conv0) */
	case 377: /* stmt: NEI1(BANDI1(dm,conv1),conv0) */
	case 356: /* stmt: EQU8(BANDU8(dm,conv1),conv1) */
	case 355: /* stmt: EQI8(BANDI8(dm,conv1),conv1) */
	case 354: /* stmt: EQU4(BANDU4(dm,conv1),conv1) */
	case 353: /* stmt: EQI4(BANDI4(dm,conv1),conv1) */
	case 352: /* stmt: EQU2(BANDU2(dm,conv1),conv1) */
	case 351: /* stmt: EQI2(BANDI2(dm,conv1),conv1) */
	case 350: /* stmt: EQU1(BANDU1(dm,conv1),conv1) */
	case 349: /* stmt: EQI1(BANDI1(dm,conv1),conv1) */
	case 348: /* stmt: EQU8(BANDU8(dm,conv1),conv0) */
	case 347: /* stmt: EQI8(BANDI8(dm,conv1),conv0) */
	case 346: /* stmt: EQU4(BANDU4(dm,conv1),conv0) */
	case 345: /* stmt: EQI4(BANDI4(dm,conv1),conv0) */
	case 344: /* stmt: EQU2(BANDU2(dm,conv1),conv0) */
	case 343: /* stmt: EQI2(BANDI2(dm,conv1),conv0) */
	case 342: /* stmt: EQU1(BANDU1(dm,conv1),conv0) */
	case 341: /* stmt: EQI1(BANDI1(dm,conv1),conv0) */
		kids[0] = LEFT_CHILD(LEFT_CHILD(p));
		kids[1] = RIGHT_CHILD(LEFT_CHILD(p));
		kids[2] = RIGHT_CHILD(p);
		break;
	case 480: /* reg: CVFF8(INDIRF4(ADDP4(dm,am))) */
		kids[0] = LEFT_CHILD(LEFT_CHILD(LEFT_CHILD(p)));
		kids[1] = RIGHT_CHILD(LEFT_CHILD(LEFT_CHILD(p)));
		break;
	default:
		fatal("_kids", "Bad rule number %d\n", eruleno);
	}
}


//*********************************
static void progbeg(long argc, char *argv[]) {

	{
		union {
			char c;
			long i;
		} u;
		u.i = 0;
		u.c = 1;
		swap = (long)((unsigned)(u.i == 1)) != IR->little_endian;
	}
	parseflags(argc, argv);

	fltreg[R14] = mkreg("$14", R14, 1, IREG);
	fltreg[R0] = mkreg("$0", R0, 1, IREG);
	fltreg[R1] = mkreg("$1", R1, 1, IREG);
	fltreg[R2] = mkreg("$2", R2, 1, IREG);
	fltreg[R3] = mkreg("$3", R3, 1, IREG);
	fltreg[R4] = mkreg("$4", R4, 1, IREG);
	fltreg[R5] = mkreg("$5", R5, 1, IREG);
	fltreg[R6] = mkreg("$6", R6, 1, IREG);
	fltreg[R7] = mkreg("$7", R7, 1, IREG);
	fltreg[R8] = mkreg("$8", R8, 1, IREG);
	fltreg[R9] = mkreg("$9", R9, 1, IREG);
	fltreg[R10] = mkreg("$10", R10, 1, IREG);
	fltreg[R11] = mkreg("$11", R11, 1, IREG);
	fltreg[R12] = mkreg("$12", R12, 1, IREG);
	fltreg[R13] = mkreg("$13", R13, 1, IREG);
	
	longlongreg[R14] = mkreg("$14", R14, 1, IREG);
	longlongreg[R0] = mkreg("$0", R0, 1, IREG);
	longlongreg[R1] = mkreg("$1", R1, 1, IREG);
	longlongreg[R2] = mkreg("$2", R2, 1, IREG);
	longlongreg[R3] = mkreg("$3", R3, 1, IREG);
	longlongreg[R4] = mkreg("$4", R4, 1, IREG);
	longlongreg[R5] = mkreg("$5", R5, 1, IREG);
	longlongreg[R6] = mkreg("$6", R6, 1, IREG);
	longlongreg[R7] = mkreg("$7", R7, 1, IREG);
	longlongreg[R8] = mkreg("$8", R8, 1, IREG);
	longlongreg[R9] = mkreg("$9", R9, 1, IREG);
	longlongreg[R10] = mkreg("$10", R10, 1, IREG);
	longlongreg[R11] = mkreg("$11", R11, 1, IREG);
	longlongreg[R12] = mkreg("$12", R12, 1, IREG);
	longlongreg[R13] = mkreg("$13", R13, 1, IREG);
	
	longreg[R14] = mkreg("$14", R14, 1, IREG);
	longreg[R0] = mkreg("$0", R0, 1, IREG);
	longreg[R1] = mkreg("$1", R1, 1, IREG);
	longreg[R2] = mkreg("$2", R2, 1, IREG);
	longreg[R3] = mkreg("$3", R3, 1, IREG);
	longreg[R4] = mkreg("$4", R4, 1, IREG);
	longreg[R5] = mkreg("$5", R5, 1, IREG);
	longreg[R6] = mkreg("$6", R6, 1, IREG);
	longreg[R7] = mkreg("$7", R7, 1, IREG);
	longreg[R8] = mkreg("$8", R8, 1, IREG);
	longreg[R9] = mkreg("$9", R9, 1, IREG);
	longreg[R10] = mkreg("$10", R10, 1, IREG);
	longreg[R11] = mkreg("$11", R11, 1, IREG);
	longreg[R12] = mkreg("$12", R12, 1, IREG);
	longreg[R13] = mkreg("$13", R13, 1, IREG);
	
	shortreg[R14] = mkreg("$14", R14, 1, IREG);
	shortreg[R0] = mkreg("$0", R0, 1, IREG);
	shortreg[R1] = mkreg("$1", R1, 1, IREG);
	shortreg[R2] = mkreg("$2", R2, 1, IREG);
	shortreg[R3] = mkreg("$3", R3, 1, IREG);
	shortreg[R4] = mkreg("$4", R4, 1, IREG);
	shortreg[R5] = mkreg("$5", R5, 1, IREG);
	shortreg[R6] = mkreg("$6", R6, 1, IREG);
	shortreg[R7] = mkreg("$7", R7, 1, IREG);
	shortreg[R8] = mkreg("$8", R8, 1, IREG);
	shortreg[R9] = mkreg("$9", R9, 1, IREG);
	shortreg[R10] = mkreg("$10", R10, 1, IREG);
	shortreg[R11] = mkreg("$11", R11, 1, IREG);
	shortreg[R12] = mkreg("$12", R12, 1, IREG);
	shortreg[R13] = mkreg("$13", R13, 1, IREG);
	
	charreg[R14] = mkreg("$14", R14, 1, IREG);
	charreg[R0] = mkreg("$0", R0, 1, IREG);
	charreg[R1] = mkreg("$1", R1, 1, IREG);
	charreg[R2] = mkreg("$2", R2, 1, IREG);
	charreg[R3] = mkreg("$3", R3, 1, IREG);
	charreg[R4] = mkreg("$4", R4, 1, IREG);
	charreg[R5] = mkreg("$5", R5, 1, IREG);
	charreg[R6] = mkreg("$6", R6, 1, IREG);
	charreg[R7] = mkreg("$7", R7, 1, IREG);
	charreg[R8] = mkreg("$8", R8, 1, IREG);
	charreg[R9] = mkreg("$9", R9, 1, IREG);
	charreg[R10] = mkreg("$10", R10, 1, IREG);
	charreg[R11] = mkreg("$11", R11, 1, IREG);
	charreg[R12] = mkreg("$12", R12, 1, IREG);
	charreg[R13] = mkreg("$13", R13, 1, IREG);
	
	charregw = mkwildcard(charreg);
	shortregw = mkwildcard(shortreg);
	longregw = mkwildcard(longreg);
	longlongregw = mkwildcard(longlongreg);
	fltregw = mkwildcard(fltreg);

	tmask[IREG] = (1<<R14) | (1<<R0) | (1<<R1) | (1<<R2) | (1<<R3) | (1<<R4) | (1<<R5) | (1<<R6) | (1<<R7) | (1<<R8) | (1<<R9) | (1<<R10) | (1<<R11) | (1<<R12) | (1<<R13);
	vmask[IREG] = 0;

	tmask[FREG] = (1<<R14) | (1<<R0) | (1<<R1) | (1<<R2) | (1<<R3) | (1<<R4) | (1<<R5) | (1<<R6) | (1<<R7) | (1<<R8) | (1<<R9) | (1<<R10) | (1<<R11) | (1<<R12) | (1<<R13);
	vmask[FREG] = 0;

	cseg = 0;

	asmresult = mkreg("$14", R14, 1, IREG);
	// output UTF-8 BOM (UTF8)
	// Nope, on second thought, don't do it.
	// UTF-8, according to the standard, shouldn't
	// have the UTF-16 BOM at the start of it,
	// even though many do. So, chew it up if it's
	// there, but don't put it out yourself.
	// print("\357\273\277");
}

//*********************************
static Symbol rmap(long opk) {

	switch (optype(opk)) {
	 case B:	// struct
	 case P:	// pointer
		return longregw;
	 case I:
	 case U:
		if( opsize(opk)==1 ) return charregw;
		else if( opsize(opk)==2 ) return shortregw;
		else if( opsize(opk)==4 ) return longregw;
		else return longlongregw;
	 case F:
		return fltregw;
	 default:
		return 0;
	}
}

//*********************************
static void segment(long n) {

	if( ((cseg==CODE || cseg==LIT || cseg==DATA) && (n==0 || n==BSS))
	 || (cseg==BSS && n!=BSS) ) print("\t@seg_end\n\t; *************************\n");	// end of segment
	if( (n==CODE || n==LIT || n==DATA) && (cseg==0 || cseg==BSS) ) print("\t@code\n");
	if( n==BSS && cseg!=BSS ) print("\t@zero\n");

	cseg = n;
}

//*********************************
static void progend(void) {
	segment(0);
	print("\t@end\n");
}

//*********************************
static void target(Node p) {

	assert(p);
	switch (specific(p->op)) {
	 case ASMRET+U:
	 case ASM0+U:
	 case ASM1+U:
	 case ASM2+U:
		setreg(p, asmresult);
		break;
	 case CALL+I:
	 case CALL+U:
	 case CALL+P:
		setreg(p, longreg[R14]);
		break;
	 case RET+I:
	 case RET+U:
	 case RET+P:
	 case RET+F:
		rtarget(p, 0, longreg[R14]);
		break;
	}
}

//****************************************
static void clobber(Node p) {

	assert(p);
	switch (specific(p->op)) {
	 case EQ+F:
	 case LE+F:
	 case GE+F:
	 case LT+F:
	 case GT+F:
	 case NE+F:
		if( specific(p->op) == EQ+F ) p->syms[1] = findlabel(genlabel(1));
		break;

	}
}

//*******************************************
static int sametree(Node p, Node q) {

	return (p == NULL && q == NULL)
		|| (p && q && p->op == q->op && p->syms[0] == q->syms[0]
				&& sametree(p->kids[0], q->kids[0])
				&& sametree(p->kids[1], q->kids[1]));
}

//*******************************************
static int memop(Node p) {

	assert(p);
	assert(generic(p->op) == ASGN);
	assert(p->kids[0]);
	assert(p->kids[1]);
	return (generic(p->kids[1]->kids[0]->op) == INDIR && sametree(p->kids[0], p->kids[1]->kids[0]->kids[0])) ? 1 : LBURG_MAX;
}

//*******************************************
// return 1 if handled, else return 0
static int emit2(Node p) {

	long sop = specific(p->op);
	long gop = generic(p->op);
	long ops = opsize(p->op);
	long opsk= 0;

	if( p->kids[0] ) opsk = opsize(p->kids[0]->op);

#define preg(f) ((f)[getregnum(p->x.kids[0])]->x.name)

//print("EMIT2 sop=%d p->op=%d\r\n", sop, p->op);

	// GEEKDO: check: shouldn't the following opsize(p->x.kids[0]->op) be opsk (above)???
	
	// GEEKDO: these conversions should check to see if the SRC & DST are the same and omit the 'mov'
	
// convert CHAR to signed
//	if (sop == CVI+I && opsize(p->x.kids[0]->op) == 1) {
	if (sop == CVI+I && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb cvss %s\n", p->syms[RX]->x.name);
// convert SHORT to signed
//	} else if (sop == CVI+I && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVI+I && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvss %s\n", p->syms[RX]->x.name);
// convert LONG to signed
//	} else if (sop == CVI+I && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVI+I && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
		print("\td cvss %s\n", p->syms[RX]->x.name);

// convert CHAR to unsigned
//	} else if (sop == CVI+U && opsize(p->x.kids[0]->op) == 1) {
	} else if (sop == CVI+U && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb ncvuu %s\n", p->syms[RX]->x.name);
// convert SHORT to unsigned
//	} else if (sop == CVI+U && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVI+U && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvuu %s\n", p->syms[RX]->x.name);
// convert LONG to unsigned
//	} else if (sop == CVI+U && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVI+U && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
		print("\td cvuu %s\n", p->syms[RX]->x.name);

// convert BYTE to signed
//	} else if (sop == CVU+I && opsize(p->x.kids[0]->op) == 1) {
	} else if (sop == CVU+I && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb cvuu %s\n", p->syms[RX]->x.name);
// convert WORD to signed
//	} else if (sop == CVU+I && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVU+I && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvuu %s\n", p->syms[RX]->x.name);
// convert DWORD to signed
//	} else if (sop == CVU+I && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVU+I && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\td cvuu %s\n", p->syms[RX]->x.name);
		
// convert BYTE to unsigned
//	} else if (sop == CVU+U && opsize(p->x.kids[0]->op) == 1) {
	} else if (sop == CVU+U && opsk == 1) {
		if( strcmp(p->syms[RX]->x.name, preg(charreg)) ) print("\tb mov %s,%s\n", p->syms[RX]->x.name, preg(charreg));
		print("\tb cvuu %s\n", p->syms[RX]->x.name);
// convert WORD to unsigned
//	} else if (sop == CVU+U && opsize(p->x.kids[0]->op) == 2) {
	} else if (sop == CVU+U && opsk == 2) {
		if( strcmp(p->syms[RX]->x.name, preg(shortreg)) ) print("\tw mov %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
		print("\tw cvuu %s\n", p->syms[RX]->x.name);
// convert DWORD to unsigned
//	} else if (sop == CVU+U && opsize(p->x.kids[0]->op) == 4) {
	} else if (sop == CVU+U && opsk == 4) {
		if( ops!=8 || p->kids[0]->op!=4870 ) {	// don't convert ASMRETU4 if output is QWORD
			if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
			print("\td cvuu %s\n", p->syms[RX]->x.name);
		}
		
// convert FLOAT to signed
	} else if( sop == CVF+I && opsk == 4) {
		if( strcmp(p->syms[RX]->x.name, preg(longreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longreg));
		print("\tcvfs %s\n", p->syms[RX]->x.name);
// convert DOUBLE to signed
	} else if( sop == CVF+I && opsk == 8) {
		if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\tq mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
		print("\tcvds %s\n", p->syms[RX]->x.name);
		
// convert CHAR to 


// GEEKDO: need to finish these!

	} else if( sop == CVI+F && opsk==1 ) {
		opsk = 1;
	} else if( sop == CVI+F && opsk==2 ) {
		opsk = 2;
	} else if( gop == CVI ) {
		if( ops==8 ) {	// target is DOUBLE
			if( opsk==4 ) {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\td cvsd %s\n", p->syms[RX]->x.name);
			} else {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\tq mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\tq cvsd %s\n", p->syms[RX]->x.name);
			}
		} else if( ops==4 ) {		// target is FLOAT
			if( opsk==4 ) {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\td cvsf %s\n", p->syms[RX]->x.name);
			} else {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\tq mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\tq cvsf %s\n", p->syms[RX]->x.name);
			}
		}
	} else if( gop == CVU ) {
		if( ops==8 ) {	// target is DOUBLE
			if( opsk==4 ) {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\td cvud %s\n", p->syms[RX]->x.name);
			} else {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\tq mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\tq cvud %s\n", p->syms[RX]->x.name);
			}
		} else if( ops==4 ) {		// target is FLOAT
			if( opsk==4 ) {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\td mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\td cvuf %s\n", p->syms[RX]->x.name);
			} else {
				if( strcmp(p->syms[RX]->x.name, preg(longlongreg)) ) print("\tq mov %s,%s\n", p->syms[RX]->x.name, preg(longlongreg));
				print("\tq cvuf %s\n", p->syms[RX]->x.name);
			}
		}
	} else if( sop == CVI+F && opsk==8 ) {
		opsk = 8;
	} else if( sop == CVI+F && opsk==1 ) {
	
	
	
	
		if( opsk == 1 ) {
			if( opsize(p->op) == 4 ) print("\tb mov %s,%s\n\tcvss %sb\n\tcvsf %s", p->syms[RX]->x.name, preg(longreg), p->syms[RX]->x.name, p->syms[RX]->x.name);
		} else if( opsk==2 ) {
		
		} else if( opsk==4 ) {
			if( ops==4 ) ;
			else if( ops==8 ) ;
			else print("Oh, crap. emit2(CVI+F) opsk=4, ops=%d\n", ops);
		} else if( opsk==8 ) {
		} else print("Oh, crap. emit2(CVI+F) opsk=%d\n", opsk);
	} else if( gop==CVI || gop==CVU || gop==LOAD ) {

		char	*dst;
		char	*src;
		long	srcopsize, dstopsize;
		long	srcreg,	dstreg;

/*		if( gop==LOAD && generic(p->kids[0]->op)==INDIR && p->kids[0]->kids[0]->op==VREG+P ) {
			return 1;
		}
*/
		if( p->x.kids[0] ) {
			srcopsize = opsize(p->x.kids[0]->op);
			srcreg = getregnum(p->x.kids[0]);
		} else {
			srcopsize = opsize(p->kids[0]->op);
			srcreg = getregnum(p->kids[0]);
		}
		dstopsize = opsize(p->op);
		dstreg = getregnum(p);

		if( generic(p->kids[0]->op)==ASMRET ) {
			return 1;
		}
		if( dstopsize==1 )		dst = charreg[dstreg]->x.name;
		else if( dstopsize==2 )	dst = shortreg[dstreg]->x.name;
		else if( dstopsize==4 )	dst = longreg[dstreg]->x.name;
		else					dst = longlongreg[dstreg]->x.name;

		if( srcopsize==1 ) 		src = charreg[srcreg]->x.name;
		else if( srcopsize==2 )	src = shortreg[srcreg]->x.name;
		else if( srcopsize==4 )	src = longreg[srcreg]->x.name;
		else					src = longlongreg[srcreg]->x.name;

//		assert(dstopsize <= srcopsize);
		if (dst != src) {
				 if( dstopsize==1 ) print("\tb mov %s,%s\n", dst, src);
			else if( dstopsize==2 ) print("\tw mov %s,%s\n", dst, src);
			else if( dstopsize==4 ) print("\td mov %s,%s\n", dst, src);
			else if( dstopsize==8 ) print("\tq mov %s,%s\n", dst, src);
		}
	} else if( gop==RET+I || gop==RET+U ) {
		// check if %0 is 'r14' (return causing "mov $14,$14"). if so, return 1, else return 0
//		if( stricmp("$14", p->syms[0]->x.name) ) return 0;
		return 1;
	} else if( gop==LABEL ) {
		print(":%s\n", p->syms[0]->x.name);
	} else if( gop==ASM0 ) {
		if( p->syms[0]->name[0]==':' ) print("%s", p->syms[0]->name);
		else print("\t%s", p->syms[0]->name);
	} else if( gop==ASMRET || gop==ASM2 ) {
		if( p->kids[0] ) emit2(p->kids[0]);
		if( p->kids[1] ) emit2(p->kids[1]);
	}
	return 1;
}

//*********************************
static void doarg(Node p) {

	assert(p && p->syms[0]);
	mkactual(4, p->syms[0]->u.c.v.ill);
}

//*********************************
static void blkfetch(long k, long off, long reg, long tmp) {}

//*********************************
static void blkstore(long k, long off, long reg, long tmp) {}

//*********************************
static void blkloop(long dreg, long doff, long sreg, long soff, long size, long tmps[]) {}

//*********************************
static void local(Symbol p) {

	if( isfloat(p->type) ) p->sclass = AUTO;
	if( askregvar(p, (*IR->x.rmap)(ttob(p->type))) == 0 ) {
		assert(p->sclass == AUTO);
		offset = roundup(offset + p->type->size, p->type->align);// < 4 ? 4 : p->type->align);
		p->x.offset = -offset;
		p->x.name = stringd(-offset);
	}
}

//***************************************************************************
static void function(Symbol f, Symbol caller[], Symbol callee[], long n) {

	long 			i, argsize;
	unsigned long	mask, maskcnt;

	usedmask[0] = usedmask[1] = 0;
	freemask[0] = freemask[1] = ~(unsigned)0;
	offset = maxoffset = maxargoffset = 12;	// for retaddr & saved FP & saved OldSP

	argsize = 0;
	for(i=0; callee[i]; i++) {
			Symbol p = callee[i];
			Symbol q = caller[i];
			assert(q);
			p->x.offset = q->x.offset = offset;
			p->x.name = q->x.name = stringf("%d", p->x.offset);
			p->sclass = q->sclass = AUTO;
			offset += roundup(q->type->size, 4);
			argsize += q->type->size;
	}
	assert(caller[i]==0);

	offset = maxoffset = 0;
	
	gencode(caller, callee);

	framesize = roundup(maxoffset, 4);

	/*
		Based on the enum{} near the start of this file, the mask bits to reg# mapping is:
		MASKBIT	REG#
			0	14 (function return values)
			1	0
			2	1
			3	2
			4	3
			5	4
			6	5
			7	6
			8	7
			9	8
			10	9
			11	10
			12	11
			13	12
			14	13
			15	14 (never used by C compiler)
	*/
	mask = usedmask[IREG] & 0x7FFE;	// never save R14 (bit0) & R15 (bit15), return value register, only 0-13
	
	print("\t@align 4\n");
	print(":%s\n", f->x.name);
	
// if framesize==0, no local vars
// if mask==0, no registers to save
	
	for(maskcnt=0, i=14; mask; i--, ++maskcnt) mask &= ~(1<<i);
	print("\tSFVR %u,%u\n", framesize, maskcnt);
/*
	if( framesize ) {
		if( mask ) print("\tSFVR %u,%u\n", framesize, mask);	// start frame with both LOCAL VARS and REGS saved
		else print("\tSFV %u\n", framesize);					// start frame with LOCAL VARS but no regs saved
	} else {
		if( mask ) print("\tSFR 0x%x\n", mask);					// start frame with no local vars but REGS saved
		else print("\tSF\n");									// start frame with no local vars and no regs saved
	}
*/		
	emitcode();

	print("\tEFR %u\n", maskcnt);
//	if( mask ) print("\tEFR 0x%x\n", mask);		// end frame with saved REGS to restore
//	else print("\tEF\n");						// end frame with no saved regs to restore
	print("\t; =========================\n");
}

//***********************************
static void defsymbol(Symbol p) {

	if (p->scope >= LOCAL && p->sclass == STATIC) p->x.name = stringf("%c%d", 'A'+Kfilenum, genlabel(1));
	else if (p->generated) p->x.name = stringf("%c%s", 'A'+Kfilenum, p->name);
	else if (p->scope == GLOBAL || p->sclass == EXTERN) p->x.name = stringf("_%s", p->name);
	else if (p->scope == CONSTANTS && (isint(p->type) || isptr(p->type))
		  && p->name[0] == '0' && p->name[1] == 'x') p->x.name = stringf("0x%s", &p->name[2]);
	else p->x.name = p->name;
}

//***********************************************
static void address(Symbol q, Symbol p, long n) {

	if( p->scope == GLOBAL || p->sclass == STATIC || p->sclass == EXTERN ) {
		q->x.name = stringf("%s%s%d", p->x.name, n >= 0 ? "+" : "", n);
	} else {
		assert(n <= INT_MAX && n >= INT_MIN);
		q->x.offset = p->x.offset + n;
		q->x.name = stringd(q->x.offset);
	}
}

//******************************************************
static void defconst(long suffix, long size, Value v) {

	assert(sizeof(float)==4 && sizeof(double)==8);
		 if (suffix == I && size == 1) print("\tb %d\n", (unsigned long)v.ull);
	else if (suffix == I && size == 2) print("\tw %d\n", (unsigned long)v.ill);
	else if (suffix == I && size == 4) print("\td %d\n",(unsigned long)v.ill);
	else if (suffix == I && size == 8) print("\tq 0x%X\n",v.ill);
	else if (suffix == U && size == 1) print("\tb 0x%x\n", (unsigned)((unsigned char)v.ull));
	else if (suffix == U && size == 2) print("\tw 0x%x\n", (unsigned)((unsigned short)v.ull));
	else if (suffix == U && size == 4) print("\td 0x%x\n", (unsigned)v.ull);
	else if (suffix == U && size == 8) print("\tq 0x%X\n",v.ull);
	else if (suffix == P && size == 4) print("\td 0x%x\n", (unsigned)v.p);
	else if (suffix == F && size == 4) {
		float f = (float)v.d;
		print("\td 0x%x\n", *(unsigned *)&f);
	} else if (suffix == F && size == 8) print("\tq 0x%X\n", *(unsigned long long *)&v.d);
	else assert(0);
}

//*********************************
static void defaddress(Symbol p) {

	print("\t& %s\n", p->x.name);
}

//*********************************
static void defstring(long n, char *str) {

	char 	*s;
	long	a, b, last;

	for(s=str, a=last=0; s<str+n; s++) if( *s>=' ' && *s<128 ) {++a; last = s-str;}
	if( 100*a/n > 90 || (a==n-1 && (*(s-1)<32 || *(s-1)>127)) ) {
		print("\tb \"");
		for(s=str; s<str+n; s++) {
			if( *s=='"' || *s=='\\' ) print("\\%c", *s);
			else if( *s>=32 && *s<128 ) print("%c", *s);
			else if( *s=='\t' ) print("\\t");
			else if( *s=='\r' ) print("\\r");
			else if( *s=='\n' ) print("\\n");
			else print("\\0%o", *s);
		}
		print("\"\n");
	} else {
		for(s=str, b=0; s<str+n; b++) {
			print("\tb ");
			for(a=0; s<str+n && a<8; ++s, ++a) {
				if( a<7 && s+1<str+n ) print("%d,", *s);
				else print("%d", *s);
			}
			print("\n");
		}
	}
}

//*********************************
static void export(Symbol p) {

// KASM assembles all files in a project at the
// same time, so in that environment, 'public'
// and 'extern' aren't needed and just clog up
// the assembly language source code needlessly.
//	print("public %s\n", p->x.name);
}

//*********************************
static void import(Symbol p) {

// See comment above in export()
//	if( p->ref > 0 ) print("extern %s\n", p->x.name);
}

//*********************************
static void global(Symbol p) {

	Type	t;

	print("\t@align %d\n", p->type->align > 4 ? 4 : p->type->align);
	print(":%s\n", p->x.name);
	for(t=p->type; isarray(t); t=t->type) ;
	if( t->op==STRUCT ) print("\t@align 1\n");
}

//*********************************
static void space(long n) {

	if( cseg==BSS ) print("\tb [%d]\n", n);
	else print("\tb [%d] 0\n", n);
}

//****************************************************
// returns 0 if cp doesn't point to a register name, or len of register name if it does
static long ASMisreg() {

	if( *cp=='$' ) {
		if( *(cp+1)>='0' && *(cp+1)<='9' ) {
			if( (*(cp+1)=='1' && *(cp+2)>='0' && *(cp+2)<='9') || (*(cp+1)=='2' && *(cp+2)>='0' && *(cp+2)<='3') ) {
				if( !isalnum(*(cp+3)) && *(cp+3)!='_' ) return 3;
			} else if( !isalnum(*(cp+2)) && *(cp+2)!='_') return 2;
		}
	}
	return 0;
}

//*********************************
Interface kint7IR = {
	{1, 1, 0},	/* char */
	{2, 2, 0},	/* short */
	{4, 4, 0},	/* int */
	{4, 4, 0},	/* long */
	{8, 4, 0},	/* long long */
	{4, 4, 1},	/* float */
	{8, 4, 1},	/* double (size, alignment, outofline) */
	{8, 4, 1},	/* long double */
	{4, 4, 0},	/* T * */
	{0, 1, 0},	/* struct */
	
	1,			/* little_endian */
	0,			/* mulops_calls */
	0,			/* wants_callb */
	1,			/* wants_argb */
	0,			/* left_to_right */
	0,			/* wants_dag */
	0,			/* unsigned_char */
	
	address,
	blockbeg,
	blockend,
	defaddress,
	defconst,
	defstring,
	defsymbol,
	emit,
	export,
	function,
	gen,
	global,
	import,
	local,
	progbeg,
	progend,
	segment,
	space,
	0, 0, 0, 0, 0, 0, 0,
	{
		1, rmap,
		blkfetch, blkstore, blkloop,
		_label,
		_rule,
		_nts,
		_kids,
		_string,
		_templates,
		_isinstruction,
		_ntname,
		emit2,
		doarg,
		target,
		clobber,
		ASMisreg,
	}
};
